﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Xml.Linq;

using Esri.ArcGISRuntime;

using Extensibility;

namespace AppUpdateAddin
{

	/*
	 * AppUpdateHelper manages application, extension, and/or Runtime updates
	 * 
	 * **** ASSUMPTIONS ****
	 * 1) Application files reside under [IMainWindow.UpdateLocation]\[IMainWindow.ApplicationName]\RuntimeVersion
	 * 2) Extensibility.dll resides under [IMainWindow.UpdateLocation]\Extensibility\RuntimeVersion
	 * 3) Extension DLLs reside under [IMainWindow.UpdateLocation]\Extensions\RuntimeVersion
	 * 4) View DLLs reside under [IMainWindow.UpdateLocation]\Views
	 * 5) A file [IMainWindow.UpdateLocation]\Deployments\Deployments.txt may exist that specifies 
	 * the latest deployment for each AppName
	 * 6) Runtime deployment packages reside in the "Deployments" folder adjacent to the dataset update folder
	 * 7) Any new Runtime deployment will be placed into the current ArcGISRuntimeEnvironment.InstallPath
	*/

	public static class AppUpdateHelper
	{

		private struct FileUpdate
		{
			public string RemoteFile;
			public string LocalFile;
		}

		private const string _sUpdateScriptName = "update.js";
		private const string _sHeader = "Application Update";
		private const string _sDeployments = "Deployments.txt";

		private static bool _bConfirmUpdates = true;

		private static string _sAppUpdatePath;
		private static string _sRuntimeVersion;
		private static string _sLocalPath;
		private static string _sSep;
		private static string _sEscSep;

		// Public methods

		public static bool CheckForUpdates()
		{

			_sAppUpdatePath = UpdateModelExt.UpdateLocation;
			if (string.IsNullOrEmpty(_sAppUpdatePath))
				return false;
			_sRuntimeVersion = MapApplication.Current.RuntimeVersion;
			Assembly a = Assembly.GetEntryAssembly();
			_sLocalPath = Path.GetDirectoryName(a.Location);
			_sSep = Path.DirectorySeparatorChar.ToString();
			_sEscSep = _sSep + _sSep;

			// Check for presence of existing update script and delete it

			string sTempPath = Path.GetTempPath();
			string sUpdateScript = Path.Combine(sTempPath, _sUpdateScriptName);
			if (File.Exists(sUpdateScript))
				File.Delete(sUpdateScript);

			// Start building update script

			StringBuilder sb = new StringBuilder();
			sb.Append("var ws = WScript.CreateObject(\"WScript.Shell\");\r\n");

			// Kill any instances of the application
			// [This is harsh, but it ensures any application instances are completely closed before copying files]
			// [NOTE: This will not work in a Visual Studio session because it runs under a vshost]

			string sLocalExe = Path.GetFileName(a.Location);
			sb.Append("ws.Run(\"taskkill /f /im " + sLocalExe + "\");\r\n");
			sb.Append("WScript.Echo(\"Click OK to perform update.\");\r\n");
			sb.Append("var fso = new ActiveXObject(\"Scripting.FileSystemObject\");\r\n");

			// Check for updates

			bool bRuntime = AppUpdateHelper.CheckForRuntimeUpdates(sb);
			bool bUpdate = AppUpdateHelper.CheckForFileUpdates(sb, bRuntime);
			if (!bUpdate)
				return false;

			// Get confirmation

			if (_bConfirmUpdates)
			{
				string sMsg = "Updates have been detected.\nWould you like to install them?";
				MessageBoxResult mbResult = MessageBox.Show(sMsg, _sHeader, MessageBoxButton.YesNo);
				if (mbResult != MessageBoxResult.Yes)
					return false;
			}
			else
			{
				string sMsg = "Updates have been detected.\nThis application will restart in order to apply them.";
				MessageBox.Show(sMsg, _sHeader, MessageBoxButton.OK);
			}

			// Finish and execute update script

			string sConfig = ConfigHelper.ConfigName;
			string sCommand = a.Location + " " + sConfig;
			string sCommandJ = sCommand.Replace(_sSep, _sEscSep);
			sb.Append("ws.Run(\"" + sCommandJ + "\");\r\n");
			string sScript = sb.ToString();
			File.WriteAllText(sUpdateScript, sScript);
#if DEBUG
			if (Debugger.IsAttached)
			{
				MessageBox.Show("Update script may not work against a debugger session.\nApplication will now signal a breakpoint.");
				Debugger.Break();
			}
#endif
			ConfigHelper.Suspend();
			Process.Start(sUpdateScript);
			MapApplication.Current.Exit();
			return true;

		}

		// Private methods

		private static bool CheckForRuntimeUpdates(StringBuilder sb)
		{

			// Get Runtime update path

			if (string.IsNullOrEmpty(DatasetHelper.DatasetUpdateFolder))
				return false;
			string sRuntimeUpdatePath = DatasetHelper.DatasetUpdateFolder.Replace("Datasets", "Deployments");
			if (!Directory.Exists(sRuntimeUpdatePath))
				return false;

			// Check for latest deployment

			string sDeploymentFile = Path.Combine(_sAppUpdatePath, "Deployments", _sDeployments);
			if (!File.Exists(sDeploymentFile))
				return false;
			Dictionary<string, string> Deployments = new Dictionary<string, string>();
			using (StreamReader sr = new StreamReader(sDeploymentFile))
			{
				while (!sr.EndOfStream)
				{
					string s = sr.ReadLine();
					string[] Toks = s.Split(':');
					string sKey = Toks[0];
					string sVal = Toks[1];
					Deployments[sKey] = sVal;
				}
			}
			string sAppName = MapApplication.Current.ApplicationName;
			if (!Deployments.ContainsKey(sAppName))
				return false;
			string[] Entries = Deployments[sAppName].Split(',');
			string sRuntimeName = Entries[0];
			string sRuntimePackage = Entries[1];

			// Check if the deployment is an earlier version
			// [This will be true if installation is a manual upgrade]

			string sDeploymentVersion = sRuntimeName.ToLower().Replace("arcgisruntime", "");
			int[] Current = ParseVersion(_sRuntimeVersion);
			int[] Deployment = ParseVersion(sDeploymentVersion);
			if (Deployment[0] < Current[0])
				return false;
			else if ((Deployment[0] == Current[0]) && (Deployment[1] < Current[1]))
				return false;

			// Check if Runtime deployment already exists

			string sRuntimeLocalPath = ArcGISRuntimeEnvironment.InstallPath;
			string sNewFullPath = Path.Combine(sRuntimeLocalPath, sRuntimeName);
			if (Directory.Exists(sNewFullPath))
				return false;

			// Copy deployment to local folder

			string sTempPath = Path.GetTempPath();
			string sLocalFileName = Path.Combine(sTempPath, sRuntimePackage);
			string sLocalFileNameJ = sLocalFileName.Replace(_sSep, _sEscSep);
			string sRemoteFileName = Path.Combine(sRuntimeUpdatePath, sRuntimePackage);
			string sRemoteFileNameJ = sRemoteFileName.Replace(_sSep, _sEscSep);
			sb.Append("fso.CopyFile(\"" + sRemoteFileNameJ + "\",\"" + sLocalFileNameJ + "\", true);\r\n");

			// Unzip deployment

			sb.Append("var sa = WScript.CreateObject(\"Shell.Application\");\r\n");
			sb.Append("var items = sa.NameSpace(\"" + sLocalFileNameJ + "\").Items();\r\n");
			string sRuntimePathJ = sRuntimeLocalPath.Replace(_sSep, _sEscSep);
			sb.Append("sa.NameSpace(\"" + sRuntimePathJ + "\").CopyHere(items);\r\n");

			// Delete zip file

			sb.Append("fso.DeleteFile(\"" + sLocalFileNameJ + "\");\r\n");

			// Check for old Runtime deployment

			string sOldDeployment = "ArcGISRuntime" + MapApplication.Current.RuntimeVersion;
			string sOldRuntimePath = Path.Combine(sRuntimeLocalPath, sOldDeployment);
			if (Directory.Exists(sOldRuntimePath))
			{

				// Delete old folder

				string sOldRuntimePathJ = sOldRuntimePath.Replace(_sSep, _sEscSep);
				sb.Append("fso.DeleteFolder(\"" + sOldRuntimePathJ + "\");\r\n");

			}

			// Delete all files in app folder

			sLocalFileName = _sLocalPath + "\\*.*";
			sLocalFileNameJ = sLocalFileName.Replace(_sSep, _sEscSep);
			sb.Append("fso.DeleteFile(\"" + sLocalFileNameJ + "\");\r\n");

			// Delete all files in app subfolders except data

			foreach (string sSubfolder in Directory.GetDirectories(_sLocalPath))
			{
				if (sSubfolder.ToLower().IndexOf("data") > -1)
					continue;
				sLocalFileName = sSubfolder + "\\*.*";
				sLocalFileNameJ = sLocalFileName.Replace(_sSep, _sEscSep);
				sb.Append("fso.DeleteFile(\"" + sLocalFileNameJ + "\");\r\n");
			}

			return true;

		}

		private static bool CheckForFileUpdates(StringBuilder sb, bool bRuntime = true)
		{

			// Check for configuration update

			List<FileUpdate> FileUpdates = new List<FileUpdate>();
			string sConfigFile = ConfigHelper.ConfigName + ".xml";
			string sRemoteConfig = Path.Combine(_sAppUpdatePath, "Config", sConfigFile);
			string sLocalConfig = Path.Combine(_sLocalPath, "Config", sConfigFile);
			AppUpdateHelper.CheckForFileUpdate(sLocalConfig, sRemoteConfig, FileUpdates, bRuntime);

			// Check for core application updates

			string sAppName = MapApplication.Current.ApplicationName;
			foreach (string sSubDir in new string[] { sAppName, "Extensibility" })
			{
				string sRemotePath = Path.Combine(_sAppUpdatePath, sSubDir, _sRuntimeVersion);
				foreach (string sRemoteFileName in Directory.GetFiles(sRemotePath))
				{
					string sLocalFileName = sRemoteFileName.Replace(sRemotePath, _sLocalPath);
					AppUpdateHelper.CheckForFileUpdate(sLocalFileName, sRemoteFileName, FileUpdates, bRuntime);
				}
			}

			// Check for extension/view updates

			if (!bRuntime)
				AppUpdateHelper.CheckForExtensionUpdates(FileUpdates, bRuntime);
			if (FileUpdates.Count == 0)
				return false;

			// Copy application files

			foreach (FileUpdate fu in FileUpdates)
			{
				string sRemoteFileName = fu.RemoteFile.Replace(_sSep, _sEscSep);
				string sLocalFileName = fu.LocalFile.Replace(_sSep, _sEscSep);
				sb.Append("fso.CopyFile(\"" + sRemoteFileName + "\",\"" + sLocalFileName + "\", true);\r\n");
			}

			return true;

		}

		private static void CheckForFileUpdate(string sLocalFileName, string sRemoteFileName, List<FileUpdate>FileUpdates, bool bForce = false)
		{
			if (!UpdateModelExt.CheckForUpdate(sLocalFileName, sRemoteFileName, true))
				if (!bForce)
					return;
			FileUpdate fu = new FileUpdate();
			fu.RemoteFile = sRemoteFileName;
			fu.LocalFile = sLocalFileName;
			FileUpdates.Add(fu);
		}

		private static void CheckForExtensionUpdates(List<FileUpdate>FileUpdates, bool bForce = false)
		{

			// Get list of extension and view assembly names

			HashSet<string> ViewSet = new HashSet<string>();
			HashSet<string> ExtSet = new HashSet<string>();
			XDocument xConfigDoc = ConfigHelper.ConfigDoc;
			XElement xRoot = xConfigDoc.Root;
			XElement xLayout = xRoot.Element("MainLayout");
			if (xLayout != null)
			{
				XElement xView = xLayout.Element("ViewAssembly");
				if (xView != null)
					ViewSet.Add(xView.Value);
			}
			XElement xExtensions = xRoot.Element("Extensions");
			if (xExtensions != null)
			{
				foreach (XElement xExtension in xExtensions.Elements("Extension"))
				{
					ExtSet.Add(xExtension.Attribute("Name").Value);
					XElement xView = xExtension.Element("ViewAssembly");
					if (xView != null)
						ViewSet.Add(xView.Value);
				}
			}

			// Check for file updates

			string sRemoteExtPath = Path.Combine(_sAppUpdatePath, "Extensions", _sRuntimeVersion);
			string sLocalExtPath = Path.Combine(_sLocalPath, "Extensions");
			string sRemoteViewPath = Path.Combine(_sAppUpdatePath, "Views");
			string sLocalViewPath = Path.Combine(_sLocalPath, "Views");
			string sRemoteFileName, sLocalFileName;
			foreach (string sView in ViewSet)
			{
				sRemoteFileName = Path.Combine(sRemoteViewPath, sView + ".dll");
				sLocalFileName = Path.Combine(sLocalViewPath, sView + ".dll");
				AppUpdateHelper.CheckForFileUpdate(sLocalFileName, sRemoteFileName, FileUpdates);
			}
			foreach (string sExt in ExtSet)
			{
				sRemoteFileName = Path.Combine(sRemoteExtPath, sExt + ".dll");
				sLocalFileName = Path.Combine(sLocalExtPath, sExt + ".dll");
				AppUpdateHelper.CheckForFileUpdate(sLocalFileName, sRemoteFileName, FileUpdates);
			}

		}

		private static int[] ParseVersion(string sVer)
		{
			string[] Tok = sVer.Split('.');
			int.TryParse(Tok[0], out int iMajor);
			int.TryParse(Tok[1], out int iMinor);
			return new int[] { iMajor, iMinor };
		}

	}

}
