﻿using Microsoft.VisualBasic.FileIO;
using System;
using System.Collections.Generic;
using System.IO;
using System.IO.Compression;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Xml.Linq;

using Extensibility;

namespace AppUpdateAddin
{

	/*
	 * DatasetHelper manages updates of local datasets.
	 * 
	 * **** ASSUMPTIONS ****
	 * 1) Datasets are always packaged in zip files
	 * 2) If multiple dataset update locations exist, they are defined in the file DataLocations.txt
	 * in the [IMainWindow.UpdateLocation]\Datasets folder.
	*/

	static class DatasetHelper
	{

		private const string _sSettingsFileName = "TaskAddin.DatasetUpdateLocation.xml";
		private const string _sDatasetLocations = "DataLocations.txt";
		private const string _sUpdatePathEntryName = "UpdateLocation";
		private const string _sTimestampFileExtension = "tms";

		public static string ApplicationUpdateFolder;
		public static string DatasetUpdateFolder;

		private static XDocument _xSettings;

		// Public methods

		public static bool GetUpdateFolders(bool bUseSettings)
		{

			// Check if app update location is available

			string sAppUpdatePath = UpdateModelExt.UpdateLocation;
			if (string.IsNullOrEmpty(sAppUpdatePath))
				return false;
			if (!Directory.Exists(sAppUpdatePath))
				return false;
			ApplicationUpdateFolder = sAppUpdatePath;
			if (!bUseSettings)
			{
				DatasetUpdateFolder = Path.Combine(sAppUpdatePath, "Datasets");
				return true;
			}
			return DatasetHelper.GetUpdatePath(true);

		}

		public static void CheckDatasets(XElement xDatasets)
		{

			if (string.IsNullOrEmpty(DatasetUpdateFolder))
				return;

			// Loop through dataset entries

			bool bDoUpdate = false;
			DateTime dtCurrentCheck = DateTime.Now;
			foreach (XElement xDataset in xDatasets.Elements())
			{

				// Check if dataset path currently exists

				string sName = xDataset.Attribute("Name").Value;
				string sDestPath = xDataset.Attribute("Destination").Value;
				bool bExists = Directory.Exists(sDestPath);
				if (!bExists)
					Directory.CreateDirectory(sDestPath);

				// Check if dataset currently exists

				string sDestDataset = sDestPath + "\\" + sName;
				bExists = Directory.Exists(sDestDataset);
				if (!bExists)
					bExists = File.Exists(sDestDataset);

				// Check for timestamp

				string sDestTimestamp = sDestPath + "\\" + sName + "." + _sTimestampFileExtension;
				DateTime dtTimestamp = ReadTimestamp(sDestTimestamp);

				// Check for update

				string sSourceZip = DatasetUpdateFolder + "\\" + sName + ".zip";
				if (!File.Exists(sSourceZip))
					continue;
				bool bUpdate = false;
				if (!bExists)
					bUpdate = true;
				else
				{
					FileInfo fiUpdate = new FileInfo(sSourceZip);
					bUpdate = (fiUpdate.LastWriteTime > dtTimestamp);
				}
				if (!bUpdate)
					continue;
				if (!bDoUpdate)
				{
					string sMsg = "Data updates have been detected.  Would you like to install them?";
					string sAppName = MapApplication.Current.ApplicationName;
					MessageBoxResult mbResult = MessageBox.Show(sMsg, sAppName, MessageBoxButton.YesNo);
					if (mbResult != MessageBoxResult.Yes)
						return;
					bDoUpdate = true;
				}

				// Delete old dataset if it exists

				if (bExists)
				{
					try
					{
						if (Directory.Exists(sDestDataset))
							Directory.Delete(sDestDataset, true);
						else
							File.Delete(sDestDataset);
					}
					catch (Exception ex)
					{
						string sMsg = "Could not delete " + sDestDataset + ":\n" + ex.Message;
						DatasetHelper.Error(sMsg);
						return;
					}
				}

				// Retrieve and deploy update

				string sErr = "";
				try
				{
					sErr = "copy " + sName + "archive";
					string sTemp = Environment.GetEnvironmentVariable("TEMP");
					string sLocalZip = sTemp + "\\" + sName + ".zip";
					FileSystem.CopyFile(sSourceZip, sLocalZip, UIOption.AllDialogs);
					sErr = "extract " + sName + "archive";
					ZipFile.ExtractToDirectory(sLocalZip, sDestPath);
					sErr = "delete " + sName + "archive";
					File.Delete(sLocalZip);
					sErr = "write " + sName + "timestamp";
					WriteTimestamp(sDestTimestamp, dtCurrentCheck);
				}
				catch (Exception ex)
				{
					String sMsg = "Could not " + sErr + ":\n" + ex.Message;
					DatasetHelper.Error(sMsg);
					return;
				}

			} // foreach xDataset

		}

		public static bool SetUpdateLocation(bool bCheckForValue = true)
		{

			// Get current setting, if present

			string sCurrentPath = "";
			if (bCheckForValue)
			{
				DatasetHelper.GetUpdatePath(false);
				if (!string.IsNullOrEmpty(DatasetUpdateFolder))
					sCurrentPath = DatasetUpdateFolder;
			}

			// Read master list of dataset locations

			string sDatasetLocationFile = Path.Combine(ApplicationUpdateFolder, "Datasets", _sDatasetLocations);
			if (!File.Exists(sDatasetLocationFile))
			{
				MapApplication.Current.Error("File not found: " + _sDatasetLocations);
				return false;
			}
			string sCurrentLocation = "";
			Dictionary<string, string> DatasetLocations = new Dictionary<string, string>();
			List<string> LocationList = new List<string>();
			using (StreamReader sr = new StreamReader(sDatasetLocationFile))
			{
				while (!sr.EndOfStream)
				{
					string s = sr.ReadLine();
					string[] Toks = s.Split(',');
					string sKey = Toks[0];
					string sVal = Toks[1];
					if (sVal == sCurrentPath)
						sCurrentLocation = sKey;
					DatasetLocations[sKey] = sVal;
					LocationList.Add(sKey);
				}
			}
			LocationList.Sort();

			// Select location

			string sLoc = null;
			PickListDialogModel model = new PickListDialogModel(LocationList, "Choose Dataset Update Location", sCurrentLocation);
			bool? bResult = MapApplication.Current.CreateChildWindow("AppUpdateAddin", "PickListDialog", model);
			if (bResult == null)
				return false;
			if (!bResult.Value)
				return false;
			sLoc = model.SelectedValue;
			if (string.IsNullOrEmpty(sLoc))
				return false;
			DatasetUpdateFolder = DatasetLocations[sLoc];

			// Save settings

			XElement xRoot = _xSettings.Root;
			XElement xUpdatePath = xRoot.Element(_sUpdatePathEntryName);
			if (xUpdatePath == null)
			{
				xUpdatePath = new XElement(_sUpdatePathEntryName);
				xRoot.Add(xUpdatePath);
			}
			xUpdatePath.Value = DatasetUpdateFolder;
			SettingsHelper.SaveSettings(_sSettingsFileName, _xSettings);
			return true;

		}

		// Private methods

		private static void Error(string sMsg)
		{
			string sAppName = MapApplication.Current.ApplicationName;
			MessageBox.Show(sMsg, sAppName, MessageBoxButton.OK, MessageBoxImage.Error);
		}

		private static bool GetUpdatePath(bool bSetIfMissing = true)
		{

			// Get settings

			if (!string.IsNullOrEmpty(DatasetUpdateFolder))
				return true;
			_xSettings = SettingsHelper.OpenSettings(_sSettingsFileName);
			XElement xRoot = _xSettings.Root;
			XElement xUpdatePath = xRoot.Element(_sUpdatePathEntryName);
			if (xUpdatePath == null)
			{
				if (!DatasetHelper.SetUpdateLocation(false))
					return false;
			}
			else
				DatasetUpdateFolder = xUpdatePath.Value;
			return Directory.Exists(DatasetUpdateFolder);

		}

		private static DateTime ReadTimestamp(string sTimeStampFile)
		{
			DateTime dtTimestamp = DateTime.Now;
			if (!File.Exists(sTimeStampFile))
			{
				WriteTimestamp(sTimeStampFile, dtTimestamp);
				return dtTimestamp;
			}
			string sTimeStamp = File.ReadAllText(sTimeStampFile);
			dtTimestamp = Convert.ToDateTime(sTimeStamp);
			return dtTimestamp;
		}

		private static void WriteTimestamp(string sTimestampFile, DateTime dtTimestamp)
		{
			if (File.Exists(sTimestampFile))
				File.Delete(sTimestampFile);
			File.WriteAllText(sTimestampFile, dtTimestamp.ToString());
		}

	}

}
