﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;
using System.Xml.Linq;

using Esri.ArcGISRuntime;

using Extensibility;

namespace AppUpdateAddin
{

	// UpdateModelExt add support for application, Runtime, dataset, and other file updates

	public class UpdateModelExt : IConfigurable
   {

		public static string UpdateLocation;

		// Bindable layout properties

		public ICommand SetLocationCommand { get; private set; }

		public Visibility LocationVisibility { get; private set; }

		// IConfigurable

		public void LoadConfiguration(string sConfigData)
		{

			if (string.IsNullOrEmpty(sConfigData))
				return;
			XDocument xDoc = XDocument.Parse(sConfigData);
			XElement xRoot = xDoc.Root;
			this.ProcessConfiguration(xRoot);

		}

		// Constructor

		public UpdateModelExt()
		{
			CommandHelper ch = new CommandHelper(true);
			ch.AddExecuteAction(this.SetLocation);
			SetLocationCommand = ch;
		}

		// Public methods

		public static bool CheckForUpdate(string sLocalFileName, string sUpdateFileName, bool bCheckOnly = false)
		{

			if (!File.Exists(sUpdateFileName))
				return false;
			bool bUpdate = false;
			if (!File.Exists(sLocalFileName))
				bUpdate = true;
			else
			{
				FileInfo fiUpdate = new FileInfo(sUpdateFileName);
				FileInfo fiLocal = new FileInfo(sLocalFileName);
				if (fiUpdate.LastWriteTime > fiLocal.LastWriteTime)
					bUpdate = true;
			}
			if (!bUpdate)
				return bUpdate;
			if (bCheckOnly)
				return bUpdate;
			try
			{
				string sDir = Path.GetDirectoryName(sLocalFileName);
				if (!Directory.Exists(sDir))
					Directory.CreateDirectory(sDir);
				File.Copy(sUpdateFileName, sLocalFileName, true);
			}
			catch (Exception ex)
			{
				string sMsg = "Error updating " + sLocalFileName + ":\n" + ex.Message;
				MapApplication.Current.Error(sMsg);
				bUpdate = false;
			}
			return bUpdate;

		}

		// Private methods

		private void ProcessConfiguration(XElement xRoot)
		{

			LocationVisibility = Visibility.Collapsed;

			// Check for master update folder

			string sUpdateLocation = MapApplication.Current.UpdateLocation;
			if (string.IsNullOrEmpty(sUpdateLocation))
				return;
			if (!Directory.Exists(sUpdateLocation))
				return;
			UpdateLocation = sUpdateLocation;

			// Check for update folders

			bool bUseSettings = false;
			XElement xSettings = xRoot.Element("SetUpdateFolder");
			if (xSettings != null)
				bUseSettings = (xSettings.Value == "True");
			if (bUseSettings)
				LocationVisibility = Visibility.Visible;
			if (!DatasetHelper.GetUpdateFolders(bUseSettings))
				return;

			// Check for core application updates

			if (AppUpdateHelper.CheckForUpdates())
				return; // Application will be updated

			// Check for deletions

			XElement xDeletions = xRoot.Element("Deletions");
			if (xDeletions != null)
			{
				foreach (XElement xDelete in xDeletions.Elements())
				{
					string sPath = xDelete.Attribute("Path").Value;
					try
					{
						if (Directory.Exists(sPath))
							Directory.Delete(sPath, true);
						else if (File.Exists(sPath))
							File.Delete(sPath);
					}
					catch (Exception ex)
					{
						string sMsg = "Error deleting " + sPath + ":\n" + ex.Message;
						MapApplication.Current.Error(sMsg);
					}
				}
			}

			// Check for file updates

			XElement xFileUpdates = xRoot.Element("FileUpdates");
			if (xFileUpdates != null)
			{
				string sAppUpdatePath = DatasetHelper.ApplicationUpdateFolder;
				foreach (XElement xFile in xFileUpdates.Elements())
				{
					string sSrc = xFile.Attribute("Source").Value;
					string sUpdateFileName = Path.Combine(sAppUpdatePath, sSrc);
					string sDest = xFile.Attribute("Destination").Value;
					string sLocalFileName = Path.Combine(sDest, Path.GetFileName(sSrc));
					UpdateModelExt.CheckForUpdate(sLocalFileName, sUpdateFileName);
				}
			}

			// Check for dataset updates

			XElement xDatasetUpdates = xRoot.Element("DatasetUpdates");
			if (xDatasetUpdates == null)
				return;
			DatasetHelper.CheckDatasets(xDatasetUpdates);

		}

		private void SetLocation(object parameter)
		{
			DatasetHelper.SetUpdateLocation(true);
		}

	}

}
