﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;

using Esri.ArcGISRuntime.Data;
using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Symbology;
using Esri.ArcGISRuntime.UI;

using Extensibility;
using EditAddin;

namespace DemoAddin
{
	public class DemoTools : INotifyPropertyChanged
	{

		private bool _bEditChecked = false;

		// Bindable layout properties

		public ICommand EditToolCommand { get; private set; }

		public Visibility EditToolVisibility { get; private set; }
		
		public bool EditToolChecked
		{
			get { return _bEditChecked; }
			set
			{
				_bEditChecked = value;
				this.RaisePropertyChanged();
			}
		}

		// INotifyPropertyChanged

		public event PropertyChangedEventHandler PropertyChanged;

		// Constructors

		public DemoTools()
		{
			EditToolVisibility = Visibility.Visible;
			MapApplication.Current.AppReady += MapApplication_AppReady;
		}

		// Public events

		public event EventHandler<NotifyCanExecuteArgs> CanExecuteEditTool;

		// Event handlers

		void ToolHelper_ToolComplete(object sender, ToolHelper.NotifyToolArgs e)
		{
			this.AddMarkup(e.Geometry);
		}

		// Private methods

		private void RaisePropertyChanged([CallerMemberName] string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private void RaiseCanExecuteEditTool()
		{
			if (this.CanExecuteEditTool == null)
				return;
			NotifyCanExecuteArgs e = new NotifyCanExecuteArgs();
			e.CanExecute = true;
			this.CanExecuteEditTool(this, e);
		}

		private void MapApplication_AppReady(object sender, EventArgs e)
		{

			// Create edit tool

			ToolHelper th = new ToolHelper(false);
			th.ExecuteCommand.AddCanExecuteEvent(this, "CanExecuteEditTool");
			th.ActivateCommand.AddExecuteAction(this.ExecuteActivate, null);
			th.DeactivateCommand.AddExecuteAction(this.ExecuteDeactivate, null);
			th.ToolComplete += ToolHelper_ToolComplete;
			EditToolCommand = th.ExecuteCommand;
			EditModelExt.Current.EditDataLoaded += EditModelExt_EditDataLoaded;

			// Set up configuration

			ToolConfiguration tc = th.Configuration;
			tc.Cursor = Cursors.Cross;
			tc.Editable = false;
			tc.SketchType = SketchCreationMode.Point;

		}

		private void EditModelExt_EditDataLoaded(object sender, EventArgs e)
		{
			this.RaiseCanExecuteEditTool();
		}

		private void ExecuteActivate(object parameter)
		{
			EditToolChecked = true;
		}

		private void ExecuteDeactivate(object parameter)
		{
			EditToolChecked = false;
		}

		private async void AddMarkup(Geometry geom)
		{

			// Get feature table

			string sTableName = "Map Change Request";
			ArcGISFeatureTable fTab = EditModelExt.Current.GetFeatureTable(sTableName);
			if (fTab == null)
			{
				MapApplication.Current.Error("Feature table not found: " + sTableName);
				return;
			}

			// Get attributes

			DemoMarkupDialogModel model = new DemoMarkupDialogModel();
			bool? bResult = MapApplication.Current.CreateChildWindow("DemoAddin", "DemoMarkupDialog", model);
			if (bResult == null)
				return;
			if (!bResult.Value)
				return;
			Dictionary<string, object> Values = model.Values;
			if (Values == null)
				return;

			// Add new feature

			try
			{
				Feature feat = fTab.CreateFeature();
				feat.Geometry = geom;
				foreach (string sKey in Values.Keys)
					feat.Attributes[sKey] = Values[sKey];
				await fTab.AddFeatureAsync(feat);
				if (fTab is ServiceFeatureTable)
				{
					ServiceFeatureTable sfTab = (ServiceFeatureTable)fTab;
					IReadOnlyList<EditResult> Results = await sfTab.ApplyEditsAsync();
					EditResult edResult = Results[0];
					string sMsg = "Edit ";
					if (!edResult.CompletedWithErrors)
						sMsg += "successful";
					else
						sMsg += "failed";
					System.Diagnostics.Debug.Print(sMsg);
				}
			}
			catch (Exception ex)
			{
				MapApplication.Current.Error("Error adding markup:\n" + ex.Message);
			}

		}

	}
}
