﻿using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Runtime.InteropServices.ComTypes;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Interop;
using System.Windows.Media;

/*
 **** DirectShow Image Capture Helpers **** 
 Adapted from code by Tamir Khason
 http://khason.net/blog/webcam-control-with-wpf-or-how-to-create-high-framerate-player-with-directshow-by-using-interopbitmap-in-wpf-application/
*/

namespace EditAddin
{

	#region CapPlayer
	public class CapPlayer : Image, IDisposable
	{

		CapDevice _device;

		public CapPlayer()
		{
			Application.Current.Exit += new ExitEventHandler(Current_Exit);
		}

		public void SetDevice(string sMoniker)
		{
			if (_device != null)
			{
				_device.OnNewBitmapReady -= new EventHandler(_device_OnNewBitmapReady);
				_device.Dispose();
			}
			_device = new CapDevice(sMoniker);
			_device.OnNewBitmapReady += new EventHandler(_device_OnNewBitmapReady);
			_device.Start();
		}

		void Current_Exit(object sender, ExitEventArgs e)
		{
			this.Dispose();
		}

		void _device_OnNewBitmapReady(object sender, EventArgs e)
		{
			this.Source = _device.BitmapSource;
		}

		#region IDisposable Members
		public void Dispose()
		{
			if (_device != null)
				_device.Dispose();
		}
		#endregion

	}
	#endregion

	#region CapDevice
	internal class CapDevice : DependencyObject, IDisposable
	{

		[DllImport("kernel32.dll", SetLastError = true)]
		static extern IntPtr CreateFileMapping(IntPtr hFile, IntPtr lpFileMappingAttributes, uint flProtect, uint dwMaximumSizeHigh, uint dwMaximumSizeLow, string lpName);

		[DllImport("kernel32.dll", SetLastError = true)]
		static extern IntPtr MapViewOfFile(IntPtr hFileMappingObject, uint dwDesiredAccess, uint dwFileOffsetHigh, uint dwFileOffsetLow, uint dwNumberOfBytesToMap);

		bool _bActive = false;
		CapGrabber capGrabber;
		static string deviceMoniker;

		public InteropBitmap BitmapSource
		{
			get { return (InteropBitmap)GetValue(BitmapSourceProperty); }
			private set { SetValue(BitmapSourcePropertyKey, value); }
		}

		private static readonly DependencyPropertyKey BitmapSourcePropertyKey =
			 DependencyProperty.RegisterReadOnly("BitmapSource", typeof(InteropBitmap), typeof(CapDevice), new UIPropertyMetadata(default(InteropBitmap)));
		public static readonly DependencyProperty BitmapSourceProperty = BitmapSourcePropertyKey.DependencyProperty;

		public CapDevice()
		{
			if (DeviceMonikers.Count() <= 0)
				throw new NotSupportedException("You should connect DirectShow source first");
			deviceMoniker = DeviceMonikers[0].MonikerString;
			Start();
		}

		public CapDevice(string moniker)
		{
			deviceMoniker = moniker;
			Start();
		}

		public static FilterInfo[] DeviceMonikers
		{
			get
			{
				List<FilterInfo> filters = new List<FilterInfo>();
				IMoniker[] ms = new IMoniker[1];
				ICreateDevEnum enumD = Activator.CreateInstance(Type.GetTypeFromCLSID(SystemDeviceEnum)) as ICreateDevEnum;
				IEnumMoniker moniker;
				Guid g = VideoInputDevice;
				if (enumD.CreateClassEnumerator(ref g, out moniker, 0) == 0)
				{

					while (true)
					{
						int r = moniker.Next(1, ms, IntPtr.Zero);
						if (r != 0 || ms[0] == null)
							break;
						filters.Add(new FilterInfo(ms[0]));
						Marshal.ReleaseComObject(ms[0]);
						ms[0] = null;

					}
				}

				return filters.ToArray();
			}
		}

		public void Start()
		{
			_bActive = true;
			if (capGrabber != null)
			{
				capGrabber.PropertyChanged -= new System.ComponentModel.PropertyChangedEventHandler(capGrabber_PropertyChanged);
				capGrabber.NewFrameArrived -= new EventHandler(capGrabber_NewFrameArrived);
				capGrabber = null;
			}
			capGrabber = new CapGrabber();
			capGrabber.PropertyChanged += new System.ComponentModel.PropertyChangedEventHandler(capGrabber_PropertyChanged);
			capGrabber.NewFrameArrived += new EventHandler(capGrabber_NewFrameArrived);
			Task t = Task.Factory.StartNew(() => this.RunWorker());

		}

		public void Stop()
		{
			_bActive = false;
		}

		void capGrabber_NewFrameArrived(object sender, EventArgs e)
		{
			if (this.Dispatcher != null)
			{
				this.Dispatcher.Invoke(System.Windows.Threading.DispatcherPriority.Render, (SendOrPostCallback)delegate
				{
					if (BitmapSource != null)
					{
						BitmapSource.Invalidate();
					}
				}, null);
			}
		}

		void capGrabber_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
		{
			this.Dispatcher.Invoke(System.Windows.Threading.DispatcherPriority.DataBind, (SendOrPostCallback)delegate
			{
				if (capGrabber.Width != default(int) && capGrabber.Height != default(int))
				{

					uint pcount = (uint)(capGrabber.Width * capGrabber.Height * PixelFormats.Bgr32.BitsPerPixel / 8);
					IntPtr section = CreateFileMapping(new IntPtr(-1), IntPtr.Zero, 0x04, 0, pcount, null);
					IntPtr map = MapViewOfFile(section, 0xF001F, 0, 0, pcount);
					BitmapSource = System.Windows.Interop.Imaging.CreateBitmapSourceFromMemorySection(section, capGrabber.Width, capGrabber.Height, PixelFormats.Bgr32,
						 capGrabber.Width * PixelFormats.Bgr32.BitsPerPixel / 8, 0) as InteropBitmap;
					capGrabber.Map = map;
					if (OnNewBitmapReady != null)
						OnNewBitmapReady(this, null);
				}
			}, null);
		}

		void RunWorker()
		{

			IGraphBuilder graph;
			ISampleGrabber grabber;
			IBaseFilter sourceObject, grabberObject;
			IMediaControl control;

			try
			{

				graph = Activator.CreateInstance(Type.GetTypeFromCLSID(FilterGraph)) as IGraphBuilder;

				sourceObject = FilterInfo.CreateFilter(deviceMoniker);

				grabber = Activator.CreateInstance(Type.GetTypeFromCLSID(SampleGrabber)) as ISampleGrabber;
				grabberObject = grabber as IBaseFilter;

				graph.AddFilter(sourceObject, "source");
				graph.AddFilter(grabberObject, "grabber");

				using (AMMediaType mediaType = new AMMediaType())
				{
					mediaType.MajorType = MediaTypes.Video;
					mediaType.SubType = MediaSubTypes.RGB32;
					grabber.SetMediaType(mediaType);

					if (graph.Connect(sourceObject.GetPin(PinDirection.Output, 0), grabberObject.GetPin(PinDirection.Input, 0)) >= 0)
					{
						if (grabber.GetConnectedMediaType(mediaType) == 0)
						{
							VideoInfoHeader header = (VideoInfoHeader)Marshal.PtrToStructure(mediaType.FormatPtr, typeof(VideoInfoHeader));
							capGrabber.Width = header.BmiHeader.Width;
							capGrabber.Height = header.BmiHeader.Height;
						}
					}
					graph.Render(grabberObject.GetPin(PinDirection.Output, 0));
					grabber.SetBufferSamples(false);
					grabber.SetOneShot(false);
					grabber.SetCallback(capGrabber, 1);

					IVideoWindow wnd = (IVideoWindow)graph;
					wnd.put_AutoShow(false);
					wnd = null;

					control = (IMediaControl)graph;
					control.Run();
					while (_bActive)
					{ 
						// Nothing to do here
					}
					control.StopWhenReady();
				}
			}
			catch (Exception ex)
			{
				System.Diagnostics.Debug.WriteLine(ex);
			}
			finally
			{
				graph = null;
				sourceObject = null;
				grabberObject = null;
				grabber = null;
				capGrabber = null;
				control = null;

			}

		}

		static readonly Guid FilterGraph = new Guid(0xE436EBB3, 0x524F, 0x11CE, 0x9F, 0x53, 0x00, 0x20, 0xAF, 0x0B, 0xA7, 0x70);

		static readonly Guid SampleGrabber = new Guid(0xC1F400A0, 0x3F08, 0x11D3, 0x9F, 0x0B, 0x00, 0x60, 0x08, 0x03, 0x9E, 0x37);

		public static readonly Guid SystemDeviceEnum = new Guid(0x62BE5D10, 0x60EB, 0x11D0, 0xBD, 0x3B, 0x00, 0xA0, 0xC9, 0x11, 0xCE, 0x86);

		public static readonly Guid VideoInputDevice = new Guid(0x860BB310, 0x5D01, 0x11D0, 0xBD, 0x3B, 0x00, 0xA0, 0xC9, 0x11, 0xCE, 0x86);

		[ComVisible(false)]
		internal class MediaTypes
		{
			public static readonly Guid Video = new Guid(0x73646976, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xAA, 0x00, 0x38, 0x9B, 0x71);

			public static readonly Guid Interleaved = new Guid(0x73766169, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xAA, 0x00, 0x38, 0x9B, 0x71);

			public static readonly Guid Audio = new Guid(0x73647561, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xAA, 0x00, 0x38, 0x9B, 0x71);

			public static readonly Guid Text = new Guid(0x73747874, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xAA, 0x00, 0x38, 0x9B, 0x71);

			public static readonly Guid Stream = new Guid(0xE436EB83, 0x524F, 0x11CE, 0x9F, 0x53, 0x00, 0x20, 0xAF, 0x0B, 0xA7, 0x70);
		}

		[ComVisible(false)]
		internal class MediaSubTypes
		{
			public static readonly Guid YUYV = new Guid(0x56595559, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xAA, 0x00, 0x38, 0x9B, 0x71);

			public static readonly Guid IYUV = new Guid(0x56555949, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xAA, 0x00, 0x38, 0x9B, 0x71);

			public static readonly Guid DVSD = new Guid(0x44535644, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xAA, 0x00, 0x38, 0x9B, 0x71);

			public static readonly Guid RGB1 = new Guid(0xE436EB78, 0x524F, 0x11CE, 0x9F, 0x53, 0x00, 0x20, 0xAF, 0x0B, 0xA7, 0x70);

			public static readonly Guid RGB4 = new Guid(0xE436EB79, 0x524F, 0x11CE, 0x9F, 0x53, 0x00, 0x20, 0xAF, 0x0B, 0xA7, 0x70);

			public static readonly Guid RGB8 = new Guid(0xE436EB7A, 0x524F, 0x11CE, 0x9F, 0x53, 0x00, 0x20, 0xAF, 0x0B, 0xA7, 0x70);

			public static readonly Guid RGB565 = new Guid(0xE436EB7B, 0x524F, 0x11CE, 0x9F, 0x53, 0x00, 0x20, 0xAF, 0x0B, 0xA7, 0x70);

			public static readonly Guid RGB555 = new Guid(0xE436EB7C, 0x524F, 0x11CE, 0x9F, 0x53, 0x00, 0x20, 0xAF, 0x0B, 0xA7, 0x70);

			public static readonly Guid RGB24 = new Guid(0xE436Eb7D, 0x524F, 0x11CE, 0x9F, 0x53, 0x00, 0x20, 0xAF, 0x0B, 0xA7, 0x70);

			public static readonly Guid RGB32 = new Guid(0xE436EB7E, 0x524F, 0x11CE, 0x9F, 0x53, 0x00, 0x20, 0xAF, 0x0B, 0xA7, 0x70);

			public static readonly Guid Avi = new Guid(0xE436EB88, 0x524F, 0x11CE, 0x9F, 0x53, 0x00, 0x20, 0xAF, 0x0B, 0xA7, 0x70);

			public static readonly Guid Asf = new Guid(0x3DB80F90, 0x9412, 0x11D1, 0xAD, 0xED, 0x00, 0x00, 0xF8, 0x75, 0x4B, 0x99);
		}

		public event EventHandler OnNewBitmapReady;

		#region IDisposable Members
		public void Dispose()
		{
			Stop();
		}
		#endregion
	}
	#endregion

	#region CapGrabber
	internal class CapGrabber : ISampleGrabberCB, INotifyPropertyChanged
	{

		[DllImport("Kernel32.dll", EntryPoint = "RtlMoveMemory")]
		private static extern void CopyMemory(IntPtr Destination, IntPtr Source, int Length);

		public CapGrabber()
		{

		}

		public IntPtr Map { get; set; }

		public int Width
		{
			get { return m_Width; }
			set
			{
				m_Width = value;
				OnPropertyChanged("Width");
			}
		}

		public int Height
		{
			get { return m_Height; }
			set
			{
				m_Height = value;
				OnPropertyChanged("Height");
			}
		}

		int m_Height = default(int);

		int m_Width = default(int);

		#region ISampleGrabberCB Members
		public int SampleCB(double sampleTime, IntPtr sample)
		{
			return 0;
		}

		public int BufferCB(double sampleTime, IntPtr buffer, int bufferLen)
		{
			if (Map != IntPtr.Zero)
			{
				int stride = Width * PixelFormats.Bgr32.BitsPerPixel / 8;
				for (IntPtr pMap = Map, pBuf = buffer + bufferLen; pBuf.ToInt64() > buffer.ToInt64(); pMap += stride, pBuf -= stride)
					CopyMemory(pMap, pBuf - stride, stride);
				//CopyMemory(Map, buffer, bufferLen);
				OnNewFrameArrived();
			}
			return 0;
		}
		#endregion

		public event EventHandler NewFrameArrived;
		void OnNewFrameArrived()
		{
			if (NewFrameArrived != null)
				NewFrameArrived(this, null);
		}

		#region INotifyPropertyChanged Members
		public event PropertyChangedEventHandler PropertyChanged;

		void OnPropertyChanged(string name)
		{
			if (PropertyChanged != null)
			{
				PropertyChanged(this, new PropertyChangedEventArgs(name));
			}
		}
		#endregion

	}
	#endregion

	#region FilterInfo
	public class FilterInfo : IComparable
	{

		[DllImport("ole32.dll")]
		public static extern int CreateBindCtx(int reserved, out IBindCtx ppbc);

		[DllImport("ole32.dll", CharSet = CharSet.Unicode)]
		public static extern int MkParseDisplayName(IBindCtx pbc, string szUserName, ref int pchEaten, out IMoniker ppmk);

		public readonly string Name;

		public readonly string MonikerString;

		public FilterInfo(string monikerString)
		{
			MonikerString = monikerString;
			Name = GetName(monikerString);
		}

		public override string ToString()
		{
			return Name;
		}

		internal FilterInfo(IMoniker moniker)
		{
			MonikerString = GetMonikerString(moniker);
			Name = GetName(moniker);
		}

		public int CompareTo(object value)
		{
			FilterInfo f = (FilterInfo)value;

			if (f == null)
				return 1;

			return (this.Name.CompareTo(f.Name));
		}

		internal static IBaseFilter CreateFilter(string filterMoniker)
		{
			object filterObject = null;
			IBindCtx bindCtx = null;
			IMoniker moniker = null;

			int n = 0;

			if (CreateBindCtx(0, out bindCtx) == 0)
			{
				if (MkParseDisplayName(bindCtx, filterMoniker, ref n, out moniker) == 0)
				{
					Guid filterId = typeof(IBaseFilter).GUID;
					moniker.BindToObject(null, null, ref filterId, out filterObject);

					Marshal.ReleaseComObject(moniker);
				}
				Marshal.ReleaseComObject(bindCtx);
			}
			return filterObject as IBaseFilter;
		}

		private string GetMonikerString(IMoniker moniker)
		{
			string str;
			moniker.GetDisplayName(null, null, out str);
			return str;
		}

		private string GetName(IMoniker moniker)
		{
			Object bagObj = null;
			IPropertyBag bag = null;

			try
			{
				Guid bagId = typeof(IPropertyBag).GUID;
				moniker.BindToStorage(null, null, ref bagId, out bagObj);
				bag = (IPropertyBag)bagObj;

				object val = "";
				int hr = bag.Read("FriendlyName", ref val, IntPtr.Zero);
				if (hr != 0)
					Marshal.ThrowExceptionForHR(hr);

				string ret = (string)val;
				if ((ret == null) || (ret.Length < 1))
					throw new ApplicationException();

				return ret;
			}
			catch (Exception)
			{
				return "";
			}
			finally
			{
				bag = null;
				if (bagObj != null)
				{
					Marshal.ReleaseComObject(bagObj);
					bagObj = null;
				}
			}
		}

		private string GetName(string monikerString)
		{
			IBindCtx bindCtx = null;
			IMoniker moniker = null;
			String name = "";
			int n = 0;

			if (CreateBindCtx(0, out bindCtx) == 0)
			{
				if (MkParseDisplayName(bindCtx, monikerString, ref n, out moniker) == 0)
				{
					name = GetName(moniker);

					Marshal.ReleaseComObject(moniker);
					moniker = null;
				}
				Marshal.ReleaseComObject(bindCtx);
				bindCtx = null;
			}
			return name;
		}

	}
	#endregion

	#region CapHelper
	internal static class CapHelper
	{
		public static IPin GetPin(this IBaseFilter filter, PinDirection dir, int num)
		{
			IPin[] pin = new IPin[1];
			IEnumPins pinsEnum = null;

			if (filter.EnumPins(out pinsEnum) == 0)
			{
				PinDirection pinDir;
				int n;

				while (pinsEnum.Next(1, pin, out n) == 0)
				{
					pin[0].QueryDirection(out pinDir);

					if (pinDir == dir)
					{
						if (num == 0)
							return pin[0];
						num--;
					}

					Marshal.ReleaseComObject(pin[0]);
					pin[0] = null;
				}
			}
			return null;
		}
	}
	#endregion

	#region COM Interfaces
	[ComImport, Guid("56A868A9-0AD4-11CE-B03A-0020AF0BA770"), InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
	internal interface IGraphBuilder
	{
		[PreserveSig]
		int AddFilter([In] IBaseFilter filter, [In, MarshalAs(UnmanagedType.LPWStr)] string name);

		[PreserveSig]
		int RemoveFilter([In] IBaseFilter filter);

		[PreserveSig]
		int EnumFilters([Out] out IntPtr enumerator);

		[PreserveSig]
		int FindFilterByName([In, MarshalAs(UnmanagedType.LPWStr)] string name, [Out] out IBaseFilter filter);

		[PreserveSig]
		int ConnectDirect([In] IPin pinOut, [In] IPin pinIn, [In, MarshalAs(UnmanagedType.LPStruct)] AMMediaType mediaType);

		[PreserveSig]
		int Reconnect([In] IPin pin);

		[PreserveSig]
		int Disconnect([In] IPin pin);

		[PreserveSig]
		int SetDefaultSyncSource();

		[PreserveSig]
		int Connect([In] IPin pinOut, [In] IPin pinIn);

		[PreserveSig]
		int Render([In] IPin pinOut);

		[PreserveSig]
		int RenderFile(
			 [In, MarshalAs(UnmanagedType.LPWStr)] string file,
			 [In, MarshalAs(UnmanagedType.LPWStr)] string playList);

		[PreserveSig]
		int AddSourceFilter(
			 [In, MarshalAs(UnmanagedType.LPWStr)] string fileName,
			 [In, MarshalAs(UnmanagedType.LPWStr)] string filterName,
			 [Out] out IBaseFilter filter);

		[PreserveSig]
		int SetLogFile(IntPtr hFile);

		[PreserveSig]
		int Abort();

		[PreserveSig]
		int ShouldOperationContinue();
	}

	[ComImport, Guid("56A86895-0AD4-11CE-B03A-0020AF0BA770"), InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
	internal interface IBaseFilter
	{
		[PreserveSig]
		int GetClassID([Out] out Guid ClassID);

		[PreserveSig]
		int Stop();

		[PreserveSig]
		int Pause();

		[PreserveSig]
		int Run(long start);

		[PreserveSig]
		int GetState(int milliSecsTimeout, [Out] out int filterState);

		[PreserveSig]
		int SetSyncSource([In] IntPtr clock);

		[PreserveSig]
		int GetSyncSource([Out] out IntPtr clock);

		[PreserveSig]
		int EnumPins([Out] out IEnumPins enumPins);

		[PreserveSig]
		int FindPin([In, MarshalAs(UnmanagedType.LPWStr)] string id, [Out] out IPin pin);

		[PreserveSig]
		int QueryFilterInfo([Out] FilterInfo filterInfo);

		[PreserveSig]
		int JoinFilterGraph([In] IFilterGraph graph, [In, MarshalAs(UnmanagedType.LPWStr)] string name);

		[PreserveSig]
		int QueryVendorInfo([Out, MarshalAs(UnmanagedType.LPWStr)] out string vendorInfo);
	}

	[ComImport, Guid("56A86891-0AD4-11CE-B03A-0020AF0BA770"), InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
	internal interface IPin
	{
		[PreserveSig]
		int Connect([In] IPin receivePin, [In, MarshalAs(UnmanagedType.LPStruct)] AMMediaType mediaType);

		[PreserveSig]
		int ReceiveConnection([In] IPin receivePin, [In, MarshalAs(UnmanagedType.LPStruct)] AMMediaType mediaType);

		[PreserveSig]
		int Disconnect();

		[PreserveSig]
		int ConnectedTo([Out] out IPin pin);

		[PreserveSig]
		int ConnectionMediaType([Out, MarshalAs(UnmanagedType.LPStruct)] AMMediaType mediaType);

		[PreserveSig]
		int QueryPinInfo([Out, MarshalAs(UnmanagedType.LPStruct)] PinInfo pinInfo);

		[PreserveSig]
		int QueryDirection(out PinDirection pinDirection);

		[PreserveSig]
		int QueryId([Out, MarshalAs(UnmanagedType.LPWStr)] out string id);

		[PreserveSig]
		int QueryAccept([In, MarshalAs(UnmanagedType.LPStruct)] AMMediaType mediaType);

		[PreserveSig]
		int EnumMediaTypes(IntPtr enumerator);

		[PreserveSig]
		int QueryInternalConnections(IntPtr apPin, [In, Out] ref int nPin);

		[PreserveSig]
		int EndOfStream();

		[PreserveSig]
		int BeginFlush();

		[PreserveSig]
		int EndFlush();

		[PreserveSig]
		int NewSegment(long start, long stop, double rate);
	}

	[ComImport, Guid("56A86892-0AD4-11CE-B03A-0020AF0BA770"), InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
	internal interface IEnumPins
	{
		[PreserveSig]
		int Next([In] int cPins,
			[Out, MarshalAs(UnmanagedType.LPArray, SizeParamIndex = 0)] IPin[] pins,
			[Out] out int pinsFetched);

		[PreserveSig]
		int Skip([In] int cPins);

		[PreserveSig]
		int Reset();

		[PreserveSig]
		int Clone([Out] out IEnumPins enumPins);
	}

	[ComImport, Guid("56A8689F-0AD4-11CE-B03A-0020AF0BA770"), InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
	internal interface IFilterGraph
	{
		[PreserveSig]
		int AddFilter([In] IBaseFilter filter, [In, MarshalAs(UnmanagedType.LPWStr)] string name);

		[PreserveSig]
		int RemoveFilter([In] IBaseFilter filter);

		[PreserveSig]
		int EnumFilters([Out] out IntPtr enumerator);

		[PreserveSig]
		int FindFilterByName([In, MarshalAs(UnmanagedType.LPWStr)] string name, [Out] out IBaseFilter filter);

		[PreserveSig]
		int ConnectDirect([In] IPin pinOut, [In] IPin pinIn, [In, MarshalAs(UnmanagedType.LPStruct)] AMMediaType mediaType);

		[PreserveSig]
		int Reconnect([In] IPin pin);

		[PreserveSig]
		int Disconnect([In] IPin pin);

		[PreserveSig]
		int SetDefaultSyncSource();
	}

	[ComImport, Guid("55272A00-42CB-11CE-8135-00AA004BB851"), InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
	internal interface IPropertyBag
	{
		[PreserveSig]
		int Read(
			 [In, MarshalAs(UnmanagedType.LPWStr)] string propertyName,
			 [In, Out, MarshalAs(UnmanagedType.Struct)] ref object pVar,
			 [In] IntPtr pErrorLog);

		[PreserveSig]
		int Write(
			 [In, MarshalAs(UnmanagedType.LPWStr)] string propertyName,
			 [In, MarshalAs(UnmanagedType.Struct)] ref object pVar);
	}

	[ComImport, Guid("6B652FFF-11FE-4FCE-92AD-0266B5D7C78F"), InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
	internal interface ISampleGrabber
	{
		[PreserveSig]
		int SetOneShot([In, MarshalAs(UnmanagedType.Bool)] bool oneShot);

		[PreserveSig]
		int SetMediaType([In, MarshalAs(UnmanagedType.LPStruct)] AMMediaType mediaType);

		[PreserveSig]
		int GetConnectedMediaType([Out, MarshalAs(UnmanagedType.LPStruct)] AMMediaType mediaType);

		[PreserveSig]
		int SetBufferSamples([In, MarshalAs(UnmanagedType.Bool)] bool bufferThem);

		[PreserveSig]
		int GetCurrentBuffer(ref int bufferSize, IntPtr buffer);

		[PreserveSig]
		int GetCurrentSample(IntPtr sample);

		[PreserveSig]
		int SetCallback(ISampleGrabberCB callback, int whichMethodToCallback);
	}

	[ComImport, Guid("0579154A-2B53-4994-B0D0-E773148EFF85"), InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
	internal interface ISampleGrabberCB
	{
		[PreserveSig]
		int SampleCB(double sampleTime, IntPtr sample);

		[PreserveSig]
		int BufferCB(double sampleTime, IntPtr buffer, int bufferLen);
	}

	[ComImport, Guid("29840822-5B84-11D0-BD3B-00A0C911CE86"), InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
	internal interface ICreateDevEnum
	{
		[PreserveSig]
		int CreateClassEnumerator([In] ref Guid type, [Out] out IEnumMoniker enumMoniker, [In] int flags);
	}

	[ComImport, Guid("56A868B4-0AD4-11CE-B03A-0020AF0BA770"), InterfaceType(ComInterfaceType.InterfaceIsDual)]
	internal interface IVideoWindow
	{
		[PreserveSig]
		int put_Caption(string caption);

		[PreserveSig]
		int get_Caption([Out] out string caption);

		[PreserveSig]
		int put_WindowStyle(int windowStyle);

		[PreserveSig]
		int get_WindowStyle(out int windowStyle);

		[PreserveSig]
		int put_WindowStyleEx(int windowStyleEx);

		[PreserveSig]
		int get_WindowStyleEx(out int windowStyleEx);

		[PreserveSig]
		int put_AutoShow([In, MarshalAs(UnmanagedType.Bool)] bool autoShow);

		[PreserveSig]
		int get_AutoShow([Out, MarshalAs(UnmanagedType.Bool)] out bool autoShow);

		[PreserveSig]
		int put_WindowState(int windowState);

		[PreserveSig]
		int get_WindowState(out int windowState);

		[PreserveSig]
		int put_BackgroundPalette([In, MarshalAs(UnmanagedType.Bool)] bool backgroundPalette);

		[PreserveSig]
		int get_BackgroundPalette([Out, MarshalAs(UnmanagedType.Bool)] out bool backgroundPalette);

		[PreserveSig]
		int put_Visible([In, MarshalAs(UnmanagedType.Bool)] bool visible);

		[PreserveSig]
		int get_Visible([Out, MarshalAs(UnmanagedType.Bool)] out bool visible);

		[PreserveSig]
		int put_Left(int left);

		[PreserveSig]
		int get_Left(out int left);

		[PreserveSig]
		int put_Width(int width);

		[PreserveSig]
		int get_Width(out int width);

		[PreserveSig]
		int put_Top(int top);

		[PreserveSig]
		int get_Top(out int top);

		[PreserveSig]
		int put_Height(int height);

		[PreserveSig]
		int get_Height(out int height);

		[PreserveSig]
		int put_Owner(IntPtr owner);

		[PreserveSig]
		int get_Owner(out IntPtr owner);

		[PreserveSig]
		int put_MessageDrain(IntPtr drain);

		[PreserveSig]
		int get_MessageDrain(out IntPtr drain);

		[PreserveSig]
		int get_BorderColor(out int color);

		[PreserveSig]
		int put_BorderColor(int color);

		[PreserveSig]
		int get_FullScreenMode(
			 [Out, MarshalAs(UnmanagedType.Bool)] out bool fullScreenMode);

		[PreserveSig]
		int put_FullScreenMode([In, MarshalAs(UnmanagedType.Bool)] bool fullScreenMode);

		[PreserveSig]
		int SetWindowForeground(int focus);

		[PreserveSig]
		int NotifyOwnerMessage(IntPtr hwnd, int msg, IntPtr wParam, IntPtr lParam);

		[PreserveSig]
		int SetWindowPosition(int left, int top, int width, int height);

		[PreserveSig]
		int GetWindowPosition(out int left, out int top, out int width, out int height);

		[PreserveSig]
		int GetMinIdealImageSize(out int width, out int height);

		[PreserveSig]
		int GetMaxIdealImageSize(out int width, out int height);

		[PreserveSig]
		int GetRestorePosition(out int left, out int top, out int width, out int height);

		[PreserveSig]
		int HideCursor([In, MarshalAs(UnmanagedType.Bool)] bool hideCursor);

		[PreserveSig]
		int IsCursorHidden([Out, MarshalAs(UnmanagedType.Bool)] out bool hideCursor);
	}

	[ComImport, Guid("56A868B1-0AD4-11CE-B03A-0020AF0BA770"), InterfaceType(ComInterfaceType.InterfaceIsDual)]
	internal interface IMediaControl
	{
		[PreserveSig]
		int Run();

		[PreserveSig]
		int Pause();

		[PreserveSig]
		int Stop();

		[PreserveSig]
		int GetState(int timeout, out int filterState);

		[PreserveSig]
		int RenderFile(string fileName);

		[PreserveSig]
		int AddSourceFilter([In] string fileName, [Out, MarshalAs(UnmanagedType.IDispatch)] out object filterInfo);

		[PreserveSig]
		int get_FilterCollection(
			 [Out, MarshalAs(UnmanagedType.IDispatch)] out object collection);

		[PreserveSig]
		int get_RegFilterCollection(
			 [Out, MarshalAs(UnmanagedType.IDispatch)] out object collection);

		[PreserveSig]
		int StopWhenReady();
	}
	#endregion

	#region Structures
	[ComVisible(false)]
	internal enum PinDirection
	{
		Input,
		Output
	}

	[ComVisible(false), StructLayout(LayoutKind.Sequential)]
	internal class AMMediaType : IDisposable
	{
		public Guid MajorType;

		public Guid SubType;

		[MarshalAs(UnmanagedType.Bool)]
		public bool FixedSizeSamples = true;

		[MarshalAs(UnmanagedType.Bool)]
		public bool TemporalCompression;

		public int SampleSize = 1;

		public Guid FormatType;

		public IntPtr unkPtr;

		public int FormatSize;

		public IntPtr FormatPtr;

		~AMMediaType()
		{
			Dispose(false);
		}

		public void Dispose()
		{
			Dispose(true);
			// remove me from the Finalization queue 
			GC.SuppressFinalize(this);
		}

		protected virtual void Dispose(bool disposing)
		{
			if (FormatSize != 0)
				Marshal.FreeCoTaskMem(FormatPtr);
			if (unkPtr != IntPtr.Zero)
				Marshal.Release(unkPtr);
		}
	}

	[ComVisible(false), StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Unicode)]
	internal class PinInfo
	{
		public IBaseFilter Filter;

		public PinDirection Direction;

		[MarshalAs(UnmanagedType.ByValTStr, SizeConst = 128)]
		public string Name;
	}

	[ComVisible(false), StructLayout(LayoutKind.Sequential)]
	internal struct VideoInfoHeader
	{
		public RECT SrcRect;

		public RECT TargetRect;

		public int BitRate;

		public int BitErrorRate;

		public long AverageTimePerFrame;

		public BitmapInfoHeader BmiHeader;
	}

	[ComVisible(false), StructLayout(LayoutKind.Sequential, Pack = 2)]
	internal struct BitmapInfoHeader
	{
		public int Size;

		public int Width;

		public int Height;

		public short Planes;

		public short BitCount;

		public int Compression;

		public int ImageSize;

		public int XPelsPerMeter;

		public int YPelsPerMeter;

		public int ColorsUsed;

		public int ColorsImportant;
	}

	[ComVisible(false), StructLayout(LayoutKind.Sequential)]
	internal struct RECT
	{
		public int Left;

		public int Top;

		public int Right;

		public int Bottom;
	}
	#endregion

}
