﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Threading;
using System.Xml.Linq;

using Esri.ArcGISRuntime.ArcGISServices;
using Esri.ArcGISRuntime.Data;
using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Mapping;
using Esri.ArcGISRuntime.Security;
using Esri.ArcGISRuntime.Tasks;
using Esri.ArcGISRuntime.Tasks.Offline;
using Esri.ArcGISRuntime.UI.Controls;

using Extensibility;
using NavAddin;

namespace EditAddin
{

	// EditHelper handles the creation of feature layers, either directly from feature services
	// or locally via a replica geodatabase, and also performs synchronization

	public class NotifyLayersLoadedArgs : EventArgs
	{
		public Dictionary<string, ArcGISFeatureTable> FeatureTables;
	}

	public class EditHelper
	{

		private class EditLayerVisibility
		{
			public bool Visible = false;
			public bool ShowLabels = false;
		}

		// For development purposes only - change to false for testing/production
		private static readonly bool _bOverrideDebugCheck = false; 
	
		private MapView _mv;
		private Dispatcher _dispatcher;

		private string _sRemoteUrl;
		private bool _bPortal = false;
		private string _sPortalUrl;
		private bool _bAutoUserName = false;
		private bool _bLocal = false;
		private string _sLocalGDB;
		private string _sGroupLayerName = "Edit Layers";
		private bool _bGroupLayerVisible = false;
		private List<string> _FeatureLayerNames;
		private Dictionary<string, GenerateLayerOption> _GenLayerOptions;
		private Dictionary<string, EditLayerVisibility> _LayerVisibilities;
		private Envelope _envAOI;

		private bool _bBusy = false;
		private GeodatabaseSyncTask _gdbSyncTask;
		private GenerateGeodatabaseJob _gdbJob;
		private SyncGeodatabaseJob _syncJob;

		// Until group layers are supported, maintain a list of layers added to map
		bool _bLayersLoaded = false;
		private List<FeatureLayer> _EditLayers;
		private List<FeatureTable> _EditTables;
		private GroupLayer _EditLayer;
		private TocDataItem _EditTOC;
		private Dictionary<string, ArcGISFeatureTable> _FeatureTables;
		private IStatus _pStatus;
		private Geodatabase _GDB;
		private bool _bAuthorized = false;

		// Public events

		public event EventHandler<NotifyLayersLoadedArgs> EditLayersLoaded;

		// Public constructor

		public EditHelper(MapView mv)
		{
			_mv = mv;
			_dispatcher = mv.Dispatcher;
		}

		// Public methods

		public static DateTime LocalDateTime(object oVal)
		{
			DateTime dtLocal = DateTime.MinValue;
			if (oVal is DateTime)
				dtLocal = ((DateTime)oVal).ToLocalTime();
			else if (oVal is DateTimeOffset)
				dtLocal = ((DateTimeOffset)oVal).ToLocalTime().DateTime;
			return dtLocal;
		}

		public void Initialize(string sConfigData)
		{
			_pStatus = MapApplication.Current;
			MapModelExt.Current.MapLoaded += Current_MapLoaded;
			this.LoadConfiguration(sConfigData);
			if (AppViewExt.Current.Map != null)
				this.CheckLayersLoaded();
		}

		public async Task SynchronizeAsync(IStatus pStatus = null)
		{

			if (!_bLocal)
				return;
			if (_bBusy)
				return;

			// Check if initialized

			if (pStatus != null)
				_pStatus = pStatus;
			if (!_bLayersLoaded)
			{
				await this.LoadFeatureLayersAsync();
				if (!_bLayersLoaded)
					return;
			}

			// Authorize if needed

			if (_bPortal)
			{
				bool bResult = await this.ConnectToPortal();
				if (!bResult)
					return;
			}

			string sErr = "";
			try
			{

				// Create a new GeodatabaseSyncTask with the uri of the feature server to pull from

				sErr = "create GeodatabaseSyncTask: \n" + _sRemoteUrl;
				Uri uri = new Uri(_sRemoteUrl);
				GeodatabaseSyncTask gdbSyncTask = await this.GetGdbTask();

				// Create parameters

				sErr = "create sync parameters";
				SyncGeodatabaseParameters gdbParams = await gdbSyncTask.CreateDefaultSyncGeodatabaseParametersAsync(_GDB);

				// Execute task

				sErr = "execute task";
				_pStatus.Status("Synchronizing...", true);
				_syncJob = gdbSyncTask.SyncGeodatabase(gdbParams, _GDB);
				_syncJob.JobChanged += Job_Updated;
				_syncJob.ProgressChanged += Job_Updated;
				_bBusy = true;
				_syncJob.Start();

			}
			catch (Exception ex)
			{
				string sMsg = "Synchronize: Unable to " + sErr + "\n" + ex.Message;
				_pStatus.Error(sMsg);
				_bBusy = false;
			}

		}

		public void DeleteGDB()
		{

			// Check if in debug session

			if (!_bOverrideDebugCheck)
			{
				if (Debugger.IsAttached)
				{
					Debug.WriteLine("EditHelper: skipping debug session");
					return;
				}
			}

			// Check for GDB

			_pStatus = MapApplication.Current;
			if (!(File.Exists(_sLocalGDB)))
			{
				_pStatus.Message("Field data not found.");
				return;
			}
			string sMsg = "Any unsynchronized edits will be lost.\nDo you want to continue?";
			MessageBoxResult mbResult = MessageBox.Show(sMsg, "Delete Field Data", MessageBoxButton.YesNo, MessageBoxImage.Warning);
			if (mbResult != MessageBoxResult.Yes)
				return;

			// Start building delete script

			string sSep = Path.DirectorySeparatorChar.ToString();
			string sEscSep = sSep + sSep;
			string sTempPath = Path.GetTempPath();
			string sDeleteScript = Path.Combine(sTempPath, "delete.js");
			StringBuilder sb = new StringBuilder();
			sb.Append("var ws = WScript.CreateObject(\"WScript.Shell\");\r\n");

			// Kill any instances of the application
			// [This is harsh, but it ensures any application instances are completely closed before deleting GDB]
			// [NOTE: This will not work in a Visual Studio session because it runs under a vshost]

			Assembly a = Assembly.GetEntryAssembly();
			string sLocalPath = Path.GetDirectoryName(a.Location);
			string sLocalExe = Path.GetFileName(a.Location);
			sb.Append("ws.Run(\"taskkill /f /im " + sLocalExe + "\");\r\n");
			sb.Append("WScript.Echo(\"Click OK to continue.\");\r\n");

			// Delete GDB

			string sLocalFileNameJ = _sLocalGDB.Replace(sSep, sEscSep);
			sb.Append("var fso = new ActiveXObject(\"Scripting.FileSystemObject\");\r\n");
			sb.Append("fso.DeleteFile(\"" + sLocalFileNameJ + "\");\r\n");

			// Finish and execute script

			string sConfig = ConfigHelper.ConfigName;
			string sCommand = a.Location + " " + sConfig;
			string sCommandJ = sCommand.Replace(sSep, sEscSep);
			sb.Append("ws.Run(\"" + sCommandJ + "\");\r\n");
			string sScript = sb.ToString();
			File.WriteAllText(sDeleteScript, sScript);
			Process.Start(sDeleteScript);
			MapApplication.Current.Exit();

		}

		// Event handlers

		private void Current_MapLoaded(object sender, EventArgs e)
		{
			this.CheckLayersLoaded();
		}

		private void Job_Updated(object sender, EventArgs e)
		{
			_dispatcher.Invoke(() => { this.UpdateJobStatus(); });
		}

		// Private methods

		private async Task<GeodatabaseSyncTask> GetGdbTask()
		{
			if (_gdbSyncTask != null)
				return _gdbSyncTask;
			Uri uri = new Uri(_sRemoteUrl);
			GeodatabaseSyncTask syncTask = await GeodatabaseSyncTask.CreateAsync(uri);
			_gdbSyncTask = syncTask;
			return _gdbSyncTask;
		}

		private void LoadConfiguration(string sConfigData)
		{
			if (string.IsNullOrEmpty(sConfigData))
				return;
			XDocument xDoc = XDocument.Parse(sConfigData);
			XElement xRoot = xDoc.Root;
			foreach (XElement xElement in xRoot.Elements())
			{
				string sName = xElement.Name.LocalName;
				switch (sName)
				{
					case "RemoteFeatureService":
						_sRemoteUrl = xElement.Attribute("Url").Value;
						break;
					case "PortalAuth":
						_bPortal = true;
						_sPortalUrl = xElement.Attribute("Url").Value;
						if (xElement.Attribute("AutoUserName") != null)
							_bAutoUserName = (xElement.Attribute("AutoUserName").Value.ToLower() == "true");
						break;
					case "LocalGeodatabase":
						_bLocal = true;
						_sLocalGDB = xElement.Attribute("Path").Value;
						break;
					case "EditLayers":

						// Assume that layers are listed in the order they are to appear in the TOC (top to bottom)

						if (xElement.Attribute("GroupLayerName") != null)
							_sGroupLayerName = xElement.Attribute("GroupLayerName").Value;
						if (xElement.Attribute("Visible") != null)
							_bGroupLayerVisible = (xElement.Attribute("Visible").Value.ToLower() == "true");
						_FeatureLayerNames = new List<string>();
						_GenLayerOptions = new Dictionary<string, GenerateLayerOption>();
						_LayerVisibilities = new Dictionary<string, EditLayerVisibility>();
						foreach (XElement xEditLayer in xElement.Elements())
						{
							string sLayerName = xEditLayer.Attribute("Name").Value;
							_FeatureLayerNames.Add(sLayerName);

							// Default layer options are:
							//		IncludeRelated = false
							//		QueryOption = UseFilter
							//		UseGeometry = true
							//		WhereClause = ""

							GenerateLayerOption glOpt = new GenerateLayerOption() { LayerId = -1 };
							bool bRelated = false;
							if (xEditLayer.Attribute("IncludeRelated") != null)
								bRelated = (xEditLayer.Attribute("IncludeRelated").Value.ToLower() == "true");
							glOpt.IncludeRelated = bRelated;
							bool bGeom = true;
							if (xEditLayer.Attribute("UseGeometry") != null)
								bGeom = (xEditLayer.Attribute("UseGeometry").Value.ToLower() == "true");
							glOpt.UseGeometry = bGeom;
							GenerateLayerQueryOption glqOpt = GenerateLayerQueryOption.UseFilter;
							string sWhereClause = "";
							XAttribute xQueryOption = xEditLayer.Attribute("QueryOption");
							if (xQueryOption != null)
							{
								switch (xQueryOption.Value)
								{
									case "None":
										glqOpt = GenerateLayerQueryOption.None;
										break;
									case "UseFilter":
										glqOpt = GenerateLayerQueryOption.UseFilter;
										sWhereClause = xEditLayer.Attribute("Where").Value;
										break;
									case "All":
										glqOpt = GenerateLayerQueryOption.All;
										break;
								}
							}
							glOpt.QueryOption = glqOpt;
							glOpt.WhereClause = sWhereClause;
							_GenLayerOptions[sLayerName] = glOpt;
							EditLayerVisibility vis = new EditLayerVisibility();
							bool bVis = false;
							if (xEditLayer.Attribute("Visible") != null)
								bVis = (xEditLayer.Attribute("Visible").Value.ToLower() == "true");
							vis.Visible = bVis;
							bool bLabels = false;
							if (xEditLayer.Attribute("ShowLabels") != null)
								bLabels = (xEditLayer.Attribute("ShowLabels").Value.ToLower() == "true");
							vis.ShowLabels = bLabels;
							_LayerVisibilities[sLayerName] = vis;
						}
						break;
					case "AOI":

						// Relica area of interest is given in DD (WGS 84)

						bool bSuccess = true;
						double dMinX = 0;
						double dMinY = 0;
						double dMaxX = 0;
						double dMaxY = 0;
						bSuccess = bSuccess && double.TryParse(xElement.Attribute("XMin").Value, out dMinX);
						bSuccess = bSuccess && double.TryParse(xElement.Attribute("YMin").Value, out dMinY);
						bSuccess = bSuccess && double.TryParse(xElement.Attribute("XMax").Value, out dMaxX);
						bSuccess = bSuccess && double.TryParse(xElement.Attribute("YMax").Value, out dMaxY);
						if (bSuccess)
							_envAOI = new Envelope(dMinX, dMinY, dMaxX, dMaxY, new SpatialReference(4326));
						break;

				} // switch sName

			} // foreach xElement

		}

		private async void CheckLayersLoaded()
		{

			// Check if layers have been created

			if (!_bLayersLoaded)
			{
				await this.LoadFeatureLayersAsync();
				if (!_bLayersLoaded)
					return;
			}

			// Check if layers are present in map

			this.CheckMap();

		}

		private async Task LoadFeatureLayersAsync()
		{

			// Get geodatabase

			if (_bLocal && (_GDB == null))
			{
				bool bResult = true;
				if (File.Exists(_sLocalGDB))
					bResult = await this.OpenGeodatabase();
				else
				{
					string sMsg = "Local field data not loaded.  Would you like to load it?";
					MessageBoxResult mbResult = MessageBox.Show(sMsg, "Field Edits", MessageBoxButton.YesNo);
					if (mbResult != MessageBoxResult.Yes)
					{
						_pStatus.Message("When desired, click \"Synchronize\" to load field data.");
						return;
					}
					await this.LoadGeodatabaseAsync();
					return;
				}
				if (!bResult)
					return;
			}

			// Create feature layers

			_pStatus.Status("Adding feature layers...", true);
			_EditLayers = new List<FeatureLayer>();
			_EditTables = new List<FeatureTable>();
			_FeatureTables = new Dictionary<string, ArcGISFeatureTable>();
			if (!_bLocal)
				await this.AddLayersFromService();
			else
				await this.AddLayersFromGeodatabase();
			if (!_bLayersLoaded)
				return;
			this.CheckMap();
			_pStatus.Status("OK", false);

		}

		private async void CheckMap()
		{

			// Check if layers and tables are present in map

			// BEGIN WORKAROUND:  RUNTIME 100.5/100.6 BUG
			bool bUseGroupLayers = MapHelper.PreserveGroupLayers;
			// END WORKAROUND

			bool bEditLayersLoaded = false;
			Map map = AppViewExt.Current.Map;
			if (bUseGroupLayers)
			{
				if (_EditLayer == null)
				{
					_EditLayer = new GroupLayer()
					{
						Name = _sGroupLayerName,
						Id = Guid.NewGuid().ToString(),
						IsVisible = _bGroupLayerVisible
					};
					foreach (FeatureLayer fLyr in _EditLayers)
						_EditLayer.Layers.Add(fLyr);
				}
				if (!map.OperationalLayers.Contains(_EditLayer))
				{
					map.OperationalLayers.Add(_EditLayer);
					await _EditLayer.LoadAsync();
					bEditLayersLoaded = true;
				}
			}
			else
			{
				foreach (FeatureLayer fLyr in _EditLayers)
				{
					if (!map.OperationalLayers.Contains(fLyr))
					{
						map.OperationalLayers.Add(fLyr);
						bEditLayersLoaded = true;
					}
				}
			}
			foreach (FeatureTable fTab in _EditTables)
			{
				if (!map.Tables.Contains(fTab))
					map.Tables.Add(fTab);
			}

			// Check if TOC item is present

			if (_EditTOC == null)
			{
				_EditTOC = new TocDataItem(!bUseGroupLayers);
				if (bUseGroupLayers)
					await _EditTOC.SetLayer(_EditLayer);
				else
				{
					_EditTOC.DisplayName = _sGroupLayerName;
					List<TocDataItem> Subitems = new List<TocDataItem>();
					foreach (FeatureLayer fLyr in _EditLayers)
					{
						TocDataItem item = new TocDataItem(true);
						await item.SetLayer(fLyr);
						Subitems.Add(item);
					}
					Subitems.Reverse();
					_EditTOC.SetSubitems(Subitems);
				}
				_EditTOC.IsChecked = _bGroupLayerVisible;
				_EditTOC.IsExpanded = true;
				MapModelExt.Current.OperationalLayers.Add(_EditTOC);
			}
			EditViewExt.Current.LocalGDB = _bLocal;

			// Fire layers loaded event

			if (!bEditLayersLoaded)
				return;
			if (this.EditLayersLoaded == null)
				return;
			NotifyLayersLoadedArgs e = new NotifyLayersLoadedArgs() { FeatureTables = _FeatureTables };
			this.EditLayersLoaded(this, e);

			// RUNTIME 100.6 QUIRK: Set edit layer label visibility after loading map

			foreach (FeatureLayer fLyr in _EditLayers)
			{
				EditLayerVisibility vis = _LayerVisibilities[fLyr.Name];
				fLyr.LabelsEnabled = vis.ShowLabels;
			}

		}

		private async Task AddLayersFromService()
		{

			// Authorize if needed

			if (_bPortal)
			{
				bool bResult = await this.ConnectToPortal();
				if (!bResult)
					return;
			}

			string sErr = "";
			try
			{

				// Get service info

				sErr = "get service info: \n" + _sRemoteUrl;
				Uri uri = new Uri(_sRemoteUrl);
				GeodatabaseSyncTask gdbSyncTask = await this.GetGdbTask();
				ArcGISFeatureServiceInfo fsInfo = gdbSyncTask.ServiceInfo;
				Dictionary<string, long> SubIds = new Dictionary<string, long>();
				foreach (IdInfo info in fsInfo.LayerInfos)
					SubIds[info.Name] = info.Id;
				foreach (IdInfo info in fsInfo.TableInfos)
					SubIds[info.Name] = info.Id;

				// Build tables and layers

				sErr = "build table or layer";
				int iCount = _FeatureLayerNames.Count;
				for (int i = iCount - 1; i > -1; i--)
				{

					string sName = _FeatureLayerNames[i];
					if (!SubIds.ContainsKey(sName))
					{
						Debug.Print("EditHelper service layer not found: " + sName);
						continue;
					}
					long iLayerID = SubIds[sName];
					string sID = "EDIT_" + Guid.NewGuid().ToString();

					// Create, initialize, and add layer to group

					string sUrl = _sRemoteUrl + "/" + iLayerID.ToString();
					ServiceFeatureTable sfTab = new ServiceFeatureTable(new Uri(sUrl));
					await sfTab.LoadAsync();
					bool bTable = (sfTab.LayerInfo.GeometryType == GeometryType.Unknown);
					_FeatureTables[sID] = sfTab;
					if (bTable)
					{
						_EditTables.Add(sfTab);
						continue;
					}
					EditLayerVisibility vis = _LayerVisibilities[sName];

					// RUNTIME 100.6 QUIRK: Set edit layer label visibility after loading map

					FeatureLayer fLyr = new FeatureLayer(sfTab)
					{
						Name = sName,
						Id = sID,
						IsVisible = vis.Visible,
						LabelsEnabled = false
					};
					_EditLayers.Add(fLyr);

				}
				_bLayersLoaded = true;

			}
			catch (Exception ex)
			{
				string sMsg = "Load service layers: Unable to " + sErr + "\n" + ex.Message;
				_pStatus.Error(sMsg);
			}

		}

		private async Task AddLayersFromGeodatabase()
		{

			// Get gdb feature tables

			Dictionary<string, GeodatabaseFeatureTable> Tables = new Dictionary<string, GeodatabaseFeatureTable>();
			try
			{
				foreach (GeodatabaseFeatureTable gfTab in _GDB.GeodatabaseFeatureTables)
				{
					await gfTab.LoadAsync();
					string sName = gfTab.LayerInfo.ServiceLayerName;
					Tables[sName] = gfTab;
				}
			}
			catch
			{
				string sMsg = "Error loading layers from geodatabase: may be corrupt.";
				_pStatus.Error(sMsg);
				return;
			}

			// Build tables and layers

			int iCount = _FeatureLayerNames.Count;
			QueryParameters qp = new QueryParameters() { WhereClause = "1 = 1" };
			for (int i = iCount - 1; i > -1; i--)
			{

				string sName = _FeatureLayerNames[i];
				if (!Tables.ContainsKey(sName))
				{
					Debug.Print("EditHelper gdb layer not found: " + sName);
					continue;
				}
				string sID = "EDIT_" + Guid.NewGuid().ToString();
				GeodatabaseFeatureTable gfTab = Tables[sName];
				bool bTable = (gfTab.LayerInfo.GeometryType == GeometryType.Unknown);
				long iRecCount = await gfTab.QueryFeatureCountAsync(qp);
				if (bTable)
					System.Diagnostics.Debug.Print("Table " + gfTab.TableName + ": " + iRecCount.ToString());
				else
					System.Diagnostics.Debug.Print("Layer " + gfTab.TableName + ": " + iRecCount.ToString());
				_FeatureTables[sID] = gfTab;
				if (bTable)
				{
					// Add table only
					_EditTables.Add(gfTab);
					continue;
				}
				DrawingInfo drInfo = gfTab.LayerInfo.DrawingInfo;
				IReadOnlyList<LabelingInfo> LabelInfos = drInfo.LabelingInfo;
				EditLayerVisibility vis = _LayerVisibilities[sName];

				// RUNTIME 100.6 QUIRK: Set edit layer label visibility after loading map

				FeatureLayer fLyr = new FeatureLayer(gfTab)
				{
					Name = sName,
					Id = sID,
					IsVisible = vis.Visible,
					LabelsEnabled = false
				};
				await fLyr.LoadAsync();
				_EditLayers.Add(fLyr);

			}
			_bLayersLoaded = true;

		}

		private async Task<bool> OpenGeodatabase()
		{
			if (!(File.Exists(_sLocalGDB)))
				return false;
			bool bResult = true;
			try
			{
				_pStatus.Status("Opening field data...", true);
				_GDB = await Geodatabase.OpenAsync(_sLocalGDB);
			}
			catch (Exception ex)
			{
				_pStatus.Error("Could not open field data:\n" + ex.Message);
			}
			return bResult;
		}

		private async Task LoadGeodatabaseAsync()
		{

			// Authorize if needed

			if (_bPortal)
			{
				bool bResult = await this.ConnectToPortal();
				if (!bResult)
					return;
			}

			string sErr = "";
			try
			{

				// Get service info

				sErr = "create GeodatabaseSyncTask: \n" + _sRemoteUrl;
				Uri uri = new Uri(_sRemoteUrl);
				GeodatabaseSyncTask gdbSyncTask = await this.GetGdbTask();
				ArcGISFeatureServiceInfo fsInfo = gdbSyncTask.ServiceInfo;
				Dictionary<string, long> SubIds = new Dictionary<string, long>();
				foreach (IdInfo info in fsInfo.LayerInfos)
					SubIds[info.Name] = info.Id;
				foreach (IdInfo info in fsInfo.TableInfos)
					SubIds[info.Name] = info.Id;

				// Set gdb parameters 

				sErr = "set gdb parameters";
				GenerateGeodatabaseParameters param = await gdbSyncTask.CreateDefaultGenerateGeodatabaseParametersAsync(_envAOI);
				SpatialReference sr = _mv.SpatialReference;
				param.OutSpatialReference = sr;
				if (!fsInfo.SyncEnabled)
				{
					_pStatus.Error("Feature service not sync enabled:\n" + _sRemoteUrl);
					return;
				}
				if (fsInfo.SyncCapabilities.SupportsSyncModelLayer)
				{

					// Set layer options

					param.LayerOptions.Clear();
					foreach (string sName in _FeatureLayerNames)
					{
						if (!SubIds.ContainsKey(sName))
						{
							Debug.Print("EditHelper service layer not found: " + sName);
							continue;
						}
						long id = SubIds[sName];
						GenerateLayerOption glOpt = _GenLayerOptions[sName];
						if (glOpt.LayerId < 0)
							glOpt.LayerId = id;
						param.LayerOptions.Add(glOpt);
					}
					if (param.LayerOptions.Count == 0)
						throw new Exception("No matching layers");

				}

				// Execute task

				sErr = "execute task";
				_pStatus.Status("Calling feature service...", true);
				_gdbJob = gdbSyncTask.GenerateGeodatabase(param, _sLocalGDB);
				_gdbJob.JobChanged += Job_Updated;
				_gdbJob.ProgressChanged += Job_Updated;
				_bBusy = true;
				_gdbJob.Start();

			}
			catch (Exception ex)
			{
				string sMsg = "Load GDB: Unable to " + sErr + "\n" + ex.Message;
				_pStatus.Error(sMsg);
			}

		}

		private async void GdbComplete()
		{

			// Open gdb and report results

			bool bResult = await this.OpenGeodatabase();
			if (!bResult)
			{
				_pStatus.Error("Could not open geodatabase.");
				return;
			}
			StringBuilder sb = new StringBuilder();
			sb.Append("Field data loaded:");
			foreach (GeodatabaseFeatureTable gfTab in _GDB.GeodatabaseFeatureTables)
			{
				await gfTab.LoadAsync();
				sb.Append("\n" + gfTab.TableName + ": " + gfTab.NumberOfFeatures.ToString() + " rows");
			}
			_pStatus.Message(sb.ToString());

			// Finally, load layers

			await this.LoadFeatureLayersAsync();

		}

		private async Task<bool> ConnectToPortal()
		{
			if (_bAuthorized)
				return true;
			try
			{

				// Open authentication dialog

				string sUser;
				string sPassword;
				AuthenticationDialogModel authModel = new AuthenticationDialogModel();
				if (_bAutoUserName)
					authModel.UserNameVisibility = Visibility.Collapsed;
				bool? bResult = MapApplication.Current.CreateChildWindow("EditAddin", "AuthenticationDialog", authModel);
				if (bResult != true)
					return false;
				if (_bAutoUserName)
					sUser = System.Environment.GetEnvironmentVariable("USERNAME");
				else
					sUser = authModel.UserName;
				sPassword = authModel.Password;

				// Try authenticating

				Uri uri = new Uri(_sPortalUrl);
				TokenCredential cred = await AuthenticationManager.Current.GenerateCredentialAsync(uri, sUser, sPassword);
				AuthenticationManager.Current.AddCredential(cred);
				_bAuthorized = true;
				return true;

			}
			catch (Exception ex)
			{
				_pStatus.Error("Portal authorization failed:\n" + ex.Message);
				return false;
			}
		}

		private void UpdateJobStatus()
		{

			// Get job status and progress

			bool bGDB = false;
			string sMsg = "";
			JobStatus jobStatus = JobStatus.NotStarted;
			int iProgress = 0;
			if (_gdbJob != null)
			{
				bGDB = true;
				sMsg = "Creating replica...";
				jobStatus = _gdbJob.Status;
				iProgress = _gdbJob.Progress;
			}
			else if (_syncJob != null)
			{
				sMsg = "Synchronizing...";
				jobStatus = _syncJob.Status;
				iProgress = _syncJob.Progress;
			}
			string sStatus = "";
			_bBusy = true;
			switch (jobStatus)
			{
				case JobStatus.NotStarted:
					sStatus = "Starting";
					break;
				case JobStatus.Started:
					sStatus = "Working (" + iProgress.ToString() + "%)";
					break;
				case JobStatus.Paused:
					sStatus = "Paused";
					break;
				case JobStatus.Succeeded:
				case JobStatus.Failed:
					_bBusy = false;
					break;
			}
			if (_bBusy)
				sMsg += sStatus;
			else
				sMsg = "OK";
			_pStatus.Status(sMsg, _bBusy);
			if (_bBusy)
				return;

			// Clean up and report completion

			if (bGDB)
			{
				_gdbJob.JobChanged -= Job_Updated;
				_gdbJob.ProgressChanged -= Job_Updated;
				_gdbJob = null;
				_pStatus = MapApplication.Current;
				this.GdbComplete();
				return;
			}
			_syncJob.JobChanged -= Job_Updated;
			_syncJob.ProgressChanged -= Job_Updated;
			_syncJob = null;
			if (jobStatus == JobStatus.Succeeded)
				_pStatus.Message("Synchronization succeeded.");
			else if (jobStatus == JobStatus.Failed)
				_pStatus.Message("Synchronization failed.");
			_pStatus = MapApplication.Current;

		}

	}

}
