﻿using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;

using Esri.ArcGISRuntime.Data;

using Extensibility;
using QueryAddin;

namespace EditAddin
{

	// EditAttachmentDialogModel adds the ability to add, capture, and remove attachments

	public class EditAttachmentDialogModel : IAttachmentDialogModel, ICommand, IDialogContent
	{

		private IdentifyResult _idResult;
		private int _iSelectedIndex;
		private ArcGISFeatureTable _afTab;
		private ArcGISFeature _feat;
		private bool _bEditable = false;

		// Bindable properties

		public ObservableCollection<AttachmentDataItem> AttachmentItems { get; private set; }

		public int SelectedIndex
		{
			get { return _iSelectedIndex; }
			set
			{
				_iSelectedIndex = value;
				if (this.CanExecuteChanged == null)
					return;
				this.CanExecuteChanged(this, new EventArgs());
			}
		}

		public ICommand Command { get { return this; } }

		public Visibility EditVisibility 
		{ 
			get 
			{
				if (_bEditable)
					return Visibility.Visible;
				else
					return Visibility.Collapsed; 
			} 
		}

		// IAttachmentDialogModel

		public void Initialize(IdentifyResult idResult)
		{
			_idResult = idResult;
			string sLayerID = idResult.LayerID;
			if (EditModelExt.Current.FeatureTables.ContainsKey(sLayerID))
			{
				_afTab = EditModelExt.Current.FeatureTables[sLayerID];
				_bEditable = (_afTab != null);
				_feat = idResult.Feature as ArcGISFeature;
			}
			AttachmentItems = new ObservableCollection<AttachmentDataItem>();
			foreach (Attachment item in idResult.Attachments)
				AttachmentItems.Add(new AttachmentDataItem(item));
			//SelectedIndex = -1;
		}

		// ICommand

		public bool CanExecute(object parameter)
		{
			string sParam = parameter as string;
			switch (sParam)
			{
				case "Open":
				case "SaveAs":
					return (SelectedIndex > -1);
				case "SaveAll":
				case "OK":
				case "Cancel":
					return true;
				case "Add":
				case "Capture":
					return _bEditable;
				case "Remove":
					return (_bEditable && (SelectedIndex > -1));
			}
			return false;
		}

		public event EventHandler CanExecuteChanged;

		public void Execute(object parameter)
		{
			AttachmentDataItem item;
			string sParam = parameter as string;
			switch (sParam)
			{
				case "Open":
					item = AttachmentItems[SelectedIndex];
					item.Open();
					return;
				case "SaveAs":
					item = AttachmentItems[SelectedIndex];
					item.Save();
					MapApplication.Current.Message("Attachment saved.");
					return;
				case "SaveAll":
					string sFolder = DialogHelper.GetFolder();
					if (string.IsNullOrEmpty(sFolder))
						return;
					foreach (AttachmentDataItem it in AttachmentItems)
						it.SaveToFolder(sFolder);
					MapApplication.Current.Message("Attachment(s) saved.");
					return;
				case "Add":
					this.AddAsync(false);
					return;
				case "Capture":
					this.AddAsync(true);
					return;
				case "Remove":
					this.RemoveAsync();
					return;
				case "OK":
					if (this.DialogOK == null)
						return;
					this.DialogOK(this, new EventArgs());
					return;
				case "Cancel":
					if (this.DialogCancel == null)
						return;
					this.DialogCancel(this, new EventArgs());
					return;
			}
		}

		// IDialogContent

		public string Title
		{
			get { return "Attachments"; }
		}

		public bool Closable
		{
			get { return true; }
		}

		public event EventHandler<EventArgs> DialogOK;

		public event EventHandler<EventArgs> DialogCancel;

		// Private methods

		private async void AddAsync(bool bCapture = false)
		{

			// Get attachment

			if (!_bEditable)
				return;
			bool bError = false;
			string sErr = "";
			string sSnapFile = "";
			string sName = "";
			string sType = "";
			byte[] data = null;
			if (bCapture)
			{

				// Capture image

				CaptureDialogModel model = new CaptureDialogModel();
				bool? bResult = MapApplication.Current.CreateChildWindow("EditAddin", "CaptureDialog", model);
				if (bResult == null)
					return;
				if (!bResult.Value)
					return;
				Snapshot snap = model.Snapshot;
				if (snap == null)
					return;
				sName = snap.Name;
				sName = sName + ".jpg";
				string sExt = ".jpg";
				sType = this.GetContentType(sExt);
				sSnapFile = snap.FilePath;
				using (Stream str = new FileStream(sSnapFile, FileMode.Open, FileAccess.Read))
				{
					data = this.ReadStream(str);
				}

			}
			else
			{
				OpenFileDialog dlg = new OpenFileDialog();
				if (dlg.ShowDialog() != true)
					return;
				sName = Path.GetFileName(dlg.FileName);
				string sExt = Path.GetExtension(dlg.FileName).ToLower();
				sType = this.GetContentType(sExt);
				using (Stream str = dlg.OpenFile())
				{
					data = this.ReadStream(str);
				}
			}

			// Add attachment

			Attachment atResult = null;
			try
			{
				atResult = await _feat.AddAttachmentAsync(sName, sType, data);
				if (atResult == null)
				{
					bError = true;
					sErr = "Error adding attachment.";
				}
				else if (_afTab is ServiceFeatureTable)
				{
					ServiceFeatureTable sfTab = (ServiceFeatureTable)_afTab;
					IReadOnlyList<EditResult> Results = await sfTab.ApplyEditsAsync();
					EditResult edResult = Results[0];
					if (edResult.CompletedWithErrors)
					{
						bError = true;
						sErr = "Error adding attachment:\n" + edResult.Error.Message;
					}
				}
			}
			catch (Exception ex)
			{
				bError = true;
				sErr = "Error adding attachment:\n" + ex.Message;
			}
			if (bCapture)
				File.Delete(sSnapFile);
			if (bError)
			{
				MapApplication.Current.Error(sErr);
				return;
			}
			this.UpdateAttachments();

		}

		private async void RemoveAsync()
		{

			// Remove selected attachment

			if (!_bEditable)
				return;
			if (SelectedIndex < 0)
				return;
			List<Attachment> Attachments = new List<Attachment>();
			Attachments.Add(AttachmentItems[SelectedIndex].Item);
			bool bError = false;
			string sErr = "";
			try
			{
				await _feat.DeleteAttachmentsAsync(Attachments);
				if (_afTab is ServiceFeatureTable)
				{
					ServiceFeatureTable sfTab = (ServiceFeatureTable)_afTab;
					IReadOnlyList<EditResult> Results = await sfTab.ApplyEditsAsync();
					EditResult edResult = Results[0];
					if (edResult.CompletedWithErrors)
					{
						bError = true;
						sErr = "Error deleting attachment:\n" + edResult.Error.Message;
					}
				}
			}
			catch (Exception ex)
			{
				bError = true;
				sErr = "Error deleting attachment:\n" + ex.Message;
			}
			if (bError)
			{
				MapApplication.Current.Error(sErr);
				return;
			}
			this.UpdateAttachments();

		}

		private async void UpdateAttachments()
		{

			// Query attachments

			if (_feat == null)
				return;
			IReadOnlyList<Attachment> Attachments = null;
			try
			{
				Attachments = await _feat.GetAttachmentsAsync();
			}
			catch (Exception ex)
			{
				System.Diagnostics.Debug.Print(ex.Message);
				return;
			}

			// Update lists

			_idResult.Attachments.Clear();
			AttachmentItems.Clear();
			foreach (Attachment item in Attachments)
			{
				_idResult.Attachments.Add(item);
				AttachmentItems.Add(new AttachmentDataItem(item));
			}

		}

		private string GetContentType(string sExt)
		{

			// Search registry for file type

			RegistryKey rkRoot = Registry.ClassesRoot;
			RegistryKey rkFileType = null;
			foreach (string sKey in rkRoot.GetSubKeyNames())
			{
				if (sKey == null)
					continue;
				if (sKey.ToLower() != sExt)
					continue;
				rkFileType = rkRoot.OpenSubKey(sKey);
				break;
			}
			if (rkFileType == null)
				return null;
			object oVal = rkFileType.GetValue("Content Type");
			if (oVal == null)
				return null;
			return oVal.ToString();

		}

		private byte[] ReadStream(Stream str)
		{
			byte[] data = null;
			using (MemoryStream ms = new MemoryStream())
			{
				str.CopyTo(ms);
				data = ms.ToArray();
			}
			return data;
		}

	}

}
