﻿using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Runtime.CompilerServices;
using System.Windows.Input;

using Extensibility;

namespace EditAddin
{

	// SynchronizeDialogModel manages field data synchronization

	public class SynchronizeDialogModel : INotifyPropertyChanged, ICommand, IDialogContent, IStatus
	{

		private bool _bBusy = false;
		private string _sStatus = "OK";

		// Bindable properties

		public string Status
		{
			get { return _sStatus; }
			set
			{
				Debug.Print(value);
				_sStatus = value;
				this.RaisePropertyChanged();
			}
		}

		public ICommand Command { get { return this; } }

		// INotifyPropertyChanged

		public event PropertyChangedEventHandler PropertyChanged;

		// ICommand

		public bool CanExecute(object parameter)
		{
			string sParam = parameter as string;
			if (string.IsNullOrEmpty(sParam))
				return false;
			switch (sParam)
			{
				case "OK":
				case "Quit":
					return !_bBusy;
			}
			return false;
		}

		public event EventHandler CanExecuteChanged;

		public void Execute(object parameter)
		{
			string sParam = parameter as string;
			if (string.IsNullOrEmpty(sParam))
				return;
			switch (sParam)
			{
				case "OK":
					if (EditViewExt.Current == null)
						return;
					_bBusy = true;
					this.RaiseCanExecuteChanged();
					EditViewExt.Current.Synchronize(this);
					return;
				case "Quit":
					if (this.DialogCancel == null)
						return;
					this.DialogCancel(this, new EventArgs());
					return;
			}
		}

		// IDialogContent

		public string Title
		{
			get { return "Synchronize"; }
		}

		public bool Closable
		{
			get { return false; }
		}

		public event EventHandler<EventArgs> DialogOK { add { } remove { } } // Not used

		public event EventHandler<EventArgs> DialogCancel;

		// IStatus

		void IStatus.Status(string sMsg)
		{
			Status = sMsg;
		}

		void IStatus.Status(string sMsg, bool bBusy)
		{
			Status = sMsg;
			_bBusy = bBusy;
			this.RaiseCanExecuteChanged();
		}

		public void Message(string sMsg)
		{
			Status = "OK";
			MapApplication.Current.Message(sMsg);
			_bBusy = false;
			this.RaiseCanExecuteChanged();
		}

		public void Error(string sErr)
		{
			Status = "ERROR";
			MapApplication.Current.Error(sErr);
			_bBusy = false;
			this.RaiseCanExecuteChanged();
		}

		//  Public methods

		public static void Open()
		{
			SynchronizeDialogModel sd = new SynchronizeDialogModel();
			MapApplication.Current.CreateChildWindow("EditAddin", "SynchronizeDialog", sd);
		}

		// Private constructor

		private SynchronizeDialogModel() { }

		// Private methods

		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private void RaiseCanExecuteChanged()
		{
			if (this.CanExecuteChanged == null)
				return;
			this.CanExecuteChanged(this, new EventArgs());
		}

	}
}
