﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Threading.Tasks;
using System.Windows.Input;
using System.Windows.Interactivity;

//using Debug = System.Diagnostics.Debug;
using Color = System.Drawing.Color;

using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Mapping;
using Esri.ArcGISRuntime.Symbology;
using Esri.ArcGISRuntime.UI;
using Esri.ArcGISRuntime.UI.Controls;

namespace Extensibility
{

	// AppViewExt provides an entry point to the map
	// and manages tool registration and activation

	public class ToolConfiguration
	{

		// RUNTIME 100.6 WORKAROUND:  Using SketchEditor in rubber band mode causes quirky behavior
		// Still present at 100.7

		public static bool RubberBandWorkaround = true;

		private SketchStyle _style = new SketchStyle();
		private SketchEditConfiguration _config = new SketchEditConfiguration();

		public Cursor Cursor = Cursors.Arrow;
		public SketchCreationMode SketchType = SketchCreationMode.Point;
		public Color SelectionColor = Color.Transparent;
		public bool CollectGeometry = true;
		// If Editable is false, EditConfiguration is ignored
		public bool Editable = true;
		// SingleTapCreatesPoint is not honored for Point, Polyline, and Polygon modes
		public bool SingleTapCreatesPoint = false;
		// PointPixelTolerance converts a MapPoint result to an Envelope based on UnitsPerPixel
		public double PointPixelTolerance = double.NaN;
		public bool DoubleTapToFinish = false;
		// RubberBand is only honored for Polyline and Polygon modes
		// If RubberBand is true, Editable and EditConfiguration are ignored,
		// and DoubleTapToFinish is assumed to be true
		public bool RubberBand = false;
		public bool OneShot = false;

		public SketchStyle SketchStyle
		{
			get { return _style; }
			set
			{

				// Clone supplied value

				SketchStyle style = value;
				_style = new SketchStyle()
				{
					FeedbackFillSymbol = style.FeedbackFillSymbol.Clone(),
					FeedbackLineSymbol = style.FeedbackLineSymbol.Clone(),
					FeedbackVertexSymbol = style.FeedbackVertexSymbol.Clone(),
					FillSymbol = style.FillSymbol.Clone(),
					LineSymbol = style.LineSymbol.Clone(),
					MidVertexSymbol = style.MidVertexSymbol.Clone(),
					RotateSymbol = style.RotateSymbol.Clone(),
					ScaleSymbol = style.ScaleSymbol.Clone(),
					SelectedMidVertexSymbol = style.SelectedMidVertexSymbol.Clone(),
					SelectedVertexSymbol = style.SelectedVertexSymbol.Clone(),
					ShowNumbersForVertices = style.ShowNumbersForVertices,
					TransformBoxSymbol = style.TransformBoxSymbol.Clone(),
					VertexSymbol = style.VertexSymbol.Clone(),
					VertexTextSymbol = style.VertexTextSymbol.Clone()
				};

			}
		}
			
		public SketchEditConfiguration EditConfiguration
		{
			get { return _config; }
			set
			{

				// Clone supplied value

				SketchEditConfiguration config = value;
				_config = new SketchEditConfiguration()
				{
					AllowMove = config.AllowMove,
					AllowRotate = config.AllowRotate,
					AllowVertexEditing = config.AllowVertexEditing,
					RequireSelectionBeforeDrag = config.RequireSelectionBeforeDrag,
					ResizeMode = config.ResizeMode,
					VertexEditMode = config.VertexEditMode
				};

			}
		}

		public ToolConfiguration Clone()
		{

			ToolConfiguration tc = new ToolConfiguration()
			{
				Cursor = this.Cursor,
				SketchType = this.SketchType,
				SelectionColor = this.SelectionColor,
				CollectGeometry = this.CollectGeometry,
				Editable = this.Editable,
				SingleTapCreatesPoint = this.SingleTapCreatesPoint,
				PointPixelTolerance = this.PointPixelTolerance,
				DoubleTapToFinish = this.DoubleTapToFinish,
				RubberBand = this.RubberBand,
				OneShot = this.OneShot,
				SketchStyle = this.SketchStyle,
				EditConfiguration = this.EditConfiguration,
			};
			return tc;
		}

	}

	public class MapChangedArgs : EventArgs
	{
		public Map Map;
	}

	public class AppViewExt : Behavior<MapView>
	{

		public Map Map { get; private set; }

		private static AppViewExt _current;
		private bool _bInitialized = false;
		private MapView _mv;
		private SketchEditor _ed;
		private SketchEditConfiguration _edConfig;
		private Dictionary<string, IMapTool> _Tools = new Dictionary<string, IMapTool>();
		private ToolConfiguration _tcDefault = null;
		private ToolConfiguration _tcCurrent = null;
		private IMapTool _pCurrentTool;
		private bool _bPointAdded = false;
		private RubberBandSketch _rb = null;

		public static AppViewExt Current
		{
			get
			{
				if (_current == null)
					_current = new AppViewExt();
				return _current;
			}
		}

		public bool Initialized { get { return _bInitialized; } }

		public BackgroundGrid MapBackground
		{
			get
			{
				if (!_bInitialized)
					return null;
				return _mv.BackgroundGrid;
			}
			set
			{
				if (!_bInitialized)
					return;
				_mv.BackgroundGrid = value;
				if (this.MapBackgroundChanged == null)
					return;
				this.MapBackgroundChanged(this, new EventArgs());
			}
		}
		public MapViewInteractionOptions InteractionOptions
		{
			get { return _mv.InteractionOptions; }
			set { _mv.InteractionOptions = value; }
		}

		// Public events

		public event EventHandler<MapChangedArgs> MapChanged; // Fired when map is replaced
		public event EventHandler<EventArgs> MapBackgroundChanged;

		// Behavior

		protected override void OnAttached()
		{

 			base.OnAttached();
			_mv = AssociatedObject;
			_ed = _mv.SketchEditor;
			if (ToolConfiguration.RubberBandWorkaround)
			{
				_rb = new RubberBandSketch(_mv);
				_rb.GeometryChanged += Workaround_GeometryChanged;
			}

			// Save default configuration

			Cursor cursor = Cursors.Arrow;
			if (_mv.Cursor != null)
				cursor = _mv.Cursor;
			SketchStyle style = new SketchStyle();
			if (_mv.SketchEditor.Style != null)
				style = _mv.SketchEditor.Style;
			SketchEditConfiguration config = new SketchEditConfiguration();
			if (_mv.SketchEditor.EditConfiguration != null)
				config = _mv.SketchEditor.EditConfiguration;
			_tcDefault = new ToolConfiguration()
			{
				Cursor = cursor,
				SelectionColor = _mv.SelectionProperties.Color,
				SketchStyle = style,
				EditConfiguration = config
			};
			_bInitialized = true;

		}

		// Public methods

		public async void SetMap(Map MyMap)
		{
			if (!_bInitialized)
				return; // This should never happen
			MapApplication.Current.Status("Loading map...", true);
			Map = MyMap;
			_mv.Map = MyMap;
			await Map.LoadAsync();

			// 100.5 QUIRK:  NavigationCompleted is no longer fired when the map is loaded

			await _mv.SetViewpointAsync(MyMap.InitialViewpoint);
			this.RaiseMapChanged();
			MapApplication.Current.Status("OK", false);
		}

		public void SetMapCursor(Cursor MyCursor)
		{
			if (!_bInitialized)
				return; // This should never happen
			_mv.Cursor = MyCursor;
		}

		public ToolConfiguration GetDefaultConfiguration()
		{
			return _tcDefault.Clone();
		}

		public void RegisterTool(IMapTool pTool)
		{
			_Tools[pTool.ID] = pTool;
		}

		public void SetTool(string sID = null)
		{

			// Activates a tool and begins collecting geometry
			// (Cancels any active tool)

			//Debug.Print("AppViewExt SetTool requested...");
			if (!_bInitialized)
				return;
			if (!this.SetToolConfig(sID))
				return;
			Task t = this.ExecuteTool();

		}

		public void Complete()
		{

			// Complete active geometry, if any

			//Debug.Print("AppViewExt Complete requested...");
			if (!_bInitialized)
				return;
			if (_pCurrentTool == null)
				return;

			if (_bPointAdded)
			{

				// Remove current mouse point

				ICommand pUndo = _ed.UndoCommand;
				if (!pUndo.CanExecute(null))
					return;
				pUndo.Execute(null);

			}

			ICommand pComplete = _ed.CompleteCommand;
			if (!pComplete.CanExecute(null))
				return;
			pComplete.Execute(null);

		}

		// Event handlers

		void MapView_GeoViewTapped(object sender, GeoViewInputEventArgs e)
		{

			if (_tcCurrent.SingleTapCreatesPoint)
			{

				// Send point as completed geometry

				_pCurrentTool.NotifyComplete(this.CheckPixelTolerance(e.Location));
				return;

			}

			// For RubberBand mode, keep the current mouse point

			_bPointAdded = false;

		}

		void MapView_GeoViewDoubleTapped(object sender, GeoViewInputEventArgs e)
		{

			// For RubberBand mode, keep the final mouse point

			ICommand pComplete = _ed.CompleteCommand;
			if (!pComplete.CanExecute(null))
				return;
			pComplete.Execute(null);
			e.Handled = true;

		}

		void MapView_MouseMove(object sender, MouseEventArgs e)
		{

			Geometry g = _ed.Geometry;
			if (g == null)
				return;
			if (g.IsEmpty)
				return;
			MapPoint pt = _mv.ScreenToLocation(e.GetPosition(_mv));

			if (_bPointAdded)
			{

				// Remove previous mouse point

				ICommand pUndo = _ed.UndoCommand;
				if (!pUndo.CanExecute(null))
					return;
				pUndo.Execute(null);

			}

			// Add mouse point

			ICommand pAdd = _ed.AddCommand;
			if (!pAdd.CanExecute(pt))
				return;
			pAdd.Execute(pt);
			_bPointAdded = true;
			_pCurrentTool.NotifyProgress(_ed.Geometry);

		}

		void SketchEditor_GeometryChanged(object sender, GeometryChangedEventArgs e)
		{
			_pCurrentTool.NotifyProgress(e.NewGeometry);
		}

		private void Workaround_GeometryChanged(object sender, GeometryChangedArgs e)
		{
			_pCurrentTool.NotifyProgress(e.Geometry);
		}

		// Private constructor

		private AppViewExt() { }

		// Private methods

		private void RaiseMapChanged()
		{
			if (this.MapChanged == null)
				return;
			MapChangedArgs e = new MapChangedArgs() { Map = Map };
			this.MapChanged(this, e);
		}

		private bool SetToolConfig(string sID)
		{

			// Cancel active tool, if any

			try
			{
				if (_tcCurrent != null)
				{
					if (ToolConfiguration.RubberBandWorkaround && _tcCurrent.RubberBand)
						_rb.Cancel();
					else if (_ed.CancelCommand.CanExecute(null))
					{
						//Debug.Print("AppViewExt Cancelling SketchEditor...");
						_ed.CancelCommand.Execute(null);
					}
				}
			}
			catch
			{
			}
			if (_pCurrentTool != null)
				_pCurrentTool.Deactivate();
			_pCurrentTool = null;

			// Reset to default if config doesn't exist or is null

			bool bTool = true;
			if (string.IsNullOrEmpty(sID))
				bTool = false;
			else if (!_Tools.ContainsKey(sID))
				bTool = false;
			else if (_Tools[sID] == null)
				bTool = false;

			ToolConfiguration tc = null;
			if (bTool)
			{
				IMapTool pTool = _Tools[sID];
				tc = pTool.Configuration;
				_pCurrentTool = pTool;
			}
			if (tc == null)
				tc = _tcDefault;

			// Set current configuration

			_mv.Cursor = tc.Cursor;
			_edConfig = null;
			if (tc.Editable && (tc.EditConfiguration != null))
				_edConfig = tc.EditConfiguration;
			if (tc.RubberBand)
			{
				_edConfig = new SketchEditConfiguration()
				{
					AllowMove = false,
					AllowRotate = false,
					AllowVertexEditing = true,
					RequireSelectionBeforeDrag = false,
					ResizeMode = SketchResizeMode.None
				};
			}
			_ed.Style = tc.SketchStyle;
			_mv.SelectionProperties.Color = tc.SelectionColor;
			_tcCurrent = tc;
			return (bTool);

		}

		private async Task ExecuteTool()
		{

			_pCurrentTool.Activate();
			if (!_tcCurrent.CollectGeometry)
				return;
			bool bUseWorkaround = ToolConfiguration.RubberBandWorkaround && _tcCurrent.RubberBand;
			bool bInterrupted = false;
			if (_tcCurrent.OneShot)
			{
				_bPointAdded = false;
				if (!bUseWorkaround)
					bInterrupted = await this.GetGeometry();
				else
					bInterrupted = await this.GetGeometryWorkaround();
				if (!bInterrupted)
					this.SetToolConfig(null);
			}
			else
			{
				while (!bInterrupted)
				{
					_bPointAdded = false;
					if (!bUseWorkaround)
						bInterrupted = await this.GetGeometry();
					else
						bInterrupted = await this.GetGeometryWorkaround();
				}
			}

		}

		private async Task<bool> GetGeometry()
		{

			// Add event listeners

			ToolConfiguration tc = _tcCurrent;
			if (tc.SingleTapCreatesPoint || tc.RubberBand)
				_mv.GeoViewTapped += MapView_GeoViewTapped;
			if (tc.DoubleTapToFinish || tc.RubberBand)
				_mv.GeoViewDoubleTapped += MapView_GeoViewDoubleTapped;
			if (tc.RubberBand)
				_mv.MouseMove += MapView_MouseMove;
			else
				_ed.GeometryChanged += SketchEditor_GeometryChanged;

			// Attempt to capture geometry

			Geometry g = null;
			bool bInterrupted = false;
			try
			{
				//Debug.Print("AppViewExt Awaiting Geometry...");
				if (_edConfig != null)
					g = await _ed.StartAsync(tc.SketchType, _edConfig);
				else
					g = await _ed.StartAsync(tc.SketchType, tc.Editable);
				//Debug.Print("Result = " + g.ToString());
			}
			catch (Exception ex)
			{
				if (ex is TaskCanceledException)
				{
					//Debug.Print("AppViewExt GetGeometry task cancelled.");
					bInterrupted = true;
				}
			}

			// Remove event listeners

			if (tc.SingleTapCreatesPoint || tc.RubberBand)
				_mv.GeoViewTapped -= MapView_GeoViewTapped;
			if (tc.DoubleTapToFinish || tc.RubberBand)
				_mv.GeoViewDoubleTapped -= MapView_GeoViewDoubleTapped;
			if (tc.RubberBand)
				_mv.MouseMove -= MapView_MouseMove;
			else
				_ed.GeometryChanged -= SketchEditor_GeometryChanged;

			// Notify complete if geom is not null

			if (bInterrupted)
				return true;
			if (g == null)
				return false;
			_pCurrentTool.NotifyComplete(this.CheckPixelTolerance(g));
			return false;

		}

		private async Task<bool> GetGeometryWorkaround()
		{

			// Set up sketch

			bool bLine = (_tcCurrent.SketchType == SketchCreationMode.Polyline);
			Symbol symDraw = _tcCurrent.SketchStyle.LineSymbol;
			if (!bLine)
				symDraw = _tcCurrent.SketchStyle.FillSymbol;
			_rb.IsLine = bLine;
			_rb.DrawSymbol = symDraw;
			_rb.DoubleTapToFinish = true;

			// Get geometry

			Geometry g = null;
			try
			{
				g = await _rb.Start();
				if (g == null)
					return true;
				_pCurrentTool.NotifyComplete(this.CheckPixelTolerance(g));
			}
			catch (Exception ex)
			{
				Debug.Print("RubberBandSketch ERROR: " + ex.Message);
				return true;
			}
			return false;

		}

		private Geometry CheckPixelTolerance(Geometry geom)
		{
			if (!(geom is MapPoint))
				return geom;
			double dTol = _tcCurrent.PointPixelTolerance;
			if (double.IsNaN(dTol))
				return geom;
			if (dTol <= 0)
				return geom;
			double d = _mv.UnitsPerPixel * dTol;
			MapPoint pt = (MapPoint)geom;
			double dXMin = pt.X - d;
			double dYMin = pt.Y - d;
			double dXMax = pt.X + d;
			double dYMax = pt.Y + d;
			Envelope env = new Envelope(dXMin, dYMin, dXMax, dYMax, pt.SpatialReference);
			return env;
		}

	}

}
