﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Interactivity;
using System.Xml.Linq;

using Esri.ArcGISRuntime.UI.Controls;

namespace Extensibility
{

	// ConfigHelper handles the actual configuration processing,
	// and manages assemblies declared by the configuration

	public static class ConfigHelper
	{

		private const string MapViewContentKey = "MapViewContainerContent";

		private struct ExtBinding
		{
			public string TargetDependencyProperty;
			public string SourceProperty;
			public BindingMode BindingMode;
		}

		private struct ExtBindingConfig
		{
			public string DataContext;
			public List<ExtBinding> Bindings;
		}

		public static string ConfigName;
		public static XDocument ConfigDoc;
		public static Dictionary<string, Assembly> Assemblies = new Dictionary<string, Assembly>();

		private static IMainWindow _pMainWindow;
		private static string _sLocalPath;
		private static MainLayoutModel _LayoutModel;
		private static IStatus _pStatus;
		private static Dictionary<string, DependencyObject> _ViewExts;
		private static Dictionary<string, ExtBindingConfig> _Bindings;
		private static bool _bSuspend = false;

		// Public methods

		public static bool FileFound(string sLocalFileName, bool bVersionSubdir)
		{

			// Check if file can be found
			// (This does not check for updates -- use AppUpdateAddin for that)

			try
			{
				if (File.Exists(sLocalFileName))
					return true;
				string sRemotePath = MapApplication.Current.UpdateLocation;
				if (!string.IsNullOrEmpty(sRemotePath))
				{
					string sLocalDir = Path.GetDirectoryName(sLocalFileName);
					string sFileName = Path.GetFileName(sLocalFileName);
					string sRemoteDir = sLocalDir.Replace(_sLocalPath, sRemotePath);
					if (bVersionSubdir)
						sRemoteDir = Path.Combine(sRemoteDir, MapApplication.Current.RuntimeVersion);
					string sRemoteFileName = Path.Combine(sRemoteDir, sFileName);
					if (File.Exists(sRemoteFileName))
					{
						if (!Directory.Exists(sLocalDir))
							Directory.CreateDirectory(sLocalDir);
						File.Copy(sRemoteFileName, sLocalFileName);
						return true;
					}
				}
				MapApplication.Current.Error("Could not find file: \n" + sLocalFileName);
			}
			catch (Exception ex)
			{
				string sMsg = "ConfigHelper FileFound:\n" + ex.Message;
				MapApplication.Current.Error(sMsg);
			}
			return false;

		}

		public static object CreateInstance(string sAssemblyName, string sClassName, string sNameSpace = "")
		{
			if (string.IsNullOrEmpty(sAssemblyName))
				return null;
			if (!Assemblies.ContainsKey(sAssemblyName))
				return null;
			if (string.IsNullOrEmpty(sClassName))
				return null;
			string sTypeName;
			if (string.IsNullOrEmpty(sNameSpace))
				sTypeName = sAssemblyName + "." + sClassName;
			else
				sTypeName = sNameSpace + "." + sClassName;
			object obj = null;
			try
			{
				Assembly asm = Assemblies[sAssemblyName];
				obj = asm.CreateInstance(sTypeName);
			}
			catch (Exception ex)
			{
				string sMsg = "ConfigHelper.CreateInstance :\n" + ex.Message;
				MapApplication.Current.Error(sMsg);
			}
			return obj;
		}

		public static void Suspend()
		{

			// Suspends further configuration processing
			// (typically in order to restart application)

			_bSuspend = true;

		}

		// Internal methods

		internal static IStatus LoadConfiguration(IMainWindow pMW)
		{

			_pMainWindow = pMW;
			_LayoutModel = MapApplication.Current.MainLayoutModel;
			_ViewExts = new Dictionary<string, DependencyObject>();
			_Bindings = new Dictionary<string,ExtBindingConfig>();
			_bSuspend = false;
			ResourceHelper.ViewAssemblies = new Dictionary<string, string>();
			ResourceHelper.ViewResources = new Dictionary<string, ResourceDictionary>();

			// Load configuration

			ConfigName = _pMainWindow.Configuration;
			Assembly a = Assembly.GetEntryAssembly();
			_sLocalPath = Path.GetDirectoryName(a.Location);
			ConfigDoc = ConfigHelper.LoadConfigDocument();
			if (ConfigDoc == null)
				return null;
			XElement xRoot = ConfigDoc.Root;

			// MainLayout (required) specifies the assembly and class name
			// for the contents of the main application window.

			XElement xLayout = xRoot.Element("MainLayout");
			if (xLayout == null)
			{
				MapApplication.Current.Error("MainLayout is not defined in the configuration");
				return null;
			}
			UserControl ucLayout = ConfigHelper.ProcessLayout(xLayout);
			if (ucLayout == null)
				return null;

			// Each extension is an assembly containing
			// view extensions and/or model extensions

			XElement xExtensions = xRoot.Element("Extensions");
			if (xExtensions != null)
			{
				foreach (XElement xExtension in xExtensions.Elements("Extension"))
				{
					if (_bSuspend)
						return null;
					ConfigHelper.ProcessExtension(xExtension);
				}
			}

			// Assign ModelManager to layout

			ucLayout.DataContext = _LayoutModel;

			// Process extension bindings

			ConfigHelper.ProcessExtBindings();

			// Done

			return _pStatus;

		}

		// Private methods

		private static XDocument LoadConfigDocument()
		{

			// Assume config files do not use version subfolder

			string sLocalFileName = Path.Combine(_sLocalPath, "Config", ConfigName + ".xml");
			if (!ConfigHelper.FileFound(sLocalFileName, false))
				return null;

			// Load configuration

			XDocument xConfigDoc = null;
			try
			{
				xConfigDoc = XDocument.Load(sLocalFileName);
			}
			catch (Exception ex)
			{
				string sMsg = "ConfigHelper LoadConfigDocument:\n" + ex.Message;
				MapApplication.Current.Error(sMsg);
				return null;
			}
			return xConfigDoc;

		}

		private static UserControl ProcessLayout(XElement xLayout)
		{

			// Title (optional) sets the title of the main window

			XElement xTitle = xLayout.Element("Title");
			if (xTitle != null)
				_pMainWindow.SetTitle(xTitle.Value);

			// Get view assembly

			XElement xView = xLayout.Element("ViewAssembly");
			if (xView == null)
			{
				MapApplication.Current.Error("ViewAssembly is not defined in the configuration");
				return null;
			}
			string sViewAssemblyName = xView.Value;
			Assembly asmView = ConfigHelper.GetViewAssembly(sViewAssemblyName, MapApplication.Current.ApplicationName);
			if (asmView == null)
				return null;

			// Get main layout name

			XElement xName = xLayout.Element("MainLayoutName");
			string sLayoutName = xName.Value;
			UserControl ucLayout = ConfigHelper.CreateInstance(sViewAssemblyName, sLayoutName) as UserControl;
			if (ucLayout == null)
			{
				string sMsg = "ConfigHelper MainLayoutName:\nNot found or not a UserControl.";
				MapApplication.Current.Error(sMsg);
				return null;
			}

			// Child windows or items

			XElement xChild;
			string sVal;

			xChild = xLayout.Element("NonModalChild");
			if (xChild != null)
			{
				sVal = xChild.Value;
				switch (sVal)
				{
					case "ChildWindow":
						MapApplication.Current.UseChildWindow = true;
						break;
					case "ChildItem":
						MapApplication.Current.UseChildWindow = false;
						break;
				}
			}
			xChild = xLayout.Element("ShowChildItemIcon");
			if (xChild != null)
			{
				sVal = xChild.Value;
				switch (sVal)
				{
					case "True":
						MapApplication.Current.ShowChildItemIcon = true;
						break;
					case "False":
						MapApplication.Current.ShowChildItemIcon = false;
						break;
				}
			}
			xChild = xLayout.Element("ShowChildItemTitle");
			if (xChild != null)
			{
				sVal = xChild.Value;
				switch (sVal)
				{
					case "True":
						MapApplication.Current.ShowChildItemTitle = true;
						break;
					case "False":
						MapApplication.Current.ShowChildItemTitle = false;
						break;
				}
			}

			// Create MapView, add it to the model manager, and attach AppViewExt

			MapView mv = new MapView();
			_LayoutModel.MapViewContainerContent = mv;
			AppViewExt ext = AppViewExt.Current;
			Interaction.GetBehaviors(mv).Add(ext);

			// Add layout to main window

			_pMainWindow.SetLayout(ucLayout);
			return ucLayout;

		}

		private static void ProcessExtension(XElement xExtension)
		{

			// Load extension assembly
			// Assume extension files use version subfolder

			string sExtAssemblyName = xExtension.Attribute("Name").Value;
			try
			{
				string sLocalFileName = Path.Combine(_sLocalPath, "Extensions", sExtAssemblyName + ".dll");
				if (!ConfigHelper.FileFound(sLocalFileName, true))
					return;
				Assembly asmExt = Assembly.LoadFrom(sLocalFileName);
				Assemblies[sExtAssemblyName] = asmExt;
			}
			catch (Exception ex)
			{
				string sMsg = "ConfigHelper ProcessExtension:\n" + ex.Message;
				System.Diagnostics.Debug.Print(sMsg);
				return;
			}

			// Load view assembly

			XElement xView = xExtension.Element("ViewAssembly");
			if (xView != null)
			{
				string sViewAssemblyName = xView.Value;
				Assembly asmView = GetViewAssembly(sViewAssemblyName, sExtAssemblyName);
			}

			// View extensions are of the type Behavior<MapView>

			XElement xViewExts = xExtension.Element("ViewExtensions");
			if (xViewExts != null)
			{
				foreach (XElement xViewExt in xViewExts.Elements("ViewExtension"))
					ConfigHelper.ProcessViewExtension(sExtAssemblyName, xViewExt);
			}

			// Model extensions are objects that can optionally implement IViewExtensionBinding

			XElement xModelExts = xExtension.Element("ModelExtensions");
			if (xModelExts != null)
			{
				foreach (XElement xModelExt in xModelExts.Elements("ModelExtension"))
					ConfigHelper.ProcessModelExtension(sExtAssemblyName, xModelExt);
			}

		}

		private static void ProcessViewExtension(string sExtAssemblyName, XElement xViewExt)
		{

			// Create instance

			string sClassName = xViewExt.Attribute("Name").Value;
			Behavior<MapView> extView = ConfigHelper.CreateInstance(sExtAssemblyName, sClassName) as Behavior<MapView>;
			if (extView == null)
			{
				string sMsg = "ConfigHelper " + sClassName + ":\nNot found or not a Behavior<MapView>.";
				MapApplication.Current.Error(sMsg);
				return;
			}
			string sKey = sExtAssemblyName + "." + sClassName;
			_ViewExts[sKey] = extView;

			// Check for data context

			XElement xDataContext = xViewExt.Element("DataContext");
			if (xDataContext != null)
			{

				// Check for bindings

				XElement xBindings = xViewExt.Element("Bindings");
				if (xBindings != null)
				{
					ExtBindingConfig config = new ExtBindingConfig();
					config.DataContext = xDataContext.Attribute("Name").Value;
					config.Bindings = new List<ExtBinding>();
					foreach (XElement xBinding in xBindings.Elements())
					{
						ExtBinding binding = new ExtBinding();
						binding.TargetDependencyProperty = xBinding.Attribute("TargetDependencyProperty").Value;
						binding.SourceProperty = xBinding.Attribute("SourceProperty").Value;
						BindingMode bm;
						string sMode = xBinding.Attribute("BindingMode").Value;
						switch (sMode)
						{
							case "OneWay":
								bm = BindingMode.OneWay;
								break;
							case "TwoWay":
								bm = BindingMode.TwoWay;
								break;
							case "OneWayToSource":
								bm = BindingMode.OneWayToSource;
								break;
							case "OneTime":
								bm = BindingMode.OneTime;
								break;
							default:
								bm = BindingMode.Default;
								break;
						}
						binding.BindingMode = bm;
						config.Bindings.Add(binding);
					}
					_Bindings[sKey] = config;
				}
			}

			// If an extension implements IConfigurable, load the configuration

			if (extView is IConfigurable)
			{
				IConfigurable pConfig = (IConfigurable)extView;
				XElement xConfigData = xViewExt.Element("ConfigData");
				if (xConfigData != null)
				{
					string sConfigData = null;
					if (xConfigData.HasElements)
						sConfigData = xConfigData.ToString();
					else
						sConfigData = xConfigData.Value;
					pConfig.LoadConfiguration(sConfigData);
				}
			}

			// Attach to MapView

			MapView mv = _LayoutModel.MapViewContainerContent;
			Interaction.GetBehaviors(mv).Add(extView);

		}

		private static void ProcessModelExtension(string sExtAssemblyName, XElement xModelExt)
		{

			// Create instance

			string sClassName = xModelExt.Attribute("Name").Value;
			object extModel = ConfigHelper.CreateInstance(sExtAssemblyName, sClassName);
			if (extModel == null)
			{
				string sMsg = "ConfigHelper " + sClassName + ":\nNot found.";
				MapApplication.Current.Error(sMsg);
				return;
			}

			// Add to ModelManager

			_LayoutModel.DataContexts[sExtAssemblyName + "." + sClassName] = extModel;

			// If an extension implements IConfigurable, load the configuration

			if (extModel is IConfigurable)
			{
				IConfigurable pConfig = (IConfigurable)extModel;
				XElement xConfigData = xModelExt.Element("ConfigData");
				if (xConfigData != null)
				{
					string sConfigData = null;
					if (xConfigData.HasElements)
						sConfigData = xConfigData.ToString();
					else
						sConfigData = xConfigData.Value;
					if (!string.IsNullOrEmpty(sConfigData))
						pConfig.LoadConfiguration(sConfigData);
				}
			}

			// If a model extension implements IStatus, it is assigned to the MapApplication
			// in order to handle status messages for the application

			if (extModel is IStatus)
				_pStatus = (IStatus)extModel;

		}

		private static void ProcessExtBindings()
		{

			// Generate bindings based on configuration

			foreach (string sKey in _Bindings.Keys)
			{
				if (!_ViewExts.ContainsKey(sKey))
					continue; // This should never happen
				DependencyObject extView = _ViewExts[sKey];
				Type t = extView.GetType();
				ExtBindingConfig config = _Bindings[sKey];
				object DataContext = _LayoutModel.DataContexts[config.DataContext];
				if (DataContext == null)
					continue;
				foreach (ExtBinding binding in config.Bindings)
				{
					FieldInfo fi = t.GetField(binding.TargetDependencyProperty);
					if (fi == null)
						continue;
					object value = fi.GetValue(extView);
					if (!(value is DependencyProperty))
						continue;
					DependencyProperty dp = (DependencyProperty)value;
					Binding b = new Binding(binding.SourceProperty);
					b.Source = DataContext;
					b.Mode = binding.BindingMode;
					BindingOperations.SetBinding(extView, dp, b);
				}
			}

		}

		private static Assembly GetViewAssembly(string sViewAssemblyName, string sAssemblyKey)
		{
			Assembly asmView = null;
			try
			{
				if (!Assemblies.ContainsKey(sViewAssemblyName))
				{

					// Assume view files do not use version subfolder

					string sLocalFileName = Path.Combine(_sLocalPath, "Views", sViewAssemblyName + ".dll");
					if (!ConfigHelper.FileFound(sLocalFileName, false))
						return null;
					asmView = Assembly.LoadFrom(sLocalFileName);
					Assemblies[sViewAssemblyName] = asmView;

				}
				ResourceHelper.ViewAssemblies[sAssemblyKey] = sViewAssemblyName;
			}
			catch (Exception ex)
			{
				string sMsg = "ConfigHelper GetViewAssembly:\n" + ex.Message;
				MapApplication.Current.Error(sMsg);
				asmView = null;
			}
			return asmView;
		}

	}

}
