﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Resources;

using Esri.ArcGISRuntime.Symbology;

namespace Extensibility
{

	// ResourceHelper helps retrieve resources from the view libraries

	public static class ResourceHelper
	{

		// MainAssemblyName is the name of the view assembly containing the main window contents
		public static string MainAssemblyName { get; set; }
		// ViewAssemblies stores the name of each view assembly associated with a key
		// [typically an extension name]
		public static Dictionary<string, string> ViewAssemblies { get; set; }
		// Stores resources retrieved by the GetResources method
		public static Dictionary<string, ResourceDictionary> ViewResources { get; set; }

		public static System.Drawing.Color ConvertColor(System.Windows.Media.Color colorWPF)
		{
			return System.Drawing.Color.FromArgb(colorWPF.A, colorWPF.R, colorWPF.G, colorWPF.B);
		}

		public static System.Windows.Media.Color ConvertColor(System.Drawing.Color colorStd)
		{
			return System.Windows.Media.Color.FromArgb(colorStd.A, colorStd.R, colorStd.G, colorStd.B);
		}

		public static ResourceDictionary GetResources(string sViewKey)
		{

			// Assumes that a ResourceDictionary named "Resources.xaml" exists in a view assembly

			if (!ViewAssemblies.ContainsKey(sViewKey))
				return null;
			string sAssemblyName = ViewAssemblies[sViewKey];
			if (ViewResources.ContainsKey(sAssemblyName))
				return ViewResources[sAssemblyName];
			ResourceDictionary rd = null;
			try
			{
				string sPath = "pack://application:,,,/" + sAssemblyName + ";component/Resources.xaml";
				Uri MyUri = new Uri(sPath);
				rd = new ResourceDictionary();
				rd.Source = MyUri;
				ViewResources[sAssemblyName] = rd;
			}
			catch (Exception ex)
			{
				string sMsg = "ResourceHelper.GetResources ERROR: " + ex.Message;
				System.Diagnostics.Debug.Print(sMsg);
			}
			return rd;

		}

		public static ImageSource GetImageSource(string sViewKey, string sIconPath)
		{

			// Assumes that image file is included in the view assembly as a Resource
			// and that the path is relative, e.g. "Images/MyImage.png"

			if (!ViewAssemblies.ContainsKey(sViewKey))
				return null;
			string sAssemblyName = ViewAssemblies[sViewKey];
			BitmapImage bmi = null;
			try
			{
				string sPath = "pack://application:,,,/" + sAssemblyName + ";component/" + sIconPath;
				Uri MyUri = new Uri(sPath);
				bmi = new BitmapImage(MyUri);
			}
			catch (Exception ex)
			{
				string sMsg = "ResourceHelper.GetImageSource ERROR: " + ex.Message;
				System.Diagnostics.Debug.Print(sMsg);
			}
			return bmi;

		}

		public static System.Windows.Controls.Image GetImage(string sViewKey, string sIconPath)
		{
			ImageSource src = GetImageSource(sViewKey, sIconPath);
			if (src == null)
				return null;
			System.Windows.Controls.Image img = new System.Windows.Controls.Image();
			img.Source = src;
			return img;
		}

		public static Cursor GetCursor(string sViewKey, string sCursorPath)
		{

			// Assumes that cursor file is included in the view assembly as a Resource
			// and that the path is relative, e.g. "Images/MyCursor.cur"

			if (!ViewAssemblies.ContainsKey(sViewKey))
				return null;
			string sAssemblyName = ViewAssemblies[sViewKey];
			Cursor c = null;
			try
			{
				string sPath = "pack://application:,,,/" + sAssemblyName + ";component/" + sCursorPath;
				Uri MyUri = new Uri(sPath);
				StreamResourceInfo info = Application.GetResourceStream(MyUri);
				c = new Cursor(info.Stream);
			}
			catch (Exception ex)
			{
				string sMsg = "ResourceHelper.GetCursor ERROR: " + ex.Message;
				System.Diagnostics.Debug.Print(sMsg);
			}
			return c;

		}

		public async static Task<PictureMarkerSymbol> GetPictureMarkerSymbol(string sViewKey, string sMarkerPath)
		{

			// Assumes that image file is included in the view assembly as a Resource
			// and that the path is relative, e.g. "Images/MyImage.png"

			if (!ViewAssemblies.ContainsKey(sViewKey))
				return null;
			string sAssemblyName = ViewAssemblies[sViewKey];
			PictureMarkerSymbol pmSym = null;
			try
			{
				string sPath = "pack://application:,,,/" + sAssemblyName + ";component/" + sMarkerPath;
				StreamResourceInfo srInfo = Application.GetResourceStream(new Uri(sPath));
				pmSym = await PictureMarkerSymbol.CreateAsync(srInfo.Stream);
			}
			catch (Exception ex)
			{
				string sMsg = "ResourceHelper.GetPictureMarkerSymbol ERROR: " + ex.Message;
				System.Diagnostics.Debug.Print(sMsg);
			}
			return pmSym;

		}

		public static SimpleFillSymbol GetFillSymbol(string sViewKey, string sEntryKey)
		{

			// Assumes that the fill symbol is in a designated ResourceDictionary for a view assembly.
			// Resource entry must be a shape, and only solid color brushes are supported.

			ResourceDictionary rd = GetResources(sViewKey);
			if (rd == null)
				return null;
			if (!rd.Contains(sEntryKey))
				return null;
			System.Windows.Shapes.Shape shape = rd[sEntryKey] as System.Windows.Shapes.Shape;
			if (shape == null)
				return null;
			SimpleFillSymbol sfSym = new SimpleFillSymbol();
			sfSym.Style = SimpleFillSymbolStyle.Null;
			SolidColorBrush b = shape.Fill as SolidColorBrush;
			if (b != null)
			{
				sfSym.Style = SimpleFillSymbolStyle.Solid;
				sfSym.Color = ResourceHelper.ConvertColor(b.Color);
			}
			SimpleLineSymbol slSym = ResourceHelper.GetLineSymbol(sViewKey, sEntryKey);
			sfSym.Outline = slSym;
			return sfSym;

		}

		public static SimpleLineSymbol GetLineSymbol(string sViewKey, string sEntryKey)
		{

			// Assumes that the line symbol is in a designated ResourceDictionary for a view assembly.
			// Resource entry must be a shape, and only solid color brushes are supported.
			// [Dash pattern conversion is not implemented.]

			ResourceDictionary rd = GetResources(sViewKey);
			if (rd == null)
				return null;
			if (!rd.Contains(sEntryKey))
				return null;
			System.Windows.Shapes.Shape shape = rd[sEntryKey] as System.Windows.Shapes.Shape;
			if (shape == null)
				return null;
			SimpleLineSymbol slSym = new SimpleLineSymbol();
			slSym.Style = SimpleLineSymbolStyle.Solid;
			slSym.Width = shape.StrokeThickness;
			SolidColorBrush b = shape.Stroke as SolidColorBrush;
			if (b != null)
				slSym.Color = ResourceHelper.ConvertColor(b.Color);
			return slSym;

		}

		public static TextSymbol GetTextSymbol(string sViewKey, string sEntryKey)
		{

			// Assumes that the text symbol is in a designated ResourceDictionary for a view assembly.
			// Resource entry must be a TextBlock style.  Foreground, if specified, must be a solid color brush.

			ResourceDictionary rd = GetResources(sViewKey);
			if (rd == null)
				return null;
			if (!rd.Contains(sEntryKey))
				return null;
			Style style = rd[sEntryKey] as Style;
			if (style == null)
				return null;
			if (style.TargetType != typeof(TextBlock))
				return null;
			TextSymbol txtSym = new TextSymbol();
			foreach (Setter setter in style.Setters)
			{
				DependencyProperty prop = setter.Property;
				object val = setter.Value;
				switch (prop.Name)
				{
					case "FontFamily":
						System.Windows.Media.FontFamily ff = (System.Windows.Media.FontFamily)val;
						txtSym.FontFamily = ff.ToString();
						break;
					case "FontSize":
						txtSym.Size = (double)val;
						break;
					case "FontStyle":
						System.Windows.FontStyle fs = (System.Windows.FontStyle)val;
						if (fs == FontStyles.Italic)
							txtSym.FontStyle = Esri.ArcGISRuntime.Symbology.FontStyle.Italic;
						else if (fs == FontStyles.Oblique)
							txtSym.FontStyle = Esri.ArcGISRuntime.Symbology.FontStyle.Oblique;
						break;
					case "FontWeight":
						System.Windows.FontWeight fw = (System.Windows.FontWeight)val;
						if (fw == FontWeights.Bold)
							txtSym.FontWeight = Esri.ArcGISRuntime.Symbology.FontWeight.Bold;
						break;
					case "TextDecorations":
						TextDecorationCollection tdc = (TextDecorationCollection)val;
						if (tdc == TextDecorations.Underline)
							txtSym.FontDecoration = Esri.ArcGISRuntime.Symbology.FontDecoration.Underline;
						if (tdc == TextDecorations.Strikethrough)
							txtSym.FontDecoration = Esri.ArcGISRuntime.Symbology.FontDecoration.LineThrough;
						break;
					case "Foreground":
						SolidColorBrush br = (SolidColorBrush)val;
						txtSym.Color = ResourceHelper.ConvertColor(br.Color);
						break;
				}

			}
			return txtSym;

		}


	}

}
