﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using Esri.ArcGISRuntime.Data;
using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Location;
using Esri.ArcGISRuntime.Mapping;
using Esri.ArcGISRuntime.Security;
using Esri.ArcGISRuntime.Symbology;
using Esri.ArcGISRuntime.Tasks;
using Esri.ArcGISRuntime.UI;
using Esri.ArcGISRuntime.UI.Controls;

namespace Extensibility
{

	public class GeometryChangedArgs: EventArgs
	{
		public Geometry Geometry;
	}

   public class RubberBandSketch
   {

		// **** WORKAROUND for 100.6 SketchEditor quirk ****

		private MapView _mv;
		private GraphicsOverlay _go;
		private bool _bLine;
		private Symbol _sym;
		private bool _bDoubleTap;
		private Graphic _gr;
		private PolylineBuilder _line;
		private PolygonBuilder _poly;
		private bool _bComplete;
		private bool _bInterrupt;

		// Public constructor

		public RubberBandSketch(MapView mv)
		{

			_mv = mv;

			// Find or create graphics overlay

			int iCount = _mv.GraphicsOverlays.Count;
			if (iCount == 0)
			{
				_go = new GraphicsOverlay();
				_mv.GraphicsOverlays.Add(_go);
			}
			else
				_go = _mv.GraphicsOverlays[iCount - 1];


			// Set default properties

			IsLine = true;
			DrawSymbol = new SimpleLineSymbol()
			{
				Style = SimpleLineSymbolStyle.Solid,
				Color = Color.Black,
				Width = 1
			};
			DoubleTapToFinish = true;
			IsBusy = false;

		}

		// Public properties
		// These are read at start, and then ignored until the next start

		public bool IsLine { get; set; }

		public Symbol DrawSymbol { get; set; }

		public bool DoubleTapToFinish { get; set; }

		public bool IsBusy { get; private set; }

		// Public methods and events

		public event EventHandler<GeometryChangedArgs> GeometryChanged;

		public async Task<Geometry> Start()
		{
			if (IsBusy)
				throw new Exception("RubberBandSketch is busy.");
			Geometry geom = await Task.Factory.StartNew(() => { return this.GetGeometry(); });
			return geom;
		}

		public void Complete()
		{
			if (!IsBusy)
				return;
			_bComplete = true;
			_bInterrupt = true;
		}

		public void Cancel()
		{
			if (!IsBusy)
				return;
			_bComplete = false;
			_bInterrupt = true;
		}

		// private constructor

		private RubberBandSketch() { }

		// private methods

		private void OnGeometryChanged()
		{
			if (this.GeometryChanged == null)
				return;
			Geometry geom = null;
			if (_bLine)
				geom = _line.ToGeometry();
			else
				geom = _poly.ToGeometry();
			GeometryChangedArgs e = new GeometryChangedArgs() { Geometry = geom };
			this.GeometryChanged(this, e);
		}

		private Geometry GetGeometry()
		{

			// Check parameters

			_bLine = IsLine;
			_sym = DrawSymbol;
			_bDoubleTap = DoubleTapToFinish;
			IsBusy = true;

			// Add event handlers

			_mv.GeoViewTapped += MapView_GeoViewTapped;
			_mv.GeoViewDoubleTapped += MapView_GeoViewDoubleTapped;
			_mv.MouseMove += MapView_MouseMove;

			// Loop

			_gr = null;
			_bComplete = false;
			_bInterrupt = false;
			while (!_bInterrupt)
			{

			}

			// Remove event Handlers

			_mv.GeoViewTapped -= MapView_GeoViewTapped;
			_mv.GeoViewDoubleTapped -= MapView_GeoViewDoubleTapped;
			_mv.MouseMove -= MapView_MouseMove;

			// Remove graphic

			if (_gr != null)
			{
				try
				{
					if (_go.Graphics.Count == 1)
						_go.Graphics.Clear();
					else
						_go.Graphics.Remove(_gr); // 100.5 BUG? Sometimes crashes
					_gr = null;
				}
				catch (Exception ex)
				{
					Debug.WriteLine("RubberBandSketch ERROR: " + ex.Message);
				}
			}

			// Check geometry

			IsBusy = false;
			Geometry geom = null;
			if (_bComplete)
				geom = this.GetCurrentGeometry();
			_line = null;
			_poly = null;
			if (geom == null)
				return null;
			if (geom.IsEmpty)
				return null;
			int iCount = 2;
			if (!_bLine)
				iCount = 3;
			Multipart mp = geom as Multipart;
			ReadOnlyPart part = mp.Parts[0];
			if (part.PointCount < iCount)
				return null;
			return geom;

		}

		private Geometry GetCurrentGeometry()
		{
			Geometry geom = null;
			if (_bLine)
			{
				if (_line != null)
					geom = _line.ToGeometry();
			}
			else
			{
				if (_poly != null)
					geom = _poly.ToGeometry();
			}
			return geom;
		}

		private IList<Part> GetCurrentParts()
		{
			IList<Part> parts = null;
			if (_bLine)
			{
				if (_line != null)
					parts = _line.Parts;
			}
			else
			{
				if (_poly != null)
					parts = _poly.Parts;
			}
			return parts;
		}

		private void AddPoint(MapPoint pt)
		{

			// Add point to builder

			Geometry geom = null;
			if (_bLine)
			{
				if (_line == null)
					_line = new PolylineBuilder(pt.SpatialReference);
				_line.AddPoint(pt);
				geom = _line.ToGeometry();
			}
			else
			{
				if (_poly == null)
					_poly = new PolygonBuilder(pt.SpatialReference);
				_poly.AddPoint(pt);
				geom = _poly.ToGeometry();
			}
			this.UpdateGraphic();

		}

		private void MovePoint(MapPoint pt)
		{

			IList<Part> parts = this.GetCurrentParts();
			if (parts == null)
			{
				this.AddPoint(pt);
				return;
			}
			Part part = parts[0];
			int iCount = part.PointCount;
			if (iCount == 1)
			{
				this.AddPoint(pt);
				return;
			}
			part.RemovePoint(iCount - 1);
			part.AddPoint(pt);
			this.UpdateGraphic();

		}

		private void UpdateGraphic()
		{

			// Create or update graphic

			Geometry geom = this.GetCurrentGeometry();
			if (geom == null)
				return;
			if (_gr == null)
			{
				_gr = new Graphic() { Symbol = _sym };
				_go.Graphics.Add(_gr);
				return;
			}
			_gr.Geometry = geom;

		}

		// event handlers

		private void MapView_GeoViewTapped(object sender, GeoViewInputEventArgs e)
		{
			//Debug.Print("GeoViewTapped.");
			MapPoint pt = e.Location;
			this.AddPoint(pt);
			e.Handled = true;
		}

		private void MapView_GeoViewDoubleTapped(object sender, GeoViewInputEventArgs e)
		{

			if (_gr == null)
				return;
			//Debug.Print("GeoViewDoubleTapped.");

			// Remove final point

			IList<Part> parts = this.GetCurrentParts();
			Part part = parts[0];
			int iCount = part.PointCount;
			part.RemovePoint(iCount - 1);

			e.Handled = true;
			_bComplete = true;
			_bInterrupt = true;

		}

		private void MapView_MouseMove(object sender, System.Windows.Input.MouseEventArgs e)
		{
			if (_gr == null)
				return;
			MapPoint pt = _mv.ScreenToLocation(e.GetPosition(_mv));
			this.MovePoint(pt);
			this.OnGeometryChanged();
		}

	}

}
