﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Reflection;
using System.Windows;
using System.Windows.Input;

using Esri.ArcGISRuntime.Geometry;

namespace Extensibility
{

	// CommandHelper is the core wrapper for ICommand support

	public class NotifyCanExecuteArgs : EventArgs
	{
		public bool CanExecute { get; set; }
	}

	public class CommandHelper : ICommand
	{

		private List<Action<object>> _ExecuteActions = new List<Action<object>>();
		private List<object> _ExecuteParameters = new List<object>();
		private bool _bEnabled;

		public CommandHelper(bool bEnabled)
		{
			_bEnabled = bEnabled;
		}

		public void AddExecuteAction(Action<object> ExecuteAction, object ExecuteParameter = null)
		{
			_ExecuteActions.Add(ExecuteAction);
			_ExecuteParameters.Add(ExecuteParameter);
		}

		public void AddCanExecuteEvent(object Sender, string EventName)
		{
			EventInfo eiEvent = Sender.GetType().GetEvent(EventName, BindingFlags.Public | BindingFlags.Instance);
			Type tDelegate = eiEvent.EventHandlerType;
			MethodInfo miEventHandler = this.GetType().GetMethod("CanExecuteHandler", BindingFlags.NonPublic | BindingFlags.Instance);
			Delegate d = Delegate.CreateDelegate(tDelegate, this, miEventHandler);
			MethodInfo miAddHandler = eiEvent.GetAddMethod();
			object[] Args = { d };
			miAddHandler.Invoke(Sender, Args);
		}

		public void RemoveCanExecuteEvent(object Sender, string EventName)
		{
			EventInfo eiEvent = Sender.GetType().GetEvent(EventName, BindingFlags.Public | BindingFlags.Instance);
			Type tDelegate = eiEvent.EventHandlerType;
			MethodInfo miEventHandler = this.GetType().GetMethod("CanExecuteHandler", BindingFlags.NonPublic | BindingFlags.Instance);
			Delegate d = Delegate.CreateDelegate(tDelegate, this, miEventHandler);
			MethodInfo miRemoveHandler = eiEvent.GetRemoveMethod();
			object[] Args = { d };
			miRemoveHandler.Invoke(Sender, Args);
		}

		public bool CanExecute(object parameter)
		{
			return _bEnabled;
		}

		public event EventHandler CanExecuteChanged;

		public void Execute(object parameter)
		{
			int iCount = _ExecuteActions.Count;
			if (iCount == 0)
				return;
			for (int i = 0; i < iCount; i++)
			{
				Action<object> ExecuteAction = _ExecuteActions[i];
				object ExecuteParameter = _ExecuteParameters[i];
				if (ExecuteParameter == null)
					ExecuteParameter = parameter;
				ExecuteAction(ExecuteParameter);
			}
		}

		private void CanExecuteHandler(object sender, NotifyCanExecuteArgs e)
		{
			if (_bEnabled == e.CanExecute)
				return;
			_bEnabled = e.CanExecute;
			if (CanExecuteChanged != null)
				this.CanExecuteChanged(this, new EventArgs());
		}

	}

	// ToolHelper encapsulates a tool that registers with AppViewExt

	public class ToolHelper : IMapTool
	{

		public class NotifyToolArgs : EventArgs
		{
			public Geometry Geometry;
		}

		public CommandHelper ExecuteCommand { get; private set; }
		public CommandHelper ActivateCommand { get; private set; }
		public CommandHelper DeactivateCommand { get; private set; }

		private string _sID;
		private ToolConfiguration _tc = AppViewExt.Current.GetDefaultConfiguration();

		public event EventHandler<NotifyToolArgs> ToolProgress;
		public event EventHandler<NotifyToolArgs> ToolComplete;

		public ToolHelper(bool bEnabled, string sAssemblyKey = null, string sCursorPath = null)
		{
			_sID = Guid.NewGuid().ToString();
			this.SetCursor(sAssemblyKey, sCursorPath);
			ExecuteCommand = new CommandHelper(bEnabled);
			ExecuteCommand.AddExecuteAction(this.Execute);
			ActivateCommand = new CommandHelper(true);
			DeactivateCommand = new CommandHelper(true);
			AppViewExt.Current.RegisterTool(this);
		}

		public void SetCursor(string sAssemblyKey, string sCursorPath)
		{
			if (string.IsNullOrEmpty(sAssemblyKey))
				return;
			if (string.IsNullOrEmpty(sCursorPath))
				return;
			Cursor c = ResourceHelper.GetCursor(sAssemblyKey, sCursorPath);
			if (c != null)
				_tc.Cursor = c;
		}

		// IMapTool

		public string ID { get { return _sID; } }

		public ToolConfiguration Configuration
		{
			get { return _tc; }
			set { _tc = value; }
		}

		public void Activate()
		{
			if (!ActivateCommand.CanExecute(null))
				return;
			ActivateCommand.Execute(null);
		}

		public void Deactivate()
		{
			if (!DeactivateCommand.CanExecute(null))
				return;
			DeactivateCommand.Execute(null);
		}

		public void NotifyProgress(Geometry g)
		{
			if (this.ToolProgress == null)
				return;
			NotifyToolArgs e = new NotifyToolArgs();
			e.Geometry = g;
			this.ToolProgress(this, e);
		}

		public void NotifyComplete(Geometry g)
		{
			if (this.ToolComplete == null)
				return;
			NotifyToolArgs e = new NotifyToolArgs();
			e.Geometry = g;
			this.ToolComplete(this, e);
		}

		// Private

		private void Execute(object parameter)
		{
			AppViewExt.Current.SetTool(_sID);
		}

	}

	// DataHelper provides a moderately safe observable dictionary for use as a bindable property
	// It should not be used for non-nullable properties

	public class DataHelper : IDictionary<string, object>, INotifyPropertyChanged
	{

		private Dictionary<string, object> _dict = new Dictionary<string, object>();

		public event PropertyChangedEventHandler PropertyChanged;

		public object this[string key]
		{
			get
			{
				if (!_dict.ContainsKey(key))
					return null;
				return _dict[key];
			}
			set
			{
				_dict[key] = value;
				this.RaisePropertyChanged(key);
			}
		}

		public void Add(string key, object value)
		{
			_dict.Add(key, value);
			this.RaisePropertyChanged(key);
		}

		public bool Remove(string key)
		{
			bool bResult = _dict.Remove(key);
			if (bResult)
				this.RaisePropertyChanged(key);
			return bResult;
		}

		public bool ContainsKey(string key)
		{
			return _dict.ContainsKey(key);
		}

		public bool TryGetValue(string key, out object value)
		{
			return _dict.TryGetValue(key, out value);
		}

		public ICollection<string> Keys
		{
			get { return _dict.Keys; }
		}

		public ICollection<object> Values
		{
			get { return _dict.Values; }
		}

		public int Count
		{
			get { return _dict.Count; }
		}

		public bool IsReadOnly
		{
			get { return false; }
		}

		public void Clear()
		{
			_dict.Clear();
		}

		public void Add(KeyValuePair<string, object> item)
		{
			throw new NotImplementedException();
		}

		public bool Contains(KeyValuePair<string, object> item)
		{
			throw new NotImplementedException();
		}

		public void CopyTo(KeyValuePair<string, object>[] array, int arrayIndex)
		{
			throw new NotImplementedException();
		}

		public bool Remove(KeyValuePair<string, object> item)
		{
			throw new NotImplementedException();
		}

		public IEnumerator<KeyValuePair<string, object>> GetEnumerator()
		{
			throw new NotImplementedException();
		}

		System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
		{
			throw new NotImplementedException();
		}

		private void RaisePropertyChanged(string sKey)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sKey));
		}

	}

}
