﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Interactivity;
using System.Windows.Media;

namespace Extensibility
{

	// MapApplication implements IStatus for the application
	// and handles the creation of child windows 

	public class MapApplication : IStatus
	{

		public MainLayoutModel MainLayoutModel { get; private set; }
		public string ApplicationName { get; private set; }
		public string RuntimeVersion { get; private set; }
		public string UpdateLocation { get; private set; }
		public ImageSource WindowIcon { get; private set; }

		private static MapApplication _current;
		private IMainWindow _pMainWindow;
		private IStatus _pStatus;
		private List<ChildWindowModel> _Windows = new List<ChildWindowModel>();

		private MapApplication() { }

		// Public events

		public event EventHandler AppReady;  // Fired when configuration is fully processed

		// Public entry point for application

		public static void Init(IMainWindow pMainWindow)
		{
			MapApplication app = new MapApplication();
			app._pMainWindow = pMainWindow;
			_current = app;
			try
			{
				app.Initialize();
			}
			catch (Exception ex)
			{
				string sMsg = "Error initializing MapApplication:\n" + ex.Message + "\n" + ex.StackTrace;
				app.Error(sMsg);
			}
		}

		public static MapApplication Current
		{
			get
			{
				if (_current == null)
				{
					MapApplication.Init(null);
				}
				return _current;
			}
		}

		// IStatus

		public void Status(string sMsg)
		{
			System.Diagnostics.Debug.Print("MapApplication.Status: " + sMsg);
			if (_pStatus == null)
				return;
			_pStatus.Status(sMsg);
		}

		public void Status(string sMsg, bool bBusy)
		{
			System.Diagnostics.Debug.Print("MapApplication.Status: " + sMsg);
			if (_pStatus == null)
				return;
			_pStatus.Status(sMsg, bBusy);
		}

		public void Message(string sMsg)
		{
			if (_pStatus != null)
			{
				_pStatus.Message(sMsg);
				return;
			}
			MessageBox.Show(sMsg, _pMainWindow.ApplicationName, MessageBoxButton.OK, MessageBoxImage.Information);
		}

		public void Error(string sErr)
		{
			if (_pStatus != null)
			{
				_pStatus.Error(sErr);
				return;
			}
			MessageBox.Show("ERROR: " + sErr, _pMainWindow.ApplicationName, MessageBoxButton.OK, MessageBoxImage.Error);
		}

		// Public methods

		public bool? CreateChildWindow(string sViewKey, string sViewName, object ViewModel)
		{

			// Get View

			if (!ResourceHelper.ViewAssemblies.ContainsKey(sViewKey))
			{
				string sErr = sViewKey + " is not a registered view assembly key.";
				sErr += "\nViewName = " + sViewName;
				sErr += "\nView Model = " + ViewModel.ToString();
				this.Error(sErr);
				return null;
			}
			string sAssembly = ResourceHelper.ViewAssemblies[sViewKey];
			UserControl ucView = ConfigHelper.CreateInstance(sAssembly, sViewName) as UserControl;
			if (ucView == null)
			{
				this.Error(sViewName + " is not found or is not a UserControl.");
				return null;
			}

			// Set up ChildWindow or ChildItem

			bool bOpenModal = (ViewModel is IDialogContent);
			if ((!bOpenModal) && (!UseChildWindow))
			{
				IWindowContent pModel = ViewModel as IWindowContent;
				ChildItem item = new ChildItem();
				item.SetContent(ucView, pModel);
				MainLayoutModel.ChildItems.Add(item);
				return null;
			}
			ChildWindow cw = new ChildWindow();
			ChildWindowModel cwm = new ChildWindowModel();
			Interaction.GetBehaviors(cw).Add(cwm);
			cwm.SetContent(ucView, ViewModel);
			if (!bOpenModal)
			{
				cw.Owner = Application.Current.MainWindow;
				_Windows.Add(cwm);
				return null;
			}
			return cw.ShowDialog();

		}

		public void Exit()
		{

			// Close main window

			Application.Current.MainWindow.Close();

		}

		// Internal properties

		internal bool UseChildWindow { get; set; }
		internal bool ShowChildItemIcon { get; set; }
		internal bool ShowChildItemTitle { get; set; }

		// Event handlers

		void MainWindow_AppClosing(object sender, EventArgs e)
		{
			this.ShutDown();
		}

		void CurrentDomain_FirstChanceException(object sender, System.Runtime.ExceptionServices.FirstChanceExceptionEventArgs e)
		{
			System.Diagnostics.Debug.Print(e.Exception.Message);
		}

		// Private methods

		private void Initialize()
		{

			MainLayoutModel = new MainLayoutModel();
			UseChildWindow = true;
			ShowChildItemIcon = true;
			ShowChildItemTitle = true;

			// Get main window

			if (_pMainWindow == null)
			{
				Application app = Application.Current;
				Window mw = app.MainWindow;
				if (!(mw is IMainWindow))
					throw new Exception("IMainWindow not found");
				_pMainWindow = (IMainWindow)mw;
			}
			ApplicationName = _pMainWindow.ApplicationName;
			RuntimeVersion = _pMainWindow.RuntimeVersion;
			UpdateLocation = _pMainWindow.UpdateLocation;
			WindowIcon = _pMainWindow.WindowIcon;
			_pMainWindow.AppClosing += MainWindow_AppClosing;
			AppDomain.CurrentDomain.FirstChanceException += CurrentDomain_FirstChanceException;

			// Process configuration

			if (string.IsNullOrEmpty(_pMainWindow.Configuration))
				throw new Exception("No configuration set");
			_pStatus = ConfigHelper.LoadConfiguration(_pMainWindow);

			// Done

			if (this.AppReady == null)
				return;
			this.AppReady(this, new EventArgs());

		}

		private void ShutDown()
		{

			// Shut down child windows and items

			foreach (ChildWindowModel cwm in _Windows)
				cwm.ShutDown();
			foreach (ChildItem ci in MainLayoutModel.ChildItems)
				ci.ShutDown();

		}

	}
}
