﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Interop;
using System.Windows.Interactivity;
using System.Windows.Media;
using System.Windows.Threading;

namespace Extensibility
{

	public class ChildWindowModel : Behavior<ChildWindow>, INotifyPropertyChanged
	{

		private const int TimerIntervalSeconds = 5;
		private const bool RefreshTimer = true;

		private double _dOpacity = 1;
		private UserControl _ucLayout;

		private IWindowContent _pWC;
		private bool _bClosable = true;
		private bool _bShutDown = false;
		private DispatcherTimer _Timer;

		// Bindable properties

		public double Opacity
		{
			get { return _dOpacity; }
			set
			{
				_dOpacity = value;
				this.RaisePropertyChanged();
			}
		}

		public UserControl ChildLayout
		{
			get { return _ucLayout; }
			set
			{
				_ucLayout = value;
				this.RaisePropertyChanged();
			}
		}

		// INotifyPropertyChanged

		public event PropertyChangedEventHandler PropertyChanged;

		// Behavior

		protected override void OnAttached()
		{
			base.OnAttached();
			AssociatedObject.Closing += AssociatedObject_Closing;
			AssociatedObject.DataContext = this;
		}

		// Public methods

		public void SetContent(UserControl MyView, object MyViewModel)
		{
			if (AssociatedObject == null)
				return; // This should never happen
			this.Init(MyView, MyViewModel);
		}

		public void ShutDown()
		{
			if (_bShutDown)
				return;
			if (_pWC != null)
				_pWC.ShutDown();
			ChildLayout = null;
			_pWC = null;
			_bShutDown = true;
		}

		// Event handlers

		void DC_OK(object sender, EventArgs e)
		{
			_bClosable = true;
			AssociatedObject.DialogResult = true;
		}

		void DC_Cancel(object sender, EventArgs e)
		{
			_bClosable = true;
			AssociatedObject.DialogResult = false;
		}

		private void WC_Show(object sender, EventArgs e)
		{
			AssociatedObject.Show();
		}

		private void WC_Hide(object sender, EventArgs e)
		{
			AssociatedObject.Hide();
		}

		void Timer_Tick(object sender, EventArgs e)
		{
			try
			{
				this.Refresh();
			}
			catch (Exception ex)
			{
				Debug.Print("ChildWindowModel Timer_Tick Error: " + ex.Message);
			}
		}

		void AssociatedObject_Closing(object sender, CancelEventArgs e)
		{
			if (this.IsModal())
			{
				if (_bClosable)
					this.ShutDown();
				else
					e.Cancel = true;
				return;
			}
			AssociatedObject.Hide();
			e.Cancel = true;
		}

		// Private methods

		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private void Init(UserControl MyView, object MyViewModel)
		{

			AssociatedObject.Icon = MapApplication.Current.WindowIcon;

			// Search resources for Window properties

			ResourceDictionary rd = MyView.Resources;
			if (rd.Contains("WindowStyle"))
			{
				object val = rd["WindowStyle"];
				if (val is WindowStyle)
					AssociatedObject.WindowStyle = (WindowStyle)val;
			}
			if (rd.Contains("Icon"))
			{
				Image img = rd["Icon"] as Image;
				if (img != null)
					AssociatedObject.Icon = img.Source;
			}
			if (rd.Contains("ResizeMode"))
			{
				object val = rd["ResizeMode"];
				if (val is ResizeMode)
					AssociatedObject.ResizeMode = (ResizeMode)val;
			}
			if (rd.Contains("SizeToContent"))
			{
				object val = rd["SizeToContent"];
				if (val is SizeToContent)
					AssociatedObject.SizeToContent = (SizeToContent)val;
			}
			if (rd.Contains("WindowSize"))
			{
				object val = rd["WindowSize"];
				if (val is Size)
				{
					Size sz = (Size)val;
					AssociatedObject.Height = sz.Height;
					AssociatedObject.Width = sz.Width;
				}
			}

			// Set up view and view model

			ChildLayout = MyView;
			ChildLayout.DataContext = MyViewModel;
			if (MyViewModel is IDialogContent)
			{
				IDialogContent pDC = (IDialogContent)MyViewModel;
				_bClosable = pDC.Closable;
				pDC.DialogOK += DC_OK;
				pDC.DialogCancel += DC_Cancel;
			}
			else if (MyViewModel is IWindowContent)
			{
				IWindowContent pWC = (IWindowContent)MyViewModel;
				pWC.ShowWindow += WC_Show;
				pWC.HideWindow += WC_Hide;
				_pWC = pWC;
			}

			// Start refresh timer

			bool bStartTimer = RefreshTimer;
			if (!bStartTimer)
				return;
			_Timer = new DispatcherTimer();
			_Timer.Interval = TimeSpan.FromSeconds(TimerIntervalSeconds);
			_Timer.Tick += Timer_Tick;
			_Timer.Start();

		}

		private void Refresh()
		{

			// This method helps eliminate visual artifacts created by other application windows.

			double dCurrentOpacity = Opacity;
			double dNewOpacity;
			if (dCurrentOpacity < 0.5)
				dNewOpacity = dCurrentOpacity + 0.01;
			else
				dNewOpacity = dCurrentOpacity - 0.01;
			Opacity = dNewOpacity;
			Opacity = dCurrentOpacity;

		}

		private bool IsModal()
		{
			return (bool)typeof(Window).GetField("_showingAsDialog", BindingFlags.Instance | BindingFlags.NonPublic).GetValue(AssociatedObject);
		}

	}

}
