﻿using System;
using System.Collections.Generic;
using System.IO;
using System.IO.Packaging;
using System.Linq;
using System.Printing;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Xps;
using System.Windows.Xps.Packaging;

using Transform = System.Windows.Media.Transform;

using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Mapping;
using Esri.ArcGISRuntime.UI;
using Esri.ArcGISRuntime.UI.Controls;

using Extensibility;
using NavAddin;

namespace LayoutAddin
{

	// LayoutHelper handles layout creation and presentation

	public class LayoutHelper
	{

		static bool bExportPDF = true;

		public UserControl LayoutControl { get { return _LayoutControl; } }

		private MapView _MainMapView;
		private Layout _Layout;
		private LayoutModel _Model;
		private UserControl _LayoutControl;
		private MapView _MapView;
		private List<MapOverlayExport> _Overlays;
		private Image _Image;

		// Constructor

		public LayoutHelper(MapView MainMapView)
		{
			_MainMapView = MainMapView;
		}

		// Public methods

		public void OpenLayout(Layout layout)
		{

			// Get layout, title, and page size

			_Model = null;
			_Layout = layout;
			_Overlays = null;
			_Image = null;
			if (_LayoutControl != null)
			{
				_LayoutControl.Loaded -= Layout_Loaded;
				_LayoutControl = null;
			}
			string sAssembly = ResourceHelper.ViewAssemblies["LayoutAddin"];
			string sLayout = layout.Name;
			_LayoutControl = ConfigHelper.CreateInstance(sAssembly, sLayout) as UserControl;
			if (_LayoutControl == null)
				return;
			_LayoutControl.Loaded += Layout_Loaded;

			// Get current scale and center point

			double dScale = _MainMapView.MapScale;
			Envelope env = _MainMapView.VisibleArea.Extent;
			MapPoint ptCenter = env.GetCenter();
			Viewpoint vpStart = new Viewpoint(ptCenter, dScale);

			// Create layout MapView

			_MapView = new MapView();
			_MapView.NavigationCompleted += MapView_NavigationCompleted;
			_MapView.BackgroundGrid = _MainMapView.BackgroundGrid;
			_MapView.Map = new Map();
			_MapView.Map.MaxScale = _MainMapView.Map.MaxScale;
			_MapView.Map.MinScale = _MainMapView.Map.MinScale;
			_MapView.Map.ReferenceScale = _MainMapView.Map.ReferenceScale;
			_MapView.Map.InitialViewpoint = vpStart;

			// Copy layers

			foreach (Layer Lyr in _MainMapView.Map.Basemap.BaseLayers)
				_MapView.Map.Basemap.BaseLayers.Add(this.CopyLayer(Lyr));
			foreach (Layer Lyr in _MainMapView.Map.OperationalLayers)
				_MapView.Map.OperationalLayers.Add(this.CopyLayer(Lyr));

			// Copy graphic overlays

			foreach (GraphicsOverlay go in _MainMapView.GraphicsOverlays)
			{
				GraphicsOverlay goCopy = new GraphicsOverlay();
				_MapView.GraphicsOverlays.Add(goCopy);
				foreach (Graphic gr in go.Graphics)
				{
					Graphic grCopy = new Graphic(gr.Geometry, gr.Symbol);
					goCopy.Graphics.Add(grCopy);
				}
			}

			// Export and add overlay items

			_Overlays = MapExportHelper.GetMapOverlays(_MainMapView);
			foreach (MapOverlayExport overlay in _Overlays)
			{
				Image img = overlay.OverlayImage;
				MapView.SetViewOverlayAnchor(img, overlay.Anchor);
				_MapView.Overlays.Items.Add(img);
			}

			// Open window

			_Model = new LayoutModel(this);
			MapApplication.Current.CreateChildWindow("LayoutAddin", "LayoutView", _Model);
			_Model.MapViewContainerContent = _MapView;
			_Model.Open(layout);
			_MapView.SetViewpointAsync(vpStart);

		}

		public string GetScaleText(LayoutProperty prop)
		{

			// If page unit is not specified, return scale value

			double dScale = _MapView.MapScale;
			string sPageUnit = "";
			if (prop.Parameters.ContainsKey("PageUnit"))
				sPageUnit = prop.Parameters["PageUnit"];
			if (string.IsNullOrEmpty(sPageUnit))
				return "1:" + dScale.ToString("N0");

			// Get distance units

			string sDistanceUnits = "";
			LinearUnit lu = null;
			if (prop.Parameters.ContainsKey("DistanceUnit"))
			{
				sDistanceUnits = prop.Parameters["DistanceUnit"];
				lu = this.GetLinearUnit(sDistanceUnits);
			}
			else
			{
				lu = _MapView.SpatialReference.Unit as LinearUnit;
				if (lu != null)
					sDistanceUnits = this.GetUnitName(lu);
			}
			if (lu == null)
				return "";

			// Get conversion

			double dVal = 0;
			switch (sPageUnit)
			{
				case "Inch":
					dVal = lu.ConvertTo(LinearUnits.Inches, dScale);
					break;
				case "Point":
					dVal = lu.ConvertTo(LinearUnits.Inches, dScale) * 72;
					break;
				case "Millimeter":
					dVal = lu.ConvertTo(LinearUnits.Millimeters, dScale);
					break;
				case "Centimeter":
					dVal = lu.ConvertTo(LinearUnits.Millimeters, dScale) / 10;
					break;
			}
			if (dVal == 0)
				return "";
			string sVal = "One " + sPageUnit + " = " + dVal.ToString("N0") + " " + sDistanceUnits;
			return sVal;

		}

		public void ExportLayout()
		{

			this.GenerateImage();
			if (bExportPDF)
			{
				try
				{
					LocalPrintServer lps = new LocalPrintServer();
					PrintQueue pq = new PrintQueue(lps, "Microsoft Print to PDF");
					PrintDialog pd = new PrintDialog() { PrintQueue = pq };
					pd.PrintVisual(_LayoutControl, _Layout.DisplayName);
				}
				catch (Exception ex)
				{
					string sMsg = "Could not print to PDF:\n";
					sMsg += ex.Message;
					MapApplication.Current.Error(sMsg);
				}
			}
			else
			{

				// Export XPS file

				string sOutFile = DialogHelper.GetOutputFile("xps");
				if (string.IsNullOrEmpty(sOutFile))
					return;
				Transform t = _LayoutControl.LayoutTransform;
				_LayoutControl.LayoutTransform = null;
				Size sz = new Size(_Layout.PageWidth, _Layout.PageHeight);
				_LayoutControl.Measure(sz);
				_LayoutControl.Arrange(new Rect(sz));
				Uri uriPath = new Uri(sOutFile);
				Package pkg = Package.Open(uriPath.LocalPath, FileMode.Create);
				XpsDocument xpsDoc = new XpsDocument(pkg);
				XpsDocumentWriter xpsWriter = XpsDocument.CreateXpsDocumentWriter(xpsDoc);
				xpsWriter.Write(_LayoutControl);
				xpsDoc.Close();
				pkg.Close();
				_LayoutControl.LayoutTransform = t;
				MapApplication.Current.Message("File created.");

			}

		}

		public void PrintLayout()
		{
			this.GenerateImage();
			PrintDialog pd = new PrintDialog();
			if (pd.ShowDialog() != true)
				return;
			pd.PrintVisual(_LayoutControl, _Layout.DisplayName);
		}

		// Event handlers

		void Layout_Loaded(object sender, RoutedEventArgs e)
		{
			if (_Model == null)
				return; // This should never happen
			_Model.ZoomToFit();
		}

		void MapView_NavigationCompleted(object sender, EventArgs e)
		{

			if (_Model == null)
				return; // This should never happen
			_MapView.NavigationCompleted -= MapView_NavigationCompleted;

			// Update any scale properties

			foreach (LayoutProperty prop in _Layout.Properties)
			{
				if (prop.Type != LayoutPropertyType.Scale)
					continue;
				_Model.LayoutProperties[prop.Name] = this.GetScaleText(prop);
			}

		}

		// Private methods

		private string GetUnitName(LinearUnit lu)
		{
			if (!string.IsNullOrEmpty(lu.Name))
				return lu.Name;
			double dMeters = lu.ToMeters(1.0);
			if (dMeters == 1.0)
				return "Meters";
			if ((Math.Floor(dMeters * 10000) / 10000) == 0.3048)
				return "Feet";
			return "";
		}

		private LinearUnit GetLinearUnit(string sUnitName)
		{
			switch (sUnitName)
			{
				case "Feet":
					return LinearUnits.Feet;
				case "Inches":
					return LinearUnits.Inches;
				case "Kilometers":
					return LinearUnits.Kilometers;
				case "Meters":
					return LinearUnits.Meters;
				case "Miles":
					return LinearUnits.Miles;
				case "Millimeters":
					return LinearUnits.Millimeters;
				case "Nautical Miles":
					return LinearUnits.NauticalMiles;
				case "Yards":
					return LinearUnits.Yards;
			}
			return null;
		}

		private Layer CopyLayer(Layer lyrIn)
		{

			// 100.x BUG:  Layer.Clone does not copy DefinitionExpression

			Layer lyrOut = lyrIn.Clone();
			if (lyrOut is GroupLayer grpOut)
			{
				// This would occur if MapHelper.PreserveGroupLayers = true
				this.FixSublayers(grpOut, lyrIn as GroupLayer);
				return lyrOut;
			}
			if (!(lyrOut is FeatureLayer fLyrOut))
				return lyrOut;
			FeatureLayer fLyrIn = lyrIn as FeatureLayer;
			fLyrOut.DefinitionExpression = fLyrIn.DefinitionExpression;
			return lyrOut;

		}

		private void FixSublayers(GroupLayer grpOut, GroupLayer grpIn)
		{
			int iCount = grpOut.Layers.Count;
			for (int i = 0; i < iCount; i++)
			{
				Layer lyrSubOut = grpOut.Layers[i];
				Layer lyrSubIn = grpIn.Layers[i];
				if (lyrSubOut is GroupLayer grpSubOut)
				{
					this.FixSublayers(grpSubOut, lyrSubIn as GroupLayer);
					continue;
				}
				if (!(lyrSubOut is FeatureLayer fLyrSubOut))
					continue;
				FeatureLayer fLyrSubIn = lyrSubIn as FeatureLayer;
				fLyrSubOut.DefinitionExpression = fLyrSubIn.DefinitionExpression;
			}
		}

		private async void GenerateImage()
		{

			// 100.x WORKAROUND:  Replace MapView with Image

			if (_Image != null)
				return;

			double dWidth = _MapView.ActualWidth;
			double dHeight = _MapView.ActualHeight;
			ImageSource src = await MapExportHelper.GetMapImage(_MapView);
			_Image = new Image()
			{
				Height = dHeight,
				Width = dWidth,
				Source = src
			};
			_Model.MapViewContainerContent = _Image;

		}

	}
}
