﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;
using System.Xml.Linq;

using Esri.ArcGISRuntime.Geometry;

using Extensibility;

namespace LayoutAddin
{

	// LayoutModel acts as a view model for the layout view

	public class LayoutModel : INotifyPropertyChanged, IWindowContent
	{

		private Layout _Layout;
		private string _sTitle;
		private object _MapContent;
		private double _dPageWidth = 1056;
		private double _dPageHeight = 816;
		private LayoutHelper _lh;
		private double _dLayoutScale = 1;
		private double _dViewWidth;
		private double _dViewHeight;

		// Bindable properties

		public ICommand Command { get; private set; }

		public ICommand ViewWidthCommand { get; private set; }

		public ICommand ViewHeightCommand { get; private set; }

		public object LayoutContent 
		{ 
			get
			{
				if (_lh == null)
					return null;
				return _lh.LayoutControl;
			}
		}

		public object MapViewContainerContent 
		{ 
			get
			{
				return _MapContent;
			}
			set
			{
				_MapContent = value;
				this.RaisePropertyChanged();
			}
		}

		public double PageWidth
		{
			get { return _dPageWidth; }
			set
			{
				_dPageWidth = value;
				this.RaisePropertyChanged();
			}
		}

		public double PageHeight
		{
			get { return _dPageHeight; }
			set
			{
				_dPageHeight = value;
				this.RaisePropertyChanged();
			}
		}

		public double LayoutScale
		{
			get { return _dLayoutScale; }
			set
			{
				_dLayoutScale = value;
				this.RaisePropertyChanged();
			}
		}

		public DataHelper LayoutProperties { get; private set; }

		// INotifyPropertyChanged

		public event PropertyChangedEventHandler PropertyChanged;

		// IWindowContent

		public string Title
		{
			get { return _sTitle; }
			set
			{
				_sTitle = value;
				this.RaisePropertyChanged();
			}
		}

		public event EventHandler<EventArgs> ShowWindow;

		public event EventHandler<EventArgs> HideWindow { add { } remove { } } // Not used

		public void ShutDown() {}

		// Constructor

		public LayoutModel(LayoutHelper lh)
		{
			_lh = lh;
			LayoutProperties = new DataHelper();
			LayoutProperties.PropertyChanged += LayoutProperties_PropertyChanged;
			CommandHelper ch;
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.Execute);
			Command = ch;
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.ExecuteViewWidth);
			ViewWidthCommand = ch;
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.ExecuteViewHeight);
			ViewHeightCommand = ch;
		}

		// Public methods

		public void Open(Layout layout)
		{

			_Layout = layout;
			Title = _Layout.DisplayName;
			PageWidth = _Layout.PageWidth;
			PageHeight = _Layout.PageHeight;
			this.RaisePropertyChanged("LayoutContent");
			this.RaisePropertyChanged("MapViewContainerContent");

			// Set up properties

			LayoutProperties.Clear();
			Dictionary<string, string> Properties = new Dictionary<string, string>();
			foreach (LayoutProperty prop in _Layout.Properties)
			{
				string sName = prop.Name;
				string sVal = prop.Value;
				switch (prop.Type)
				{
					case LayoutPropertyType.Date:
						if (prop.Parameters.ContainsKey("FormatString"))
						{
							string sFormat = prop.Parameters["FormatString"];
							sVal = DateTime.Now.ToString(sFormat);
						}
						else
							sVal = DateTime.Now.Date.ToString();
						break;
					case LayoutPropertyType.Scale:
						sVal = _lh.GetScaleText(prop);
						break;
				}
				LayoutProperties[sName] = sVal;
			}

			// Show window

			if (this.ShowWindow == null)
				return;
			this.ShowWindow(this, new EventArgs());

		}

		public void ZoomToFit()
		{
			if ((PageWidth == 0) || (PageHeight == 0))
				return;
			double dW = _dViewWidth / PageWidth;
			double dH = _dViewHeight / PageHeight;
			double dScale = Math.Min(dW, dH) * 0.9;
			if (dScale == 0)
				return;
			LayoutScale = dScale;
		}

		// Event handlers

		void LayoutProperties_PropertyChanged(object sender, PropertyChangedEventArgs e)
		{
			this.RaisePropertyChanged("LayoutProperties");
		}

		// Private methods

		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private void Execute(object parameter)
		{
			string sParam = parameter as string;
			if (string.IsNullOrEmpty(sParam))
				return;
			switch (sParam)
			{
				case "ZoomIn":
					LayoutScale *= 1.5;
					break;
				case "ZoomOut":
					LayoutScale *= 0.75;
					break;
				case "ZoomFit":
					this.ZoomToFit();
					break;
				case "ZoomActual":
					LayoutScale = 1;
					break;
				case "Export":
					_lh.ExportLayout();
					break;
				case "Print":
					_lh.PrintLayout();
					break;
			}
		}

		private void ExecuteViewWidth(object parameter)
		{
			if (!(parameter is double))
				return;
			_dViewWidth = (double)parameter;
		}

		private void ExecuteViewHeight(object parameter)
		{
			if (!(parameter is double))
				return;
			_dViewHeight = (double)parameter;
		}

	}

}
