﻿using System;
using System.IO;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;
using System.Windows.Interactivity;
using System.Windows.Media;
using System.Windows.Media.Imaging;

using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.UI;
using Esri.ArcGISRuntime.UI.Controls;

using Extensibility;

namespace NavAddin
{

	// MapViewExt is a binding target and adds the following dependency properties:
	//    MapCoordinateText, String, OneWayToSource
	//    MapScale, double, OneWayToSource

	public class MapViewExt : Behavior<MapView>
	{

		private enum CoordinateUnits
		{
			MapUnits,
			DD,
			DMS
		}

		public static MapViewExt Current { get; private set; }

//		public event EventHandler<NotifyCanExecuteArgs> CanExportOrCopy;

		private bool _bMapReady = false;
		private CoordinateUnits _units = CoordinateUnits.MapUnits;

		// Behavior

		protected override void OnAttached()
		{
			base.OnAttached();
			Current = this;

			// Set event listeners

			AssociatedObject.DrawStatusChanged += AssociatedObject_DrawStatusChanged;
			AssociatedObject.NavigationCompleted += AssociatedObject_NavigationCompleted;
			AssociatedObject.MouseMove += AssociatedObject_MouseMove;

		}

		// MapCoordinateText property
		// This provides to the binding source the coordinate of the current mouse position

		public string MapCoordinateText
		{
			get { return (string)this.GetValue(MapCoordinateTextProperty); }
			set { this.SetValue(MapCoordinateTextProperty, value); }
		}

		public static readonly DependencyProperty MapCoordinateTextProperty =
			DependencyProperty.Register("MapCoordinateText", typeof(string), typeof(MapViewExt));

		// MapScale property
		// This provides to the binding source the current map scale as it changes

		public double MapScale
		{
			get { return (double)this.GetValue(MapScaleProperty); }
			set { this.SetValue(MapScaleProperty, value); }
		}

		public static readonly DependencyProperty MapScaleProperty =
			DependencyProperty.Register("MapScale", typeof(double), typeof(MapViewExt));

		// Public methods

		public async Task<bool> ExportMap(string sOutFile)
		{
			if (!_bMapReady)
				return false;
			if (string.IsNullOrEmpty(sOutFile))
				return false;
			try
			{
				ImageSource src = await MapExportHelper.GetMapImage(AssociatedObject);
				BmpBitmapEncoder bmpEncoder = new BmpBitmapEncoder();
				bmpEncoder.Frames.Add(BitmapFrame.Create(src as BitmapSource));
				using (Stream strOut = new FileStream(sOutFile, FileMode.Create))
				{
					bmpEncoder.Save(strOut);
				}
				MapApplication.Current.Message("File saved.");
				return true;
			}
			catch (Exception ex)
			{
				string sMsg = "Could not export map:\n" + ex.Message;
				MapApplication.Current.Error(sMsg);
				return false;
			}
		}

		public async Task<bool> CopyMap()
		{
			if (!_bMapReady)
				return false;
			ImageSource src = await MapExportHelper.GetMapImage(AssociatedObject);
			Clipboard.SetImage(src as BitmapSource);
			MapApplication.Current.Message("Map copied.");
			return true;
		}

		public void SetCoordinateUnits(string sParam)
		{
			switch (sParam)
			{
				case "MapUnits":
					_units = CoordinateUnits.MapUnits;
					break;
				case "DD":
					_units = CoordinateUnits.DD;
					break;
				case "DMS":
					_units = CoordinateUnits.DMS;
					break;
			}
		}

		// Event listeners

		void AssociatedObject_DrawStatusChanged(object sender, DrawStatusChangedEventArgs e)
		{
			if (e.Status != DrawStatus.Completed)
				return;
			this.CheckScale();
		}

		void AssociatedObject_NavigationCompleted(object sender, EventArgs e)
		{
			this.CheckScale();
		}

		private void AssociatedObject_MouseMove(object sender, MouseEventArgs e)
		{

			if (!_bMapReady)
				return;
			MapPoint pt = AssociatedObject.ScreenToLocation(e.GetPosition(AssociatedObject));
			if (pt == null)
				return;
			string sCoord = "";
			string sUnits = "";
			string sFormat = "N2";
			switch (_units)
			{
				case CoordinateUnits.MapUnits:
					sUnits = this.GetUnitText(AssociatedObject);
					if (sUnits == "Degrees")
						sFormat = "N6";
					sCoord = pt.X.ToString(sFormat) + "   " + pt.Y.ToString(sFormat) + "  " + sUnits;
					break;
				case CoordinateUnits.DD:
					sUnits = "Degrees";
					sFormat = "N6";
					MapPoint ptDD = GeometryEngine.Project(pt, SpatialReferences.Wgs84) as MapPoint;
					if (ptDD == null)
						return;
					sCoord = ptDD.X.ToString(sFormat) + "   " + ptDD.Y.ToString(sFormat) + "  " + sUnits;
					break;
				case CoordinateUnits.DMS:
					sCoord = this.GetDMS(pt);
					break;
			}
			if (string.IsNullOrEmpty(sCoord))
				return;
			MapCoordinateText = sCoord;

		}

		// Private methods

		private void CheckScale()
		{
			double dScale = AssociatedObject.MapScale;
			if (double.IsNaN(dScale))
				return;
			if (dScale <= 0)
				return;
			if (MapScale == dScale)
				return;
			MapScale = dScale;
			if (_bMapReady)
				return;
			_bMapReady = true;
		}

		private string GetUnitText(MapView mv)
		{

			SpatialReference sr = mv.SpatialReference;
			if (sr == null)
				return "";
			Unit unit = sr.Unit;
			if (unit == null)
				return "";
			if (!string.IsNullOrEmpty(unit.Name))
				return unit.Name;
			LinearUnit lu = unit as LinearUnit;
			if (lu == null)
				return "";
			double dMeters = lu.ToMeters(1.0);
			if (dMeters == 1.0)
				return "Meters";
			if ((Math.Floor(dMeters * 10000) / 10000) == 0.3048)
				return "Feet";
			return "";

		}

		private string GetDMS(MapPoint pt)
		{

			MapPoint ptDD = GeometryEngine.Project(pt, SpatialReferences.Wgs84) as MapPoint;
			if (ptDD == null)
				return "";
			return this.ToDMS(ptDD.X, true) + " " + this.ToDMS(ptDD.Y, false);

		}

		private string ToDMS(double d, bool bIsLongitude)
		{
			int iD = (int)Math.Floor(Math.Abs(d));
			string sDegrees = iD.ToString() + "°";
			double dMS = (Math.Abs(d) - (double)iD) * 60;
			int iM = (int)Math.Floor(Math.Abs(dMS));
			sDegrees += iM.ToString() + "'";
			double dS = (dMS - iM) * 60;
			sDegrees += dS.ToString("N3") + '"';
			if (bIsLongitude)
			{
				if (d < 0)
					sDegrees += "W";
				else
					sDegrees += "E";
			}
			else
			{
				if (d < 0)
					sDegrees += "S";
				else
					sDegrees += "N";
			}
			return sDegrees;
		}


	}

}
