﻿using System;
using System.Windows;
using System.Windows.Interactivity;

using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Mapping;
using Esri.ArcGISRuntime.UI;
using Esri.ArcGISRuntime.UI.Controls;

using Extensibility;

namespace NavAddin
{

	// NavViewExt is a binding target and adds the following dependency properties:
	//    NavScale, Double, TwoWay
	//    NavExtent, Envelope, TwoWay
	//    NavUnitsPerPixel, Double, OneWayToSource
	// NavViewExt also provides access to the MapView's LocationDisplay

	public class NavViewExt : Behavior<MapView>
	{

		public static NavViewExt Current { get; private set; }

		public LocationDisplay LocationDisplay
		{
			get
			{
				if (AssociatedObject == null)
					return null;
				return AssociatedObject.LocationDisplay;
			}
		}

		// Behavior

		protected override void OnAttached()
		{
			base.OnAttached();
			Current = this;

			// Set event listeners

			AssociatedObject.DrawStatusChanged += AssociatedObject_DrawStatusChanged;
			AssociatedObject.NavigationCompleted += AssociatedObject_NavigationCompleted;

		}

		// NavScale property
		// This allows the binding source to get and/or set the MapView scale

		public double NavScale
		{
			get { return (double)this.GetValue(NavScaleProperty); }
			set { this.SetValue(NavScaleProperty, value); }
		}

		public static readonly DependencyProperty NavScaleProperty =
			DependencyProperty.Register("NavScale", typeof(double), typeof(NavViewExt),
			new FrameworkPropertyMetadata(double.NaN, FrameworkPropertyMetadataOptions.AffectsRender,
				OnNavScaleChanged));

		private static void OnNavScaleChanged(DependencyObject obj, DependencyPropertyChangedEventArgs eventArgs)
		{
			NavViewExt mvb = (NavViewExt)obj;
			MapView mv = mvb.AssociatedObject;
			double dScale = (double)eventArgs.NewValue;
			if (double.IsNaN(dScale))
				return;
			if (dScale <= 0)
				return;
			if (dScale == mv.MapScale)
				return;
			mv.SetViewpointScaleAsync(dScale);
		}

		// NavExtent property
		// This allows the binding source to get and/or set the MapView extent

		public Envelope NavExtent
		{
			get { return (Envelope)this.GetValue(NavExtentProperty); }
			set { this.SetValue(NavExtentProperty, value); }
		}

		public static readonly DependencyProperty NavExtentProperty =
			DependencyProperty.Register("NavExtent", typeof(Envelope), typeof(NavViewExt),
			new FrameworkPropertyMetadata(null, FrameworkPropertyMetadataOptions.AffectsRender,
				OnNavExtentChanged));

		private static void OnNavExtentChanged(DependencyObject obj, DependencyPropertyChangedEventArgs eventArgs)
		{
			NavViewExt mvb = (NavViewExt)obj;
			MapView mv = mvb.AssociatedObject;
			Envelope envExtent = (Envelope)eventArgs.NewValue;
			if (envExtent == null)
				return;
			Envelope envCurrent = mv.VisibleArea.Extent;
			if (HistoryHelper.Equals(envExtent, envCurrent, mv.UnitsPerPixel))
				return;
			mv.SetViewpointAsync(new Viewpoint(envExtent));
		}

		// NavUnitsPerPixel property
		// This provides to the binding source the current MapView UnitsPerPixel

		public double NavUnitsPerPixel
		{
			get { return (double)this.GetValue(NavUnitsPerPixelProperty); }
			set { this.SetValue(NavUnitsPerPixelProperty, value); }
		}

		public static readonly DependencyProperty NavUnitsPerPixelProperty =
			DependencyProperty.Register("NavUnitsPerPixel", typeof(double), typeof(NavViewExt));

		// Event handlers

		void AssociatedObject_DrawStatusChanged(object sender, DrawStatusChangedEventArgs e)
		{
			if (e.Status != DrawStatus.Completed)
				return;
			double dScale = AssociatedObject.MapScale;
			if (double.IsNaN(dScale))
				return;
			if (dScale <= 0)
				return;
			this.NavigationCompleted();
		}

		void AssociatedObject_NavigationCompleted(object sender, EventArgs e)
		{
			this.NavigationCompleted();
		}

		// Private methods

		private void NavigationCompleted()
		{
			if (AssociatedObject.VisibleArea == null)
				return;
			this.SetValue(NavScaleProperty, AssociatedObject.MapScale);
			this.SetValue(NavExtentProperty, AssociatedObject.VisibleArea.Extent);
			this.SetValue(NavUnitsPerPixelProperty, AssociatedObject.UnitsPerPixel);
			NavModelExt.Current.NavigationCompleted();
		}

	}

}
