﻿using System;
using System.IO.Ports;
using System.Threading.Tasks;
using System.Windows.Threading;

using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Location;

namespace NavAddin
{

	// GPSLocationProvider acts as a location data source for GPS units connected via a serial port

	class GPSLocationProvider : LocationDataSource, IDisposable
	{

		private const int MyBaudRate = 4800;
		private const int MyDataBits = 8;
		private const Parity MyParity = Parity.None;
		private const StopBits MyStopBits = StopBits.One;
		private const int MyReadTimeout = 1000;

		public bool Initialized { get; private set; }
		public string SerialPortName { get; private set; }
		public new string Status { get; private set; }
		public double Longitude { get; private set; }
		public double Latitude { get; private set; }
		public double Velocity { get; private set; }
		public double Course { get; private set; }
		public double PDOP { get; private set; }
		public double HDOP { get; private set; }
		public int NumSat { get; private set; }

		private SerialPort _SerialPort;
		private DispatcherTimer _Timer;
		private SpatialReference _srGeo = SpatialReferences.Wgs84;
		private MapPoint _ptLast;

		public GPSLocationProvider()
		{
			Initialized = false;
			Longitude = double.NaN;
			Latitude = double.NaN;
			Status = "N/A";
			_Timer = new DispatcherTimer();
			_Timer.Interval = TimeSpan.FromSeconds(2);
			_Timer.Tick += _Timer_Tick;
			_Timer.IsEnabled = false;
		}

		public async Task<bool> InitializeAsync()
		{
			if (Initialized)
				return true;
			Task t = Task.Factory.StartNew(() =>
			{
				this.IntInitialize();
			});
			await t;
			return Initialized;
		}

		// Overrides

		new public event EventHandler<Location> LocationChanged;

		protected override Task OnStartAsync()
		{
			return Task.Factory.StartNew(() =>
			{
				if (Initialized)
				{
					_SerialPort.Open();
					_Timer.Start();
				}
			});
		}

		protected override Task OnStopAsync()
		{
			return Task.Factory.StartNew(() =>
			{
				if (Initialized)
				{
					_Timer.Stop();
					_SerialPort.Close();
				}
			});
		}

		// IDisposable

		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		protected virtual void Dispose(bool disposing)
		{
			if (disposing)
			{
				// Dispose managed resources
				if (_SerialPort != null)
				{
					if (_SerialPort.IsOpen)
						_SerialPort.Close();
					_SerialPort.Dispose();
				}
			}
			// No native resources
		}

		// Private methods

		private void UpdateLocation()
		{

			// 100.1 WORKAROUND:  Setting velocity to 0 can lead to
			// initial location display problems.

			double dV = 0.01;
			if (Velocity > 0)
				dV = Velocity;
			double dHA = 0;
			if (!double.IsNaN(HDOP))
				dHA = HDOP;
			double dC = 0;
			if (!double.IsNaN(Course))
				dC = Course;

			// Get location

			MapPoint pt = new MapPoint(Longitude, Latitude, _srGeo);
			bool bUnknown = (double.IsNaN(pt.X) || (double.IsNaN(pt.Y)));
			if (!bUnknown)
				_ptLast = pt;
			Location locNew = new Location(pt, dHA, dV, dC, bUnknown);

			// Send point to display

			bool bSendToDisplay = true;
			Location locDisplay = locNew;
			if (bUnknown)
			{

				// Send previous coordinate with isLastKnown tag

				if (_ptLast == null)
					bSendToDisplay = false;
				else
					locDisplay = new Location(_ptLast, dHA, dV, dC, bUnknown);

			}
			if (bSendToDisplay)
				base.UpdateLocation(locDisplay);

			// Raise location changed

			if (this.LocationChanged == null)
				return;
			this.LocationChanged(this, locNew);

		}

		private void IntInitialize()
		{

			// Find GPS port

			string sPort = "";
			string[] PortNames = SerialPort.GetPortNames();
			foreach (string sP in PortNames)
			{
				if (!this.TestPort(sP))
					continue;
				sPort = sP;
				break;
			}
			if (sPort == "")
				return;

			// Set up serial port

			_SerialPort = new SerialPort(sPort);
			_SerialPort.BaudRate = MyBaudRate;
			_SerialPort.DataBits = MyDataBits;
			_SerialPort.Parity = MyParity;
			_SerialPort.StopBits = MyStopBits;
			_SerialPort.ReadTimeout = MyReadTimeout;
			SerialPortName = sPort;
			Initialized = true;

		}

		private bool TestPort(string sPort)
		{

			bool bSuccess = false;

			// Try to open serial port

			SerialPort SerialPort1 = new SerialPort(sPort);
			SerialPort1.BaudRate = MyBaudRate;
			SerialPort1.DataBits = MyDataBits;
			SerialPort1.Parity = MyParity;
			SerialPort1.StopBits = MyStopBits;
			SerialPort1.ReadTimeout = MyReadTimeout;
			bool bError = false;
			try
			{
				SerialPort1.Open();
			}
			catch
			{
				bError = true;
			}
			if (bError || (!SerialPort1.IsOpen))
			{
				SerialPort1.Close();
				return false;
			}

			// Try to read GPRMC sentence

			try
			{
				SerialPort1.DiscardInBuffer();
				string sSentence = SerialPort1.ReadTo("GPRMC");
				bSuccess = true;
			}
			catch
			{
				bSuccess = false;
			}
			SerialPort1.Close();
			return bSuccess;

		}

		void _Timer_Tick(object sender, EventArgs e)
		{
			Task t = Task.Factory.StartNew(() =>
			{
				this.ReadAndProcess();
			});
		}

		private void ReadAndProcess()
		{

			// TO DO: Get velocity and course
			// Read GSA for PDOP and GSV for # satellites

			double dOldLon = Longitude;
			double dOldLat = Latitude;

			// Collect GPRMC sentence for lat, lon, velocity, and course

			string sGPRMC = this.GetSentence("GPRMC");
			if (sGPRMC == "")
				return;
			this.ParseGPRMC(sGPRMC);
			if ((dOldLon == Longitude) || (dOldLat == Latitude))
				return;

			// Collect GPGSA for PDOP, HDOP, and # satellites

			string sGPGSA = this.GetSentence("GPGSA");
			this.ParseGPGSA(sGPGSA);
			this.UpdateLocation();

		}

		private string GetSentence(string sPrefix)
		{
			string sSentence = "";
			try
			{
				_SerialPort.DiscardInBuffer();
				sSentence = _SerialPort.ReadTo(sPrefix);
				sSentence = _SerialPort.ReadTo("$");
				sSentence = "$" + sPrefix + sSentence;
			}
			catch
			{
				sSentence = "";
			}
			return sSentence;
		}

		private void ParseGPRMC(string sGPRMC)
		{

			string sTok;
			string[] TokList = sGPRMC.Split(new char[] { ',' });
			double dVal, dDeg, dMin;
			Latitude = double.NaN;
			Longitude = double.NaN;
			Velocity = double.NaN;
			Course = double.NaN;

			// Status

			sTok = TokList[1];
			string sStat = sTok.Substring(0, 2) + ":" + sTok.Substring(2, 2) + ":" + sTok.Substring(4, 2);
			sTok = TokList[2];
			if (sTok == "A")
				sStat += "; Valid";
			else
				sStat += "; Warning";
			Status = sStat;

			// Latitude

			sTok = TokList[3];
			if (double.TryParse(sTok, out dVal))
			{
				dDeg = Math.Floor(dVal / 100);
				dMin = dVal - (dDeg * 100);
				Latitude = dDeg + (dMin / 60);
				sTok = TokList[4];
				if (sTok == "S")
					Latitude = -Latitude;
			}

			// Longitide

			sTok = TokList[5];
			if (double.TryParse(sTok, out dVal))
			{
				dDeg = Math.Floor(dVal / 100);
				dMin = dVal - (dDeg * 100);
				Longitude = dDeg + (dMin / 60);
				sTok = TokList[6];
				if (sTok == "W")
					Longitude = -Longitude;
			}

			// Speed

			sTok = TokList[7];
			if (double.TryParse(sTok, out dVal))
				Velocity = LinearUnits.NauticalMiles.ToMeters(dVal) / 3600;

			// Track

			sTok = TokList[8];
			if (double.TryParse(sTok, out dVal))
				Course = dVal;

		}

		private void ParseGPGSA(string sGPGSA)
		{

			string sTok;
			string[] TokList = sGPGSA.Split(new char[] { ',' });
			double dVal;
			PDOP = double.NaN;
			HDOP = double.NaN;
			NumSat = 0;

			// PDOP

			sTok = TokList[15];
			if (double.TryParse(sTok, out dVal))
				PDOP = dVal;

			// HDOP

			sTok = TokList[16];
			if (double.TryParse(sTok, out dVal))
				HDOP = dVal;


			// # Satellites

			for (int i = 3; i < 15; i++)
			{
				sTok = TokList[i];
				if (string.IsNullOrEmpty(sTok))
					break;
				NumSat++;
			}

		}

	}
}
