﻿using System;
using System.Collections.Generic;

using Esri.ArcGISRuntime.Geometry;

using Extensibility;

namespace NavAddin
{

	// HistoryHelper helps manage and navigate the extent history list

	public class HistoryHelper
	{

		private List<Envelope> _ExtentHistory = new List<Envelope>();
		private int _iCurrentExtentIndex = -1;
		private bool _bCanZoomPrevious = false;
		private bool _bCanZoomNext = false;
		private Envelope _envFullExtent;

		// Public events

		public event EventHandler<NotifyCanExecuteArgs> CanZoomPrevious;
		public event EventHandler<NotifyCanExecuteArgs> CanZoomNext;

		// Public methods

		public bool AddExtent(Envelope envNew, double dMargin)
		{

			if (envNew == null)
				return false;
			if (_iCurrentExtentIndex == -1)
			{
				_ExtentHistory.Add(envNew);
				_iCurrentExtentIndex = 0;
				_envFullExtent = envNew;
				UserSettingsHelper.SetInitialView();
				return true;
			}
			Envelope env = _ExtentHistory[_iCurrentExtentIndex];
			if (HistoryHelper.Equals(env, envNew, dMargin))
				return false;
			_iCurrentExtentIndex += 1;
			if (_ExtentHistory.Count - _iCurrentExtentIndex > 0)
				_ExtentHistory.RemoveRange(_iCurrentExtentIndex, (_ExtentHistory.Count - _iCurrentExtentIndex));
			if (_bCanZoomNext)
			{
				_bCanZoomNext = false;
				this.RaiseCanZoomNext();
			}
			env = envNew;
			envNew = new Envelope(env.XMin, env.YMin, env.XMax, env.YMax, env.SpatialReference);
			_ExtentHistory.Add(envNew);
			if (!_bCanZoomPrevious)
			{
				_bCanZoomPrevious = true;
				this.RaiseCanZoomPrevious();
			}
			return true;

		}

		public Envelope GetPrevious()
		{
			if (!_bCanZoomPrevious)
				return null;
			if (_iCurrentExtentIndex == 0)
				return null;
			_iCurrentExtentIndex -= 1;
			if (!_bCanZoomNext)
			{
				_bCanZoomNext = true;
				this.RaiseCanZoomNext();
			}
			if (_iCurrentExtentIndex == 0)
			{
				if (_bCanZoomPrevious)
				{
					_bCanZoomPrevious = false;
					this.RaiseCanZoomPrevious();
				}
			}
			Envelope env = _ExtentHistory[_iCurrentExtentIndex];
			Envelope envNew = new Envelope(env.XMin, env.YMin, env.XMax, env.YMax, env.SpatialReference);
			return envNew;
		}

		public Envelope GetNext()
		{
			if (!_bCanZoomNext)
				return null;
			if (_iCurrentExtentIndex == _ExtentHistory.Count - 1)
				return null;
			_iCurrentExtentIndex += 1;
			if (!_bCanZoomPrevious)
			{
				_bCanZoomPrevious = true;
				this.RaiseCanZoomPrevious();
			}
			if (_iCurrentExtentIndex == _ExtentHistory.Count - 1)
			{
				if (_bCanZoomNext)
				{
					_bCanZoomNext = false;
					this.RaiseCanZoomNext();
				}
			}
			Envelope env = _ExtentHistory[_iCurrentExtentIndex];
			Envelope envNew = new Envelope(env.XMin, env.YMin, env.XMax, env.YMax, env.SpatialReference);
			return envNew;
		}

		public Envelope GetFull()
		{
			Envelope env = _envFullExtent;
			if (env == null)
				return null;
			Envelope envNew = new Envelope(env.XMin, env.YMin, env.XMax, env.YMax, env.SpatialReference);
			return envNew;
		}

		public void Reset()
		{
			_ExtentHistory.Clear();
			_iCurrentExtentIndex = -1;
			_envFullExtent = null;
		}

		public static bool Equals(Envelope env1, Envelope env2, double dMargin)
		{
			if ((env1.SpatialReference != null) && (env2.SpatialReference != null))
			{
				if (!env1.SpatialReference.IsEqual(env2.SpatialReference))
					return false;
			}
			double[] Vals1 = { env1.XMin, env1.YMin, env1.XMax, env1.YMax };
			double[] Vals2 = { env2.XMin, env2.YMin, env2.XMax, env2.YMax };
			for (int i = 0; i < 4; i++)
			{
				double dVal1 = Vals1[i];
				double dVal2 = Vals2[i];
				if (Math.Abs(dVal2 - dVal1) > dMargin)
					return false;
			}
			return true;
		}

		// Private methods

		private void RaiseCanZoomPrevious()
		{
			if (this.CanZoomPrevious == null)
				return;
			NotifyCanExecuteArgs e = new NotifyCanExecuteArgs();
			e.CanExecute = _bCanZoomPrevious;
			this.CanZoomPrevious(this, e);
		}

		private void RaiseCanZoomNext()
		{
			if (this.CanZoomNext == null)
				return;
			NotifyCanExecuteArgs e = new NotifyCanExecuteArgs();
			e.CanExecute = _bCanZoomNext;
			this.CanZoomNext(this, e);
		}



	}

}
