﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Xaml;

using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.UI;
using Esri.ArcGISRuntime.UI.Controls;

namespace NavAddin
{

	public struct MapOverlayExport
	{
		public Image OverlayImage;
		public MapPoint Anchor;
		public MapPoint TopLeft;
	}

	public static class MapExportHelper
	{

		// Export bitmap from map with XAML graphics overlays

		public static async Task<ImageSource> GetMapImage(MapView mv)
		{

			RuntimeImage ri = await mv.ExportImageAsync();
			ImageSource src = await ri.ToImageSourceAsync();
			if (mv.Overlays.Items.Count == 0)
				return src; // No XAML overlays

			// Create canvas

			double dWidth = mv.ActualWidth;
			double dHeight = mv.ActualHeight;
			Rect rMap = new Rect(0, 0, dWidth, dHeight);
			Size szMap = new Size(dWidth, dHeight);
			Canvas c = new Canvas();

			// Add map image

			Image imgMap = new Image()
			{
				Height = dHeight,
				Width = dWidth,
				Source = src
			};
			imgMap.Measure(szMap);
			imgMap.Arrange(rMap);
			imgMap.UpdateLayout();
			Canvas.SetTop(imgMap, 0);
			Canvas.SetLeft(imgMap, 0);
			c.Children.Add(imgMap);

			// Add map overlays

			List<MapOverlayExport> Overlays = GetMapOverlays(mv);
			foreach (MapOverlayExport overlay in Overlays)
			{

				// Get Image and location

				Image img = overlay.OverlayImage;
				MapPoint ptMap = overlay.TopLeft;
				Point ptScreen = mv.LocationToScreen(ptMap);

				// Create and place image of element

				Canvas.SetTop(img, ptScreen.Y);
				Canvas.SetLeft(img, ptScreen.X);
				c.Children.Add(img);
				img.UpdateLayout();

			}
			c.Measure(szMap);
			c.Arrange(rMap);
			c.UpdateLayout();

			// Create RenderTargetBitmap

			RenderTargetBitmap rtBitmap = new RenderTargetBitmap(
				(int)dWidth, (int)dHeight, 96d, 96d, PixelFormats.Pbgra32);
			rtBitmap.Render(c);
			return rtBitmap;

		}

		public static List<MapOverlayExport> GetMapOverlays(MapView mv)
		{

			List<MapOverlayExport> Overlays = new List<MapOverlayExport>();
			foreach (object obj in mv.Overlays.Items)
			{

				// Get element and location

				if (!(obj is FrameworkElement elem))
				{ 
					Debug.Print("MapExportHelper: Non-FrameworkElement encountered.");
					continue;
				}
				double dW = elem.ActualWidth;
				double dH = elem.ActualHeight;
				if ((dH == 0) || (dW == 0))
				{
					Debug.Print("MapExportHelper: Unsupported FrameworkElement encountered.");
					continue;
				}

				// Remove RenderTransform and RenderTransformOrigin

				Transform tRender = elem.RenderTransform;
				Point ptOrigin = elem.RenderTransformOrigin;
				elem.RenderTransform = null;
				elem.RenderTransformOrigin = new Point(0,0);
				elem.Measure(new Size(dW, dH));
				elem.Arrange(new Rect(0, 0, dW, dH));
				elem.UpdateLayout();

				// Create image of element

				ImageSource src = null;
				if (elem is Image imgSrc)
					src = imgSrc.Source;
				else
				{
					RenderTargetBitmap bmp = new RenderTargetBitmap(
						(int)dW, (int)dH, 96d, 96d, PixelFormats.Pbgra32);
					bmp.Render(elem);
					src = bmp;
				}
				Image img = new Image()
				{
					Height = dH,
					Width = dW,
					Source = src,
					HorizontalAlignment = elem.HorizontalAlignment,
					VerticalAlignment = elem.VerticalAlignment,
					RenderTransform = tRender,
					RenderTransformOrigin = ptOrigin
				};

				// Restore RenderTransform and RenderTransformOrigin

				elem.RenderTransform = tRender;
				elem.RenderTransformOrigin = ptOrigin;

				// Find top left location in map coordinates

				MapPoint ptMap = MapView.GetViewOverlayAnchor(elem);
				Point ptScreen = mv.LocationToScreen(ptMap);
				double dY = 0;
				double dX = 0;
				switch (elem.VerticalAlignment)
				{
					case VerticalAlignment.Center:
					case VerticalAlignment.Stretch:
						dY = -dH / 2;
						break;
					case VerticalAlignment.Top:
						dY = -dH;
						break;
				}
				switch (elem.HorizontalAlignment)
				{
					case HorizontalAlignment.Center:
					case HorizontalAlignment.Stretch:
						dX = -dW / 2;
						break;
					case HorizontalAlignment.Left:
						dX = -dW;
						break;
				}
				Point ptTopLeftScreen = new Point(ptScreen.X + dX, ptScreen.Y + dY);
				MapPoint ptTopLeftMap = mv.ScreenToLocation(ptTopLeftScreen);

				// Add exported overlay to list

				Overlays.Add(new MapOverlayExport()
				{
					OverlayImage = img,
					Anchor = ptMap,
					TopLeft = ptTopLeftMap
				});

			}

			return Overlays;

		}

	}
}
