﻿using System;
using System.Collections.Generic;
using System.IO;
using System.IO.Compression;
using System.Linq;
using System.Threading.Tasks;
using System.Xml.Linq;
using System.Web.Script.Serialization;

using Color = System.Drawing.Color;
using ColorConverter = System.Windows.Media.ColorConverter;
using MessageBox = System.Windows.MessageBox;

using Esri.ArcGISRuntime.Data;
using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Mapping;
using Esri.ArcGISRuntime.UI;

using Extensibility;

namespace NavAddin
{

	// MapHelper can either create a simple online map (for testing)
	// or create a map containing supported layer types

	public static class MapHelper
	{

		// 100.5/100.6 BUG: GroupLayers have bugs associated with them,
		// so the workaround is to imitate them in the TOC instead.
		// 100.7: FIXED - set PreserveGroupLayers to true

		public static readonly bool PreserveGroupLayers = true;

		private class TocEntry
		{
			public string ID = Guid.NewGuid().ToString();
			public TocDataItem Item;
			public int ExpansionLevel = -1;
			public ILayerContent Layer;
		}

		private static List<TocEntry> _BaseLayers;
		private static List<TocEntry> _OperationalLayers;
		private static Dictionary<string, Dictionary<string, Layer>> _MobileMapLayers;

		public static void OpenSimpleTestMap()
		{
			MapModelExt.Current.BaseLayers.Clear();
			MapModelExt.Current.OperationalLayers.Clear();
			Map MyMap = new Map(Basemap.CreateStreets());
			AppViewExt.Current.SetMap(MyMap);
			MapModelExt.Current.NotifyMapLoaded();
		}

		public static async void OpenWebMap(XElement xMap)
		{
			MapApplication.Current.Status("Initializing map...", true);
			try
			{
				await MapHelper.ProcessWebMap(xMap);
			}
			catch (Exception ex)
			{
				string sMsg = "Error loading map:\n" + ex.Message;
				MapApplication.Current.Error(sMsg);
				return;
			}
			MapApplication.Current.Status("OK", false);
		}

		public static async void OpenBasicMapAsync(XElement xMap)
		{
			MapApplication.Current.Status("Initializing map...", true);
			try
			{
				await MapHelper.ProcessBasicMap(xMap);
			}
			catch (Exception ex)
			{
				string sMsg = "Error loading map:\n" + ex.Message;
				MapApplication.Current.Error(sMsg);
				return;
			}
			MapApplication.Current.Status("OK", false);
		}

		public static async void OpenMobileMapPackage(string sPath)
		{
			MapApplication.Current.Status("Initializing map...", true);
			try
			{
				XElement xMap = new XElement("Map");
				xMap.SetAttributeValue("Type", "BasicMap");
				XElement xLayer = new XElement("MobileMapPackageLayer");
				xLayer.SetAttributeValue("Path", sPath);
				xLayer.SetAttributeValue("MapIndex", "0");
				xLayer.SetAttributeValue("AddBasemap", "True");
				xLayer.SetAttributeValue("IsExpanded", "True");
				xLayer.SetAttributeValue("ExpansionLevel", "0");
				xMap.Add(xLayer);
				await MapHelper.ProcessBasicMap(xMap);
			}
			catch (Exception ex)
			{
				string sMsg = "Error loading map:\n" + ex.Message;
				MapApplication.Current.Error(sMsg);
				return;
			}
			MapApplication.Current.Status("OK", false);
		}

		// Private methods

		private static async Task ProcessWebMap(XElement xMap)
		{

			// Set background

			Color colorBackground = Color.White;
			XAttribute xBackground = xMap.Attribute("Background");
			if (xBackground != null)
			{
				System.Windows.Media.Color colWPF = (System.Windows.Media.Color)ColorConverter.ConvertFromString(xBackground.Value);
				colorBackground = ResourceHelper.ConvertColor(colWPF);
			}
			BackgroundGrid bg = new BackgroundGrid()
			{
				Color = colorBackground,
				GridLineColor = Color.Transparent
			};
			AppViewExt.Current.MapBackground = bg;

			// Create and load map

			XAttribute xUrl = xMap.Attribute("Url");
			if (xUrl == null)
				return;
			Uri uri = new Uri(xUrl.Value);
			Map MyMap = new Map(uri);
			AppViewExt.Current.SetMap(MyMap);

			// Initialize TOC

			await MyMap.LoadAsync();
			MapModelExt.Current.BaseLayers.Clear();
			MapModelExt.Current.OperationalLayers.Clear();
			foreach (Layer lyr in MyMap.Basemap.BaseLayers)
			{
				TocDataItem tocItem = new TocDataItem(!PreserveGroupLayers);
				await lyr.LoadAsync();
				await tocItem.SetLayer(lyr);
				MapModelExt.Current.BaseLayers.Add(tocItem);
			}
			foreach (Layer lyr in MyMap.OperationalLayers)
			{
				TocDataItem tocItem = new TocDataItem(!PreserveGroupLayers);
				await lyr.LoadAsync();
				await tocItem.SetLayer(lyr);
				MapModelExt.Current.BaseLayers.Add(tocItem);
			}
			MapModelExt.Current.NotifyMapLoaded();

		}

		private static async Task ProcessBasicMap(XElement xMap)
		{

			Map MyMap = new Map();

			// Map settings

			Color colorBackground = Color.White;
			XAttribute xBackground = xMap.Attribute("Background");
			if (xBackground != null)
			{
				System.Windows.Media.Color colWPF = (System.Windows.Media.Color)ColorConverter.ConvertFromString(xBackground.Value);
				colorBackground = ResourceHelper.ConvertColor(colWPF);
			}
			BackgroundGrid bg = new BackgroundGrid()
			{
				Color = colorBackground,
				GridLineColor = Color.Transparent
			};
			AppViewExt.Current.MapBackground = bg;
			XAttribute xMaxScale = xMap.Attribute("MaxScale");
			if (xMaxScale != null)
				MyMap.MaxScale = Convert.ToDouble(xMaxScale.Value);
			XAttribute xMinScale = xMap.Attribute("MinScale");
			if (xMinScale != null)
				MyMap.MinScale = Convert.ToDouble(xMinScale.Value);
			XAttribute xRefScale = xMap.Attribute("RefScale");
			if (xRefScale != null)
				MyMap.ReferenceScale = Convert.ToDouble(xRefScale.Value);
			bool bRotationEnabled = false;
			XAttribute xRotation = xMap.Attribute("RotationEnabled");
			if (xRotation != null)
				bRotationEnabled = Convert.ToBoolean(xRotation.Value);
			if (!bRotationEnabled)
			{
				MapViewInteractionOptions opt = new MapViewInteractionOptions();
				opt.IsRotateEnabled = false;
				AppViewExt.Current.InteractionOptions = opt;
			}
			XAttribute xExtent = xMap.Attribute("Extent");

			// Map layers
			// 100.5 QUIRK:  Strange results unless TOC is initialized after map is loaded
			// TO DO?: Populate map layers first and then build TOC

			SpatialReference srExtent = null;
			_BaseLayers = new List<TocEntry>();
			_OperationalLayers = new List<TocEntry>();
			_MobileMapLayers = new Dictionary<string, Dictionary<string, Layer>>();
			foreach (XElement xLayer in xMap.Elements())
			{
				SpatialReference sr = null;
				switch (xLayer.Name.LocalName)
				{
					case "TileCacheLayer":
						sr = await MapHelper.AddTileCacheLayer(MyMap, xLayer);
						break;
					case "MobileMapPackageLayer":
						sr = await MapHelper.AddMobileMapPackageLayer(MyMap, xLayer);
						break;
					case "WebBaseImageLayer":
						sr = await MapHelper.AddMapImageLayer(MyMap, xLayer, true);
						break;
					case "WebMapImageLayer":
						sr = await MapHelper.AddMapImageLayer(MyMap, xLayer, false);
						break;
					case "WebFeatureLayer":
						sr = await MapHelper.AddWebFeatureLayer(MyMap, xLayer);
						break;
				}
				if ((srExtent == null) && (sr != null))
					srExtent = sr;
			}
			if (srExtent == null)
			{
				MapApplication.Current.Error("No map layers.");
				return;
			}
			if (xExtent != null)
			{
				string[] Tok = xExtent.Value.Split(new char[] { ',', ' ' }, StringSplitOptions.RemoveEmptyEntries);
				Envelope envExtent = new Envelope(Convert.ToDouble(Tok[0]), Convert.ToDouble(Tok[1]), Convert.ToDouble(Tok[2]), Convert.ToDouble(Tok[3]), srExtent);
				MyMap.InitialViewpoint = new Viewpoint(envExtent);
			}

			// Assign map

			AppViewExt.Current.SetMap(MyMap);

			// Initialize TOC

			await MyMap.LoadAsync();
			MapModelExt.Current.BaseLayers.Clear();
			MapModelExt.Current.OperationalLayers.Clear();
			foreach (TocEntry entry in _BaseLayers)
			{
				await entry.Item.SetLayer(entry.Layer, entry.ExpansionLevel);
				MapModelExt.Current.BaseLayers.Add(entry.Item);
			}
			foreach (TocEntry entry in _OperationalLayers)
			{
				if (entry.Layer is Layer)
					await entry.Item.SetLayer(entry.Layer, entry.ExpansionLevel);
				else if (entry.Layer is MobileLayerTree mltMap)
				{
					Dictionary<string, Layer> Layers = _MobileMapLayers[entry.ID];
					int iLevel = entry.ExpansionLevel - 1;
					List<TocDataItem> subitems = await MobileMapHelper.GetTOC(mltMap, Layers, iLevel);
					entry.Item.SetSubitems(subitems, true);
				}
				MapModelExt.Current.OperationalLayers.Add(entry.Item);
			}
			_BaseLayers = null;
			_OperationalLayers = null;
			_MobileMapLayers = null;
			MapModelExt.Current.NotifyMapLoaded();

		}

		private static async Task<SpatialReference> AddTileCacheLayer(Map MyMap, XElement xLayer)
		{

			// 100.x BUG: Runtime currently will not display tile packages/caches
			// that have only one scale level defined

			try
			{
				string sPath = xLayer.Attribute("Path").Value;
				if (!(Directory.Exists(sPath) || File.Exists(sPath)))
				{
					MapApplication.Current.Error("Not found:" + sPath);
					return null;
				}
				if (!TileCacheHelper.CheckTileCache(sPath))
					return null;
				TileCache tc = new TileCache(sPath);
				ArcGISTiledLayer lyr = new ArcGISTiledLayer(tc)
				{
					Id = "TILE_" + Guid.NewGuid().ToString()
				};
				await lyr.LoadAsync();
				MyMap.Basemap.BaseLayers.Add(lyr);
				TocEntry entry = MapHelper.GetTOC(xLayer, lyr);
				if (entry != null)
					_BaseLayers.Add(entry);
				return lyr.SpatialReference;
			}
			catch (Exception ex)
			{
				string sMsg = "Could not load tile package or cache:\n" + ex.Message;
				MapApplication.Current.Error(sMsg);
				return null;
			}

		}

		private static async Task<SpatialReference> AddMapImageLayer(Map MyMap, XElement xLayer, bool bBasemap)
		{

			try
			{

				string sUrl = xLayer.Attribute("Url").Value;
				ArcGISMapImageLayer lyr = new ArcGISMapImageLayer(new Uri(sUrl))
				{
					Id = "MAP_" + Guid.NewGuid().ToString()
				};
				await lyr.LoadAsync();
				if (bBasemap)
					MyMap.Basemap.BaseLayers.Add(lyr);
				else
					MyMap.OperationalLayers.Add(lyr);
				TocEntry entry = MapHelper.GetTOC(xLayer, lyr);
				if (entry != null)
				{
					if (bBasemap)
						_BaseLayers.Add(entry);
					else
						_OperationalLayers.Add(entry);
				}
				return lyr.SpatialReference;

			}
			catch (Exception ex)
			{
				string sMsg = "Could not load web layer:\n" + ex.Message;
				MapApplication.Current.Error(sMsg);
				return null;
			}

		}

		private static async Task<SpatialReference> AddWebFeatureLayer(Map MyMap, XElement xLayer)
		{

			try
			{

				string sUrl = xLayer.Attribute("Url").Value;
				ServiceFeatureTable sfTab = new ServiceFeatureTable(new Uri(sUrl));
				await sfTab.LoadAsync();
				if (sfTab.GeometryType == GeometryType.Unknown)
				{
					MyMap.Tables.Add(sfTab);
					return null;
				}
				FeatureLayer fLyr = new FeatureLayer(sfTab)
				{
					Id = "WEB_" + Guid.NewGuid().ToString()
				};
				MyMap.OperationalLayers.Add(fLyr);
				TocEntry entry = MapHelper.GetTOC(xLayer, fLyr);
				if (entry != null)
					_OperationalLayers.Add(entry);
				return fLyr.SpatialReference;

			}
			catch (Exception ex)
			{
				string sMsg = "Could not load web layer:\n" + ex.Message;
				MapApplication.Current.Error(sMsg);
				return null;
			}

		}

		private static async Task<SpatialReference> AddMobileMapPackageLayer(Map MyMap, XElement xLayer)
		{

			// Check for DisableAdvancedSymbology option

			HashSet<string> DisableList = new HashSet<string>();
			XElement xDisableList = xLayer.Element("DisableAdvancedSymbology");
			if (xDisableList != null)
			{
				foreach (XElement xLayerName in xDisableList.Elements())
				{
					XAttribute xName = xLayerName.Attribute("Name");
					DisableList.Add(xName.Value);
				}
			}

			// Process layer

			SpatialReference sr = null;
			List<Layer> MapLayers = null;
			List<MobileLayerTree> MapEntries = null;
			MobileBasemapLayer baseLayer = null;
			List<FeatureTable> MapTables = new List<FeatureTable>();
			try
			{

				XAttribute xPath = xLayer.Attribute("Path");
				if (xPath == null)
					return null;
				string sPath = xPath.Value;

				// Get map

				Map map = null;
				int iMapIndex = 0;
				MobileMapPackage mmpk = await MobileMapPackage.OpenAsync(sPath);
				IReadOnlyList<Map> MapList = mmpk.Maps;
				if (MapList.Count > 1)
				{
					XAttribute xMapIndex = xLayer.Attribute("MapIndex");
					if (xMapIndex != null)
						iMapIndex = Convert.ToInt32(xMapIndex.Value);
				}
				map = MapList[iMapIndex];
				sr = map.SpatialReference;

				// Set display name if not specified

				XAttribute xDisplayName = xLayer.Attribute("DisplayName");
				if (xDisplayName == null)
					xLayer.SetAttributeValue("DisplayName", mmpk.Item.Title);

				// Get operational layers

				if (PreserveGroupLayers)
				{
					MapLayers = new List<Layer>();
					foreach (Layer lyr in map.OperationalLayers.ToList())
					{
						map.OperationalLayers.Remove(lyr);
						MapLayers.Add(lyr);
					}
				}
				else
				{
					MapResult result = MobileMapHelper.ParseMap(map);
					MapLayers = result.Layers;
					MapEntries = result.TreeEntries;
				}

				// Disable advanced symbology

				MapHelper.CheckLayers(MapLayers, DisableList);

				// Get basemap layer

				if (map.Basemap.BaseLayers.Count > 0)
				{
					Layer lyr = map.Basemap.BaseLayers[0];
					map.Basemap.BaseLayers.Remove(lyr);
					baseLayer = lyr as MobileBasemapLayer;
				}

				// Get tables

				foreach (FeatureTable fTab in map.Tables.ToList())
				{
					map.Tables.Remove(fTab);
					MapTables.Add(fTab);
				}

			}
			catch (Exception ex)
			{
				string sMsg = "Could not load mobile map package:\n" + ex.Message;
				MapApplication.Current.Error(sMsg);
				return null;
			}

			bool bLegend = true;
			XAttribute xLegend = xLayer.Attribute("Legend");
			if (xLegend != null)
				bLegend = Convert.ToBoolean(xLegend.Value);

			// Add items to map

			ILayerContent pLayer = null;
			Dictionary<string, Layer> MobileMapLayers = null;
			if (PreserveGroupLayers)
			{
				GroupLayer glParent = new GroupLayer()
				{
					Id = Guid.NewGuid().ToString()
				};
				foreach (Layer lyr in MapLayers)
					glParent.Layers.Add(lyr);
				MyMap.OperationalLayers.Add(glParent);
				pLayer = glParent;
			}
			else
			{
				MobileMapLayers = new Dictionary<string, Layer>();
				foreach (Layer lyr in MapLayers)
				{
					MyMap.OperationalLayers.Add(lyr);
					MobileMapLayers[lyr.Id] = lyr;
				}
				MobileLayerTree mltParent = new MobileLayerTree();
				foreach (MobileLayerTree mltChild in MapEntries)
					mltParent.Sublayers.Add(mltChild);
				pLayer = mltParent;
			}
			foreach (FeatureTable fTab in MapTables)
				MyMap.Tables.Add(fTab);
			bool bBasemap = false;
			XAttribute xBasemap = xLayer.Attribute("AddBasemap");
			if (xBasemap != null)
				bBasemap = Convert.ToBoolean(xBasemap.Value);
			bBasemap = bBasemap && (baseLayer != null);
			if (bBasemap)
				MyMap.Basemap.BaseLayers.Add(baseLayer);
			if (!bLegend)
				return sr;

			// Create TOC item

			TocEntry entry = MapHelper.GetTOC(xLayer, pLayer);
			_OperationalLayers.Add(entry);
			if (!PreserveGroupLayers)
				_MobileMapLayers[entry.ID] = MobileMapLayers;
			if (bBasemap)
			{
				entry = MapHelper.GetTOC(xLayer, baseLayer);
				_BaseLayers.Add(entry);
			}
			return sr;

		}

		private static TocEntry GetTOC(XElement xLayer, ILayerContent pLayer)
		{

			bool bLegend = true;
			XAttribute xLegend = xLayer.Attribute("Legend");
			if (xLegend != null)
				bLegend = Convert.ToBoolean(xLegend.Value);
			if (!bLegend)
				return null;

			// Override properties

			bool bLayer = (pLayer is Layer);
			string sID = null;
			XAttribute xID = xLayer.Attribute("ID");
			if (xID != null)
				sID = xID.Value;
			string sDisplayName = null;
			XAttribute xDisplayName = xLayer.Attribute("DisplayName");
			if (xDisplayName != null)
				sDisplayName = xDisplayName.Value;
			bool? bVisible = null;
			XAttribute xVisible = xLayer.Attribute("IsVisible");
			if (xVisible != null)
				bVisible = Convert.ToBoolean(xVisible.Value);
			double? dMinScale = null;
			XAttribute xMinScale = xLayer.Attribute("MinScale");
			if (xMinScale != null)
				dMinScale = Convert.ToDouble(xMinScale.Value);
			double? dMaxScale = null;
			XAttribute xMaxScale = xLayer.Attribute("MaxScale");
			if (xMaxScale != null)
				dMaxScale = Convert.ToDouble(xMaxScale.Value);
			bool bExpanded = false;
			XAttribute xExpanded = xLayer.Attribute("IsExpanded");
			if (xExpanded != null)
				bExpanded = Convert.ToBoolean(xExpanded.Value);
			int iExpansionLevel = -1;
			XAttribute xExpansionLevel = xLayer.Attribute("ExpansionLevel");
			if (xExpansionLevel != null)
				iExpansionLevel = Convert.ToInt32(xExpansionLevel.Value);

			// Initialize layer

			if (bLayer)
			{
				Layer layer = pLayer as Layer;
				if (!string.IsNullOrEmpty(sID))
					layer.Id = sID;
				if (!string.IsNullOrEmpty(sDisplayName))
					layer.Name = sDisplayName;
				if (bVisible != null)
					layer.IsVisible = bVisible.Value;
				if (dMinScale != null)
					layer.MinScale = dMinScale.Value;
				if (dMaxScale != null)
					layer.MaxScale = dMaxScale.Value;
			}

			// Create item

			// BEGIN WORKAROUD
			TocDataItem item = new TocDataItem(!PreserveGroupLayers);
			// END WORKAROUND
			if (!bLayer)
			{
				if (!string.IsNullOrEmpty(sDisplayName))
					item.DisplayName = sDisplayName;
				if (bVisible == null)
					bVisible = true;
				item.IsChecked = bVisible.Value;
				if (dMinScale != null)
					item.MinScale = dMinScale.Value;
				if (dMaxScale != null)
					item.MaxScale = dMaxScale.Value;
			}
			item.IsExpanded = bExpanded;

			TocEntry entry = new TocEntry()
			{
				Item = item,
				ExpansionLevel = iExpansionLevel,
				Layer = pLayer
			};
			return entry;

		}

		private static void CheckLayers(List<Layer> Layers, HashSet<string> DisableList)
		{
			foreach (Layer lyr in Layers)
			{
				if (lyr is GroupLayer grpLyr)
				{
					CheckLayers(grpLyr.Layers.ToList(), DisableList);
				}
				else if (lyr is FeatureLayer fLyr)
				{
					if (!DisableList.Contains(lyr.Name))
						continue;
					GeodatabaseFeatureTable gfTab = fLyr.FeatureTable as GeodatabaseFeatureTable;
					gfTab.UseAdvancedSymbology = false;
				}
			}

		}

	}

}
