﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Xml.Linq;

using Esri.ArcGISRuntime.Geometry;

using Extensibility;

namespace NavAddin
{

	public enum UserSettingsType
	{
		View,
		Visibility
	}

	public static class UserSettingsHelper
	{

		private const string _sSettingsFileName = "NavAddin.Settings.xml";

		private static XElement _xView;
		private static XElement _xVisibility;
		private static XDocument _xSettings;

		// Public methods

		public static void SetInitialView()
		{
			_xView = new XElement("Default");
			SaveView(_xView);
		}

		public static void SetInitialVisibility()
		{
			_xVisibility = new XElement("Default");
			SaveVisibility(_xVisibility);
		}

		public static XDocument OpenSettings()
		{
			_xSettings = SettingsHelper.OpenSettings(_sSettingsFileName);
			return _xSettings;
		}

		public static List<string> GetSettings(UserSettingsType opt)
		{

			if (_xSettings == null)
				return null;
			List<string> Items = new List<string>();

			// Find category element

			XElement xCategory = FindCategory(opt, false);
			if (xCategory == null)
				return Items;

			// Build item name list

			foreach (XElement xItem in xCategory.Elements())
			{
				XElement xName = xItem.Element("name");
				if (xName == null)
					continue; // this should never happen
				Items.Add(xName.Value);
			}
			if (Items.Count > 1)
				Items.Sort();
			return Items;

		}

		public static void Save(string sName, UserSettingsType opt)
		{
			if (_xSettings == null)
				return;
			XElement xItem = FindElement(sName, opt, true);
			switch (opt)
			{
				case UserSettingsType.View:
					SaveView(xItem);
					break;
				case UserSettingsType.Visibility:
					SaveVisibility(xItem);
					break;
			}
		}

		public static void Delete(string sName, UserSettingsType opt)
		{
			if (_xSettings == null)
				return;
			XElement xItem = FindElement(sName, opt, false);
			if (xItem == null)
				return;
			xItem.Remove();
		}

		public static void Restore(string sName, UserSettingsType opt)
		{
			if (_xSettings == null)
				return;
			XElement xItem = FindElement(sName, opt, false);
			if (xItem == null)
				return;
			switch (opt)
			{
				case UserSettingsType.View:
					RestoreView(xItem);
					break;
				case UserSettingsType.Visibility:
					RestoreVisibility(xItem);
					break;
			}
		}

		public static void Reset(UserSettingsType opt)
		{
			switch (opt)
			{
				case UserSettingsType.View:
					RestoreView(_xView);
					break;
				case UserSettingsType.Visibility:
					RestoreVisibility(_xVisibility);
					break;
			}
		}

		public static void SaveSettings()
		{
			SettingsHelper.SaveSettings(_sSettingsFileName, _xSettings);
		}

		// Private methods

		private static void SaveView(XElement xItem)
		{

			// A View item has 4 subitems: "xmin", "ymin", "xmax", "ymax"

			Envelope env = NavModelExt.Current.ViewExtent;
			if (env == null)
				return;
			ClearItem(xItem);
			XElement xXMin = new XElement("xmin");
			xXMin.Value = env.XMin.ToString();
			xItem.Add(xXMin);
			XElement xYMin = new XElement("ymin");
			xYMin.Value = env.YMin.ToString();
			xItem.Add(xYMin);
			XElement xXMax = new XElement("xmax");
			xXMax.Value = env.XMax.ToString();
			xItem.Add(xXMax);
			XElement xYMax = new XElement("ymax");
			xYMax.Value = env.YMax.ToString();
			xItem.Add(xYMax);

		}

		private static void SaveVisibility(XElement xItem)
		{

			// A Visibility item has one subitem per layer item in the legend
			// Each subitem has these attributes:
			//    "header" = the label of this item
			//    "checked" = "true" or "false"
			//    "expanded" = "true" or "false"
			// Each subitem may have its own subitems corresponding to child layer items

			if (!MapModelExt.Current.EnableTOC)
				return;
			ObservableCollection<TocDataItem> TOC = MapModelExt.Current.TOC;
			ClearItem(xItem);
			foreach (TocDataItem item in TOC)
			{
				XElement xSubItem = SaveItemVisibility(item);
				xItem.Add(xSubItem);
			}
			return;

		}

		private static XElement SaveItemVisibility(TocDataItem item)
		{
			XElement xSubitem = new XElement("item");
			xSubitem.SetAttributeValue("header", item.Header);
			string sVal;
			if (item.IsChecked)
				sVal = "true";
			else
				sVal = "false";
			xSubitem.SetAttributeValue("checked", sVal);
			if (item.IsExpanded)
				sVal = "true";
			else
				sVal = "false";
			xSubitem.SetAttributeValue("expanded", sVal);
			if (item.Items != null)
			{
				foreach (TocDataItem si in item.Items)
				{
					XElement xSI = SaveItemVisibility(si);
					xSubitem.Add(xSI);
				}
			}
			return xSubitem;
		}

		private static void RestoreView(XElement xItem)
		{
			if (xItem == null)
				return;
			double dXMin = 0;
			double dYMin = 0;
			double dXMax = 0;
			double dYMax = 0;
			SpatialReference sr = AppViewExt.Current.Map.SpatialReference;
			foreach (XElement xSubItem in xItem.Elements())
			{
				string sItemName = xSubItem.Name.LocalName;
				switch (sItemName)
				{
					case "xmin":
						dXMin = Convert.ToDouble(xSubItem.Value);
						break;
					case "ymin":
						dYMin = Convert.ToDouble(xSubItem.Value);
						break;
					case "xmax":
						dXMax = Convert.ToDouble(xSubItem.Value);
						break;
					case "ymax":
						dYMax = Convert.ToDouble(xSubItem.Value);
						break;
				}
			}
			Envelope env = new Envelope(dXMin, dYMin, dXMax, dYMax, sr);
			NavModelExt.Current.ViewExtent = env;
		}

		private static void RestoreVisibility(XElement xItem)
		{
			if (xItem == null)
				return;
			ObservableCollection<TocDataItem> TOC = MapModelExt.Current.TOC;
			foreach (TocDataItem item in TOC)
				RestoreItemVisibility(xItem, item);
		}

		private static void RestoreItemVisibility(XElement xParent, TocDataItem item)
		{
			string sHeader = item.Header;
			XElement xItem = null;
			foreach (XElement xChild in xParent.Elements())
			{
				XAttribute xHeader = xChild.Attribute("header");
				if (xHeader == null)
					continue; // Skip the "name" subitem
				if (xHeader.Value != sHeader)
					continue;
				xItem = xChild;
				break;
			}
			if (xItem == null)
				return; // This might happen if the map contents are different from what was saved
			XAttribute xChecked = xItem.Attribute("checked");
			item.IsChecked = (xChecked.Value == "true");
			XAttribute xExpanded = xItem.Attribute("expanded");
			item.IsExpanded = (xExpanded.Value == "true");
			if (item.Items == null)
				return;
			foreach (TocDataItem ti in item.Items)
				RestoreItemVisibility(xItem, ti);
		}

		private static XElement FindCategory(UserSettingsType opt, bool bAdd)
		{

			// Find or add category element

			XElement xRoot = _xSettings.Root;
			XElement xCategory = null;
			string sCategory = "";
			switch (opt)
			{
				case UserSettingsType.View:
					sCategory = "View";
					break;
				case UserSettingsType.Visibility:
					sCategory = "Visibility";
					break;
			}
			foreach (XElement xChild in xRoot.Elements())
			{
				string sChildName = xChild.Name.LocalName;
				if (sChildName != sCategory)
					continue;
				xCategory = xChild;
				break;
			}
			if (xCategory == null)
			{
				if (!bAdd)
					return null;
				xCategory = new XElement(sCategory);
				xRoot.Add(xCategory);
			}
			return xCategory;

		}

		private static XElement FindElement(string sName, UserSettingsType opt, bool bAdd)
		{

			// Find or add category element

			XElement xCategory = FindCategory(opt, bAdd);
			if (xCategory == null)
				return null;

			// Find or add item element

			XElement xItem = null;
			foreach (XElement xChild in xCategory.Elements())
			{
				XElement xName = xChild.Element("name");
				if (xName == null)
					continue; // This should never happen
				if (xName.Value != sName)
					continue;
				xItem = xChild;
				break;
			}
			if (xItem == null)
			{
				if (!bAdd)
					return null;
				xItem = new XElement("item");
				XElement xName = new XElement("name");
				xName.Value = sName;
				xItem.Add(xName);
				xCategory.Add(xItem);
			}
			return xItem;

		}

		private static void ClearItem(XElement xItem)
		{

			// Clear all subitems except name

			List<XElement> RemoveList = new List<XElement>();
			foreach (XElement xSubItem in xItem.Elements())
			{
				string sItemName = xSubItem.Name.LocalName;
				if (sItemName == "name")
					continue;
				RemoveList.Add(xSubItem);
			}
			foreach (XElement xSubItem in RemoveList)
				xSubItem.Remove();

		}

	}

}
