﻿using System;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Input;
using System.Xml.Linq;

using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Location;
using Esri.ArcGISRuntime.UI;
using Esri.ArcGISRuntime.Symbology;

using Extensibility;

namespace NavAddin
{

	// GPSDialogModel controls the LocationDisplay, and also supports a serial GPS

	public class GPSDialogModel : INotifyPropertyChanged, IWindowContent, IDisposable
	{

		private const string _sSettingsFileName = "NavAddin.GPS.xml";
		private const string _sExternalSettingEntryName = "UseExternalGPS";
		private const string _sUseExternalValue = "Y";
		private const string _sDoNotUseExternalValue = "N";
		private const int _iInitialZoomScale = 1000;

		private static GPSDialogModel _current;

		private XDocument _xSettings;
		private PictureMarkerSymbol _symDefault;
		private PictureMarkerSymbol _symCourse;
		private LocationDisplay _ld;
		private LocationDataSource _InternalDataSource;
		private GPSLocationProvider _GPS;
		private string _sLat;
		private string _sLon;
		private string _sStatus;
		private bool _bUseSerial = false;
		private bool _bAutoPan = true;
		private LocationDisplayAutoPanMode _apMode = LocationDisplayAutoPanMode.Recenter;
		private bool _bZoomFirst = true;
		private bool _bActive = false;

		// Bindable properties

		public string Latitude
		{
			get { return _sLat; }
			set
			{
				_sLat = value;
				this.RaisePropertyChanged();
			}
		}

		public string Longitude
		{
			get { return _sLon; }
			set
			{
				_sLon = value;
				this.RaisePropertyChanged();
			}
		}

		public string Status
		{
			get { return _sStatus; }
			set
			{
				_sStatus = value;
				this.RaisePropertyChanged();
			}
		}

		public bool AutoPan
		{
			get { return _bAutoPan; }
			set
			{
				_bAutoPan = value;
				this.RaisePropertyChanged();
				if (_bAutoPan)
					_apMode = LocationDisplayAutoPanMode.Recenter;
				else
					_apMode = LocationDisplayAutoPanMode.Off;
				if (_ld == null)
					return;
				_ld.AutoPanMode = _apMode;
			}
		}

		public bool ExternalGPS
		{
			get { return _bUseSerial; }
			set
			{
				_bUseSerial = value;
				this.RaisePropertyChanged();
				this.SaveSettings();
				if (_ld == null)
					return;
				_ld.IsEnabled = false;
				if (_bUseSerial)
					_ld.DataSource = _GPS;
				else
					_ld.DataSource = _InternalDataSource;
				if (!_bActive)
					return;
				Task t = this.StartGPS();
			}
		}

		public ICommand Start { get; private set; }

		public ICommand Stop { get; private set; }

		// INotifyPropertyChanged

		public event PropertyChangedEventHandler PropertyChanged; 

		// IWindowContent

		public string Title { get { return "GPS Display"; } }

		public event EventHandler<EventArgs> ShowWindow;

		public event EventHandler<EventArgs> HideWindow { add { } remove { } } // Not used

		public void ShutDown()
		{
			this.Dispose();
		}

		// IDisposable

		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		protected virtual void Dispose(bool disposing)
		{
			if (disposing)
			{
				if (_ld != null)
				{
					_ld.IsEnabled = false;
					_ld.DataSource = _InternalDataSource;
				}
				// Dispose managed resources
				if (_GPS != null)
					_GPS.Dispose();
			}
			// No native resources
		}

		// Public events

		public event EventHandler<NotifyCanExecuteArgs> CanStartGPS;
		public event EventHandler<NotifyCanExecuteArgs> CanStopGPS;

		// Public methods

		public static void Open()
		{
			if (_current == null)
			{
				try
				{
					_current = new GPSDialogModel();
					MapApplication.Current.CreateChildWindow("NavAddin", "GPSDialog", _current);
				}
				catch (Exception ex)
				{
					string sMsg = "Could not create GPS controller:\n" + ex.Message;
					MapApplication.Current.Error(sMsg);
					return;
				}
			}
			_current.RaiseShowWindow();
		}

		// Event handlers

		void LocationDisplay_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
		{
			if (e.PropertyName != "AutoPanMode")
				return;
			if (_ld.AutoPanMode == _apMode)
				return;
			if (_apMode != LocationDisplayAutoPanMode.Off)
			{
				this.Status = "Restoring auto-pan";
				// Disable initial zoom				
				_ld.InitialZoomScale = double.NaN; 
			}
			_ld.AutoPanMode = _apMode;
		}

		void LocationChanged(object sender, Location loc)
		{
			if (!_bActive)
				return;
			MapPoint pt = loc.Position;
			if (pt == null)
				return;
			this.Longitude = pt.X.ToString("F6");
			this.Latitude = pt.Y.ToString("F6");
			if (_bUseSerial)
				this.Status = _GPS.Status;
			else
				this.Status = "Running...";
			if (double.IsNaN(pt.X) || double.IsNaN(pt.Y))
				return;
			if (_bZoomFirst)
			{
				_bZoomFirst = false;
				NavModelExt.Current.ViewScale = 1000;
			}
		}

		// Private constructors

		private GPSDialogModel()
		{
			this.Initialize();
		}

		// Private methods

		private async void Initialize()
		{

			// Get saved GPS settings

			this.OpenSettings();

			// Get resources

			_symDefault = await ResourceHelper.GetPictureMarkerSymbol("NavAddin", "Images/GPSLocation.png");
			_symDefault.Angle = 0;
			_symDefault.Width = 18;
			_symDefault.Height = 18;
			_symCourse = await ResourceHelper.GetPictureMarkerSymbol("NavAddin", "Images/GPSCourse.png");
			_symCourse.Angle = 0;
			_symCourse.Width = 18;
			_symCourse.Height = 18;

			// Set up location display

			_ld = NavViewExt.Current.LocationDisplay;
			_ld.AutoPanMode = _apMode;
			_ld.PropertyChanged += LocationDisplay_PropertyChanged;
			_ld.IsEnabled = false;
			_ld.InitialZoomScale = _iInitialZoomScale;
			_ld.DefaultSymbol = _symDefault;
			_ld.CourseSymbol = _symCourse;
			_InternalDataSource = _ld.DataSource;
			_InternalDataSource.LocationChanged += LocationChanged;
			_GPS = new GPSLocationProvider();
			_GPS.LocationChanged += LocationChanged;
			if (_bUseSerial)
				_ld.DataSource = _GPS;

			// Properties

			Latitude = "NaN";
			Longitude = "NaN";
			Status = "N/A";

			// Commands

			CommandHelper ch;
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.ExecuteStart);
			ch.AddCanExecuteEvent(this, "CanStartGPS");
			Start = ch;
			ch = new CommandHelper(false);
			ch.AddExecuteAction(this.ExecuteStop);
			ch.AddCanExecuteEvent(this, "CanStopGPS");
			Stop = ch;

		}

		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private void RaiseShowWindow()
		{
			if (this.ShowWindow == null)
				return;
			this.ShowWindow(this, new EventArgs());
		}

		private void RaiseCanStartGPS(bool bCanExecute)
		{
			if (this.CanStartGPS == null)
				return;
			NotifyCanExecuteArgs e = new NotifyCanExecuteArgs();
			e.CanExecute = bCanExecute;
			this.CanStartGPS(this, e);
		}

		private void RaiseCanStopGPS(bool bCanExecute)
		{
			if (this.CanStopGPS == null)
				return;
			NotifyCanExecuteArgs e = new NotifyCanExecuteArgs();
			e.CanExecute = bCanExecute;
			this.CanStopGPS(this, e);
		}

		private void ExecuteStart(object parameter)
		{
			_bZoomFirst = _bAutoPan;
			this.RaiseCanStartGPS(false);
			this.RaiseCanStopGPS(true);
			Task t = this.StartGPS();
		}

		private void ExecuteStop(object parameter)
		{
			this.RaiseCanStopGPS(false);
			this.RaiseCanStartGPS(true);
			if (_ld != null)
				_ld.IsEnabled = false;
			this.Status = "Stopped";
			_bActive = false;
		}

		private async Task StartGPS()
		{
			_bActive = true;
			if (_bUseSerial && !_GPS.Initialized)
			{
				this.Status = "Initializing...";
				bool bResult = await _GPS.InitializeAsync();
				if (!bResult)
				{
					this.Status = "Initialization failed.";
					this.RaiseCanStopGPS(false);
					this.RaiseCanStartGPS(true);
					_bActive = false;
					return;
				}
			}
			this.Status = "Starting...";
			_ld.IsEnabled = true;
		}

		private void OpenSettings()
		{
			_xSettings = SettingsHelper.OpenSettings(_sSettingsFileName);
			XElement xRoot = _xSettings.Root;
			XElement xGPS = xRoot.Element(_sExternalSettingEntryName);
			if (xGPS == null)
				return;
			string sVal = xGPS.Value;
			_bUseSerial = (sVal == _sUseExternalValue);
		}

		private void SaveSettings()
		{
			string sVal;
			if (_bUseSerial)
				sVal = _sUseExternalValue;
			else
				sVal = _sDoNotUseExternalValue;
			XElement xRoot = _xSettings.Root;
			XElement xGPS = xRoot.Element(_sExternalSettingEntryName);
			if (xGPS == null)
			{
				xGPS = new XElement(_sExternalSettingEntryName);
				xRoot.Add(xGPS);
			}
			xGPS.Value = sVal;
			SettingsHelper.SaveSettings(_sSettingsFileName, _xSettings);
		}

	}

}

