﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Windows;
using System.Xml.Linq;

using Esri.ArcGISRuntime.Mapping;

using Extensibility;

namespace NavAddin
{

	// MapModelExt manages the Map, TOC, and status functions

	public class MapModelExt : IConfigurable, IStatus, INotifyPropertyChanged, IWindowContent
	{

		public static MapModelExt Current { get; private set; }

		public bool EnableTOC { get; private set; }

		// OperationalLayers and BaseLayers determine which layers should appear in the TOC

		public ObservableCollection<TocDataItem> OperationalLayers { get; private set; }
		public ObservableCollection<TocDataItem> BaseLayers { get; private set; }

		private XElement _xMap;
		private string _sStatus;
		private Visibility _visCoord = Visibility.Collapsed;
		private string _sCoord;
		private Visibility _visProgress = Visibility.Hidden;
		private double _dMapScale;
		private DateTime _dtTOC = DateTime.MinValue;
		private bool _bUseChildWindow = false;
		private bool _bChildWindowCreated = false;
		private bool _bAutoShow = false;
		private bool _bFirst = true;

		// Bindable layout properties

		public ObservableCollection<TocDataItem> TOC { get; private set; }

		public string StatusText
		{
			get { return _sStatus; }
			set
			{
				_sStatus = value;
				this.RaisePropertyChanged();
			}
		}

		public Visibility CoordinateVisibility
		{
			get { return _visCoord; }
			set
			{
				_visCoord = value;
				this.RaisePropertyChanged();
			}
		}

		public string CoordinateText
		{
			get { return _sCoord; }
			set
			{
				_sCoord = value; // This will come from MapViewExt
				this.RaisePropertyChanged();
			}
		}

		public Visibility ProgressVisibility
		{
			get { return _visProgress; }
			set
			{
				_visProgress = value;
				this.RaisePropertyChanged();
			}
		}

		// Bindable view properties

		public string ViewMapCoordinateText
		{
			get { return CoordinateText; }
			set { CoordinateText = value; }
		}

		public double ViewMapScale
		{
			get { return _dMapScale; }
			set
			{
				if (value == _dMapScale)
					return;
				_dMapScale = value;
				if (!EnableTOC)
					return;
				foreach (TocDataItem item in TOC)
					item.CheckVisibility(_dMapScale);
			}
		}

		// IConfigurable

		public void LoadConfiguration(string sConfigData)
		{
			if (string.IsNullOrEmpty(sConfigData))
				return;
			XDocument xDoc = XDocument.Parse(sConfigData);
			XElement xRoot = xDoc.Root;
			_xMap = xRoot.Element("Map");
			XElement xTOC = xRoot.Element("EnableTOC");
			if (xTOC == null)
				return;
			EnableTOC = (xTOC.Value == "True");
			if (!EnableTOC)
				return;
			XElement xWindow = xRoot.Element("ChildWindowTOC");
			if (xWindow != null)
				_bUseChildWindow = (xWindow.Value == "True");
			XElement xShow = xRoot.Element("ShowTOC");
			if (xShow != null)
				_bAutoShow = (xShow.Value == "True");
		}

		// IStatus

		public void Status(string sMsg)
		{
			StatusText = sMsg;
		}

		public void Status(string sMsg, bool bBusy)
		{
			if (bBusy)
				ProgressVisibility = Visibility.Visible;
			else
				ProgressVisibility = Visibility.Hidden;
			this.Status(sMsg);
		}

		public void Message(string sMsg)
		{
			ProgressVisibility = Visibility.Hidden;
			this.Status("OK", false);
			MessageBox.Show(sMsg, MapApplication.Current.ApplicationName, MessageBoxButton.OK, MessageBoxImage.Information);
		}

		public void Error(string sErr)
		{
			ProgressVisibility = Visibility.Hidden;
			this.Status("ERROR", false);
			MessageBox.Show(sErr, MapApplication.Current.ApplicationName, MessageBoxButton.OK, MessageBoxImage.Error);
		}

		// INotifyPropertyChanged

		public event PropertyChangedEventHandler PropertyChanged;

		// IWindowContent

		public string Title { get { return "TOC"; } }
		public event EventHandler<EventArgs> ShowWindow;
		public event EventHandler<EventArgs> HideWindow;
		public void ShutDown() { }

		// Constructors

		public MapModelExt()
		{
			Current = this;
			EnableTOC = false;
			OperationalLayers = new ObservableCollection<TocDataItem>();
			OperationalLayers.CollectionChanged += Layers_CollectionChanged;
			BaseLayers = new ObservableCollection<TocDataItem>();
			BaseLayers.CollectionChanged += Layers_CollectionChanged;
			MapApplication.Current.AppReady += MapApplication_AppReady;
			TOC = new ObservableCollection<TocDataItem>();
		}

		// Public events

		public event EventHandler<EventArgs> MapLoaded;

		// Public methods

		public void NotifyMapLoaded()
		{
			if (this.MapLoaded == null)
				return;
			this.MapLoaded(this, new EventArgs());
		}

		public void ShowTOC()
		{
			if (!_bUseChildWindow)
				return;
			if (!_bChildWindowCreated)
			{
				MapApplication.Current.CreateChildWindow("NavAddin", "TOC", this);
				_bChildWindowCreated = true;
			}
			if (this.ShowWindow == null)
				return;
			this.ShowWindow(this, new EventArgs());
		}

		public void HideTOC()
		{
			if (!_bUseChildWindow)
				return;
			if (this.HideWindow == null)
				return;
			this.HideWindow(this, new EventArgs());
		}

		// Event handlers

		void MapApplication_AppReady(object sender, EventArgs e)
		{

			// The Map can be of the one of the following types:
			// "SimpleTestMap" - A simple online test map
			// "BasicMap" - Allows specification of layers and other properties

			if (_xMap != null)
			{
				string sMapType = _xMap.Attribute("Type").Value;
				switch (sMapType)
				{
					case "SimpleTestMap":
						MapHelper.OpenSimpleTestMap();
						break;
					case "WebMap":
						MapHelper.OpenWebMap(_xMap);
						break;
					case "BasicMap":
						MapHelper.OpenBasicMapAsync(_xMap);
						break;
				}
			}

		}

		void Layers_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
		{
			if (!EnableTOC)
				return;
			this.UpdateTOC();
			if (!_bUseChildWindow)
				return;
			if (!_bAutoShow)
				return;
			if (!_bFirst)
				return;
			_bFirst = true;
			this.ShowTOC();
		}

		// Private methods

		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private void UpdateTOC()
		{
			TOC.Clear();
			foreach (TocDataItem item in BaseLayers)
				TOC.Insert(0, item);
			foreach (TocDataItem item in OperationalLayers)
				TOC.Insert(0, item);
			UserSettingsHelper.SetInitialVisibility();
		}

	}

}
