﻿using System;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;
using System.Xml.Linq;

using Esri.ArcGISRuntime.Geometry;

using Extensibility;

namespace NavAddin
{

	// NavCommands manages commands typically associated with buttons or menu items

	public class SetCoordinateUnitsCommand : ICommand
	{
		public event EventHandler CanExecuteChanged { add { } remove { } }

		public bool CanExecute(object parameter)
		{
			return true;
		}

		public void Execute(object parameter)
		{
			MapViewExt.Current.SetCoordinateUnits(parameter.ToString());
		}

	}

	public class NavCommands: INotifyPropertyChanged, IConfigurable
	{

		private enum NavCommandType
		{
			OpenMap,
			ExportMap,
			CopyMap,
			UserSettings,
			ZoomInFixed,
			ZoomOutFixed,
			ZoomFull,
			ZoomPrev,
			ZoomNext,
			GPS,
			ShowTOC,
			HideTOC,
			ToggleCoordinateDisplay,
			Exit
		}

		// Bindable layout properties

		public ICommand OpenMapCommand { get; private set; }

		public Visibility OpenMapVisibility { get; private set; }

		public ICommand ExportMapCommand { get; private set; }

		public ICommand CopyMapCommand { get; private set; }

		public ICommand UserSettingsCommand { get; private set; }

		public ICommand ZoomInFixedCommand { get; private set; }

		public ICommand ZoomOutFixedCommand { get; private set; }

		public ICommand ZoomFullCommand { get; private set; }

		public ICommand ZoomPrevCommand { get; private set; }

		public ICommand ZoomNextCommand { get; private set; }

		public ICommand GPSCommand { get; private set; }

		public Visibility GPSVisibility { get; private set; }

		public ICommand ShowTOCCommand { get; private set; }

		public ICommand HideTOCCommand { get; private set; }

		public ICommand ToggleCoordinatesCommand { get; private set; }

		public ICommand SetUnitsCommand { get; private set; }

		public ICommand ExitCommand { get; private set; }

		// INotifyMapChanged

		public event PropertyChangedEventHandler PropertyChanged;

		// IConfigurable

		public void LoadConfiguration(string sConfigData)
		{
			if (string.IsNullOrEmpty(sConfigData))
				return;
			XDocument xDoc = XDocument.Parse(sConfigData);
			XElement xRoot = xDoc.Root;
			XElement xOpenMap = xRoot.Element("EnableOpenMap");
			bool bValue = true;
			if (xOpenMap != null)
				bValue = (xOpenMap.Value == "True");
			if (bValue)
				OpenMapVisibility = Visibility.Visible;
			else
				OpenMapVisibility = Visibility.Collapsed;
			bValue = true;
			XElement xGPS = xRoot.Element("EnableGPS");
			if (xGPS != null)
				bValue = (xGPS.Value == "True");
			if (bValue)
				GPSVisibility = Visibility.Visible;
			else
				GPSVisibility = Visibility.Collapsed;
		}

		// Constructors

		public NavCommands()
		{
			CommandHelper ch;
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.ExecuteNavCommand, NavCommandType.OpenMap);
			OpenMapCommand = ch;
			OpenMapVisibility = Visibility.Visible;
			ch = new CommandHelper(false);
			ch.AddExecuteAction(this.ExecuteNavCommand, NavCommandType.ExportMap);
			ch.AddCanExecuteEvent(NavModelExt.Current, "CanExecuteNavCommand");
			ExportMapCommand = ch;
			ch = new CommandHelper(false);
			ch.AddExecuteAction(this.ExecuteNavCommand, NavCommandType.CopyMap);
			ch.AddCanExecuteEvent(NavModelExt.Current, "CanExecuteNavCommand");
			CopyMapCommand = ch;
			ch = new CommandHelper(false);
			ch.AddExecuteAction(this.ExecuteNavCommand, NavCommandType.UserSettings);
			ch.AddCanExecuteEvent(NavModelExt.Current, "CanExecuteNavCommand");
			UserSettingsCommand = ch;
			ch = new CommandHelper(false);
			ch.AddExecuteAction(this.ExecuteNavCommand, NavCommandType.ZoomInFixed);
			ch.AddCanExecuteEvent(NavModelExt.Current, "CanExecuteNavCommand");
			ZoomInFixedCommand = ch;
			ch = new CommandHelper(false);
			ch.AddCanExecuteEvent(NavModelExt.Current, "CanExecuteNavCommand");
			ch.AddExecuteAction(this.ExecuteNavCommand, NavCommandType.ZoomOutFixed);
			ZoomOutFixedCommand = ch;
			ch = new CommandHelper(false);
			ch.AddExecuteAction(this.ExecuteNavCommand, NavCommandType.ZoomFull);
			ch.AddCanExecuteEvent(NavModelExt.Current, "CanExecuteNavCommand");
			ZoomFullCommand = ch;
			ch = new CommandHelper(false);
			ch.AddExecuteAction(this.ExecuteNavCommand, NavCommandType.ZoomPrev);
			ch.AddCanExecuteEvent(NavModelExt.Current.History, "CanZoomPrevious");
			ZoomPrevCommand = ch;
			ch = new CommandHelper(false);
			ch.AddExecuteAction(this.ExecuteNavCommand, NavCommandType.ZoomNext);
			ch.AddCanExecuteEvent(NavModelExt.Current.History, "CanZoomNext");
			ZoomNextCommand = ch;
			ch = new CommandHelper(false);
			ch.AddExecuteAction(this.ExecuteNavCommand, NavCommandType.GPS);
			ch.AddCanExecuteEvent(NavModelExt.Current, "CanExecuteNavCommand");
			GPSCommand = ch;
			GPSVisibility = Visibility.Visible;
			ch = new CommandHelper(false);
			ch.AddExecuteAction(this.ExecuteNavCommand, NavCommandType.ShowTOC);
			ch.AddCanExecuteEvent(NavModelExt.Current, "CanExecuteNavCommand");
			ShowTOCCommand = ch;
			ch = new CommandHelper(false);
			ch.AddExecuteAction(this.ExecuteNavCommand, NavCommandType.HideTOC);
			ch.AddCanExecuteEvent(NavModelExt.Current, "CanExecuteNavCommand");
			HideTOCCommand = ch;
			ch = new CommandHelper(false);
			ch.AddExecuteAction(this.ExecuteNavCommand, NavCommandType.ToggleCoordinateDisplay);
			ch.AddCanExecuteEvent(NavModelExt.Current, "CanExecuteNavCommand");
			ToggleCoordinatesCommand = ch;
			SetUnitsCommand = new SetCoordinateUnitsCommand();
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.ExecuteNavCommand, NavCommandType.Exit);
			ExitCommand = ch;
		}

		// Private methods

		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private void ExecuteNavCommand(object parameter)
		{
			if (!(parameter is NavCommandType))
				return; // This should never happen
			string sPath;
			double dNewScale;
			Envelope envNew;
			Task t;
			switch ((NavCommandType)parameter)
			{
				case NavCommandType.OpenMap:
					sPath = DialogHelper.GetInputFile("mmpk");
					if (string.IsNullOrEmpty(sPath))
						return;
					MapHelper.OpenMobileMapPackage(sPath);
					return;
				case NavCommandType.ExportMap:
					string sOutFile = DialogHelper.GetOutputFile("bmp");
					if (string.IsNullOrEmpty(sOutFile))
						return;
					t = MapViewExt.Current.ExportMap(sOutFile);
					return;
				case NavCommandType.CopyMap:
					t = MapViewExt.Current.CopyMap();
					return;
				case NavCommandType.UserSettings:
					UserSettingsDialogModel.Open();
					return;
				case NavCommandType.ZoomInFixed:
					dNewScale = NavModelExt.Current.ViewScale / 2;
					NavModelExt.Current.ViewScale = dNewScale;
					return;
				case NavCommandType.ZoomOutFixed:
					dNewScale = NavModelExt.Current.ViewScale * 2;
					NavModelExt.Current.ViewScale = dNewScale;
					return;
				case NavCommandType.ZoomFull:
					envNew = NavModelExt.Current.History.GetFull();
					if (envNew == null)
						return;
					NavModelExt.Current.ViewExtent = envNew;
					return;
				case NavCommandType.ZoomPrev:
					envNew = NavModelExt.Current.History.GetPrevious();
					if (envNew == null)
						return;
					NavModelExt.Current.ViewExtent = envNew;
					return;
				case NavCommandType.ZoomNext:
					envNew = NavModelExt.Current.History.GetNext();
					if (envNew == null)
						return;
					NavModelExt.Current.ViewExtent = envNew;
					return;
				case NavCommandType.GPS:
					GPSDialogModel.Open();
					return;
				case NavCommandType.ShowTOC:
					MapModelExt.Current.ShowTOC();
					return;
				case NavCommandType.HideTOC:
					MapModelExt.Current.HideTOC();
					return;
				case NavCommandType.ToggleCoordinateDisplay:
					if (MapModelExt.Current.CoordinateVisibility == Visibility.Collapsed)
						MapModelExt.Current.CoordinateVisibility = Visibility.Visible;
					else
						MapModelExt.Current.CoordinateVisibility = Visibility.Collapsed;
					return;
				case NavCommandType.Exit:
					MapApplication.Current.Exit();
					return;
			}
		}

	}

}
