﻿using System;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Windows;
using System.Windows.Input;

using Esri.ArcGISRuntime.Geometry;

using Extensibility;

namespace NavAddin
{

	// NavModelExt manages map navigation tasks

	public class NavModelExt : INotifyPropertyChanged
	{

		public static NavModelExt Current { get; private set; }

		public HistoryHelper History;

		private ScaleHelper _sh;
		private bool _bMapReady = false;
		private int _iScaleItemIndex = -1;
		private double _dViewScale;
		private bool _bScaleVisible = false;
		private Visibility _visScale = Visibility.Collapsed;
		private Envelope _envViewExtent;
		private bool _bOverride = false;

		// Bindable layout properties

		public bool MapReady
		{
			get { return _bMapReady; }
			set
			{
				_bMapReady = value;
				this.RaisePropertyChanged();
				this.RaiseCanExecuteNavCommand();
			}
		}

		public ObservableCollection<string> ScaleItemsSource { get; private set; }

		public int ScaleSelectedIndex
		{
			get { return _iScaleItemIndex; }
			set
			{
				if (value == _iScaleItemIndex)
					return;
				_iScaleItemIndex = value;
				this.RaisePropertyChanged();
				if (_bOverride)
				{
					// Value was set internally
					_bOverride = false;
					return;
				}
				double dNewScale = _sh.GetScale(_iScaleItemIndex);
				if (double.IsNaN(dNewScale))
					return;
				this.ViewScale = dNewScale;
			}
		}

		public Visibility ScaleVisibility
		{
			get { return _visScale; }
			set
			{
				_visScale = value;
				this.RaisePropertyChanged();
			}
		}

		public bool IsScaleVisible
		{
			get { return _bScaleVisible; }
			set
			{
				_bScaleVisible = value;
				if (_bScaleVisible)
					ScaleVisibility = Visibility.Visible;
				else
					ScaleVisibility = Visibility.Collapsed;
			}
		}

		public string ScaleText { get; set; }

		public ICommand ScaleEnteredCommand { get; private set; }

		// Bindable view properties

		public double ViewScale
		{
			get { return _dViewScale; }
			set
			{
				if (value == ViewScale)
					return;
				_dViewScale = value;
				this.RaisePropertyChanged();
			}
		}

		public Envelope ViewExtent
		{
			get { return _envViewExtent; }
			set
			{
				_envViewExtent = value;
				this.RaisePropertyChanged();
			}
		}

		public double ViewUnitsPerPixel { get; set; }

		// INotifyPropertyChanged

		public event PropertyChangedEventHandler PropertyChanged; 

		// Public events

		public event EventHandler<NotifyCanExecuteArgs> CanExecuteNavCommand;

		// Public methods

		public void NavigationCompleted()
		{
			bool bResult = History.AddExtent(ViewExtent, ViewUnitsPerPixel);
			bResult = _sh.SetScale(ViewScale);
			if (ScaleSelectedIndex == 0)
				return;
			// Prevent unwanted feedback
			_bOverride = true; 
			ScaleSelectedIndex = 0;
			if (!MapReady)
				MapReady = true;
		}

		// Constructors

		public NavModelExt()
		{

			Current = this;

			// Set up helpers

			History = new HistoryHelper();
			_sh = new ScaleHelper();
			ScaleItemsSource = _sh.ScaleItems;
			ScaleSelectedIndex = 0;

			// Set event listeners

			AppViewExt.Current.MapChanged += App_MapChanged;

			// Commands

			CommandHelper ch;
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.ExecuteScaleEntered);
			ScaleEnteredCommand = ch;

		}

		// Event listeners

		void App_MapChanged(object sender, MapChangedArgs e)
		{
			History.Reset();
		}

		// Private methods

		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private void RaiseCanExecuteNavCommand()
		{
			if (this.CanExecuteNavCommand == null)
				return;
			NotifyCanExecuteArgs e = new NotifyCanExecuteArgs();
			e.CanExecute = MapReady;
			this.CanExecuteNavCommand(this, e);
		}

		private void ExecuteScaleEntered(object parameter)
		{
			if (!MapReady)
				return;
			string sNewScale = ScaleText as string;
			double dNewScale = _sh.GetScale(sNewScale);
			if (double.IsNaN(dNewScale))
				return;
			this.ViewScale = dNewScale;
		}

	}

}
