﻿using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Windows.Input;

using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Symbology;
using Esri.ArcGISRuntime.UI;

using Extensibility;

namespace NavAddin
{

	// NavTools manages Zoom In, Zoom Out, and Pan tool functions

	public class NavTools : INotifyPropertyChanged
	{

		private enum NavToolType
		{
			None,
			ZoomIn,
			ZoomOut,
			Pan
		}

		private enum NavZoomType
		{
			ZoomIn,
			ZoomOut
		}

		private bool _bZoomInChecked = false;
		private bool _bZoomOutChecked = false;
		private bool _bPanChecked = false;
		private NavToolType _ToolType = NavToolType.None;
		private SimpleFillSymbol _sfZoom;

		// Bindable layout properties

		public ICommand ZoomInCommand { get; private set; }

		public bool ZoomInToolChecked
		{
			get { return _bZoomInChecked; }
			set
			{
				_bZoomInChecked = value;
				this.RaisePropertyChanged();
			}
		}

		public ICommand ZoomOutCommand { get; private set; }

		public bool ZoomOutToolChecked
		{
			get { return _bZoomOutChecked; }
			set
			{
				_bZoomOutChecked = value;
				this.RaisePropertyChanged();
			}
		}

		public ICommand PanCommand { get; private set; }

		public bool PanToolChecked
		{
			get { return _bPanChecked; }
			set
			{
				_bPanChecked = value;
				this.RaisePropertyChanged();
			}
		}

		// INotifyPropertyChanged

		public event PropertyChangedEventHandler PropertyChanged;

		// Constructors

		public NavTools()
		{
			this.Init();
		}

		// Event handlers

		void ToolHelper_ToolComplete(object sender, ToolHelper.NotifyToolArgs e)
		{
			switch (_ToolType)
			{
				case NavToolType.ZoomIn:
					NavModelExt.Current.ViewExtent = this.GetZoomExtent(e.Geometry, NavZoomType.ZoomIn);
					break;
				case NavToolType.ZoomOut:
					NavModelExt.Current.ViewExtent = this.GetZoomExtent(e.Geometry, NavZoomType.ZoomOut);
					break;
			}
		}

		// Private methods

		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private void Init()
		{

			// Get zoom tool symbol

			_sfZoom = ResourceHelper.GetFillSymbol("NavAddin", "NavAddin.ZoomToolSymbol");
			if (_sfZoom == null)
			{
				_sfZoom = new SimpleFillSymbol();
				_sfZoom.Style = SimpleFillSymbolStyle.Null;
				SimpleLineSymbol slSym = new SimpleLineSymbol();
				slSym.Style = SimpleLineSymbolStyle.Solid;
				slSym.Width = 1;
				slSym.Color = System.Drawing.Color.Black;
				_sfZoom.Outline = slSym;
			}

			// Create tools

			ToolHelper th;
			th = this.CreateTool(NavToolType.ZoomIn);
			ZoomInCommand = th.ExecuteCommand;
			th = this.CreateTool(NavToolType.ZoomOut);
			ZoomOutCommand = th.ExecuteCommand;
			th = this.CreateTool(NavToolType.Pan);
			PanCommand = th.ExecuteCommand;

		}

		private void ExecuteActivate(object parameter)
		{
			_ToolType = (NavToolType)parameter;
			switch (_ToolType)
			{
				case NavToolType.ZoomIn:
					ZoomInToolChecked = true;
					break;
				case NavToolType.ZoomOut:
					ZoomOutToolChecked = true;
					break;
				case NavToolType.Pan:
					PanToolChecked = true;
					break;
			}
		}

		private void ExecuteDeactivate(object parameter)
		{
			NavToolType tt = (NavToolType)parameter;
			switch (tt)
			{
				case NavToolType.ZoomIn:
					ZoomInToolChecked = false;
					break;
				case NavToolType.ZoomOut:
					ZoomOutToolChecked = false;
					break;
				case NavToolType.Pan:
					PanToolChecked = false;
					break;
			}
			_ToolType = NavToolType.None;
		}

		private ToolHelper CreateTool(NavToolType tt)
		{

			// Create ToolHelper

			ToolHelper th = new ToolHelper(false);
			th.ExecuteCommand.AddCanExecuteEvent(NavModelExt.Current, "CanExecuteNavCommand");
			th.ActivateCommand.AddExecuteAction(this.ExecuteActivate, tt);
			th.DeactivateCommand.AddExecuteAction(this.ExecuteDeactivate, tt);
			switch(tt)
			{
				case NavToolType.ZoomIn:
					th.SetCursor("NavAddin", "Images/ZoomIn.cur");
					break;
				case NavToolType.ZoomOut:
					th.SetCursor("NavAddin", "Images/ZoomOut.cur");
					break;
				case NavToolType.Pan:
					th.SetCursor("NavAddin", "Images/Pan.cur");
					break;
			}
			th.ToolComplete += ToolHelper_ToolComplete;

			// Set up tool configuration

			ToolConfiguration tc = th.Configuration;
			if (tt == NavToolType.Pan)
			{
				tc.CollectGeometry = false;
				return th;
			}
			tc.SketchType = SketchCreationMode.Rectangle;
			tc.Editable = false;
			tc.SingleTapCreatesPoint = true;
			tc.SketchStyle.FillSymbol = _sfZoom;
			tc.SketchStyle.LineSymbol = null;
			tc.SketchStyle.MidVertexSymbol = null;
			tc.SketchStyle.VertexSymbol = null;
			return th;

		}

		private Envelope GetZoomExtent(Geometry g, NavZoomType zt)
		{

			if (g is MapPoint)
				return this.GetZoomExtent((MapPoint)g, zt);
			Envelope envZoom = g.Extent;
			if (zt == NavZoomType.ZoomIn)
				return envZoom;

			// Create zoomed out extent

			Envelope envCurrent = NavModelExt.Current.ViewExtent;
			SpatialReference sr = envCurrent.SpatialReference;
			MapPoint ptZ = envZoom.GetCenter();
			double dW = envCurrent.Width * envCurrent.Width / envZoom.Width / 2;
			double dH = envCurrent.Height * envCurrent.Height / envZoom.Height / 2;
			Envelope envZoomOut = new Envelope(ptZ.X - dW, ptZ.Y - dH, ptZ.X + dW, ptZ.Y + dH, sr);
			return envZoomOut;

		}

		private Envelope GetZoomExtent(MapPoint ptZ, NavZoomType zt)
		{
			Envelope envCurrent = NavModelExt.Current.ViewExtent;
			EnvelopeBuilder eb = new EnvelopeBuilder(envCurrent);
			eb.CenterAt(ptZ);
			if (zt == NavZoomType.ZoomIn)
				eb.Expand(0.5);
			else
				eb.Expand(2);
			return eb.ToGeometry();
		}

	}

}
