﻿using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Threading.Tasks;
using System.Windows;

using SolidColorBrush = System.Windows.Media.SolidColorBrush;
using ImageSource = System.Windows.Media.ImageSource;

using Esri.ArcGISRuntime.Mapping;
using Esri.ArcGISRuntime.Symbology;
using Esri.ArcGISRuntime.UI;

using Extensibility;

namespace NavAddin
{

	/*
	 
		TocDataItem is a model for TreeView (and other ItemsControl) items,
		to support the display of TOC/Legend info, as well as control layer visibility.
		
		Example of XAML usage:
	 
		<UserControl.Resources>
			<Style x:Key="LayerTreeItemStyle" TargetType="TreeViewItem">
				<Setter Property="IsExpanded" Value="{Binding RelativeSource={RelativeSource Self}, Path=Header.IsExpanded, Mode=TwoWay}" />
				<Setter Property="HorizontalContentAlignment" Value="Left" />
				<Setter Property="VerticalContentAlignment" Value="Center" />
			</Style>
			<Style x:Key="LayerCheckBoxStyle" TargetType="CheckBox">
				<Setter Property="VerticalAlignment" Value="Center" />
				<Setter Property="Margin" Value="1" />
				<Setter Property="Visibility" Value="{Binding LayerVisibility}" />
				<Setter Property="IsChecked" Value="{Binding IsChecked, Mode=TwoWay}" />
				<Style.Triggers>
					<DataTrigger Binding="{Binding Path=IsEnabled}" Value="True">
						<Setter Property="Opacity" Value="1.0" />
					</DataTrigger>
					<DataTrigger Binding="{Binding Path=IsEnabled}" Value="False">
						<Setter Property="Opacity" Value="0.5" />
					</DataTrigger>
				</Style.Triggers>
			</Style>
			<Style x:Key="LayerImageStyle" TargetType="Image">
				<Setter Property="VerticalAlignment" Value="Center" />
				<Setter Property="Visibility" Value="{Binding LegendVisibility}" />
				<Setter Property="Source" Value="{Binding SymbolSwatch}" />
			</Style>
			<Style x:Key="LayerTextBlockStyle" TargetType="TextBlock">
				<Setter Property="VerticalAlignment" Value="Center" />
				<Setter Property="Margin" Value="1" />
				<Setter Property="Text" Value="{Binding Header}" />
				<Style.Triggers>
					<DataTrigger Binding="{Binding Path=IsEnabled}" Value="True">
						<Setter Property="Opacity" Value="1.0" />
					</DataTrigger>
					<DataTrigger Binding="{Binding Path=IsEnabled}" Value="False">
						<Setter Property="Opacity" Value="0.5" />
					</DataTrigger>
				</Style.Triggers>
			</Style>
			<HierarchicalDataTemplate x:Key="LayerTreeItemTemplate" ItemsSource="{Binding Path=Items}">
				<StackPanel Orientation="Horizontal">
					<CheckBox Style="{StaticResource LayerCheckBoxStyle}" />
					<Image Style="{StaticResource LayerImageStyle}" />
					<TextBlock Style="{StaticResource LayerTextBlockStyle}" />
				</StackPanel>
			</HierarchicalDataTemplate>
		</UserControl.Resources>
		...
		<TreeView Grid.Column="0" DataContext="{Binding DataContexts[NavAddin.MapModelExt]}" ItemsSource="{Binding TOC}"
                  ItemTemplate="{StaticResource LayerTreeItemTemplate}" ItemContainerStyle="{StaticResource LayerTreeItemStyle}" />
	 
	*/

	public class TocDataItem : INotifyPropertyChanged
	{

		/*
		 * 100.5/100.6 BUG: Although group layers are supported, use of them causes
		 * GeodatabaseFeatureTable.QueryRelatedFeaturesAsync in a sublayer to return bad results.
		 * Using TocDataItem(true) reverts TocDataItem to an older behavior
		 * that imitates group layers.
		*/

		// BEGIN WORKAROUND
		private bool _bImitateGroupLayer = false;
		private bool _bExternalVis = true;
		private bool _bInternalVis = true;
		private TocDataItem _parent;
		public bool IsTopmost { get { return (_parent == null); } }
		// END WORKAROUND

		private bool _bIsLegendItem = false;
		private bool _bHasLegend = false;
		private bool _bIsEnabled = true;
		private bool _bIsExpanded = false;

		private ObservableCollection<TocDataItem> _items;
		private readonly ReadOnlyObservableCollection<TocDataItem> _propitems;
		private ILayerContent _pLayer;
		private SolidColorBrush _legendBackground;

		public string DisplayName { get; set; }
		public double MinScale { get; set; }
		public double MaxScale { get; set; }
		public bool HasLayerContent { get { return (_pLayer != null); } }
		public bool IsGroupLayer
		{
			get
			{
				if (!HasLayerContent) return false;
				return (_pLayer is GroupLayer);
			}
		}

		// Bindable properties

		public string Header { get { return DisplayName; } }

		public ReadOnlyObservableCollection<TocDataItem> Items { get { return _propitems; } }

		public Visibility LayerVisibility
		{
			get
			{
				if (_bIsLegendItem)
					return Visibility.Hidden;
				else
					return Visibility.Visible;
			}
		}

		public bool IsChecked
		{
			get
			{
				// BEGIN WORKAROUND
				if (_bImitateGroupLayer)
					return _bExternalVis;
				// END WORKAROUND
				if (!HasLayerContent)
					return false;
				return _pLayer.IsVisible;
			}
			set
			{
				// BEGIN WORKAROUND
				if (_bImitateGroupLayer)
				{
					if (HasLayerContent)
					{
						if (!_pLayer.CanChangeVisibility)
							return;
					}
					if (value == _bExternalVis)
						return;
					this.SetExternalVisibility(value);
					this.RaisePropertyChanged();
					return;
				}
				// END WORKAROUND
				if (!HasLayerContent)
					return;
				if (!_pLayer.CanChangeVisibility)
					return;
				if (value == _pLayer.IsVisible)
					return;
				_pLayer.IsVisible = value;
				this.RaisePropertyChanged();
			}
		}

		public bool IsEnabled
		{
			get { return _bIsEnabled; }
			set
			{
				if (value == _bIsEnabled)
					return;
				_bIsEnabled = value;
				this.RaisePropertyChanged();
			}
		}

		public bool IsExpanded
		{
			get { return _bIsExpanded; }
			set
			{
				if (value == _bIsExpanded)
					return;
				_bIsExpanded = value;
				this.RaisePropertyChanged();
			}
		}

		public Visibility LegendVisibility
		{
			get
			{
				if (_bIsLegendItem)
					return Visibility.Visible;
				else
					return Visibility.Hidden;
			}
		}

		public SolidColorBrush LegendBackground
		{
			get { return _legendBackground; }
			set
			{
				_legendBackground = value;
				this.RaisePropertyChanged();
			}
		}

		public ImageSource SymbolSwatch { get; protected set; }

		// INotifyPropertyChanged

		public event PropertyChangedEventHandler PropertyChanged;

		// Constructors

		public TocDataItem()
		{
			_items = new ObservableCollection<TocDataItem>();
			_propitems = new ReadOnlyObservableCollection<TocDataItem>(_items);
			DisplayName = "";
			MinScale = double.NaN;
			MaxScale = double.NaN;
		}

		// BEGIN WORKAROUND
		public TocDataItem(bool bImitateGroupLayer) : this()
		{
			_bImitateGroupLayer = bImitateGroupLayer;
		}
		// END WORKAROUND

		// Public methods

		public async Task SetLayer(ILayerContent pLyr, int iExpansionLevel = -1, int iCurrentLevel = 0)
		{
			await this.Init(pLyr, iExpansionLevel, iCurrentLevel);
		}

		// BEGIN WORKAROUND
		public void SetSubitems(List<TocDataItem> items, bool bCheck = false)
		{
			_items.Clear();
			foreach (TocDataItem item in items)
			{
				item._parent = this;
				_items.Add(item);
			}
			if (!bCheck)
				return;
			this.SetExternalVisibility(_bExternalVis);
		}
		// END WORKAROUND

		public List<Layer> GetMapLayers()
		{
			List<Layer> MapLayers = new List<Layer>();
			if (!HasLayerContent)
				return MapLayers;
			if (!IsGroupLayer)
			{
				if (_pLayer is Layer lyr)
					MapLayers.Add(lyr);
				return MapLayers;
			}
			int iCount = _items.Count;
			for (int i = iCount - 1; i >= 0; i--)
			{
				TocDataItem item = _items[i];
				MapLayers.AddRange(item.GetMapLayers());
			}
			return MapLayers;
		}

		public void CheckVisibility(double dScale)
		{
			if (double.IsNaN(dScale) || (dScale == 0))
				return;
			// BEGIN WORKAROUND
			if (_bImitateGroupLayer)
			{
				if (!IsTopmost)
					return;
				this.SetInternalVisibility(dScale);
				this.SetEnabled(dScale);
				return;
			}
			// END WORKAROUND
			IsEnabled = this.VisibleAtScale(dScale);
			foreach (TocDataItem item in _items)
				item.CheckVisibility(dScale);
		}

		// Event handlers

		private void MapBackgroundChanged(object sender, System.EventArgs e)
		{

			// Delete and recreate legend

			if (!_bHasLegend)
				return;
			_items.Clear();
			Task t = this.AddLegend();

		}

		// Private methods

		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private async Task Init(ILayerContent pLyr, int iExpansionLevel, int iCurrentLevel)
		{

			// Create layer item

			_pLayer = pLyr;
			_items.Clear();
			_bIsLegendItem = false;
			DisplayName = _pLayer.Name;
			IsChecked = _pLayer.IsVisible;
			if (iCurrentLevel <= iExpansionLevel)
				IsExpanded = true;

			// Set default min and max scale
			// This is necessary because ILayerContent.IsVisibleAtScale doesn't seem
			// to be independent of IsVisible

			if (pLyr is Layer lyr)
			{
				MinScale = lyr.MinScale;
				MaxScale = lyr.MaxScale;
			}
			else if (pLyr is ArcGISMapImageSublayer subLyr)
			{
				MinScale = subLyr.MinScale;
				MaxScale = subLyr.MaxScale;
			}

			// Get legend info and add legend items if present
			// [If legend items are present, assume there are no sublayers]

			await this.AddLegend();
			if (_bHasLegend)
			{
				AppViewExt.Current.MapBackgroundChanged += MapBackgroundChanged;
				return;
			}

			// Add sublayer items, if any

			if (_pLayer.SublayerContents.Count == 0)
				return;
			_items.Clear();
			List<TocDataItem> Subitems = new List<TocDataItem>();
			foreach (ILayerContent pSublayer in _pLayer.SublayerContents)
			{
				TocDataItem item = new TocDataItem();
				await item.Init(pSublayer, iExpansionLevel, iCurrentLevel + 1);
				Subitems.Add(item);
			}
			if (_pLayer is GroupLayer)
				Subitems.Reverse();
			foreach (TocDataItem item in Subitems)
				_items.Add(item);

		}

		private async Task AddLegend()
		{

			if (_pLayer is GroupLayer)
				return;
			IReadOnlyList<LegendInfo> Infos = await _pLayer.GetLegendInfosAsync();
			if (Infos == null)
				return;

			// Add legend items

			int iLegendCount = 0;
			foreach (LegendInfo info in Infos)
			{
				TocDataItem legItem = new TocDataItem()
				{
					DisplayName = info.Name,
					_bIsLegendItem = true
				};
				Symbol sym = info.Symbol;
				RuntimeImage ri = await sym.CreateSwatchAsync(96);
				legItem.SymbolSwatch = await ri.ToImageSourceAsync();
				System.Windows.Media.Color color = ResourceHelper.ConvertColor(AppViewExt.Current.MapBackground.Color);
				legItem.LegendBackground = new SolidColorBrush(color);
				_items.Add(legItem);
				iLegendCount++;
			}
			_bHasLegend = (iLegendCount > 0);

		}

		private bool VisibleAtScale(double dScale)
		{
			bool bInsideMin = (double.IsNaN(MinScale) || (MinScale == 0));
			if (!bInsideMin)
				bInsideMin = (dScale <= MinScale);
			bool bInsideMax = (double.IsNaN(MaxScale) || (MaxScale == 0));
			if (!bInsideMax)
				bInsideMax = (dScale >= MaxScale);
			bool bVis = (bInsideMin && bInsideMax);
			if (!HasLayerContent)
				return bVis;
			// IsVisibleAtScale is always false for ArcGISMapImageSublayer when default visibility is false
			if (_pLayer is ArcGISMapImageSublayer)
				return bVis;
			if (bVis)
				bVis = _pLayer.IsVisibleAtScale(dScale);
			return bVis;
		}

		// BEGIN WORKAROUND
		private void SetExternalVisibility(bool bVis)
		{
			_bExternalVis = bVis;
			bool bIntVis = bVis;
			if (!IsTopmost)
			{
				if (!_parent._bInternalVis)
					return;
			}
			this.SetInternalVisibility(bVis);
		}

		private void SetInternalVisibility(bool bVis)
		{
			if (bVis)
				_bInternalVis = _bExternalVis;
			else
				_bInternalVis = false;
			if (HasLayerContent)
			{
				_pLayer.IsVisible = _bInternalVis;
				return;
			}
			foreach (TocDataItem item in _items)
				item.SetInternalVisibility(_bInternalVis);
		}

		private void SetInternalVisibility(double dScale)
		{
			if (!_bExternalVis)
			{
				this.SetInternalVisibility(false);
				return;
			}
			if (HasLayerContent)
			{
				_bInternalVis = true;
				_pLayer.IsVisible = true;
				return;
			}
			_bInternalVis = this.VisibleAtScale(dScale);
			foreach (TocDataItem item in _items)
			{
				if (_bInternalVis)
					item.SetInternalVisibility(dScale);
				else
					item.SetInternalVisibility(false);
			}
		}

		private void SetEnabled(double dScale)
		{
			IsEnabled = this.VisibleAtScale(dScale);
			foreach (TocDataItem item in _items)
				item.SetEnabled(dScale);
		}
		// END WORKAROUND

	}

}
