﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Windows;
using System.Windows.Input;
using System.Xml.Linq;

using Extensibility;

namespace NavAddin
{

	// UserSettingsDialogModel enables saving and restoring map extent and TOC settings

	public class UserSettingsDialogModel : INotifyPropertyChanged, IWindowContent, ICommand
	{

		private static UserSettingsDialogModel _current;

		private XDocument _xSettings;
		private bool _bChanged = false;
		private bool _bViewSet = true;
		private bool _bVisibilitySet = false;
		private int _iSelectedIndex = -1;
		private string _sRecordName;

		// Bindable Properties

		public Visibility TOCVisibility { get; private set; }

		public bool ViewSet
		{
			get { return _bViewSet; }
			set
			{
				_bViewSet = value;
				this.Refresh();
			}
		}

		public bool VisibilitySet
		{
			get { return _bVisibilitySet; }
			set
			{
				_bVisibilitySet = value;
				this.Refresh();
			}
		}

		public ObservableCollection<string> RecordsSource { get; private set; }

		public int RecordsSelectedIndex
		{
			get { return _iSelectedIndex; }
			set
			{
				_iSelectedIndex = value;
				this.RaiseCanExecuteChanged();
				if (_iSelectedIndex == -1)
					return;
				RecordName = RecordsSource[_iSelectedIndex];
			}
		}

		public string RecordName 
		{
			get { return _sRecordName; }
			set
			{
				_sRecordName = value;
				this.RaisePropertyChanged();
			}
		}

		public ICommand Command { get { return this; } }

		// INotifyPropertyChanges

		public event PropertyChangedEventHandler PropertyChanged;

		// IWindowContent

		public string Title { get { return "User Settings"; } }

		public event EventHandler<EventArgs> ShowWindow;

		public event EventHandler<EventArgs> HideWindow;

		public void ShutDown() 
		{
			if (!_bChanged)
				return;
			UserSettingsHelper.SaveSettings();
		}

		// ICommand

		public bool CanExecute(object parameter) 
		{
			string sParam = parameter as string;
			switch (sParam)
			{
				case "Restore":
				case "Delete":
					return (RecordsSelectedIndex > -1);
				case "Save":
				case "Reset":
				case "Quit":
					return true;
			}
			return false;
		}

		public event EventHandler CanExecuteChanged;

		public void Execute(object parameter)
		{
			string sParam = parameter as string;
			UserSettingsType opt = UserSettingsType.View;
			if (VisibilitySet)
				opt = UserSettingsType.Visibility;
			switch (sParam)
			{
				case "Save":
					if (string.IsNullOrEmpty(RecordName))
					{
						MessageBox.Show("Invalid name");
						return;
					}
					UserSettingsHelper.Save(RecordName, opt);
					this.Refresh();
					break;
				case "Restore":
					UserSettingsHelper.Restore(RecordName, opt);
					break;
				case "Delete":
					UserSettingsHelper.Delete(RecordName, opt);
					this.Refresh();
					break;
				case "Reset":
					UserSettingsHelper.Reset(opt);
					break;
				case "Quit":
					this.RaiseHideWindow();
					break;
			}
			
		}

		// Public methods

		public static void Open()
		{
			if (_current == null)
			{
				try
				{
					_current = new UserSettingsDialogModel();
					MapApplication.Current.CreateChildWindow("NavAddin", "UserSettingsDialog", _current);
				}
				catch (Exception ex)
				{
					string sMsg = "Could not create User Settings Manager:\n" + ex.Message;
					MapApplication.Current.Error(sMsg);
					return;
				}
			}
			_current.RaiseShowWindow();
		}

		// Event handlers

		void _xSettings_Changed(object sender, XObjectChangeEventArgs e)
		{
			_bChanged = true;
		}

		// Private constructor

		private UserSettingsDialogModel()
		{
			_xSettings = UserSettingsHelper.OpenSettings();
			_xSettings.Changed += _xSettings_Changed;
			if (MapModelExt.Current.EnableTOC)
				TOCVisibility = Visibility.Visible;
			else
				TOCVisibility = Visibility.Hidden;
			RecordsSource = new ObservableCollection<string>();
			this.Refresh();
		}

		// Private methods

		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private void RaiseShowWindow()
		{
			if (this.ShowWindow == null)
				return;
			this.ShowWindow(this, new EventArgs());
		}

		private void RaiseHideWindow()
		{
			if (this.HideWindow == null)
				return;
			this.HideWindow(this, new EventArgs());
		}

		private void RaiseCanExecuteChanged()
		{
			if (this.CanExecuteChanged == null)
				return;
			this.CanExecuteChanged(this, new EventArgs());
		}

		private void Refresh()
		{
			RecordsSource.Clear();
			RecordsSelectedIndex = -1;
			RecordName = "";
			UserSettingsType opt = UserSettingsType.View;
			if (VisibilitySet)
				opt = UserSettingsType.Visibility;
			List<string> Items = UserSettingsHelper.GetSettings(opt);
			foreach (string sItem in Items)
				RecordsSource.Add(sItem);
			this.RaiseCanExecuteChanged();
		}

	}

}
