﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Interactivity;
using System.Windows.Input;

using Esri.ArcGISRuntime.Data;
using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Mapping;
using Esri.ArcGISRuntime.UI;
using Esri.ArcGISRuntime.UI.Controls;

using Extensibility;

namespace QueryAddin
{

	// QueryViewExt is a binding target and adds the following dependency properties:
	//		QuerySpatialRefChangedCommand, ICommand, OneWay
	//		QueryScaleChangedCommand, ICommand, OneWay
	// QueryViewExt also manages flash and graphic operations

	public class QueryViewExt : Behavior<MapView>
	{

		public static QueryViewExt Current { get; private set; }

		public Envelope MapExtent
		{
			get
			{
				if (AssociatedObject == null)
					return null;
				return AssociatedObject.VisibleArea.Extent;
			}
		}

		public double MapScale
		{
			get
			{
				if (AssociatedObject == null)
					return double.NaN;
				return AssociatedObject.MapScale;
			}
		}

		public Size DisplaySize
		{
			get
			{
				if (AssociatedObject == null)
					return new Size(double.NaN, double.NaN);
				return new Size(AssociatedObject.ActualWidth, AssociatedObject.ActualHeight);
			}
		}

		private GraphicsOverlay _ov;
		private FlashHelper _fh;
		private GraphicHelper _gh;
		private double _dCurrentScale = double.NaN;

		// Behavior

		protected override void OnAttached()
		{

			base.OnAttached();
			Current = this;
			AssociatedObject.SpatialReferenceChanged += AssociatedObject_SpatialReferenceChanged;
			AssociatedObject.NavigationCompleted += AssociatedObject_NavigationCompleted;

			// Create graphics overlay

			_ov = new GraphicsOverlay();
			_ov.Id = Guid.NewGuid().ToString();
			GraphicsOverlayCollection gc = AssociatedObject.GraphicsOverlays;
			gc.Add(_ov);
			gc.CollectionChanged += GraphicsOverlay_CollectionChanged;
			_fh = new FlashHelper(AssociatedObject, _ov);
			_gh = new GraphicHelper(AssociatedObject, _ov);

		}

		// QuerySpatialRefChangedCommand property
		// This receives from the binding source an ICommand to be executed when the spatial reference changes

		public ICommand QuerySpatialRefChangedCommand
		{
			get { return (ICommand)this.GetValue(QuerySpatialRefChangedCommandProperty); }
			set { this.SetValue(QuerySpatialRefChangedCommandProperty, value); }
		}

		public static readonly DependencyProperty QuerySpatialRefChangedCommandProperty =
			DependencyProperty.Register("QuerySpatialRefChangedCommand", typeof(ICommand), typeof(QueryViewExt));

		// QueryScaleChangedCommand property
		// This receives from the binding source an ICommand to be executed when the map scale changes

		public ICommand QueryScaleChangedCommand
		{
			get { return (ICommand)this.GetValue(QueryScaleChangedCommandProperty); }
			set { this.SetValue(QueryScaleChangedCommandProperty, value); }
		}

		public static readonly DependencyProperty QueryScaleChangedCommandProperty =
			DependencyProperty.Register("QueryScaleChangedCommand", typeof(ICommand), typeof(QueryViewExt));

		// Public methods

		public async Task<List<IdentifyLayerResult>> Identify(MapPoint pt, IdentifyOption opt)
		{
			// 100.x LIMITATION: MapView.IdentifyLayerAsync only works if geom is a point
			return await this.DoIdentify(pt, opt);
		}

		public void Flash(FlashParameters fp)
		{
			if (_fh == null)
				return;
			_fh.Flash(fp);
		}

		public void ProcessGraphic(GraphicParameters gp)
		{
			if (_gh == null)
				return;
			_gh.ProcessRequest(gp);
		}

		// Event handlers

		void GraphicsOverlay_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
		{

			// Make sure overlay is topmost
			// WARNING:  If another extension tries this, it will hang the application

			GraphicsOverlayCollection gc = AssociatedObject.GraphicsOverlays;
			int iCount = gc.Count;
			int iPos = -1;
			for (int i = 0; i < iCount; i++)
			{
				GraphicsOverlay ov = gc[i];
				if (ov.Id != _ov.Id)
					continue;
				iPos = i;
				break;
			}
			if (iPos == iCount - 1)
				return;
			gc.Move(iPos, iCount - 1);
			
		}

		void AssociatedObject_SpatialReferenceChanged(object sender, EventArgs e)
		{

			// Notify spatial reference changed

			SpatialReference sr = AssociatedObject.SpatialReference;
			if (QuerySpatialRefChangedCommand == null)
				return;
			if (!QuerySpatialRefChangedCommand.CanExecute(sr))
				return;
			QuerySpatialRefChangedCommand.Execute(sr);

		}

		private void AssociatedObject_NavigationCompleted(object sender, EventArgs e)
		{

			// Notify if map scale changed

			double dScale = AssociatedObject.MapScale;
			if (double.IsNaN(dScale))
				return;
			if (dScale <= 0)
				return;
			if (dScale == _dCurrentScale)
				return;
			_dCurrentScale = dScale;
			if (QueryScaleChangedCommand == null)
				return;
			if (!QueryScaleChangedCommand.CanExecute(dScale))
				return;
			QueryScaleChangedCommand.Execute(dScale);

		}

		// Private methods

		private async Task<List<IdentifyLayerResult>> DoIdentify(MapPoint ptLocation, IdentifyOption opt)
		{

			double dPixelTol = 6;
			long iMaxRecs = 50;

			bool bVis = (opt != IdentifyOption.All);
			bool bTopmost = (opt == IdentifyOption.Topmost);
			System.Windows.Point ptScreen = AssociatedObject.LocationToScreen(ptLocation);
			double dScale = AssociatedObject.MapScale;
			List<IdentifyLayerResult> Results = new List<IdentifyLayerResult>();
			LayerCollection Layers = AssociatedObject.Map.OperationalLayers;
			int iCount = Layers.Count;
			for (int i = iCount - 1; i > -1; i--)
			{
				Layer lyr = Layers[i];
				if (bVis)
				{
					if (!lyr.IsVisible)
						continue;
					if (!lyr.IsVisibleAtScale(dScale))
						continue;
				}
				IdentifyLayerResult result = await AssociatedObject.IdentifyLayerAsync(lyr, ptScreen, dPixelTol, false, iMaxRecs);
				if (result == null)
					continue;
				if ((result.GeoElements.Count == 0) && (result.SublayerResults.Count == 0))
					continue;
				Results.Add(result);
				if (bTopmost)
					break;
			}
			return Results;

		}

	}


}
