﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Threading.Tasks;

using Esri.ArcGISRuntime.ArcGISServices;
using Esri.ArcGISRuntime.Data;
using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Mapping;
using Esri.ArcGISRuntime.Mapping.Popups;

using Extensibility;

namespace QueryAddin
{

	// LayerHelper checks for queryable layers when the map changes

	public interface IQueryLayerHelper
	{
		Task<List<IQueryableLayer>> GetQueryableLayers(Map map);
	}

	public class BasicQueryLayerHelper : IQueryLayerHelper
	{

		// 100.x LIMITATION:  Geometric operations on curve geometries,
		// including spatial queries, is not supported.

		private const string CurveError = "Not implemented: operation does not support curves yet";

		private List<IQueryableLayer> _LayerList;
		private Dictionary<string, IQueryableLayer> _QueryableLayers = new Dictionary<string, IQueryableLayer>();

		// IQueryLayerHelper

		public async Task<List<IQueryableLayer>> GetQueryableLayers(Map map)
		{

			_LayerList = new List<IQueryableLayer>();

			// Get operational layers

			List<Layer> OpLayers = map.OperationalLayers.ToList();
			int iCount = OpLayers.Count;
			for (int i = iCount - 1; i >= 0; i-- )
			{

				Layer lyr = OpLayers[i];
				string sID = lyr.Id;
				if (_QueryableLayers.ContainsKey(sID))
				{
					_LayerList.Add(_QueryableLayers[sID]);
					continue;
				}

				if (lyr is GroupLayer)
				{
					IReadOnlyList<IQueryableLayer> Layers = await GetQueryableLayersFromGroup((GroupLayer)lyr);
					foreach (IQueryableLayer pLyr in Layers)
						_LayerList.Add(pLyr);
					continue;
				}

				if (lyr is ArcGISMapImageLayer)
				{
					IReadOnlyList<IQueryableLayer> Layers = await WebQueryLayerHelper.GetQueryableLayers((ArcGISMapImageLayer)lyr);
					foreach (IQueryableLayer pLyr in Layers)
						_LayerList.Add(pLyr);
					continue;
				}

				if (!(lyr is FeatureLayer))
				{
					Debug.WriteLine("BasicQueryLayerHelper layer not supported: " + lyr.Name);
					continue;
				}

				BasicQueryableLayer bqLyr = await BasicQueryableLayer.CreateFromFeatureLayer((FeatureLayer)lyr);
				if (bqLyr == null)
					continue;
				_LayerList.Add(bqLyr);
				_QueryableLayers[sID] = bqLyr;

			}

			// Get tables (assumes that names and service layer ids are unique)

			foreach (FeatureTable fTab in map.Tables.ToList())
			{
				if (!(fTab is ArcGISFeatureTable))
					continue;
				ArcGISFeatureTable afTab = (ArcGISFeatureTable)fTab;
				string sID = afTab.TableName + ";" + afTab.ServiceLayerId.ToString();
				if (_QueryableLayers.ContainsKey(sID))
				{
					_LayerList.Add(_QueryableLayers[sID]);
					continue;
				}
				IQueryableLayer pLyr = await BasicQueryableLayer.CreateFromFeatureTable(afTab);
				_LayerList.Add(pLyr);
				_QueryableLayers[sID] = pLyr;
			}
			return _LayerList;

		}

		// Public methods

		public static List<QueryFieldInfo> GetQueryFieldInfos(IReadOnlyList<Field> Fields, string sSubtypeField = null, IReadOnlyList<FeatureSubtype> Subtypes = null)
		{

			// Get info for standard fields

			string sSubtypeName = sSubtypeField.ToUpper();
			List<QueryFieldInfo> QueryFields = new List<QueryFieldInfo>();
			QueryFieldInfo qfSubtype = null;
			foreach (Field f in Fields)
			{

				QueryFieldInfo qfInfo = new QueryFieldInfo() { Name = f.Name };
				if (string.IsNullOrEmpty(f.Alias))
					qfInfo.Alias = f.Name; // This can happen with related record queries
				else
					qfInfo.Alias = f.Alias;
				qfInfo.Type = f.FieldType;
				if (f.Domain is CodedValueDomain cvd)
				{
					Dictionary<string, string> CodedVals = new Dictionary<string, string>();
					foreach (CodedValue cv in cvd.CodedValues)
						CodedVals[cv.Code.ToString()] = cv.Name;
					qfInfo.CodedValues = CodedVals;
				}
				QueryFields.Add(qfInfo);
				if (string.IsNullOrEmpty(sSubtypeField))
					continue;

				// Check for subtype

				if ((qfInfo.Name.ToUpper() != sSubtypeName) && (qfInfo.Alias.ToUpper() != sSubtypeName))
					continue;
				qfSubtype = qfInfo;

			}
			if (qfSubtype == null)
				return QueryFields;
			if (qfSubtype.CodedValues != null)
				return QueryFields; // Coded values already assigned from domain
			if (Subtypes == null)
				return QueryFields;
			if (Subtypes.Count == 0)
				return QueryFields;

			// Get coded values for subtypes

			Dictionary<string, string> cvs = new Dictionary<string, string>();
			foreach (FeatureSubtype fType in Subtypes)
				cvs[fType.Code.ToString()] = fType.Name;
			qfSubtype.CodedValues = cvs;
			return QueryFields;

		}

		public static List<QueryFieldInfo> GetQueryFieldInfos(ArcGISFeatureLayerInfo flInfo, IPopupSource pPopupSrc = null)
		{

			// If popup definition is present, use it to define the list of fields

			IReadOnlyList<Field> FieldList = flInfo.Fields;
			if (pPopupSrc != null)
			{
				PopupDefinition popDef = pPopupSrc.PopupDefinition;
				if (popDef != null)
				{
					Dictionary<string, Field> Fields = new Dictionary<string, Field>();
					foreach (Field f in flInfo.Fields)
						Fields[f.Name] = f;
					List<Field> VisibleFields = new List<Field>();
					foreach (PopupField pf in popDef.Fields)
					{
						string sFieldName = pf.FieldName;
						if (!Fields.ContainsKey(sFieldName))
							continue;
						VisibleFields.Add(Fields[sFieldName]);
					}
					FieldList = VisibleFields.AsReadOnly();
				}
			}
			return BasicQueryLayerHelper.GetQueryFieldInfos(FieldList, flInfo.SubtypeField, flInfo.FeatureSubtypes);

		}

		public static string GetDefaultDisplayField(List<QueryFieldInfo> Fields)
		{

			if (Fields == null)
				return null;

			// Find first text field whose name contains "NAME"

			foreach (QueryFieldInfo info in Fields)
			{
				if (info.Type != FieldType.Text)
					continue;
				if (info.Name.ToUpper().Contains("NAME"))
					return info.Name;
			}

			// Else find first text field

			foreach (QueryFieldInfo info in Fields)
			{
				if (info.Type == FieldType.Text)
					return info.Name;
			}

			// Else find first integer field

			foreach (QueryFieldInfo info in Fields)
			{
				if ((info.Type == FieldType.Int32) || (info.Type == FieldType.OID))
					return info.Name;
			}

			// Else return first field

			if (Fields.Count > 0)
				return Fields[0].Name;
			return null; // This should never happen

		}

		public static QueryRelatedResult CreateQueryRelatedResult(RelatedFeatureQueryResult relResult)
		{
			ArcGISFeatureTable fTabRel = relResult.RelatedTable;
			ArcGISFeatureLayerInfo flInfo = fTabRel.LayerInfo;
			QueryRelatedResult qrResult = new QueryRelatedResult()
			{
				Name = fTabRel.TableName,
				IsPopupEnabled = false,
				PopupDefinition = null,
				DisplayField = null,
				Fields = null
			};
			if (fTabRel.IsPopupEnabled && (fTabRel.PopupDefinition != null))
			{
				qrResult.IsPopupEnabled = true;
				qrResult.PopupDefinition = fTabRel.PopupDefinition;
			}
			else
			{
				qrResult.DisplayField = flInfo.DisplayFieldName;
				qrResult.Fields = BasicQueryLayerHelper.GetQueryFieldInfos(flInfo);
			}
			qrResult.RelatedFeatures = new List<Feature>();
			foreach (ArcGISFeature featRel in relResult.OfType<ArcGISFeature>())
				qrResult.RelatedFeatures.Add(featRel);
			return qrResult;
		}

		public static async Task<List<Feature>> QueryGeneral(ArcGISFeatureTable fTab, string sWhere, Geometry geom)
		{

			// Perform query

			List<Feature> Features = new List<Feature>();
			QueryParameters qp = new QueryParameters();
			bool bSpatial = false;
			if ((fTab.GeometryType != GeometryType.Unknown) && (geom != null))
			{
				qp.Geometry = geom;
				qp.SpatialRelationship = SpatialRelationship.Intersects;
				bSpatial = true;
			}
			if (!string.IsNullOrEmpty(sWhere))
				qp.WhereClause = sWhere;
			else if (!bSpatial)
				qp.WhereClause = "1 = 1";
			try
			{
				IEnumerable<Feature> Results = await fTab.QueryFeaturesAsync(qp);
				if (Results == null)
					return Features;
				foreach (ArcGISFeature feat in Results.OfType<ArcGISFeature>())
					Features.Add(feat);
			}
			catch (Exception ex)
			{
				string sMsg = "BasicQueryLayerHelper.QueryGeneral: " + ex.Message;
				Debug.WriteLine(sMsg);

				// 100.x LIMITATION

				if (ex.Message == CurveError)
					MapApplication.Current.Error("Curve geometry encountered.");

			}
			return Features;

		}

		public static async Task<List<string>> QueryUniqueValues(ArcGISFeatureTable fTab, string sFieldName, string sWhereClause)
		{

			// ReturnDistinctValues option not available

			List<string> UniqueValues = new List<string>();
			try
			{
				string sWhere = "1 = 1";
				if (!string.IsNullOrEmpty(sWhereClause))
					sWhere = sWhereClause;
				QueryParameters qp = new QueryParameters() { WhereClause = sWhere };
				IEnumerable<Feature> Results;
				Results = await fTab.QueryFeaturesAsync(qp);
				if (Results == null)
					return UniqueValues;
				HashSet<string> UniqueValueSet = new HashSet<string>();
				foreach (Feature feat in Results)
					UniqueValueSet.Add(feat.Attributes[sFieldName].ToString());
				UniqueValues = UniqueValueSet.ToList();
				if (UniqueValues.Count > 1)
					UniqueValues.Sort();
			}
			catch (Exception ex)
			{
				string sMsg = "BasicQueryLayerHelper.QueryUniqueValues: " + ex.Message;
				Debug.WriteLine(sMsg);
			}
			return UniqueValues;

		}

		public static async Task<List<Attachment>> QueryAttachments(ArcGISFeature feat)
		{

			// Check if table has attachments

			List<Attachment> Attachments = new List<Attachment>();
			if (feat == null)
				return Attachments;
			if (!(feat.FeatureTable is ArcGISFeatureTable fTab))
				return Attachments;
			if (!fTab.HasAttachments)
				return Attachments;

			// Query attachments

			try
			{
				IReadOnlyList<Attachment> AttachmentResults = await feat.GetAttachmentsAsync();
				if (AttachmentResults != null)
				{
					foreach (Attachment item in AttachmentResults)
						Attachments.Add(item);
				}
			}
			catch (Exception ex)
			{
				string sMsg = "BasicQueryLayerHelper.QueryAttachments: " + ex.Message;
				Debug.WriteLine(sMsg);
			}
			return Attachments;

		}

		public static async Task<List<QueryRelatedResult>> QueryRelated(ArcGISFeature feat)
		{

			// Get related features
			// 100.5/100.6 BUG: QueryRelatedFeaturesAsync will not work for a FeatureLayer
			// if it is a sublayer of a GroupLayer.
			// The workaround is to strip out group layers and imitate them instead.
			// 100.7: FIXED

			List<QueryRelatedResult> RelatedResults = new List<QueryRelatedResult>();
			if (feat == null)
				return RelatedResults;
			if (!(feat.FeatureTable is ArcGISFeatureTable fTab))
				return RelatedResults;
			foreach (RelationshipInfo relInfo in fTab.LayerInfo.RelationshipInfos)
			{
				RelatedQueryParameters param = new RelatedQueryParameters(relInfo);
				IReadOnlyList<RelatedFeatureQueryResult> relResults = null;
				try
				{
					relResults = await fTab.QueryRelatedFeaturesAsync(feat, param);
				}
				catch (Exception ex)
				{
					string sMsg = "BasicQueryLayerHelper.QueryRelated: " + ex.Message;
					Debug.WriteLine(sMsg);
				}
				if (relResults == null)
					continue;
				if (relResults.Count == 0)
					continue;
				RelatedResults.Add(BasicQueryLayerHelper.CreateQueryRelatedResult(relResults[0]));
			}
			return RelatedResults;

		}

		private static async Task<IReadOnlyList<IQueryableLayer>> GetQueryableLayersFromGroup(GroupLayer gLyr)
		{
			List<IQueryableLayer> Layers = new List<IQueryableLayer>();
			List<Layer> Sublayers = new List<Layer>();
			foreach (Layer lyr in gLyr.Layers)
				Sublayers.Add(lyr);
			Sublayers.Reverse();
			foreach (Layer lyr in Sublayers)
				await BasicQueryLayerHelper.ProcessSublayer(Layers, lyr);
			return Layers.AsReadOnly();
		}

		private static async Task ProcessSublayer(List<IQueryableLayer> Layers, Layer lyr)
		{
			if (lyr is GroupLayer gLyr)
			{
				IReadOnlyList<IQueryableLayer> Sublayers = await BasicQueryLayerHelper.GetQueryableLayersFromGroup(gLyr);
				Layers.AddRange(Sublayers);
				return;
			}
			if (!(lyr is FeatureLayer fLyr))
			{
				Debug.WriteLine("BasicQueryLayerHelper layer not supported: " + lyr.Name);
				return;
			}
			BasicQueryableLayer bqLyr = await BasicQueryableLayer.CreateFromFeatureLayer((FeatureLayer)lyr);
			if (bqLyr == null)
				return;
			Layers.Add(bqLyr);
		}

	}
}
