﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Text;
using System.Threading.Tasks;

using Esri.ArcGISRuntime.Data;

namespace QueryAddin
{

	// FindEngine supports find operations

	public enum FindOption
	{
		Visible,
		All,
		Layer
	}

	public delegate void FindStatus(string sMsg);

	public class FindResult
	{

		public string Value { get; private set; }
		public string LayerName { get; private set; }
		public string FieldName { get; private set; }
		public bool HasGeometry 
		{ 
			get 
			{  
				if (Feature == null)
					return false;
				else if (Feature.Geometry == null)
					return false;
				else if (Feature.Geometry.IsEmpty)
					return false;
				return true;
			} 
		}
		public Feature Feature { get; private set; }
		public IQueryableLayer Layer { get; private set; }

		public FindResult(string val, string fieldName, Feature feat, IQueryableLayer pLyr)
		{
			Value = val;
			LayerName = pLyr.DisplayName;
			FieldName = fieldName;
			Feature = feat;
			Layer = pLyr;
		}

	}

	public class FindEngine
	{

		private const bool UseWebQuery = false;

		private List<IQueryableLayer> _Layers = new List<IQueryableLayer>();
		private ObservableCollection<string> _LayerItems = new ObservableCollection<string>();
		private ObservableCollection<string> _FieldItems = new ObservableCollection<string>();
		private int _iLayerIndex;
		private int _iFieldIndex;
		private string _sSearchText;
		private bool _bContains;
		private FindOption _findOpt = FindOption.Visible;
		private bool _bBusy = false;
		private HashSet<string> _QueriedServices = new HashSet<string>();

		// Public methods

		public void Initialize(List<IQueryableLayer> QueryableLayers)
		{

			// Assumes map layers are already initialized

			_Layers.Clear();
			_LayerItems.Clear();
			_FieldItems.Clear();
			foreach (IQueryableLayer pLyr in QueryableLayers)
				_Layers.Add(pLyr);
			if (_Layers.Count == 0)
				return;
			_LayerItems.Add("<Visible layers>");
			_LayerItems.Add("<All layers>");
			foreach (IQueryableLayer pLyr in _Layers)
				_LayerItems.Add(pLyr.DisplayName);
			_iLayerIndex = 0;
			this.UpdateFieldList();

		}

		public ObservableCollection<string> LayerItemList { get { return _LayerItems; } }

		public ObservableCollection<string> FieldItemList { get { return _FieldItems; } }
		
		public void SetLayerItem(int Index)
		{
			_iLayerIndex = Index;
			if (_iLayerIndex < 0)
				_iLayerIndex = 0;
			if (_iLayerIndex >= _LayerItems.Count)
				_iLayerIndex = _LayerItems.Count - 1;
			switch (_iLayerIndex)
			{
				case 0:
					_findOpt = FindOption.Visible;
					break;
				case 1:
					_findOpt = FindOption.All;
					break;
				default:
					_findOpt = FindOption.Layer;
					break;
			}
			this.UpdateFieldList();
		}

		public void SetFieldItem(int Index)
		{ 
			// Set to -1 for all fields
			_iFieldIndex = Index;
			if (_iFieldIndex < -1)
				_iFieldIndex = -1;
			if (_iFieldIndex >= _FieldItems.Count)
				_iFieldIndex = _FieldItems.Count - 1;
		}

		public async Task<List<FindResult>> FindAsync(string SearchText, bool Contains, FindStatus callback = null)
		{

			if (_bBusy)
				return null;
			_bBusy = true;
			_sSearchText = SearchText;
			_bContains = Contains;
			_QueriedServices.Clear();

			List<FindResult> ResultList = new List<FindResult>();
			IQueryLayerHelper pLayerHelper = QueryModelExt.Current.LayerHelper;
			List<IQueryableLayer> LayerList = this.GetSearchLayers();
			bool bWeb = UseWebQuery;
			foreach (IQueryableLayer pLyr in LayerList)
			{
				string sMsg = "Querying " + pLyr.DisplayName + "...";
				callback?.Invoke(sMsg);
				Debug.Print(sMsg);
				List<FindResult> LayerResults = null;
				if (bWeb && (pLyr is WebQueryableLayer) && (_findOpt != FindOption.Layer))
				{
					WebQueryableLayer wqLyr = (WebQueryableLayer)pLyr;
					if (_QueriedServices.Contains(wqLyr.MapImageLayerId))
						continue;
					_QueriedServices.Add(wqLyr.MapImageLayerId);
					LayerResults = await this.FindFromWebLayerAsync(wqLyr.MapImageLayerId, LayerList);
				}
				else
					LayerResults = await this.FindFromLayerAsync(pLyr);
				if (LayerResults == null)
					continue;
				foreach (FindResult result in LayerResults)
					ResultList.Add(result);
			}
			_bBusy = false;
			return ResultList;

		}

		public void ScaleChanged()
		{
			if (_findOpt == FindOption.Visible)
				this.UpdateFieldList();
		}

		// Private methods

		private void UpdateFieldList()
		{

			// Build list of unique field aliases

			_FieldItems.Clear();
			_iFieldIndex = -1;
			HashSet<string> AliasSet = new HashSet<string>();
			List<IQueryableLayer> LayerList = this.GetSearchLayers();
			foreach (IQueryableLayer pLyr in LayerList)
			{
				foreach (QueryFieldInfo info in pLyr.Fields)
					AliasSet.Add(info.Alias);
			}
			foreach (string sName in AliasSet)
				_FieldItems.Add(sName);

		}

		private List<IQueryableLayer> GetSearchLayers()
		{

			List<IQueryableLayer> LayerList = new List<IQueryableLayer>();
			if (_findOpt == FindOption.Layer)
			{
				int i = _iLayerIndex - 2;
				int iMax = _Layers.Count - 1;
				if ((i < 0) || (i > iMax))
				{
					// This could happen if a map were added without any queryable layers
					string sMsg = "ERROR: FindEngine: index " + i.ToString() + " outside range 0-" + iMax.ToString();
					Debug.Print(sMsg);
					return LayerList;
				}
				LayerList.Add(_Layers[i]);
				return LayerList;
			}

			double dScale = QueryViewExt.Current.MapScale;
			bool bVis = (_findOpt == FindOption.Visible);
			foreach (IQueryableLayer pLyr in _Layers)
			{
				if (bVis && (!pLyr.IsVisible(dScale)))
					continue;
				LayerList.Add(pLyr);
			}
			return LayerList;
		}

		private async Task<List<FindResult>> FindFromLayerAsync(IQueryableLayer pLyr)
		{

			// Get field info and build field list

			Dictionary<string, QueryFieldInfo> FieldInfos = new Dictionary<string, QueryFieldInfo>();
			List<string> FieldNames = new List<string>();
			bool bAllFields = (_iFieldIndex == -1);
			foreach (QueryFieldInfo info in pLyr.Fields)
			{
				FieldInfos[info.Alias] = info;
				if (!bAllFields)
					continue;
				FieldNames.Add(info.Alias);
			}
			if (!bAllFields)
			{
				string sName = _FieldItems[_iFieldIndex];
				if (!FieldInfos.ContainsKey(sName))
					return null;
				QueryFieldInfo info = FieldInfos[sName];
				FieldNames.Add(info.Alias);
			}

			// Build and execute query

			string sWhere = this.BuildQuery(FieldInfos, FieldNames);
			List<Feature> QueryResults = await pLyr.Query(sWhere);
			if (QueryResults == null)
				return null;

			// Process results

			List<FindResult> FindResults = new List<FindResult>();
			foreach (Feature Feat in QueryResults)
			{
				FindResult result = this.ProcessReturnFeature(Feat, FieldInfos, FieldNames, pLyr);
				if (result == null)
					continue;
				FindResults.Add(result);
			}
			return FindResults;

		}

		private async Task<List<FindResult>> FindFromWebLayerAsync(string sMapLayerId, List<IQueryableLayer> SearchLayers)
		{

			// Get sublayer list

			List<long> SublayerList = new List<long>();
			Dictionary<long, WebQueryableLayer> Sublayers = new Dictionary<long, WebQueryableLayer>();
			double dScale = QueryViewExt.Current.MapScale;
			bool bVis = (_findOpt == FindOption.Visible);
			foreach (IQueryableLayer pLyr in SearchLayers)
			{
				if (!(pLyr is WebQueryableLayer))
					continue;
				WebQueryableLayer wqLyr = (WebQueryableLayer)pLyr;
				if (wqLyr.MapImageLayerId != sMapLayerId)
					continue;
				if (bVis)
				{
					if (!wqLyr.IsVisible(dScale))
						continue;
				}
				SublayerList.Add(wqLyr.SublayerId);
				Sublayers[wqLyr.SublayerId] = wqLyr;
			}

			// If only one sublayer, don't bother calling the map service find operation

			if (SublayerList.Count == 1)
				return await this.FindFromLayerAsync(Sublayers[SublayerList[0]]);

			// Perform web find

			WebQueryLayerHelper wh = WebQueryLayerHelper.Instances[sMapLayerId];
			WebFindParameters param = new WebFindParameters()
			{
				SearchText = _sSearchText,
				Contains = _bContains,
				Layers = SublayerList
			};
			if (_iFieldIndex > -1)
			{
				param.Fields = new List<string>();
				param.Fields.Add(FieldItemList[_iFieldIndex]);
			}
			Dictionary<long, List<WebFindResult>> WebResults = await WebUtilities.Find(wh.ServiceUrl, param);
			if (WebResults == null)
				return null;

			// Build find results

			List<long> SublayerHits = new List<long>();
			foreach (long i in WebResults.Keys)
				SublayerHits.Add(i);
			if (SublayerHits.Count > 1)
				SublayerHits.Sort();
			List<FindResult> FindResults = new List<FindResult>();
			foreach (long i in SublayerHits)
			{
				List<WebFindResult> wfResults = WebResults[i];
				IQueryableLayer pLyr = Sublayers[i];
				foreach (WebFindResult wfResult in wfResults)
				{
					FindResult result = new FindResult(wfResult.Value, wfResult.FoundFieldName, wfResult.Feature, pLyr);
					FindResults.Add(result);
				}
			}
			return FindResults;

		}

		private string BuildQuery(Dictionary<string, QueryFieldInfo> FieldInfos, List<string> FieldNames)
		{

			StringBuilder sb = new StringBuilder();
			bool bFirst = true;
			foreach (string sAlias in FieldNames)
			{

				if (!FieldInfos.ContainsKey(sAlias))
					continue;
				QueryFieldInfo info = FieldInfos[sAlias];
				FieldType fType = info.Type;
				string sName = info.Name;

				string sSubClause = "";
				string sQueryName;
				string sQueryVal;
				DateTime dtVal;
				decimal dVal;
				switch (fType)
				{

					case FieldType.Text:
					case FieldType.Guid:
					case FieldType.GlobalID:

						// String fields can honor both partial match and case sensitivity options
						// Guid fields are automatically treated as case insensitive

						if (_bContains || (fType != FieldType.Text))
						{
							sQueryName = "lower(" + sName + ")";
							sQueryVal = _sSearchText.ToLower();
						}
						else
						{
							sQueryName = sName;
							sQueryVal = _sSearchText;
						}
						if (_bContains)
							sSubClause = sQueryName + " LIKE '%" + sQueryVal + "%'";
						else
							sSubClause = sQueryName + " = '" + sQueryVal + "'";
						break;

					case FieldType.Date:

						// GDB dates are compared using the SQLite datetime function

						if (DateTime.TryParse(_sSearchText, out dtVal))
							sSubClause = "datetime(" + sName + ") = datetime('" + dtVal.ToString("yyyy-MM-dd HH:mm:ss") + "')";
						break;

					case FieldType.Float64:
					case FieldType.Float32:
					case FieldType.Int32:
					case FieldType.Int16:
					case FieldType.OID:
						if (Decimal.TryParse(_sSearchText, out dVal))
							sSubClause = sName + " = " + _sSearchText;
						break;

				}
				if (sSubClause == "")
					continue;
				if (bFirst)
					bFirst = false;
				else
					sb.Append(" OR ");
				sb.Append(sSubClause);

			}
			string sWhere = sb.ToString();
			return sWhere;

		}

		private FindResult ProcessReturnFeature(Feature Feat, Dictionary<string, QueryFieldInfo> FieldInfos, List<string> FieldNames, IQueryableLayer pLyr)
		{

			// Find matching field and value

			string sFieldName = "";
			string sFieldVal = "";
			foreach (string sAlias in FieldNames)
			{

				if (!FieldInfos.ContainsKey(sAlias))
					continue;
				QueryFieldInfo info = FieldInfos[sAlias];
				FieldType fType = info.Type;
				string sName = info.Name;
				if (!Feat.Attributes.ContainsKey(sName))
					continue;
				object oVal = Feat.Attributes[sName];
				if (oVal == null)
					continue;

				// If only one field is searched, that's the result

				string sVal = oVal.ToString();
				if (FieldNames.Count == 1)
				{
					sFieldName = sName;
					sFieldVal = sVal;
					break;
				}

				// Otherwise, see if attribute value matches

				string sFeatVal;
				string sSearchVal;
				DateTime dtVal;
				decimal dVal;
				switch (fType)
				{

					case FieldType.Text:
					case FieldType.Guid:
					case FieldType.GlobalID:
						if (_bContains || (fType != FieldType.Text))
						{
							sFeatVal = sVal.ToLower();
							sSearchVal = _sSearchText.ToLower();
						}
						else
						{
							sFeatVal = sVal;
							sSearchVal = _sSearchText;
						}
						if ((sFeatVal == sSearchVal) || (_bContains && sFeatVal.Contains(sSearchVal)))
						{
							sFieldName = sName;
							sFieldVal = sVal;
						}
						break;

					case FieldType.Date:
						if (DateTime.TryParse(_sSearchText, out dtVal))
							if ((DateTime)oVal == dtVal)
							{
								sFieldName = sName;
								sFieldVal = sVal;
							}
						break;

					case FieldType.Float64:
					case FieldType.Float32:
					case FieldType.Int32:
					case FieldType.Int16:
					case FieldType.OID:
						if (decimal.TryParse(_sSearchText, out dVal))
							if (Convert.ToDecimal(oVal) == dVal)
							{
								sFieldName = sName;
								sFieldVal = sVal;
							}
						break;
				}
				if (sFieldName != "")
					break;

			} // switch fType
			if (sFieldName == "")
				return null;

			// Create and return result

			FindResult result = new FindResult(sFieldVal, sFieldName, Feat, pLyr);
			return result;

		}

	}


}
