﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Windows;
using System.Windows.Threading;

using Color = System.Drawing.Color;

using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Mapping;
using Esri.ArcGISRuntime.Symbology;
using Esri.ArcGISRuntime.UI;
using Esri.ArcGISRuntime.UI.Controls;

using Extensibility;

namespace QueryAddin
{

	// FlashHelper performs feature zoom/pan and flash operations

	public enum FlashNavigation
	{
		None,
		Pan,
		Zoom
	}

	public class FlashParameters
	{
		public Geometry Geometry = null;
		public bool Flash = false;
		public bool ShowFlashLines = false;
		public FlashNavigation Navigate = FlashNavigation.None;
	}

	public class FlashHelper
	{

		private const string ColorResourceKey = "QueryAddin.FlashColor";

		private MapView _mv;
		private GraphicsOverlay _ov;
		private double _dPointScale = 1000;
		private Color _color;
		private FlashParameters _param = null;

		// Constructor

		public FlashHelper(MapView mv, GraphicsOverlay ov, double dPointScale = 1000)
		{

			_mv = mv;
			_ov = ov;
			_dPointScale = dPointScale;

			// Get flash color or create default

			bool bUseDefault = true;
			ResourceDictionary rd = ResourceHelper.GetResources("QueryAddin");
			if (rd.Contains(ColorResourceKey))
			{
				object val = rd[ColorResourceKey];
				if (val is System.Windows.Media.Color)
				{
					_color = ResourceHelper.ConvertColor((System.Windows.Media.Color)val);
					bUseDefault = false;
				}
			}
			if (bUseDefault)
			{
				// ArcReader: R = 0x0, G = 0x64, B = 0x0
				_color = Color.FromArgb(0xFF, 0xFF, 0x0, 0x0);
			}

		}

		// Public methods

		public void Flash(FlashParameters param)
		{
			if (param == null)
				return;
			if (param.Geometry == null)
				return;
			if (param.Geometry.IsEmpty)
				return;
			_param = param;
			if (param.Navigate != FlashNavigation.None)
			{
				this.Navigate();
				return;
			}
			this.FlashGeometry();
		}

		// Event handlers

		void MapView_DrawStatusChanged(object sender, DrawStatusChangedEventArgs e)
		{
			if (e.Status != DrawStatus.Completed)
				return;
			_mv.DrawStatusChanged -= MapView_DrawStatusChanged;
			this.FlashGeometry();
		}

		// Private methods

		private void Navigate()
		{
			Envelope envCurrent = _mv.VisibleArea.Extent;
			Envelope envNew = null;
			if (_param == null)
				return;
			if (_param.Navigate == FlashNavigation.Pan)
				envNew = this.GetPanExtent();
			else
				envNew = this.GetZoomExtent();
			if (envNew == null)
				return;
			if (this.Equals(envNew, envCurrent))
			{
				this.FlashGeometry();
				return;
			}
			_mv.DrawStatusChanged += MapView_DrawStatusChanged;
			_mv.SetViewpointAsync(new Viewpoint(envNew));
			return;
		}

		private void FlashGeometry()
		{

			if (_param == null)
				return;
			if (!_param.Flash)
			{
				_param = null;
				return;
			}
			bool bFlashLines = _param.ShowFlashLines;
			Envelope envExtent = _mv.VisibleArea.Extent;

			// 100.5 BUG: Equivalent spatial references are not being considered equal
			// 100.7: FIXED [Keep code anyway because it's a good practice]
			// 100.12: IT'S BACK

			Geometry geom = _param.Geometry;
			if (!envExtent.SpatialReference.Equals(geom.SpatialReference))
				geom = GeometryEngine.Project(_param.Geometry, envExtent.SpatialReference);
			_param = null;
			if (GeometryEngine.Disjoint(envExtent, geom.Extent))
				return;

			// Create graphics

			Graphic grFeat = this.GetGeomGraphic(geom);
			Graphic grLines = null;
			if (bFlashLines)
				grLines = this.GetFlashLinesGraphic(geom, envExtent);

			// Add and remove graphic(s)
			// TO DO:  It would be cool to try some real animation

			int iCounter = 0;
			DispatcherTimer timer = new DispatcherTimer();
			timer.Interval = new TimeSpan(0, 0, 0, 0, 100); // 100 ms
			timer.Tick += delegate(object s, EventArgs args)
			{
				bool bStop = false;
				switch (iCounter)
				{
					case 0:
						_ov.Graphics.Add(grFeat);
						break;
					case 3:
						if (bFlashLines)
							_ov.Graphics.Add(grLines);
						else
							bStop = true;
						break;
					case 4:
						bStop = true;
						break;
				}
				iCounter++;
				if (bStop)
				{
					_ov.Graphics.Remove(grFeat);
					if (bFlashLines)
						_ov.Graphics.Remove(grLines);
					timer.Stop();
				}

			};
			timer.Start();

		}

		private Envelope GetPanExtent()
		{
			Envelope envCurrent = _mv.VisibleArea.Extent;
			Geometry geomPan = GeometryEngine.Project(_param.Geometry, envCurrent.SpatialReference);
			MapPoint ptPan = geomPan.Extent.GetCenter();
			EnvelopeBuilder eb = new EnvelopeBuilder(envCurrent);
			eb.CenterAt(ptPan);
			return eb.ToGeometry();
		}

		private Envelope GetZoomExtent()
		{
			Envelope envCurrent = _mv.VisibleArea.Extent;
			Geometry geomZoom = GeometryEngine.Project(_param.Geometry, envCurrent.SpatialReference);
			Envelope envZoom = geomZoom.Extent;
			EnvelopeBuilder eb = new EnvelopeBuilder(envZoom.SpatialReference);
			if ((envZoom.Width > 0) && (envZoom.Height > 0))
			{
				eb.ReplaceGeometry(envZoom);
				eb.Expand(1.1);
				return eb.ToGeometry();
			}
			eb.ReplaceGeometry(_mv.VisibleArea.Extent);
			eb.CenterAt(envZoom.GetCenter());
			double dSF = 1;
			double dW = envZoom.Width * 1.1 / eb.Width;
			double dH = envZoom.Height * 1.1 / eb.Height;
			if (dH > dSF)
				dSF = dH;
			else
				dSF = dW;
			if (dSF == 0)
				dSF = _dPointScale / _mv.MapScale;
			eb.Expand(dSF);
			return eb.ToGeometry();
		}

		private bool Equals(Envelope env1, Envelope env2)
		{
			double dMargin = _mv.UnitsPerPixel;
			if (!env1.SpatialReference.IsEqual(env2.SpatialReference))
				return false;
			double[] Vals1 = { env1.XMin, env1.YMin, env1.XMax, env1.YMax };
			double[] Vals2 = { env2.XMin, env2.YMin, env2.XMax, env2.YMax };
			for (int i = 0; i < 4; i++)
			{
				double dVal1 = Vals1[i];
				double dVal2 = Vals2[i];
				if (Math.Abs(dVal2 - dVal1) > dMargin)
					return false;
			}
			return true;
		}

		private Graphic GetGeomGraphic(Geometry geom)
		{

			// Create symbol

			Geometry g = geom;
			Symbol symFeat;

			if ((g is MapPoint) || (g is Multipoint))
			{
				SimpleMarkerSymbol sm = new SimpleMarkerSymbol();
				sm.Color = _color;
				sm.Style = SimpleMarkerSymbolStyle.Circle;
				sm.Size = 16;
				symFeat = sm;
			}
			else if (g is Polyline)
			{
				SimpleLineSymbol sl = new SimpleLineSymbol();
				sl.Color = _color;
				sl.Width = 4;
				symFeat = sl;
			}
			else
			{
				if (!(g is Polygon))
					g = geom.Extent;
				SimpleFillSymbol sf = new SimpleFillSymbol();
				sf.Color = _color;
				SimpleLineSymbol sl = new SimpleLineSymbol();
				sl.Color = _color;
				sl.Width = 4;
				sf.Outline = sl;
				symFeat = sf;
			}

			// Create graphic 

			Graphic grFeat = new Graphic();
			grFeat.Geometry = g;
			grFeat.Symbol = symFeat;
			return grFeat;

		}

		private Graphic GetFlashLinesGraphic(Geometry geom, Envelope envExtent)
		{

			// Create crosshair geometry

			MapPoint pt = GeometryHelper.GetCenter(geom);
			PolylineBuilder pb = new PolylineBuilder(envExtent.SpatialReference);
			pb.AddPart(new List<LineSegment>());
			pb.AddPoint(envExtent.XMin, pt.Y);
			pb.AddPoint(envExtent.XMax, pt.Y);
			pb.AddPart(new List<LineSegment>());
			pb.AddPoint(pt.X, envExtent.YMin);
			pb.AddPoint(pt.X, envExtent.YMax);
			Polyline plFlash = pb.ToGeometry();

			// Create flash lines graphic

			Graphic grLines = new Graphic();
			grLines.Geometry = plFlash;
			SimpleLineSymbol sl = new SimpleLineSymbol();
			sl.Color = _color;
			sl.Width = 2;
			grLines.Symbol = sl;
			return grLines;

		}




	}

}
