﻿using System.Collections.Generic;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;

using Color = System.Drawing.Color;

using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Mapping;
using Esri.ArcGISRuntime.Symbology;
using Esri.ArcGISRuntime.UI;
using Esri.ArcGISRuntime.UI.Controls;

using Extensibility;

namespace QueryAddin
{

	// GraphicHelper performs Measure, GoToXY, Highlight, Erase, and Text graphical operations

	public enum GraphicAction
	{
		AddXYPoint,
		AddXYLabeledPoint,
		AddXYCallout,
		AddMeasure,
		AddHighlightYellow,
		AddHighlightBlue,
		AddHighlightGreen,
		AddMapText,
		ClearXY,
		ClearMeasure,
		EraseHighlightEntire,
		EraseHighlightSmall,
		EraseHighlightMedium,
		EraseHighlightLarge,
		ClearHighlight,
		ClearText
	}

	public struct GraphicParameters
	{
		public GraphicAction Action;
		public Geometry Geometry;
		public string Label;
		public MapTextModel MapText;
	}

	public class GraphicHelper
	{

		private enum ClearType
		{
			XY,
			Measure,
			Highlight,
			Text
		}

		private const string XYTag = "GoToXYResult";
		private const string MeasureTag = "MeasureResult";
		private const string HighlightTag = "Highlight";
		private const string TextTag = "MapText";

		private MapView _mv;
		private GraphicsOverlay _gLayer;
		private OverlayItemsControl _overlays;
		private Style _styleCallout = null;
		private Color _colorXY = Color.Black;
		private TextSymbol _tsXY = null;
		private Color _colorMeasureLabel = Color.Red;
		private Color _colorMeasureOutline = Color.White;
		private TextSymbol _tsMeasure = null;
		private SimpleLineSymbol _symMeasureLine;
		private SimpleFillSymbol _symMeasurePolygon;
		private Style _styleMapText = null;
		private bool _bMoving = false;
		private MapTextModel _mtMoving;
		private TextBlock _tbMoving;
		private bool _bEditorEnabled;
		private bool _bEditorVisible;

		// Constructors

		public GraphicHelper(MapView mv, GraphicsOverlay gLayer)
		{

			_mv = mv;
			_gLayer = gLayer;
			_overlays = mv.Overlays;

			// Get callout style

			ResourceDictionary rd = ResourceHelper.GetResources("QueryAddin");
			string sKey = "QueryAddin.GoToXYCalloutStyle";
			if (rd.Contains(sKey))
			{
				object val = rd[sKey];
				_styleCallout = val as Style;
			}

			// Get graphic symbols

			sKey = "QueryAddin.GoToXYColor";
			if (rd.Contains(sKey))
			{
				object obj = rd[sKey];
				if (obj is Color)
					_colorXY = (Color)obj;
			}
			_tsXY = ResourceHelper.GetTextSymbol("QueryAddin", "QueryAddin.GoToXYTextStyle");
			if (_tsXY == null)
			{
				_tsXY = new TextSymbol()
				{
					FontFamily = "Arial",
					Size = 12,
					Color = _colorXY
				};
			}
			sKey = "QueryAddin.MeasureSumLabelColor";
			if (rd.Contains(sKey))
			{
				object obj = rd[sKey];
				if (obj is Color)
					_colorMeasureLabel = (Color)obj;
			}
			sKey = "MapViewBackground";
			if (rd.Contains(sKey))
			{
				object obj = rd[sKey];
				if (obj is Color)
					_colorMeasureOutline = (Color)obj;
			}
			_tsMeasure = ResourceHelper.GetTextSymbol("QueryAddin", "QueryAddin.MeasureSumTextStyle");
			if (_tsMeasure == null)
			{
				_tsMeasure = new TextSymbol()
				{
					FontFamily = "Arial",
					Size = 12,
					FontWeight = Esri.ArcGISRuntime.Symbology.FontWeight.Bold,
					Color = _colorMeasureLabel
				};
			}
			_tsMeasure.HaloColor = _colorMeasureOutline;
			_tsMeasure.HaloWidth = 2;
			_symMeasureLine = ResourceHelper.GetLineSymbol("QueryAddin", "QueryAddin.MeasureSumLineSymbol");
			if (_symMeasureLine == null)
			{
				SimpleLineSymbol sym = new SimpleLineSymbol();
				sym.Style = SimpleLineSymbolStyle.Solid;
				sym.Color = Color.Red;
				sym.Width = 2;
				_symMeasureLine = sym;
			}
			_symMeasurePolygon = ResourceHelper.GetFillSymbol("QueryAddin", "QueryAddin.MeasureSumPolygonSymbol");
			if (_symMeasurePolygon == null)
			{
				SimpleFillSymbol sym = new SimpleFillSymbol();
				sym.Style = SimpleFillSymbolStyle.Null;
				sym.Outline = _symMeasureLine;
				_symMeasurePolygon = sym;
			}

			// Get MapText style

			sKey = "QueryAddin.MapTextStyle";
			if (rd.Contains(sKey))
			{
				object val = rd[sKey];
				_styleMapText = val as Style;
			}

			// Set up label definition class for measures

			LabelDefinition ld = this.GetMeasureLabelDef();
			_gLayer.LabelDefinitions.Add(ld);
			_gLayer.LabelsEnabled = true;

		}

		// Public methods

		public void ProcessRequest(GraphicParameters gp)
		{
			if (_gLayer == null)
				return;
			switch (gp.Action)
			{
				case GraphicAction.AddXYPoint:
					this.AddXYGraphic(gp.Geometry, false);
					break;
				case GraphicAction.AddXYLabeledPoint:
					this.AddXYGraphic(gp.Geometry, true, gp.Label);
					break;
				case GraphicAction.AddXYCallout:
					this.AddXYCallout(gp.Geometry, gp.Label);
					break;
				case GraphicAction.AddMeasure:
					this.AddMeasureSumGraphic(gp.Geometry, gp.Label);
					break;
				case GraphicAction.AddHighlightYellow:
					this.AddHighlightGraphic(HighlighterType.Yellow, gp.Geometry);
					break;
				case GraphicAction.AddHighlightBlue:
					this.AddHighlightGraphic(HighlighterType.Blue, gp.Geometry);
					break;
				case GraphicAction.AddHighlightGreen:
					this.AddHighlightGraphic(HighlighterType.Green, gp.Geometry);
					break;
				case GraphicAction.AddMapText:
					this.AddMapText(gp.Geometry, gp.MapText);
					break;
				case GraphicAction.ClearXY:
					this.ClearGraphics(ClearType.XY);
					break;
				case GraphicAction.EraseHighlightEntire:
					this.EraseHighlightGraphics(EraserType.Entire, gp.Geometry);
					break;
				case GraphicAction.EraseHighlightSmall:
					this.EraseHighlightGraphics(EraserType.Small, gp.Geometry);
					break;
				case GraphicAction.EraseHighlightMedium:
					this.EraseHighlightGraphics(EraserType.Medium, gp.Geometry);
					break;
				case GraphicAction.EraseHighlightLarge:
					this.EraseHighlightGraphics(EraserType.Large, gp.Geometry);
					break;
				case GraphicAction.ClearHighlight:
					this.ClearGraphics(ClearType.Highlight);
					break;
				case GraphicAction.ClearMeasure:
					this.ClearGraphics(ClearType.Measure);
					break;
			}

		}

		// Event handlers

		void MapTextModel_ActionRequest(object sender, NotifyMapTextActionArgs e)
		{
			MapTextModel mt = sender as MapTextModel;
			if (mt == null)
				return; // This should never happen
			switch (e.Action)
			{
				case MapTextAction.UpdateLayout:
					break;
				case MapTextAction.Move:
					this.MoveMapText(mt);
					break;
				case MapTextAction.Delete:
					this.DeleteMapText(mt);
					break;
			}
		}

		void MapView_MouseMove(object sender, MouseEventArgs e)
		{
			if (!_bMoving)
				return;
			if (_tbMoving == null)
				return;
			MapPoint pt = _mv.ScreenToLocation(e.GetPosition(_mv));
			MapView.SetViewOverlayAnchor(_tbMoving, pt);
			e.Handled = true;
		}

		void MapView_MouseUp(object sender, MouseButtonEventArgs e)
		{
			if (!_bMoving)
				return;
			_bMoving = false;
			_mv.MouseMove -= MapView_MouseMove;
			_mv.MouseUp -= MapView_MouseUp;
			_mtMoving.MapTextIsSelected = false;
			_mtMoving = null;
			_tbMoving = null;
			e.Handled = true;
			SketchEditor ed = _mv.SketchEditor;
			ed.IsVisible = _bEditorVisible;
			ed.IsEnabled = _bEditorEnabled;
		}

		// Private methods

		private LabelDefinition GetMeasureLabelDef()
		{

			// See https://developers.arcgis.com/net/latest/wpf/guide/add-labels.htm
			// and https://developers.arcgis.com/web-map-specification/objects/labelingInfo/
			// for more info

			StringBuilder sb = new StringBuilder();
			sb.Append("{\"labelExpressionInfo\":{\"expression\":\"return $feature." + MeasureTag + ";\"},");
			sb.Append("\"labelPlacement\":\"esriServerLinePlacementAboveStart\",");
			sb.Append("\"symbol\":" + _tsMeasure.ToJson() + "}");
			string sJSON = sb.ToString();
			return LabelDefinition.FromJson(sJSON);

		}

		private void ClearGraphics(ClearType ct)
		{

			string sTag = null;
			bool bGraphics = false;
			bool bOverlays = false;
			switch (ct)
			{
				case ClearType.XY:
					sTag = XYTag;
					bGraphics = true;
					bOverlays = true;
					break;
				case ClearType.Measure:
					sTag = MeasureTag;
					bGraphics = true;
					break;
				case ClearType.Highlight:
					sTag = HighlightTag;
					bGraphics = true;
					break;
				case ClearType.Text:
					sTag = TextTag;
					bOverlays = true;
					break;
			}

			// Delete graphics

			if (bGraphics)
			{
				List<Graphic> GraphicDeleteList = new List<Graphic>();
				foreach (Graphic g in _gLayer.Graphics)
				{
					if (!g.Attributes.ContainsKey(sTag))
						continue;
					GraphicDeleteList.Add(g);
				}
				foreach (Graphic g in GraphicDeleteList)
					_gLayer.Graphics.Remove(g);
			}

			// Delete overlays

			if (bOverlays)
			{
				List<FrameworkElement> OverlayDeleteList = new List<FrameworkElement>();
				foreach (UIElement uie in _overlays.Items)
				{
					if (!(uie is FrameworkElement))
						continue;
					FrameworkElement fe = (FrameworkElement)uie;
					if (fe.Tag.ToString() == XYTag)
						OverlayDeleteList.Add(fe);
				}
				foreach (FrameworkElement fe in OverlayDeleteList)
					_overlays.Items.Remove(fe);
			}

		}

		private void AddXYGraphic(Geometry geom, bool bLabel, string sLabelText = "")
		{

			// Add graphic to the results layer

			if (_gLayer == null)
				return;
			MapPoint ptMap = geom as MapPoint;
			if (ptMap == null)
				return;
			Graphic gr = new Graphic();
			if (bLabel)
			{
				SimpleMarkerSymbol sm = new SimpleMarkerSymbol();
				sm.Style = SimpleMarkerSymbolStyle.Circle;
				sm.Size = 12;
				sm.Color = _colorXY;
				TextSymbol ts = new TextSymbol();
				ts.Text = sLabelText;
				ts.FontFamily = _tsXY.FontFamily;
				ts.Size = _tsXY.Size;
				ts.Color = _tsXY.Color;
				ts.HorizontalAlignment = Esri.ArcGISRuntime.Symbology.HorizontalAlignment.Center;
				ts.VerticalAlignment = Esri.ArcGISRuntime.Symbology.VerticalAlignment.Top;
				ts.OffsetY = -(1 + (ts.Size / 2));
				CompositeSymbol cs = new CompositeSymbol();
				cs.Symbols.Add(sm);
				cs.Symbols.Add(ts);
				gr.Symbol = cs;
			}
			else
			{
				SimpleMarkerSymbol sm = new SimpleMarkerSymbol();
				sm.Style = SimpleMarkerSymbolStyle.Circle;
				sm.Size = 12;
				sm.Color = _colorXY;
				gr.Symbol = sm;
			}
			gr.Geometry = new MapPoint(ptMap.X, ptMap.Y, ptMap.SpatialReference);
			gr.Attributes[XYTag] = sLabelText;
			_gLayer.Graphics.Add(gr);

		}

		private void AddXYCallout(Geometry geom, string sVal)
		{

			if (_overlays == null)
				return;
			if (_styleCallout == null)
				return;
			MapPoint ptMap = geom as MapPoint;
			if (ptMap == null)
				return;

			// Create label and assign callout style

			Label lblCallout = new Label();
			lblCallout.Style = _styleCallout;
			lblCallout.Content = sVal;
			lblCallout.Tag = XYTag;
			MapView.SetViewOverlayAnchor(lblCallout, ptMap);
			_overlays.Items.Add(lblCallout);

		}

		private void AddMeasureSumGraphic(Geometry geom, string sVal)
		{

			// Add graphic to the results layer

			Symbol sym = null;
			if (geom.GeometryType == GeometryType.Polyline)
				sym = _symMeasureLine;
			else if (geom.GeometryType == GeometryType.Polygon)
				sym = _symMeasurePolygon;
			else
				return;
			Graphic grFeat = new Graphic();
			grFeat.Geometry = geom;
			grFeat.Symbol = sym;
			grFeat.Attributes[MeasureTag] = sVal;
			_gLayer.Graphics.Add(grFeat);

		}

		private void AddHighlightGraphic(HighlighterType ht, Geometry geom)
		{

			// Add graphic to the results layer

			Graphic gr = new Graphic();
			gr.Geometry = geom;
			switch (ht)
			{
				case HighlighterType.Yellow:
					gr.Symbol = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.HighlightYellow);
					break;
				case HighlighterType.Blue:
					gr.Symbol = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.HighlightBlue);
					break;
				case HighlighterType.Green:
					gr.Symbol = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.HighlightGreen);
					break;
			}
			gr.Attributes.Add(HighlightTag, "");
			_gLayer.Graphics.Add(gr);

		}

		private void EraseHighlightGraphics(EraserType et, Geometry geom)
		{

			// Create buffer geometry if needed

			Geometry geomErase = geom;
			bool bEntire = (et == EraserType.Entire);
			if (!bEntire)
			{
				SimpleLineSymbol slSym = null;
				switch (et)
				{
					case EraserType.Small:
						slSym = (SimpleLineSymbol)QueryToolHelper.GetDrawSymbol(QueryToolHelperType.EraseSmall);
						break;
					case EraserType.Medium:
						slSym = (SimpleLineSymbol)QueryToolHelper.GetDrawSymbol(QueryToolHelperType.EraseMedium);
						break;
					case EraserType.Large:
						slSym = (SimpleLineSymbol)QueryToolHelper.GetDrawSymbol(QueryToolHelperType.EraseLarge);
						break;
				}
				double dWidth = slSym.Width;
				geomErase = GeometryEngine.Buffer(geom, dWidth * _mv.UnitsPerPixel / 2);
			}

			// Build list of graphics to process

			List<Graphic> EraseList = new List<Graphic>();
			foreach (Graphic gr in _gLayer.Graphics)
			{
				if (!gr.Attributes.ContainsKey(HighlightTag))
					continue;
				if (GeometryEngine.Disjoint(gr.Geometry, geomErase))
					continue;
				EraseList.Add(gr);
			}

			// Perform erase

			foreach (Graphic gr in EraseList)
			{
				if (bEntire)
				{
					_gLayer.Graphics.Remove(gr);
					continue;
				}
				Geometry geomNew = GeometryEngine.Difference(gr.Geometry, geomErase);
				gr.Geometry = geomNew;
			}

		}

		private void AddMapText(Geometry geom, MapTextModel mt)
		{

			if (_overlays == null)
				return;
			if (_styleMapText == null)
				return;
			MapPoint ptMap = geom as MapPoint;
			if (ptMap == null)
				return;

			// Create TextBlock and assign callout style

			TextBlock tbMapText = new TextBlock();
			tbMapText.Tag = TextTag;
			tbMapText.Style = _styleMapText;
			tbMapText.DataContext = mt;
			MapView.SetViewOverlayAnchor(tbMapText, ptMap);
			_overlays.Items.Add(tbMapText);
			mt.ActionRequest += MapTextModel_ActionRequest;

		}

		private void UpdateMapTextLayout(MapTextModel mt)
		{
			TextBlock tbMapText = this.GetMapTextOverlay(mt);
			if (tbMapText == null)
				return; // This should never happen
			tbMapText.UpdateLayout();
		}

		private void MoveMapText(MapTextModel mt)
		{
			TextBlock tbMapText = this.GetMapTextOverlay(mt);
			if (tbMapText == null)
				return; // This should never happen
			_bMoving = true;
			_mtMoving = mt;
			_mtMoving.MapTextIsSelected = true;
			_tbMoving = tbMapText;
			_mv.MouseMove += MapView_MouseMove;
			_mv.MouseUp += MapView_MouseUp;
			SketchEditor ed = _mv.SketchEditor;
			_bEditorVisible = ed.IsVisible;
			ed.IsVisible = false;
			_bEditorEnabled = ed.IsEnabled;
			ed.IsEnabled = false;
		}

		private void DeleteMapText(MapTextModel mt)
		{
			mt.ActionRequest -= MapTextModel_ActionRequest;
			TextBlock tbMapText = this.GetMapTextOverlay(mt);
			if (tbMapText == null)
				return; // This should never happen
			tbMapText.DataContext = null;
			_overlays.Items.Remove(tbMapText);
		}

		private TextBlock GetMapTextOverlay(MapTextModel mt)
		{
			TextBlock tbMapText = null;
			string sID = mt.ID;
			foreach (object ovl in _overlays.Items)
			{
				TextBlock tb = ovl as TextBlock;
				if (tb == null)
					continue;
				object dc = tb.DataContext;
				MapTextModel m = dc as MapTextModel;
				if (m == null)
					continue;
				if (m.ID != sID)
					continue;
				tbMapText = tb;
				break;
			}
			return tbMapText;
		}


	}

}
