﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Threading.Tasks;

using Esri.ArcGISRuntime.ArcGISServices;
using Esri.ArcGISRuntime.Data;
using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Mapping;
using Esri.ArcGISRuntime.Mapping.Popups;

using Extensibility;

namespace QueryAddin
{

	// IdentifyEngine supports identify operations

	public enum ResultCollectionType
	{
		Tree,
		List
	}

	public enum IdentifyOption
	{
		Topmost,
		Visible,
		All,
		Layer
	}

	public delegate void IdentifyStatus(string sMsg);

	public class IdentifyEngine
	{

		// OverrideRelatedQuery prevents querying related features
		private const bool OverrideRelatedQuery = false;

		// UseWebQuery improves performance when querying map services
		private const bool UseWebQuery = true;

		private ResultCollectionType _rcType;
		private List<IQueryableLayer> _Layers = new List<IQueryableLayer>();
		private ObservableCollection<string> _LayerItems = new ObservableCollection<string>();
		private int _iLayerIndex;
		private Geometry _geomQuery;
		private bool _bQueryRelated;
		private IdentifyOption _idOption;
		private bool _bBusy = false;
		private HashSet<string> _QueriedServices = new HashSet<string>();

		public bool IsBusy { get { return _bBusy; } }
	
		// Public methods

		public void Initialize(List<IQueryableLayer> QueryableLayers, ResultCollectionType rcType)
		{

			// Assumes map layers are already initialized

			_rcType = rcType;
			if (_rcType == ResultCollectionType.List)
				_bQueryRelated = false;
			_Layers.Clear();
			_LayerItems.Clear();
			foreach (IQueryableLayer pLyr in QueryableLayers)
				_Layers.Add(pLyr);
			if (_Layers.Count == 0)
				return;
			_LayerItems.Add("<Top-most layer>");
			_LayerItems.Add("<Visible layers>");
			_LayerItems.Add("<All layers>");
			foreach (IQueryableLayer pLyr in _Layers)
				_LayerItems.Add(pLyr.DisplayName);
			_iLayerIndex = 0;
			
		}

		public ObservableCollection<string> LayerItemList
		{
			get { return _LayerItems; }
		}

		public void SetLayerItem(int Index)
		{
			_iLayerIndex = Index;
			switch (_iLayerIndex)
			{
				case 0:
					_idOption = IdentifyOption.Topmost;
					break;
				case 1:
					_idOption = IdentifyOption.Visible;
					break;
				case 2:
					_idOption = IdentifyOption.All;
					break;
				default:
					_idOption = IdentifyOption.Layer;
					break;
			}
		}

		public async Task<List<IdentifyResult>> IdentifyAsync(Geometry Geom, bool QueryRelated, IdentifyStatus callback = null)
		{

			if (_bBusy)
				return null;
			_bBusy = true;
			_geomQuery = Geom;
			_QueriedServices.Clear();
			if (_rcType == ResultCollectionType.Tree)
				_bQueryRelated = QueryRelated;
			else
				_bQueryRelated = false;
			bool bOverride = OverrideRelatedQuery;
			if (bOverride)
				_bQueryRelated = false;
			List<IdentifyResult> ResultList = new List<IdentifyResult>();
			List<IQueryableLayer> LayerList = this.GetSearchLayers();
			bool bWeb = UseWebQuery;
			foreach (IQueryableLayer pLyr in LayerList)
			{
				string sMsg = "Querying " + pLyr.DisplayName + "...";
				callback?.Invoke(sMsg);
				Debug.Print(sMsg);
				List<IdentifyResult> LayerResults = new List<IdentifyResult>();
				if (bWeb && (pLyr is WebQueryableLayer) && (_idOption != IdentifyOption.Layer))
				{
					WebQueryableLayer wqLyr = (WebQueryableLayer)pLyr;
					if (_QueriedServices.Contains(wqLyr.MapImageLayerId))
						continue;
					_QueriedServices.Add(wqLyr.MapImageLayerId);
					LayerResults = await this.IdentifyFromWebLayerAsync(wqLyr.MapImageLayerId, LayerList);
				}
				else
				{
					IdentifyResult idResult = await this.IdentifyFromLayerAsync(pLyr);
					if (idResult == null)
						continue;
					LayerResults.Add(idResult);
				}
				if (LayerResults == null)
					continue;
				foreach (IdentifyResult idResult in LayerResults)
				{
					if (_rcType == ResultCollectionType.Tree)
						ResultList.Add(idResult);
					else
					{
						foreach (IdentifyResult idFeatureResult in idResult.RelatedRecords)
							ResultList.Add(idFeatureResult);
					}
					if (_idOption == IdentifyOption.Topmost)
					{
						_bBusy = false;
						return ResultList;
					}
				}
			}
			_bBusy = false;
			return ResultList;
		}

		public async Task<IdentifyResult> IdentifyFeature(IQueryableLayer pLyr, Feature Feat, bool QueryRelated)
		{

			if (_bBusy)
				return null;
			_bBusy = true;
			IdentifyResult idResult;
			if (pLyr.IsPopupEnabled)
				idResult = this.BuildResult(Feat, pLyr.PopupDefinition);
			else
				idResult = this.BuildResult(Feat, pLyr.Fields, pLyr.DisplayField);
			await this.QueryAttachments(idResult, pLyr);
			if (QueryRelated && (_rcType == ResultCollectionType.Tree))
			{
				idResult.RelatedRecords = new ObservableCollection<IdentifyResult>();
				await this.QueryRelatedRecords(idResult, pLyr);
			}
			IdentifyResult idLayerResult = new IdentifyResult()
			{
				Value = pLyr.DisplayName,
				RelatedRecords = new ObservableCollection<IdentifyResult>()
			};
			idLayerResult.RelatedRecords.Add(idResult);
			idResult.Parent = idLayerResult;
			_bBusy = false;
			return idLayerResult;

		}

		public async Task<Geometry> FindFeatureGeometry(Geometry Geom)
		{
			if (_bBusy)
				return null;
			_bBusy = true;
			_geomQuery = Geom;
			_bQueryRelated = false;
			double dScale = QueryViewExt.Current.MapScale;
			List<IQueryableLayer> LayerList = this.GetSearchLayers();
			foreach (IQueryableLayer pLyr in LayerList)
			{
				if (!pLyr.IsVisible(dScale))
					continue;
				IdentifyResult idLayerResult = await this.IdentifyFromLayerAsync(pLyr);
				if (idLayerResult == null)
					continue;
				IdentifyResult idFeatureResult = idLayerResult.RelatedRecords[0];
				_bBusy = false;
				return idFeatureResult.Feature.Geometry;
			}
			return null;
		}

		// Private methods

		private List<IQueryableLayer> GetSearchLayers()
		{

			List<IQueryableLayer> LayerList = new List<IQueryableLayer>();
			if (_idOption == IdentifyOption.Layer)
			{
				int i = _iLayerIndex - 3;
				int iMax = _Layers.Count - 1;
				if ((i < 0) || (i > iMax))
				{
					// This could happen if a map were added without any queryable layers
					string sMsg = "ERROR: IdentifyEngine: index " + i.ToString() + " outside range 0-" + iMax.ToString();
					Debug.Print(sMsg);
					return LayerList;
				}
				LayerList.Add(_Layers[i]);
				return LayerList;
			}

			double dScale = QueryViewExt.Current.MapScale;
			bool bVis = ((_idOption == IdentifyOption.Topmost) || (_idOption == IdentifyOption.Visible));
			foreach (IQueryableLayer pLyr in _Layers)
			{
				if (bVis && (!pLyr.IsVisible(dScale)))
					continue;
				LayerList.Add(pLyr);
			}
			return LayerList;

		}

		private async Task<IdentifyResult> IdentifyFromLayerAsync(IQueryableLayer pLyr)
		{

			// Query layer and build results

			List<Feature> FeatureList = await pLyr.Query(null, _geomQuery);
			if (FeatureList == null)
				return null;
			if (FeatureList.Count == 0)
				return null;
			IdentifyResult idLayerResult = new IdentifyResult()
			{
				Value = pLyr.DisplayName,
				RelatedRecords = new ObservableCollection<IdentifyResult>()
			};
			foreach (Feature feat in FeatureList)
			{
				if (feat is ArcGISFeature agFeat)
				{
					if (agFeat.LoadStatus != Esri.ArcGISRuntime.LoadStatus.Loaded)
						await agFeat.LoadAsync();
				}
				IdentifyResult idResult;
				if (pLyr.IsPopupEnabled)
					idResult = BuildResult(feat, pLyr.PopupDefinition);
				else
					idResult = this.BuildResult(feat, pLyr.Fields, pLyr.DisplayField);
				idResult.LayerID = pLyr.LayerID;
				idLayerResult.RelatedRecords.Add(idResult);
				idResult.Parent = idLayerResult;
				await this.QueryAttachments(idResult, pLyr);
				if (!_bQueryRelated)
					continue;
				await this.QueryRelatedRecords(idResult, pLyr);
			}
			return idLayerResult;

		}

		private async Task<List<IdentifyResult>> IdentifyFromWebLayerAsync(string sMapLayerId, List<IQueryableLayer> SearchLayers)
		{

			// Get sublayer list

			List<IdentifyResult> ResultList = new List<IdentifyResult>();
			List<long> SublayerList = new List<long>();
			Dictionary<long, WebQueryableLayer> Sublayers = new Dictionary<long, WebQueryableLayer>();
			double dScale = QueryViewExt.Current.MapScale;
			bool bVis = ((_idOption == IdentifyOption.Topmost) || (_idOption == IdentifyOption.Visible));
			foreach (IQueryableLayer pLyr in SearchLayers)
			{
				if (!(pLyr is WebQueryableLayer))
					continue;
				WebQueryableLayer wqLyr = (WebQueryableLayer)pLyr;
				if (wqLyr.MapImageLayerId != sMapLayerId)
					continue;
				if (bVis)
				{
					if (!wqLyr.IsVisible(dScale))
						continue;
				}
				SublayerList.Add(wqLyr.SublayerId);
				Sublayers[wqLyr.SublayerId] = wqLyr;
			}

			// If only one sublayer, don't bother calling the map service identify operation

			if (SublayerList.Count == 1)
			{
				IdentifyResult idResult = await this.IdentifyFromLayerAsync(Sublayers[SublayerList[0]]);
				if (idResult == null)
					return null;
				ResultList.Add(idResult);
				return ResultList;
			}

			// Perform web identify

			WebQueryLayerHelper wh = WebQueryLayerHelper.Instances[sMapLayerId];
			WebIdentifyParameters param = new WebIdentifyParameters()
			{
				Geometry = _geomQuery,
				Layers = WebIdentifyOption.All,
				LayerIds = SublayerList,
				Tolerance = 0,
				MapExtent = QueryViewExt.Current.MapExtent,
				DisplayHeight = QueryViewExt.Current.DisplaySize.Height,
				DisplayWidth = QueryViewExt.Current.DisplaySize.Width
			};
			switch (_idOption)
			{
				case IdentifyOption.Topmost:
					param.Layers = WebIdentifyOption.Topmost;
					break;
				case IdentifyOption.Visible:
					param.Layers = WebIdentifyOption.Visible;
					break;
				default:
					param.Layers = WebIdentifyOption.All;
					break;
			}
			Dictionary<long, List<Feature>> WebResults = await WebUtilities.Identify(wh.ServiceUrl, param);

			// Build identify results

			List<long> SublayerHits = new List<long>();
			foreach (long i in WebResults.Keys)
				SublayerHits.Add(i);
			if (SublayerHits.Count > 1)
				SublayerHits.Sort();
			foreach (long i in SublayerHits)
			{
				List<Feature> FeatureList = WebResults[i];
				IQueryableLayer pLyr = Sublayers[i];
				IdentifyResult idLayerResult = new IdentifyResult()
				{
					Value = pLyr.DisplayName,
					RelatedRecords = new ObservableCollection<IdentifyResult>()
				};
				foreach (Feature feat in FeatureList)
				{
					IdentifyResult idResult;
					if (pLyr.IsPopupEnabled)
						idResult = this.BuildResult(feat, pLyr.PopupDefinition);
					else
						idResult = this.BuildResult(feat, pLyr.Fields, pLyr.DisplayField);
					idResult.LayerID = pLyr.LayerID;
					idLayerResult.RelatedRecords.Add(idResult);
					idResult.Parent = idLayerResult;
					await this.QueryAttachments(idResult, pLyr);
					if (!_bQueryRelated)
						continue;
					await this.QueryRelatedRecords(idResult, pLyr);
				}
				ResultList.Add(idLayerResult);
			}
			return ResultList;

		}

		private async Task QueryAttachments(IdentifyResult idResult, IQueryableLayer pLyr)
		{

			// Get attachments (if any) and add to results

			if (!pLyr.HasAttachments)
				return;
			idResult.Attachments = new ObservableCollection<Attachment>();
			List<Attachment> AttachmentResults = await pLyr.GetAttachments(idResult.Feature);
			if (AttachmentResults == null)
				return;
			foreach (Attachment item in AttachmentResults)
				idResult.Attachments.Add(item);

		}

		private async Task QueryRelatedRecords(IdentifyResult idResult, IQueryableLayer pLyr)
		{

			// Get related records and build results

			List<QueryRelatedResult> RelatedResults = await pLyr.GetRelatedFeatures(idResult.Feature);
			foreach (QueryRelatedResult result in RelatedResults)
			{
				List<Feature> Features = result.RelatedFeatures;
				if (Features == null)
					continue;
				if (Features.Count == 0)
					continue;
				IdentifyResult idLayerResult = new IdentifyResult()
				{
					Value = result.Name,
					RelatedRecords = new ObservableCollection<IdentifyResult>()
				};
				foreach (Feature feat in Features)
				{
					IdentifyResult idRelResult;
					if (result.IsPopupEnabled)
						idRelResult = this.BuildResult(feat, result.PopupDefinition);
					else
						idRelResult = this.BuildResult(feat, result.Fields, result.DisplayField);
					idLayerResult.RelatedRecords.Add(idRelResult);
					idRelResult.Parent = idLayerResult;
				}
				if (idResult.RelatedRecords == null)
					idResult.RelatedRecords = new ObservableCollection<IdentifyResult>();
				idResult.RelatedRecords.Add(idLayerResult);
				idLayerResult.Parent = idResult;
			}

		}

		private IdentifyResult BuildResult(Feature feat, PopupDefinition popDef)
		{
			Popup popup = new Popup(feat, popDef);
			PopupManager popMan = new PopupManager(popup);
			IdentifyResult idResult = new IdentifyResult()
			{
				Feature = feat,
				Value = popMan.Title,
				Attributes = new ObservableCollection<Attribute>()
			};
			foreach (PopupFieldValue popVal in popMan.DisplayedFields)
			{
				Attribute att = new Attribute()
				{
					Key = popVal.Field.Label,
					Value = this.StringValue(popVal)
				};
				idResult.Attributes.Add(att);
			}
			return idResult;
		}

		private IdentifyResult BuildResult(Feature feat, IReadOnlyList<QueryFieldInfo> Fields, string sDisplayField)
		{
			if (Fields == null)
				return this.BuildResult(feat, sDisplayField);
			IdentifyResult idResult = new IdentifyResult()
			{
				Value = "Record",
				Feature = feat,
				Attributes = new ObservableCollection<Attribute>()
			};
			IDictionary<string, object> attr = feat.Attributes;
			foreach (QueryFieldInfo info in Fields)
			{
				string sName = info.Name;
				string sAlias = info.Alias;
				Dictionary<string, string> CodedVals = info.CodedValues;
				string sVal = "<null>";
				object oVal = null;
				if (attr.ContainsKey(sName))
					oVal = attr[sName];
				else if (attr.ContainsKey(sAlias))
					oVal = attr[sAlias];
				if (oVal != null)
				{
					sVal = this.StringValue(oVal);
					if (CodedVals != null)
					{
						if (CodedVals.ContainsKey(sVal))
							sVal = CodedVals[sVal];
					}
				}
				if (sName == sDisplayField)
					idResult.Value = sVal;
				Attribute att = new Attribute()
				{
					Key = sAlias,
					Value = sVal
				};
				idResult.Attributes.Add(att);
			}
			return idResult;
		}

		private IdentifyResult BuildResult(Feature feat, string sDisplayField)
		{
			IdentifyResult idResult = new IdentifyResult()
			{
				Value = "Record",
				Feature = feat,
				Attributes = new ObservableCollection<Attribute>()
			};
			IDictionary<string, object> attr = feat.Attributes;
			List<string> KeyList = new List<string>();
			foreach (string sKey in attr.Keys)
				KeyList.Add(sKey);
			KeyList.Sort();
			foreach (string sKey in KeyList)
			{
				string sVal = "<null>";
				object oVal = feat.Attributes[sKey];
				if (oVal != null)
					sVal = this.StringValue(oVal);
				if (sKey == sDisplayField)
					idResult.Value = sVal;
				Attribute att = new Attribute()
				{
					Key = sKey,
					Value = sVal
				};
				idResult.Attributes.Add(att);
			}
			return idResult;
		}

		private string StringValue(PopupFieldValue popVal)
		{
			object oVal = popVal.Value;
			if (oVal == null)
				return "<null>";
			else if ((oVal is DateTime) || (oVal is DateTimeOffset))
				return this.StringValue(oVal);
			return popVal.FormattedValue;
		}

		private string StringValue(object oVal)
		{
			if (oVal == null)
				return "<null>";
			if ((oVal is DateTime) || (oVal is DateTimeOffset))
			{

				// Check if value is date/time or just date

				DateTime dtLocal;
				if (oVal is DateTime)
					dtLocal = (DateTime)oVal;
				else
					dtLocal = ((DateTimeOffset)oVal).DateTime;
				if (dtLocal.TimeOfDay == new TimeSpan(0, 0, 0))
					return dtLocal.ToString("d");
				dtLocal = dtLocal.ToLocalTime();
				if (dtLocal.TimeOfDay == new TimeSpan(0, 0, 0))
					return dtLocal.ToString("d");
			}
			return oVal.ToString();
		}

	}

}
