﻿using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;

using Color = System.Drawing.Color;

using Esri.ArcGISRuntime.Symbology;
using Esri.ArcGISRuntime.UI;

using Extensibility;

namespace QueryAddin
{

	// QueryToolHelper gets various resources for query tools

	public enum QueryToolHelperType
	{
		None,
		Identify,
		MeasureLine,
		MeasureArea,
		MeasureFeature,
		HighlightYellow,
		HighlightBlue,
		HighlightGreen,
		EraseEntire,
		EraseSmall,
		EraseMedium,
		EraseLarge,
		Text
	}

	public static class QueryToolHelper
	{

		private const double _HighlightWidth = 6;
		private const double _SmallEraserWidth = 2;
		private const double _MediumEraserWidth = 6;
		private const double _LargeEraserWidth = 10;

		private static Dictionary<QueryToolHelperType, Cursor> _Cursors = new Dictionary<QueryToolHelperType, Cursor>();
		private static Dictionary<QueryToolHelperType, Symbol> _DrawSymbols = new Dictionary<QueryToolHelperType, Symbol>();

		// Public methods

		public static Cursor GetCursor(QueryToolHelperType tt)
		{

			if (_Cursors.ContainsKey(tt))
				return _Cursors[tt];
			string sPath = "";
			switch (tt)
			{
				case QueryToolHelperType.Identify:
					sPath = "Images/Identify.cur";
					break;
				case QueryToolHelperType.MeasureLine:
				case QueryToolHelperType.MeasureArea:
				case QueryToolHelperType.MeasureFeature:
					sPath = "Images/Measure.cur";
					break;
				case QueryToolHelperType.HighlightYellow:
					sPath = "Images/HighlighterYellow.cur";
					break;
				case QueryToolHelperType.HighlightBlue:
					sPath = "Images/HighlighterBlue.cur";
					break;
				case QueryToolHelperType.HighlightGreen:
					sPath = "Images/HighlighterGreen.cur";
					break;
				case QueryToolHelperType.EraseEntire:
					sPath = "Images/Eraser.cur";
					break;
				case QueryToolHelperType.EraseSmall:
					sPath = "Images/EraserSmall.cur";
					break;
				case QueryToolHelperType.EraseMedium:
					sPath = "Images/EraserMedium.cur";
					break;
				case QueryToolHelperType.EraseLarge:
					sPath = "Images/EraserLarge.cur";
					break;
				case QueryToolHelperType.Text:
					sPath = "Images/Text.cur";
					break;
			}
			Cursor c = ResourceHelper.GetCursor("QueryAddin", sPath);
			_Cursors[tt] = c;
			return c;

		}

		public static Symbol GetDrawSymbol(QueryToolHelperType tt)
		{

			if (_DrawSymbols.ContainsKey(tt))
				return _DrawSymbols[tt];

			SimpleFillSymbol sfDef = new SimpleFillSymbol();
			sfDef = new SimpleFillSymbol();
			sfDef.Style = SimpleFillSymbolStyle.Null;
			SimpleLineSymbol slDef = new SimpleLineSymbol();
			slDef.Style = SimpleLineSymbolStyle.Solid;
			slDef.Width = 1;
			slDef.Color = Color.Black;
			sfDef.Outline = slDef;

			Symbol sym = null;
			switch (tt)
			{
				case QueryToolHelperType.Identify:
					sym = ResourceHelper.GetFillSymbol("QueryAddin", "QueryAddin.IdentifyDrawSymbol");
					if (sym == null)
						sym = sfDef;
					break;
				case QueryToolHelperType.MeasureLine:
					sym = ResourceHelper.GetLineSymbol("QueryAddin", "QueryAddin.MeasureLineSymbol");
					if (sym == null)
						sym = slDef;
					break;
				case QueryToolHelperType.MeasureArea:
					sym = ResourceHelper.GetFillSymbol("QueryAddin", "QueryAddin.MeasurePolygonSymbol");
					if (sym == null)
						sym = sfDef;
					break;
				case QueryToolHelperType.HighlightYellow:
					sym = ResourceHelper.GetLineSymbol("QueryAddin", "QueryAddin.HighlightYellowSymbol");
					if (sym == null)
						sym = QueryToolHelper.GetDefaultHighlighterSymbol(HighlighterType.Yellow);
					break;
				case QueryToolHelperType.HighlightBlue:
					sym = ResourceHelper.GetLineSymbol("QueryAddin", "QueryAddin.HighlightBlueSymbol");
					if (sym == null)
						sym = QueryToolHelper.GetDefaultHighlighterSymbol(HighlighterType.Blue);
					break;
				case QueryToolHelperType.HighlightGreen:
					sym = ResourceHelper.GetLineSymbol("QueryAddin", "QueryAddin.HighlightGreenSymbol");
					if (sym == null)
						sym = QueryToolHelper.GetDefaultHighlighterSymbol(HighlighterType.Green);
					break;
				case QueryToolHelperType.EraseEntire:
					sym = ResourceHelper.GetFillSymbol("QueryAddin", "QueryAddin.EraseEntireDrawSymbol");
					if (sym == null)
						sym = sfDef;
					break;
				case QueryToolHelperType.EraseSmall:
					sym = QueryToolHelper.GetDefaultEraserLineSymbol(EraserType.Small);
					break;
				case QueryToolHelperType.EraseMedium:
					sym = QueryToolHelper.GetDefaultEraserLineSymbol(EraserType.Medium);
					break;
				case QueryToolHelperType.EraseLarge:
					sym = QueryToolHelper.GetDefaultEraserLineSymbol(EraserType.Large);
					break;
			}
			_DrawSymbols[tt] = sym;
			return sym;

		}

		public static Image GetHighlighterIcon(HighlighterType ht)
		{
			string sResult = "";
			switch (ht)
			{
				case HighlighterType.Yellow:
					sResult = "QueryAddin.HighlightYellowIcon";
					break;
				case HighlighterType.Blue:
					sResult = "QueryAddin.HighlightBlueIcon";
					break;
				case HighlighterType.Green:
					sResult = "QueryAddin.HighlightGreenIcon";
					break;
			}
			if (string.IsNullOrEmpty(sResult))
				return null;
			ResourceDictionary rd = ResourceHelper.GetResources("QueryAddin");
			if (!rd.Contains(sResult))
				return null;
			return rd[sResult] as Image;
		}

		public static Image GetEraserIcon(EraserType et)
		{
			string sResult = "";
			switch (et)
			{
				case EraserType.Entire:
					sResult = "QueryAddin.EraseEntireIcon";
					break;
				case EraserType.Small:
					sResult = "QueryAddin.EraseSmallIcon";
					break;
				case EraserType.Medium:
					sResult = "QueryAddin.EraseMediumIcon";
					break;
				case EraserType.Large:
					sResult = "QueryAddin.EraseLargeIcon";
					break;
			}
			if (string.IsNullOrEmpty(sResult))
				return null;
			ResourceDictionary rd = ResourceHelper.GetResources("QueryAddin");
			if (!rd.Contains(sResult))
				return null;
			return rd[sResult] as Image;
		}

		public static string GetHighlighterToolTip(HighlighterType ht)
		{
			string sResult = "";
			switch (ht)
			{
				case HighlighterType.Yellow:
					sResult = "Yellow Highlighter";
					break;
				case HighlighterType.Blue:
					sResult = "Blue Highlighter";
					break;
				case HighlighterType.Green:
					sResult = "Green Highlighter";
					break;
			}
			return sResult;
		}

		public static string GetEraserToolTip(EraserType et)
		{
			string sResult = "";
			switch (et)
			{
				case EraserType.Entire:
					sResult = "Erase Entire Ink Stroke";
					break;
				case EraserType.Small:
					sResult = "Erase Using Small Eraser";
					break;
				case EraserType.Medium:
					sResult = "Erase Using Medium Eraser";
					break;
				case EraserType.Large:
					sResult = "Erase Using Large Eraser";
					break;
			}
			return sResult;
		}

		// Private methods

		private static SimpleLineSymbol GetDefaultHighlighterSymbol(HighlighterType ht)
		{
			Color col = Color.Black;
			switch (ht)
			{
				case HighlighterType.Yellow:
					col = Color.Yellow;
					break;
				case HighlighterType.Blue:
					col = Color.Cyan;
					break;
				case HighlighterType.Green:
					col = Color.Lime;
					break;
			}
			col = Color.FromArgb(127, col);
			SimpleLineSymbol slSym = new SimpleLineSymbol();
			slSym.Style = SimpleLineSymbolStyle.Solid;
			slSym.Width = _HighlightWidth;
			slSym.Color = col;
			return slSym;
		}

		private static SimpleLineSymbol GetDefaultEraserLineSymbol(EraserType et)
		{
			Color col = Color.White;
			BackgroundGrid bg = AppViewExt.Current.MapBackground;
			if (bg != null)
				col = bg.Color;
			SimpleLineSymbol slSym = new SimpleLineSymbol();
			slSym.Style = SimpleLineSymbolStyle.Solid;
			slSym.Color = col;
			switch (et)
			{
				case EraserType.Small:
					slSym.Width = _SmallEraserWidth;
					break;
				case EraserType.Medium:
					slSym.Width = _MediumEraserWidth;
					break;
				case EraserType.Large:
					slSym.Width = _LargeEraserWidth;
					break;

			}
			return slSym;
		}



	}

}
