﻿using System;

using Esri.ArcGISRuntime.Geometry;

namespace QueryAddin
{

	// UnitsHelper performs various display and conversion functions

	public enum IdentifyUnits
	{
		None,
		Feet,
		Meters,
		DD,
		DMS
	}

	public enum MeasureLinearUnits
	{
		None,
		Kilometers,
		Meters,
		Decimeters,
		Centimeters,
		Millimeters,
		Miles,
		NauticalMiles,
		Yards,
		Feet,
		Inches
	}

	public enum MeasureAreaUnits
	{
		None,
		SquareKilometers,
		Hectares,
		Ares,
		SquareMeters,
		SquareMiles,
		Acres,
		SquareYards,
		SquareFeet
	}

	public struct Measurement
	{
		public double Value;
		public string Units;
	}

	public static class UnitsHelper
	{

		public static string GetLocationString(MapPoint pt, IdentifyUnits iuTarget)
		{

			// Return X and Y by default

			string sLoc = pt.X.ToString("N3") + " " + pt.Y.ToString("N3");
			if (iuTarget == IdentifyUnits.None)
				return sLoc;
			SpatialReference sr = pt.SpatialReference;
			if (sr == null)
				return sLoc;
			LinearUnit luSource = sr.Unit as LinearUnit;
			if (luSource == null)
				return sLoc;

			// Convert to appropriate units

			bool bGeographic = false;
			double dMeters = luSource.ToMeters(1.0);
			bool bMeters = (luSource.FromMeters(1.0) == 1.0);
			bool bFeet = ((Math.Floor(dMeters * 10000) / 10000) == 0.3048);
			bool bDMS = false;
			double dX = 0;
			double dY = 0;
			switch (iuTarget)
			{
				case IdentifyUnits.Feet:
					if (bFeet)
						return sLoc;
					dX = luSource.ConvertTo(LinearUnits.Feet, pt.X);
					dY = luSource.ConvertTo(LinearUnits.Feet, pt.Y);
					break;
				case IdentifyUnits.Meters:
					if (bMeters)
						return sLoc;
					dX = luSource.ToMeters(pt.X);
					dY = luSource.ToMeters(pt.Y);
					break;
				case IdentifyUnits.DD:
					bGeographic = true;
					break;
				case IdentifyUnits.DMS:
					bGeographic = true;
					bDMS = true;
					break;
			}
			if (bGeographic)
			{
				MapPoint ptDD = GeometryEngine.Project(pt, SpatialReferences.Wgs84) as MapPoint;
				dX = ptDD.X;
				dY = ptDD.Y;
			}
			if (!bDMS)
				sLoc = dX.ToString("N3") + " " + dY.ToString("N3");
			else
				sLoc = GeometryHelper.ToDMS(dX, true) + " " + GeometryHelper.ToDMS(dY, false);
			return sLoc;

		}

		public static bool GetStandardLinearUnit(SpatialReference sr, out LinearUnit lu)
		{
			lu = null;
			if (sr == null)
				return false;
			lu = sr.Unit as LinearUnit;
			if (lu == null)
				return false;
			if (lu.UnitId > -1)
				return true;
			double dMeters = lu.ToMeters(1.0);
			if (dMeters == 1.0)
			{
				lu = LinearUnits.Meters;
				return true;
			}
			if ((Math.Floor(dMeters * 10000) / 10000) == 0.3048) 
			{
				lu = LinearUnits.Feet;
				return true;
			}
			return false;
		}

		public static string GetMeasureFeatureLabel(Geometry g, MeasureLinearUnit mluOut, MeasureAreaUnit mauOut)
		{

			// Get length and area

			Measurement m = UnitsHelper.GetMeasureLength(g, mluOut);
			double dLength = m.Value;
			if (dLength == 0)
				return null;
			m = UnitsHelper.GetMeasureArea(g, mauOut);
			double dArea = m.Value;

			// Format output

			string sLabel = null;
			if (g.Dimension == GeometryDimension.Curve)
				sLabel = "L:" + dLength.ToString("N0");
			else if (g.Dimension >= GeometryDimension.Area)
				sLabel = "P:" + dLength.ToString("N0") + "\nA:" + dArea.ToString("N0");
			return sLabel;

		}

		public static Measurement GetMeasureLength(Geometry g, MeasureLinearUnit mluOut)
		{

			// Check input

			Measurement m = new Measurement();
			m.Value = 0;
			m.Units = "";
			if (g == null)
				return m;
			if (g.IsEmpty)
				return m;
			if (g.Dimension < GeometryDimension.Curve)
				return m;

			// Get length and convert to units

			double dLength = 0;
			LinearUnit luIn = null;
			SpatialReference sr = g.SpatialReference;
			if (sr.IsGeographic)
			{
				dLength = GeometryEngine.LengthGeodetic(g);
				luIn = LinearUnits.Meters;
			}
			else
			{
				dLength = GeometryEngine.Length(g);
				luIn = sr.Unit as LinearUnit;
			}
			if (dLength == 0)
				return m;
			if (luIn == null)
			{
				m.Value = dLength;
				return m;
			}
			MeasureLinearUnit mluIn = new MeasureLinearUnit(luIn);
			m.Value = mluIn.ConvertTo(dLength, mluOut);
			m.Units = mluOut.Name;
			return m;

		}

		public static Measurement GetMeasureArea(Geometry g, MeasureAreaUnit mauOut)
		{

			// Check input

			Measurement m = new Measurement();
			m.Value = 0;
			m.Units = "";
			if (g == null)
				return m;
			if (g.IsEmpty)
				return m;
			if (g.Dimension < GeometryDimension.Area)
				return m;

			// Get area and convert to units

			double dArea = 0;
			MeasureAreaUnit mauIn = null;
			SpatialReference sr = g.SpatialReference;
			if (sr.IsGeographic)
			{
				dArea = GeometryEngine.AreaGeodetic(g);
				mauIn = new MeasureAreaUnit(AreaUnits.SquareMeters);
			}
			else
			{
				dArea = GeometryEngine.Area(g);
				LinearUnit luIn;
				bool? bResult = UnitsHelper.GetStandardLinearUnit(sr, out luIn);
				if (bResult == true)
					mauIn = new MeasureAreaUnit(AreaUnit.CreateFromLinearUnit(luIn));
			}
			if (dArea == 0)
				return m;
			if (mauIn == null)
			{
				m.Value = dArea;
				return m;
			}
			m.Value = mauIn.ConvertTo(dArea, mauOut);
			m.Units = mauOut.Name;
			return m;

		}

	}

	public class MeasureLinearUnit
	{

		private bool _bDecimeters = false;
		private LinearUnit _lu;

		public MeasureLinearUnit(MeasureLinearUnits unit)
		{
			switch (unit)
			{
				case MeasureLinearUnits.Centimeters:
					_lu = LinearUnits.Centimeters;
					break;
				case MeasureLinearUnits.Decimeters:
					_bDecimeters = true;
					_lu = LinearUnits.Meters;
					break;
				case MeasureLinearUnits.Feet:
					_lu = LinearUnits.Feet;
					break;
				case MeasureLinearUnits.Inches:
					_lu = LinearUnits.Inches;
					break;
				case MeasureLinearUnits.Kilometers:
					_lu = LinearUnits.Kilometers;
					break;
				case MeasureLinearUnits.Meters:
					_lu = LinearUnits.Meters;
					break;
				case MeasureLinearUnits.Miles:
					_lu = LinearUnits.Miles;
					break;
				case MeasureLinearUnits.Millimeters:
					_lu = LinearUnits.Millimeters;
					break;
				case MeasureLinearUnits.NauticalMiles:
					_lu = LinearUnits.NauticalMiles;
					break;
				case MeasureLinearUnits.Yards:
					_lu = LinearUnits.Yards;
					break;
			}
		}

		public MeasureLinearUnit(LinearUnit lu)
		{
			_lu = lu;
		}

		public string Name
		{
			get
			{
				if (_bDecimeters)
					return "Decimeters";
				return _lu.Name;
			}
		}

		public double ConvertTo(double inDistance, MeasureLinearUnit outUnit)
		{
			if (outUnit == null)
				return inDistance;
			double dVal = _lu.ConvertTo(outUnit._lu, inDistance);
			if (_bDecimeters)
				dVal /= 10;
			if (outUnit._bDecimeters)
				dVal *= 10;
			return dVal;
		}

	}

	public class MeasureAreaUnit
	{

		private bool _bAres = false;
		private AreaUnit _au;

		public MeasureAreaUnit(MeasureAreaUnits unit)
		{
			switch (unit)
			{
				case MeasureAreaUnits.SquareKilometers:
					_au = AreaUnits.SquareKilometers;
					break;
				case MeasureAreaUnits.Hectares:
					_au = AreaUnits.Hectares;
					break;
				case MeasureAreaUnits.Ares:
					_bAres = true;
					_au = AreaUnits.SquareMeters;
					break;
				case MeasureAreaUnits.SquareMeters:
					_au = AreaUnits.SquareMeters;
					break;
				case MeasureAreaUnits.SquareMiles:
					_au = AreaUnits.SquareMiles;
					break;
				case MeasureAreaUnits.Acres:
					_au = AreaUnits.Acres;
					break;
				case MeasureAreaUnits.SquareYards:
					_au = AreaUnits.SquareYards;
					break;
				case MeasureAreaUnits.SquareFeet:
					_au = AreaUnits.SquareFeet;
					break;
			}

		}

		public MeasureAreaUnit(AreaUnit au)
		{
			_au = au;
		}

		public string Name
		{
			get
			{
				if (_bAres)
					return "Ares";
				return _au.Name;
			}
		}

		public double ConvertTo(double inArea, MeasureAreaUnit outUnit)
		{
			if (outUnit == null)
				return inArea;
			double dVal = _au.ConvertTo(outUnit._au, inArea);
			if (_bAres)
				dVal /= 100;
			if (outUnit._bAres)
				dVal *= 100;
			return dVal;
		}

	}




}
