﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading.Tasks;
using System.Web;
using System.Web.Script.Serialization;

using Esri.ArcGISRuntime.ArcGISServices;
using Esri.ArcGISRuntime.Data;
using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Mapping;

using Extensibility;

namespace QueryAddin
{

	// WebQueryLayerHelper manages WebQueryableLayer queries

	public class WebQueryLayerHelper
	{

		private const int MaxRecs = 500;

		public static Dictionary<string, WebQueryLayerHelper> Instances { get; private set; }

		private int _iMaxRecords;
		private List<IQueryableLayer> _Layers = new List<IQueryableLayer>();
		private Dictionary<long, WebSublayerInfo> _Infos;
		private Dictionary<long, ServiceFeatureTable> _Tables;
		private Dictionary<long, ArcGISMapImageSublayer> _ParentTable;

		public ArcGISMapImageLayer MapImageLayer { get; private set; }
		public string MapImageLayerId { get; private set; }
		public string ServiceUrl { get; private set; }
		public IReadOnlyList<IQueryableLayer> QueryableLayers { get { return _Layers.AsReadOnly(); } }

		// Public methods

		public static async Task<IReadOnlyList<IQueryableLayer>> GetQueryableLayers(ArcGISMapImageLayer miLyr)
		{
			string sID = miLyr.Id;
			WebQueryLayerHelper wh = null;
			if (WebQueryLayerHelper.Instances == null)
				WebQueryLayerHelper.Instances = new Dictionary<string, WebQueryLayerHelper>();
			else if (WebQueryLayerHelper.Instances.ContainsKey(sID))
				wh = WebQueryLayerHelper.Instances[sID];
			if (wh == null)
			{
				wh = new WebQueryLayerHelper();
				WebQueryLayerHelper.Instances[sID] = wh;
			}
			return await wh.IntGetQueryableLayers(miLyr);
		}

		public async Task<WebSublayerInfo> GetSublayerInfo(long iID)
		{
			if (_Infos == null)
				return null;
			if (_Infos.ContainsKey(iID))
				return _Infos[iID];
			WebSublayerInfo wsInfo = await WebUtilities.GetSublayerInfo(ServiceUrl, iID);
			_Infos[iID] = wsInfo;
			return wsInfo;
		}

		public async Task<ServiceFeatureTable> GetSublayerTable(long iID)
		{
			if (_Tables == null)
				return null;
			if (_Tables.ContainsKey(iID))
				return _Tables[iID];
			try
			{

				// Get table

				string sUrl = ServiceUrl + "/" + iID.ToString();
				ServiceFeatureTable sfTab = new ServiceFeatureTable(new Uri(sUrl));
				Map map = AppViewExt.Current.Map;
				if (!map.Tables.Contains(sfTab))
					map.Tables.Add(sfTab);

				// Get related tables

				await sfTab.LoadAsync();
				ArcGISFeatureLayerInfo info = sfTab.LayerInfo;
				foreach (RelationshipInfo relInfo in info.RelationshipInfos)
				{
					sUrl = ServiceUrl + "/" + relInfo.RelatedTableId.ToString();
					ServiceFeatureTable sfTabRel = new ServiceFeatureTable(new Uri(sUrl));
					if (!map.Tables.Contains(sfTabRel))
						map.Tables.Add(sfTabRel);
				}
				return sfTab;

			}
			catch (Exception ex)
			{
				Debug.WriteLine("WebQueryLayerHelper.GetSublayerTable: " + ex.Message);
				return null;
			}
		}

		public async Task<List<Feature>> QueryGeneral(WebQueryableLayer wqLyr, string sWhere, Geometry geom)
		{
			ServiceFeatureTable sfTab = await this.GetSublayerTable(wqLyr.SublayerId);
			if (sfTab == null)
				return null;
			QueryParameters qp = new QueryParameters();
			bool bSpatial = false;
			if ((sfTab.GeometryType != GeometryType.Unknown) && (geom != null))
			{
				qp.Geometry = geom;
				qp.SpatialRelationship = SpatialRelationship.Intersects;
				bSpatial = true;
			}
			if (!string.IsNullOrEmpty(sWhere))
				qp.WhereClause = sWhere;
			else if (!bSpatial)
				qp.WhereClause = "1 = 1";
			List<Feature> Features = new List<Feature>();
			try
			{
				// 100.x BUG:  This will fail for non-standardized queries such as nested clauses
				IEnumerable<Feature> Results = await sfTab.QueryFeaturesAsync(qp, QueryFeatureFields.LoadAll);
				foreach (Feature feat in Results)
					Features.Add(feat);
			}
			catch (Exception ex)
			{
				// Try REST API workaround
				Debug.Print(ex.Message);
				if ((geom == null) && (!string.IsNullOrEmpty(sWhere)))
					Features = await this.WebQuery(wqLyr.SublayerId, sWhere);
			}
			return Features;
		}

		public async Task<List<string>> QueryUniqueValues(WebQueryableLayer wqLyr, string sField, string sWhereClause = null)
		{
			return await WebUtilities.QueryUniqueValues(wqLyr.Info, sField, sWhereClause);
		}

		public async Task<List<QueryRelatedResult>> QueryRelated(ArcGISFeature feat)
		{
			ServiceFeatureTable sfTab = feat.FeatureTable as ServiceFeatureTable;
			if (sfTab == null)
				return null;
			ArcGISFeatureLayerInfo info = sfTab.LayerInfo;
			List<QueryRelatedResult> RelatedResults = new List<QueryRelatedResult>();
			foreach (RelationshipInfo relInfo in info.RelationshipInfos)
			{
				RelatedQueryParameters param = new RelatedQueryParameters(relInfo);
				IReadOnlyList<RelatedFeatureQueryResult> relResults = null; 
				try
				{
					// 100.5 BUG:  This may fail for relationships involving a relationship table
					// if there are no related features ["Item not found in database"]
					// 100.6 BUG: This will not return results if the feature layer
					// is a sublayer of a group layer
					relResults = await sfTab.QueryRelatedFeaturesAsync(feat, param, QueryFeatureFields.LoadAll);
				}
				catch (Exception ex)
				{
					Debug.Print("WebQueryLayerHelper.QueryRelated: " + ex.Message);
					relResults = null;
					//await this.WebQueryRelated(feat, relInfo, RelatedResults);
				}
				if (relResults == null)
					continue;
				if (relResults.Count == 0)
					continue;
				RelatedResults.Add(BasicQueryLayerHelper.CreateQueryRelatedResult(relResults[0]));
			}
			return RelatedResults;
		}

		public bool ParentVisible(long iSublayerID, double dScale)
		{
			if (!_ParentTable.ContainsKey(iSublayerID))
				return true; // Topmost
			ArcGISMapImageSublayer parentLyr = _ParentTable[iSublayerID];
			if (!parentLyr.IsVisible)
				return false;
			if (!parentLyr.IsVisibleAtScale(dScale))
				return false;
			return ParentVisible(parentLyr.Id, dScale);
		}

		// Private methods

		private async Task<IReadOnlyList<IQueryableLayer>> IntGetQueryableLayers(ArcGISMapImageLayer miLyr)
		{

			if (QueryableLayers.Count > 0)
				return QueryableLayers;
			MapImageLayer = miLyr;
			MapImageLayerId = miLyr.Id;
			ServiceUrl = miLyr.ServiceInfo.Source.AbsoluteUri;
			_iMaxRecords = miLyr.ServiceInfo.MaxRecordCount;
			_Infos = new Dictionary<long, WebSublayerInfo>();
			_Tables = new Dictionary<long, ServiceFeatureTable>();
			_ParentTable = new Dictionary<long, ArcGISMapImageSublayer>();
			try
			{

				// Process sublayers

				foreach (ArcGISMapImageSublayer subLyr in miLyr.Sublayers)
					await this.ProcessSublayer(subLyr);

				// Process tables

				foreach (IdInfo ii in miLyr.ServiceInfo.TableInfos)
				{
					string sName = ii.Name;
					long iID = ii.Id;
					WebQueryableLayer wqLyr = await WebQueryableLayer.CreateFromSublayerId(iID, this);
					_Layers.Add(wqLyr);
				}

				return QueryableLayers;

			}
			catch (Exception ex)
			{
				Debug.WriteLine("WebQueryLayerHelper.GetQueryableLayers: " + ex.Message);
				return null;
			}

		}

		private async Task ProcessSublayer(ArcGISMapImageSublayer subLyr)
		{

			long iID = subLyr.Id;
			WebSublayerInfo wsInfo = await this.GetSublayerInfo(iID);
			ArcGISMapServiceSublayerInfo info = wsInfo.ServiceInfo;
			if (info.SublayerType == ArcGISMapServiceSublayerType.FeatureLayer)
			{
				WebQueryableLayer wqLyr = await WebQueryableLayer.CreateFromSublayer(subLyr, this);
				_Layers.Add(wqLyr);
			}

			// Process child sublayers and tables

			foreach (ArcGISMapImageSublayer sub in subLyr.Sublayers)
			{
				_ParentTable[sub.Id] = subLyr;
				await this.ProcessSublayer(sub);
			}

		}

		private async Task<List<Feature>> WebQuery(long iSublayerID, string sWhere)
		{
			List<string> OutFields = new List<string>();
			OutFields.Add("*");
			WebQueryParameters param = new WebQueryParameters()
			{
				SublayerInfo = await WebUtilities.GetSublayerInfo(ServiceUrl, iSublayerID),
				Where = sWhere,
				OutFields = OutFields
			};
			return await WebUtilities.Query(param);
		}

		/*private async Task WebQueryRelated(ArcGISFeature feat, RelationshipInfo relInfo, List<QueryRelatedResult> RelatedResults)
		{

			// Set up and execute web query

			ServiceFeatureTable sfTab = feat.FeatureTable as ServiceFeatureTable;
			ArcGISFeatureLayerInfo flInfo = sfTab.LayerInfo;
			long iLayerID = flInfo.ServiceLayerId;
			string sOIDField = sfTab.ObjectIdField;
			if (!feat.Attributes.ContainsKey(sOIDField))
				return;
			long iOID = -1;
			if (!long.TryParse(feat.Attributes[sOIDField].ToString(), out iOID))
				return;
			List<long> OIDs = new List<long>();
			OIDs.Add(iOID);
			List<string> OutFields = new List<string>();
			OutFields.Add("*");
			WebQueryRelatedParameters param = new WebQueryRelatedParameters()
			{
				SublayerInfo = await WebUtilities.GetSublayerInfo(ServiceUrl, iLayerID),
				RelationshipId = relInfo.Id,
				ObjectIds = OIDs,
				OutFields = OutFields
			};
			Dictionary<long, List<Feature>> WebResults = await WebUtilities.QueryRelatedRecords(param);
			if (WebResults == null)
				return;
			if (!WebResults.ContainsKey(iOID))
				return;
			List<Feature> RelFeatures = WebResults[iOID];
			if (RelFeatures.Count == 0)
				return;

			// Build QueryRelatedResult

			ServiceFeatureTable sfRelTab = await this.GetSublayerTable(relInfo.RelatedTableId);
			ArcGISFeatureLayerInfo flRelInfo = sfRelTab.LayerInfo;
			QueryRelatedResult qrResult = new QueryRelatedResult()
			{
				Name = sfRelTab.TableName,
				DisplayField = flRelInfo.DisplayFieldName,
				Fields = BasicQueryLayerHelper.GetQueryFieldInfos(flRelInfo),
				RelatedFeatures = RelFeatures
			};
			RelatedResults.Add(qrResult);

		}*/

	}

}
