﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading.Tasks;
using System.Web;
using System.Web.Script.Serialization;

using Esri.ArcGISRuntime.ArcGISServices;
using Esri.ArcGISRuntime.Data;
using Esri.ArcGISRuntime.Geometry;

/**** WebUtilities provides various REST API workarounds for Runtime 100.x ****/

namespace QueryAddin
{

	public class WebTableLayerParameter
	{
		// LayerId should be unique for each layer
		public long LayerId = 32768;
		public string WorkspaceId;
		public string DataSourceName;
	}

	public class WebSublayerInfo
	{
		public string MapServiceUrl;
		public long SublayerId = -1;
		public WebTableLayerParameter TableLayerParameter;
		public ArcGISMapServiceSublayerInfo ServiceInfo;
		public bool SupportsDistinct = false;
		public bool SupportsPagination = false;
	}

	public class WebQueryParameters
	{
		public WebSublayerInfo SublayerInfo;
		public string Where;
		public List<long> ObjectIds;
		public bool ReturnGeometry = true;
		public List<string> OutFields;
		public bool ReturnUnique = false;
	}

	public class WebQueryRelatedParameters
	{
		public WebSublayerInfo SublayerInfo;
		public List<long> ObjectIds;
		public long RelationshipId;
		public bool ReturnGeometry = true;
		public List<string> OutFields;
	}

	public enum WebIdentifyOption
	{
		Topmost,
		Visible,
		All
	}

	public class WebIdentifyParameters
	{
		public Geometry Geometry;
		public WebIdentifyOption Layers;
		public List<long> LayerIds;
		public int Tolerance;
		public Envelope MapExtent;
		public double DisplayHeight;
		public double DisplayWidth;
		public double DPI = 96;
		public bool ReturnGeometry = true;
	}

	public class WebFindParameters
	{
		public string SearchText;
		public bool Contains;
		public List<string> Fields;
		public List<long> Layers;
		public bool ReturnGeometry = true;
	}

	public class WebFindResult
	{
		public Feature Feature;
		public string FoundFieldName;
		public string Value;
	}

	public class WebUtilities
	{

		private const int MaxRecs = 500;
		private static JavaScriptSerializer js = new JavaScriptSerializer();

		// Public methods

		public static async Task<Dictionary<string, long>> GetSublayerIds(string sServiceUrl)
		{

			// Convenience method 

			try
			{
				return await WebUtilities.WebGetSublayerIds(sServiceUrl);
			}
			catch (Exception ex)
			{
				Debug.WriteLine("WebUtilities.GetSublayerInfo: " + ex.Message);
				return null;
			}

		}

		public static async Task<WebSublayerInfo> GetSublayerInfo(string sServiceUrl, long iSublayerId)
		{

			// Adds info not provided in ArcGISMapServiceSublayerInfo 

			try
			{
				return await WebUtilities.WebGetSublayerInfo(sServiceUrl, iSublayerId);
			}
			catch (Exception ex)
			{
				Debug.WriteLine("WebUtilities.GetSublayerInfo: " + ex.Message);
				return null;
			}

		}

		public static async Task<WebSublayerInfo> GetSublayerInfo(string sServiceUrl, WebTableLayerParameter param)
		{

			// Support for dynamic layer 

			try
			{
				return await WebUtilities.WebGetSublayerInfo(sServiceUrl, param);
			}
			catch (Exception ex)
			{
				Debug.WriteLine("WebUtilities.GetSublayerInfo: " + ex.Message);
				return null;
			}

		}

		public static async Task<List<Feature>> Query(WebQueryParameters param)
		{

			// Allows querying dynamic layers, unique values, complex non-standardized where clauses 

			try
			{
				return await WebUtilities.WebQuery(param);
			}
			catch (Exception ex)
			{
				Debug.WriteLine("WebUtilities.Query: " + ex.Message);
				return null;
			}

		}

		public static async Task<List<string>> QueryUniqueValues(WebSublayerInfo wsInfo, string sField, string sWhereClause = null)
		{

			// Convenience method 

			try
			{
				return await WebUtilities.WebGetUniqueValues(wsInfo, sField, sWhereClause);
			}
			catch (Exception ex)
			{
				Debug.WriteLine("WebUtilities.QueryUniqueValues: " + ex.Message);
				return null;
			}

		}

		public static async Task<Dictionary<long, List<Feature>>> QueryRelatedRecords(WebQueryRelatedParameters param)
		{

			// Allows querying related records via a relationship table that is not present in the map service 

			try
			{
				return await WebUtilities.WebQueryRelatedRecords(param);
			}
			catch (Exception ex)
			{
				Debug.WriteLine("WebUtilities.QueryelatedRecords: " + ex.Message);
				return null;
			}

		}

		public static async Task<Dictionary<long, List<Feature>>> Identify(string sServiceUrl, WebIdentifyParameters param)
		{

			// Supports all geometry types, not just Point 

			try
			{
				return await WebUtilities.WebIdentify(sServiceUrl, param);
			}
			catch (Exception ex)
			{
				Debug.WriteLine("WebUtilities.Identify: " + ex.Message);
				return null;
			}

		}

		public static async Task<Dictionary<long, List<WebFindResult>>> Find(string sServiceUrl, WebFindParameters param)
		{

			// Not implemented in Runtime 100.x

			try
			{
				return await WebUtilities.WebFind(sServiceUrl, param);
			}
			catch (Exception ex)
			{
				Debug.WriteLine("WebUtilities.Find: " + ex.Message);
				return null;
			}

		}

		// Private methods

		private static async Task<string> GetResponseString(string sRequest)
		{
			string sResponse;
			using (Stream s = await WebUtilities.GetResponseStream(sRequest))
			{
				using (StreamReader sr = new StreamReader(s))
				{
					sResponse = sr.ReadToEnd();
				}
			}
			return sResponse;
		}

		private static async Task<Stream> GetResponseStream(string sRequest)
		{
			HttpWebRequest r = (HttpWebRequest)WebRequest.Create(sRequest);
			r.UseDefaultCredentials = true;
			r.PreAuthenticate = true;
			r.Credentials = CredentialCache.DefaultCredentials;
			HttpWebResponse result = await r.GetResponseAsync() as HttpWebResponse;
			return result.GetResponseStream();
		}

		private static async Task<Dictionary<string, long>> WebGetSublayerIds(string sMapServiceUrl)
		{

			// Get basic service info

			string sRequest = sMapServiceUrl + "?f=json";
			string sJSON = await WebUtilities.GetResponseString(sRequest);
			ArcGISMapServiceInfo msInfo = ArcGISMapServiceInfo.FromJson(sJSON);
			Dictionary<string, long> SublayerIDs = new Dictionary<string, long>();
			foreach (IdInfo info in msInfo.LayerInfos)
				SublayerIDs[info.Name] = info.Id;
			return SublayerIDs;

		}

		private static async Task<WebSublayerInfo> WebGetSublayerInfo(string sMapServiceUrl, long iSublayerId)
		{

			// Get basic service info

			string sRequest = sMapServiceUrl + "/" + iSublayerId.ToString() + "?f=json";
			string sJSON = await WebUtilities.GetResponseString(sRequest);
			ArcGISMapServiceSublayerInfo slInfo = ArcGISMapServiceSublayerInfo.FromJson(sJSON);
			WebSublayerInfo wsInfo = new WebSublayerInfo();
			wsInfo.MapServiceUrl = sMapServiceUrl;
			wsInfo.SublayerId = iSublayerId;
			wsInfo.ServiceInfo = slInfo;
			Dictionary<string, object> dict = (Dictionary<string, object>)js.DeserializeObject(sJSON);

			// Get advanced query info

			if (!slInfo.SupportsAdvancedQueries)
				return wsInfo;
			Dictionary<string, object> aqc = dict["advancedQueryCapabilities"] as Dictionary<string, object>;
			if (aqc.ContainsKey("supportsDistinct"))
				wsInfo.SupportsDistinct = (aqc["supportsDistinct"].ToString().ToLower() == "true");
			if (aqc.ContainsKey("supportsPagination"))
				wsInfo.SupportsPagination = (aqc["supportsPagination"].ToString().ToLower() == "true");
			return wsInfo;

		}

		private static async Task<WebSublayerInfo> WebGetSublayerInfo(string sMapServiceUrl, WebTableLayerParameter param)
		{

			// Get basic service info

			string sLayerParameter = WebUtilities.BuildLayerParameter(param);
			string sRequest = sMapServiceUrl + "/dynamicLayer?layer=" + sLayerParameter + "&f=json";
			string sJSON = await WebUtilities.GetResponseString(sRequest);
			ArcGISMapServiceSublayerInfo slInfo = ArcGISMapServiceSublayerInfo.FromJson(sJSON);
			WebSublayerInfo wsInfo = new WebSublayerInfo();
			wsInfo.MapServiceUrl = sMapServiceUrl;
			wsInfo.TableLayerParameter = param;
			wsInfo.ServiceInfo = slInfo;
			Dictionary<string, object> dict = (Dictionary<string, object>)js.DeserializeObject(sJSON);

			// Get advanced query info

			if (!slInfo.SupportsAdvancedQueries)
				return wsInfo;
			Dictionary<string, object> aqc = dict["advancedQueryCapabilities"] as Dictionary<string, object>;
			if (aqc.ContainsKey("supportsDistinct"))
				wsInfo.SupportsDistinct = (aqc["supportsDistinct"].ToString() == "true");
			if (aqc.ContainsKey("supportsPagination"))
				wsInfo.SupportsPagination = (aqc["supportsPagination"].ToString() == "true");
			return wsInfo;


		}

		private static async Task<List<Feature>> WebQuery(WebQueryParameters param)
		{

			// Build request prefix

			WebSublayerInfo wsInfo = param.SublayerInfo;
			StringBuilder sb = new StringBuilder();
			sb.Append(wsInfo.MapServiceUrl);
			if (wsInfo.TableLayerParameter == null)
				sb.Append("/" + wsInfo.SublayerId + "/query?");
			else
			{
				string sLayer = WebUtilities.BuildLayerParameter(wsInfo.TableLayerParameter);
				sb.Append("/dynamiclayer/query?layer=" + sLayer + "&");
			}
			string sRequestStart = sb.ToString();

			// Build outFields parameter

			string sOutFields = "";
			sb = new StringBuilder();
			if (param.OutFields != null)
			{
				sb.Append("&outFields=");
				bool bFirst = true;
				foreach (string sField in param.OutFields)
				{
					if (bFirst)
						bFirst = false;
					else
						sb.Append(",");
					sb.Append(sField);
				}
				sOutFields = sb.ToString();
			}

			// Get distinct values or Object IDs

			List<Feature> Features = new List<Feature>();
			List<long> ObjectIDs = null;
			if (param.ObjectIds == null)
			{

				// Handle distinct queries separately

				if (param.ReturnUnique && param.SublayerInfo.SupportsDistinct)
					return await WebUtilities.QueryDistinct(param, sRequestStart, sOutFields);

				// Otherwise get object IDs to override maximum returned records

				ObjectIDs = await WebUtilities.QueryObjectIDs(param, sRequestStart);

			}
			else
				ObjectIDs = param.ObjectIds;
			int iCount = ObjectIDs.Count;
			if (iCount == 0)
				return Features;

			// Loop through object IDs

			List<object> webFeatures = new List<object>();
			List<object> webFields = null;
			long iMaxRecs = wsInfo.ServiceInfo.MaxRecordCount / 2;
			if (iMaxRecs > MaxRecs)
				iMaxRecs = MaxRecs;
			List<long> OIDList = new List<long>();
			int iSubCount = 0;
			for (int i = 0; i < iCount; i++)
			{

				OIDList.Add(ObjectIDs[i]);
				iSubCount++;
				if ((iSubCount < iMaxRecs) && (i < iCount - 1))
					continue;

				// Build and execute request

				sb = new StringBuilder();
				sb.Append(sRequestStart);
				sb.Append("objectIds=");
				bool bFirst = true;
				foreach (long iOID in OIDList)
				{
					if (bFirst)
						bFirst = false;
					else
						sb.Append(",");
					sb.Append(iOID.ToString());
				}
				sb.Append(sOutFields);
				sb.Append("&returnGeometry=" + param.ReturnGeometry.ToString().ToLower());
				sb.Append("&f=json");
				string sRequest = sb.ToString();
				string sJSON = await WebUtilities.GetResponseString(sRequest);

				// Gather results

				Dictionary<string, object> dict = (Dictionary<string, object>)js.DeserializeObject(sJSON);
				object[] features = (object[])dict["features"];
				foreach (object feature in features)
					webFeatures.Add(feature);
				if (webFields == null)
					webFields = ((object[])dict["fields"]).ToList();
				OIDList.Clear();
				iSubCount = 0;

			}
			GeometryType geomType = wsInfo.ServiceInfo.GeometryType;
			SpatialReference sr = null;
			if (geomType != GeometryType.Unknown)
				sr = wsInfo.ServiceInfo.Extent.SpatialReference;
			FeatureCollectionTable fcTab = WebUtilities.CreateFeatureTable(webFields, geomType, sr);
			return WebUtilities.CreateFeatures(webFeatures, fcTab);

		}

		private static async Task<List<string>> WebGetUniqueValues(WebSublayerInfo wsInfo, string sField, string sWhereClause = null)
		{

			// Set up WebQueryParameters

			WebQueryParameters param = new WebQueryParameters();
			param.SublayerInfo = wsInfo;
			param.Where = sWhereClause;
			param.ReturnUnique = true;
			param.OutFields = new List<string>();
			param.OutFields.Add(sField);

			// Perform query and gather unique values

			HashSet<string> Values = new HashSet<string>();
			List<Feature> Features = await WebUtilities.WebQuery(param);
			foreach (Feature feat in Features)
				Values.Add(feat.Attributes[sField].ToString());
			List<string> SortedValues = Values.ToList();
			if (SortedValues.Count > 1)
				SortedValues.Sort();
			return SortedValues;

		}

		private static async Task<Dictionary<long, List<Feature>>> WebQueryRelatedRecords(WebQueryRelatedParameters param)
		{

			// Build query and execute request

			WebSublayerInfo wsInfo = param.SublayerInfo;
			string sServiceUrl = wsInfo.MapServiceUrl;
			long iID = wsInfo.SublayerId;
			StringBuilder sbRequest = new StringBuilder();
			sbRequest.Append(sServiceUrl);
			if (wsInfo.TableLayerParameter == null)
				sbRequest.Append("/" + iID + "/queryRelatedRecords?");
			else
			{
				string sLayer = WebUtilities.BuildLayerParameter(wsInfo.TableLayerParameter);
				sbRequest.Append("/dynamiclayer/queryRelatedRecords?layer=" + sLayer + "&");
			}
			sbRequest.Append("objectIds=");
			bool bFirst = true;
			if (param.ObjectIds == null)
				return null;
			foreach (long iOID in param.ObjectIds)
			{
				if (bFirst)
					bFirst = false;
				else
					sbRequest.Append(",");
				sbRequest.Append(iOID.ToString());
			}
			sbRequest.Append("&relationshipId=" + param.RelationshipId.ToString());
			if (param.OutFields != null)
			{
				sbRequest.Append("&outFields=");
				bFirst = true;
				foreach (string sField in param.OutFields)
				{
					if (bFirst)
						bFirst = false;
					else
						sbRequest.Append(",");
					sbRequest.Append(sField);
				}
			}
			sbRequest.Append("&returnGeometry=" + param.ReturnGeometry.ToString().ToLower());
			sbRequest.Append("&f=json");
			string sRequest = sbRequest.ToString();
			string sJSON = await WebUtilities.GetResponseString(sRequest);

			// Gather results

			Dictionary<long, List<Feature>> Results = new Dictionary<long, List<Feature>>();
			GeometryType geomType = wsInfo.ServiceInfo.GeometryType;
			SpatialReference sr = null;
			if (geomType != GeometryType.Unknown)
				sr = wsInfo.ServiceInfo.Extent.SpatialReference;
			Dictionary<string, object> dict = (Dictionary<string, object>)js.DeserializeObject(sJSON);
			object[] fields = (object[])dict["fields"];
			FeatureCollectionTable fcTab = WebUtilities.CreateFeatureTable(fields.ToList(), geomType, sr);
			object[] groups = (object[])dict["relatedRecordGroups"];
			foreach (Dictionary<string, object> group in groups.OfType<Dictionary<string, object>>())
			{
				long iOID = -1;
				if (!long.TryParse(group["objectId"].ToString(), out iOID))
					continue;
				object[] features = (object[])group["relatedRecords"];
				List<Feature> Features = WebUtilities.CreateFeatures(features.ToList(), fcTab);
				Results[iOID] = Features;
			}
			return Results;

		}

		private static async Task<Dictionary<long, List<Feature>>> WebIdentify(string sMapServiceUrl, WebIdentifyParameters param)
		{

			// Build and execute request

			StringBuilder sbRequest = new StringBuilder();
			sbRequest.Append(sMapServiceUrl + "/identify");
			sbRequest.Append("?geometry=" + param.Geometry.ToJson());
			sbRequest.Append("&geometryType=esriGeometry" + param.Geometry.GeometryType.ToString());
			sbRequest.Append("&layers=");
			switch (param.Layers)
			{
				case WebIdentifyOption.Topmost:
					sbRequest.Append("top");
					break;
				case WebIdentifyOption.Visible:
					sbRequest.Append("visible");
					break;
				case WebIdentifyOption.All:
					sbRequest.Append("all");
					break;
			}
			if (param.LayerIds != null)
			{
				bool bFirst = true;
				foreach (long iID in param.LayerIds)
				{
					if (bFirst)
					{
						sbRequest.Append(":");
						bFirst = false;
					}
					else
						sbRequest.Append(",");
					sbRequest.Append(iID.ToString());
				}
			}
			sbRequest.Append("&tolerance=" + param.Tolerance.ToString());
			Envelope env = param.MapExtent;
			sbRequest.Append("&mapExtent=" + env.XMin.ToString() + "," + env.YMin.ToString());
			sbRequest.Append("," + env.XMax.ToString() + "," + env.YMax.ToString());
			sbRequest.Append("&imageDisplay=" + param.DisplayWidth.ToString());
			sbRequest.Append("," + param.DisplayHeight.ToString() + "," + param.DPI.ToString());
			sbRequest.Append("&returnGeometry=" + param.ReturnGeometry.ToString().ToLower());
			sbRequest.Append("&f=json");
			string sRequest = sbRequest.ToString();
			string sJSON = await WebUtilities.GetResponseString(sRequest);

			// Gather results

			Dictionary<string, object> dict = (Dictionary<string, object>)js.DeserializeObject(sJSON);
			object[] results = (object[])dict["results"];
			Dictionary<long, List<object>> layers = new Dictionary<long, List<object>>();
			foreach (Dictionary<string, object> result in results.OfType<Dictionary<string, object>>())
			{
				long iLayerID = -1;
				if (!long.TryParse(result["layerId"].ToString(), out iLayerID))
					continue;
				List<object> features;
				if (!layers.ContainsKey(iLayerID))
				{
					features = new List<object>();
					layers[iLayerID] = features;
				}
				else
					features = layers[iLayerID];
				Dictionary<string, object> feature = new Dictionary<string, object>();
				if (result.ContainsKey("attributes"))
					feature["attributes"] = result["attributes"];
				if (result.ContainsKey("geometry"))
					feature["geometry"] = result["geometry"];
				features.Add(feature);
			}

			// Build feature lists

			Dictionary<long, List<Feature>> Results = new Dictionary<long, List<Feature>>();
			foreach (long iLayerID in layers.Keys)
			{
				FeatureCollectionTable fcTab = await WebUtilities.CreateAliasTable(sMapServiceUrl, iLayerID);
				List<Feature> Features = WebUtilities.CreateFeatures(layers[iLayerID], fcTab);
				Results[iLayerID] = Features;
			}
			return Results;

		}

		private static async Task<Dictionary<long, List<WebFindResult>>> WebFind(string sMapServiceUrl, WebFindParameters param)
		{

			// Build and execute request

			StringBuilder sbRequest = new StringBuilder();
			sbRequest.Append(sMapServiceUrl + "/find");
			sbRequest.Append("?searchText=" + HttpUtility.UrlEncode(param.SearchText));
			if (!param.Contains)
				sbRequest.Append("&contains=false");
			if (param.Fields != null)
			{
				sbRequest.Append("&searchFields=");
				bool bFirst = true;
				foreach (string sField in param.Fields)
				{
					if (bFirst)
						bFirst = false;
					else
						sbRequest.Append(",");
					sbRequest.Append(sField);
				}
			}
			List<long> Sublayers = param.Layers;
			if (Sublayers != null)
			{
				if (Sublayers.Count > 0)
				{
					sbRequest.Append("&layers=");
					bool bFirst = true;
					foreach (long iID in Sublayers)
					{
						if (bFirst)
							bFirst = false;
						else
							sbRequest.Append(",");
						sbRequest.Append(iID.ToString());
					}
				}
			}
			sbRequest.Append("&returnGeometry=" + param.ReturnGeometry.ToString().ToLower());
			sbRequest.Append("&f=json");
			string sRequest = sbRequest.ToString();
			string sJSON = await WebUtilities.GetResponseString(sRequest);

			// Gather results

			Dictionary<string, object> dict = (Dictionary<string, object>)js.DeserializeObject(sJSON);
			object[] results = (object[])dict["results"];
			Dictionary<long, List<object>> layers = new Dictionary<long, List<object>>();
			foreach (Dictionary<string, object> result in results.OfType<Dictionary<string, object>>())
			{
				long iLayerID = -1;
				if (!long.TryParse(result["layerId"].ToString(), out iLayerID))
					continue;
				List<object> items;
				if (!layers.ContainsKey(iLayerID))
				{
					items = new List<object>();
					layers[iLayerID] = items;
				}
				else
					items = layers[iLayerID];
				items.Add(result);
			}

			// Build result lists

			Dictionary<long, List<WebFindResult>> Results = new Dictionary<long, List<WebFindResult>>();
			foreach (long iLayerID in layers.Keys)
			{
				FeatureCollectionTable fcTab = await WebUtilities.CreateAliasTable(sMapServiceUrl, iLayerID);
				List<WebFindResult> Items = new List<WebFindResult>();
				Results[iLayerID] = Items;
				List<object> items = layers[iLayerID];
				foreach (Dictionary<string, object> item in items.OfType<Dictionary<string, object>>())
				{
					WebFindResult wfResult = new WebFindResult();
					wfResult.FoundFieldName = item["foundFieldName"].ToString();
					wfResult.Value = item["value"].ToString();
					wfResult.Feature = WebUtilities.CreateFeature(item, fcTab);
					Items.Add(wfResult);
				}
			}
			return Results;

		}

		private static string BuildLayerParameter(WebTableLayerParameter param)
		{
			StringBuilder sbLayer = new StringBuilder();
			sbLayer.Append("{\"id\":" + param.LayerId.ToString() + ",\"source\":{");
			sbLayer.Append("\"type\":\"dataLayer\",\"dataSource\":{");
			sbLayer.Append("\"type\":\"table\",\"workspaceId\":\"" + param.WorkspaceId + "\",");
			sbLayer.Append("\"dataSourceName\":\"" + param.DataSourceName + "\"}}}");
			string sLayer = sbLayer.ToString();
			return sLayer;
		}

		private static async Task<List<Feature>> QueryDistinct(WebQueryParameters param, string sRequestStart, string sOutFields)
		{

			// Build and execute request

			StringBuilder sb = new StringBuilder();
			sb.Append(sRequestStart);
			string sWhere = "1 = 1";
			if (!string.IsNullOrEmpty(param.Where))
				sWhere = param.Where;
			sb.Append("where=" + HttpUtility.UrlEncode(sWhere));
			sb.Append(sOutFields);
			sb.Append("&returnGeometry=false&returnDistinctValues=true&f=json");
			string sRequest = sb.ToString();
			string sJSON = await WebUtilities.GetResponseString(sRequest);

			// Create features

			Dictionary<string, object> dict = (Dictionary<string, object>)js.DeserializeObject(sJSON);
			List<Feature> Results = new List<Feature>();
			object[] features = dict["features"] as object[];
			object[] fields = (object[])dict["fields"];
			FeatureCollectionTable fcTab = WebUtilities.CreateFeatureTable(fields.ToList());
			return WebUtilities.CreateFeatures(features.ToList(), fcTab);

		}

		private static async Task<List<long>> QueryObjectIDs(WebQueryParameters param, string sRequestStart)
		{

			StringBuilder sb = new StringBuilder();
			sb.Append(sRequestStart);
			string sWhere = "1 = 1";
			if (!string.IsNullOrEmpty(param.Where))
				sWhere = param.Where;
			sb.Append("where=" + HttpUtility.UrlEncode(sWhere));
			sb.Append("&returnIdsOnly=true&f=json");
			string sRequest = sb.ToString();
			string sJSON = await WebUtilities.GetResponseString(sRequest);

			Dictionary<string, object> dict = (Dictionary<string, object>)js.DeserializeObject(sJSON);
			if (!dict.ContainsKey("objectIds"))
				return null;
			object[] oids = dict["objectIds"] as object[];
			if (oids == null)
				return null;
			List<long> ObjectIDs = new List<long>();
			foreach (object id in oids)
				ObjectIDs.Add(Convert.ToInt64(id));
			return ObjectIDs;

		}

		private static FeatureCollectionTable CreateFeatureTable(List<object> fields, GeometryType geomType = GeometryType.Unknown, SpatialReference sr = null)
		{
			List<Field> Fields = new List<Field>();
			foreach (object field in fields)
			{
				string sField = js.Serialize(field);
				Field f = Field.FromJson(sField);
				Fields.Add(f);
			}
			FeatureCollectionTable fcTab = new FeatureCollectionTable(Fields, geomType, sr);
			return fcTab;
		}

		private static async Task<FeatureCollectionTable> CreateAliasTable(string sMapServiceUrl, long iLayerID)
		{
			WebSublayerInfo wsInfo = await WebUtilities.GetSublayerInfo(sMapServiceUrl, iLayerID);
			ArcGISMapServiceSublayerInfo slInfo = wsInfo.ServiceInfo;
			GeometryType geomType = slInfo.GeometryType;
			SpatialReference sr = null;
			if (geomType != GeometryType.Unknown)
				sr = slInfo.Extent.SpatialReference;
			List<Field> Fields = new List<Field>();
			foreach (Field field in slInfo.Fields)
			{
				if (field.FieldType == FieldType.Geometry)
					continue;
				if (field.FieldType == FieldType.OID)
					continue;
				Field fieldNew = new Field(FieldType.Text, field.Alias, field.Alias, 50);
				Fields.Add(fieldNew);
			}
			FeatureCollectionTable fcTab = new FeatureCollectionTable(Fields, geomType, sr);
			return fcTab;
		}

		private static List<Feature> CreateFeatures(List<object> features, FeatureCollectionTable fcTab)
		{
			List<Feature> Features = new List<Feature>();
			foreach (Dictionary<string, object> f in features.OfType<Dictionary<string, object>>())
			{
				Feature feat = WebUtilities.CreateFeature(f, fcTab);
				Features.Add(feat);
			}
			return Features;
		}

		private static Feature CreateFeature(Dictionary<string, object> feature, FeatureCollectionTable fcTab)
		{
			Dictionary<string, object> att = null;
			if (feature.ContainsKey("attributes"))
				att = feature["attributes"] as Dictionary<string, object>;
			Geometry geom = null;
			if (feature.ContainsKey("geometry"))
			{
				string sGeom = js.Serialize(feature["geometry"]);
				geom = Geometry.FromJson(sGeom);
			}
			Feature feat = fcTab.CreateFeature(att, geom);
			return feat;
		}

	}

}
