﻿using Microsoft.Win32;
using System;
using System.Drawing;
using System.IO;
using System.Runtime.InteropServices;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;

using Esri.ArcGISRuntime.Data;

using Extensibility;

namespace QueryAddin
{

	// AttachmentDataItem encapsulates an attachment item that can be opened or saved

	public class AttachmentDataItem
	{

		private class NativeMethods
		{
			[DllImport("shell32.dll", CharSet = CharSet.Auto)]
			internal static extern uint ExtractIconEx(string szFileName, int nIconIndex, IntPtr[] phiconLarge, IntPtr[] phiconSmall, uint nIcons);

			[DllImport("user32.dll", CharSet = CharSet.Auto)]
			internal static extern bool DestroyIcon(IntPtr hIcon);
		}

		// Bindable properties

		public Attachment Item { get; private set; }
		public ImageSource Icon { get; set; }
		public string Name { get; private set; }
		public string Size { get; private set; }
		public string Header { get; private set; }
		public ICommand OpenCommand { get; private set; }

		// Constructors

		public AttachmentDataItem(Attachment item)
		{
			Item = item;
			string sName = item.Name;
			int iPos = sName.LastIndexOf('.');
			if (iPos > -1)
			{
				string sExt = sName.Substring(iPos).ToLower();
				Icon ico = this.GetIcon(sExt);
				if (ico != null)
				{
					BitmapSource bs = System.Windows.Interop.Imaging.CreateBitmapSourceFromHIcon(ico.Handle, Int32Rect.Empty, BitmapSizeOptions.FromEmptyOptions());
					Icon = bs;
				}
			}
			Name = sName;
			Size = (item.Size / 1024).ToString();
			Header = sName;
			CommandHelper ch = new CommandHelper(true);
			ch.AddExecuteAction(this.Open);
			OpenCommand = ch;
		}

		// Public methods

		public async void Open(object parameter = null)
		{
			string sOutDir = System.Environment.GetEnvironmentVariable("TEMP");
			string sOutFile = sOutDir + "\\" + Item.Name;
			await this.SaveAsync(sOutFile);
			System.Diagnostics.Process.Start(sOutFile);
		}

		public void Save()
		{
			string sName = Name;
			string sExt = null;
			int iPos = sName.LastIndexOf('.');
			if (iPos > -1)
				sExt = sName.Substring(iPos + 1);
			string sOutFile = DialogHelper.GetOutputFile(sExt, sName);
			if (sOutFile == "")
				return;
			Task t = this.SaveAsync(sOutFile);
		}

		public void SaveToFolder(string sFolder)
		{
			string sOutFile = Path.Combine(sFolder, Name);
			Task t = this.SaveAsync(sOutFile);
		}

		// Private methods

		private Icon GetIcon(string sExt)
		{

			// Search registry for default icon info

			RegistryKey rkRoot = Registry.ClassesRoot;
			RegistryKey rkFileType = null;
			foreach (string sKey in rkRoot.GetSubKeyNames())
			{
				if (sKey == null)
					continue;
				if (sKey.ToLower() != sExt)
					continue;
				rkFileType = rkRoot.OpenSubKey(sKey);
				break;
			}
			if (rkFileType == null)
				return null;
			object oVal = rkFileType.GetValue("");
			if (oVal == null)
				return null;
			string sDefaultIcon = oVal.ToString() + "\\DefaultIcon";
			RegistryKey rkFileIcon = rkRoot.OpenSubKey(sDefaultIcon);
			if (rkFileIcon == null)
				return null;
			oVal = rkFileIcon.GetValue("");
			if (oVal == null)
				return null;
			string sInfo = oVal.ToString().Replace("\"", "");
			int iPos = sInfo.LastIndexOf(',');
			if (iPos == -1)
				return null;
			string sFileName = sInfo.Substring(0, iPos);
			string sIndex = sInfo.Substring(iPos + 1);
			int iIndex = 0;
			int.TryParse(sIndex, out iIndex);
			if (iIndex < 0)
				iIndex = 0;

			// Extract icon

			Icon ico = null;
			uint iCount = 0;
			IntPtr[] hLargeIcon = new IntPtr[1] { IntPtr.Zero };
			IntPtr[] hSmallIcon = new IntPtr[1] { IntPtr.Zero };
			try
			{
				iCount = NativeMethods.ExtractIconEx(sFileName, iIndex, hLargeIcon, hSmallIcon, 1);
				if ((iCount > 0) && (hSmallIcon[0] != IntPtr.Zero))
					ico = (Icon)System.Drawing.Icon.FromHandle(hSmallIcon[0]).Clone();
			}
			catch
			{
				ico = null;
			}
			finally
			{
				foreach (IntPtr ptr in hLargeIcon)
				{
					if (ptr != IntPtr.Zero)
						NativeMethods.DestroyIcon(ptr);
				}
				foreach (IntPtr ptr in hSmallIcon)
				{
					if (ptr != IntPtr.Zero)
						NativeMethods.DestroyIcon(ptr);
				}
			}
			return ico;

		}

		private async Task SaveAsync(string sOutFile)
		{
			Stream s = await Item.GetDataAsync();
			MemoryStream ms = (MemoryStream)s;
			FileStream fs = new FileStream(sOutFile, FileMode.Create, FileAccess.Write);
			ms.WriteTo(fs);
			fs.Close();
			ms.Close();
		}

	}


}
