﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;

using Esri.ArcGISRuntime.ArcGISServices;
using Esri.ArcGISRuntime.Data;
using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Mapping;
using Esri.ArcGISRuntime.Mapping.Popups;

namespace QueryAddin
{

	// IQueryableLayer provides a basic interface for query operations

	public interface IQueryableLayer
	{
		string DisplayName { get; }
		string LayerID { get; }
		bool IsVisible(double dScale);
		bool IsPopupEnabled { get; }
		PopupDefinition PopupDefinition { get; }
		string DisplayField { get; }
		IReadOnlyList<QueryFieldInfo> Fields { get; }
		Task<List<Feature>> Query(string sWhere = null, Geometry geom = null);
		bool HasAttachments { get; }
		Task<List<Attachment>> GetAttachments(Feature feat);
		Task<List<QueryRelatedResult>> GetRelatedFeatures(Feature feat);
		Task<List<string>> GetUniqueValues(string sField, string sWhereClause = null);
	}

	public class QueryFieldInfo
	{
		public string Name;
		public string Alias;
		public FieldType Type;
		public Dictionary<string, string> CodedValues;
	}

	public struct QueryRelatedResult
	{
		public string Name;
		public bool IsPopupEnabled;
		public PopupDefinition PopupDefinition;
		public string DisplayField;
		public List<QueryFieldInfo> Fields;
		public List<Feature> RelatedFeatures;
	}

	public abstract class QueryableLayer
	{
		protected abstract Task<List<Feature>> QueryGeneral(string sWhere = null, Geometry geom = null);
		protected abstract Task<List<string>> QueryUniqueValues(string sField, string sWhereClause = null);
		protected abstract Task<List<Attachment>> QueryAttachments(Feature feat);
		protected abstract Task<List<QueryRelatedResult>> QueryRelated(Feature feat);
		protected abstract bool CheckVisibility(double dScale);
	}

	// BasicQueryableLayer is based on an ArcGISFeatureTable

	public class BasicQueryableLayer : QueryableLayer, IQueryableLayer
	{

		protected ILayerContent _pLyr;
		protected ArcGISFeatureTable _fTab;
		protected string _sName;
		protected string _sID;
		protected bool _bPopup;
		protected PopupDefinition _popDef;
		protected string _sDisplayField;
		protected List<QueryFieldInfo> _Fields;

		// IQueryableLayer

		public string DisplayName { get { return _sName; } }

		public string LayerID { get { return _sID; } }

		public bool IsVisible(double dScale)
		{
			return this.CheckVisibility(dScale);
		}

		public ArcGISFeatureTable FeatureTable { get { return _fTab; } }

		public bool IsPopupEnabled { get { return _bPopup; } }

		public PopupDefinition PopupDefinition { get { return _popDef; } }

		public string DisplayField { get { return _sDisplayField; } }

		public IReadOnlyList<QueryFieldInfo> Fields { get { return _Fields.AsReadOnly(); } }

		public async Task<List<Feature>> Query(string sWhere = null, Geometry geom = null)
		{
			return await this.QueryGeneral(sWhere, geom);
		}

		public bool HasAttachments { get; protected set; }

		public async Task<List<string>> GetUniqueValues(string sField, string sWhereClause = null)
		{
			return await this.QueryUniqueValues(sField, sWhereClause);
		}

		public async Task<List<Attachment>> GetAttachments(Feature feat)
		{
			return await this.QueryAttachments(feat);
		}

		public bool HasRelationships { get; protected set; }

		public async Task<List<QueryRelatedResult>> GetRelatedFeatures(Feature feat)
		{
			return await this.QueryRelated(feat);
		}

		// Public methods

		public static async Task<BasicQueryableLayer> CreateFromFeatureLayer(FeatureLayer fLyr)
		{
			FeatureTable fTab = fLyr.FeatureTable;
			if (!(fTab is ArcGISFeatureTable))
				return null;
			BasicQueryableLayer qLyr = await BasicQueryableLayer.CreateFromFeatureTable((ArcGISFeatureTable)fTab, fLyr);
			qLyr._sName = fLyr.Name;
			qLyr._pLyr = fLyr;
			qLyr._sID = fLyr.Id;
			return qLyr;
		}

		public static async Task<BasicQueryableLayer> CreateFromFeatureTable(ArcGISFeatureTable fTab, IPopupSource pPopupSrc = null)
		{
			BasicQueryableLayer qLyr = new BasicQueryableLayer();
			await fTab.LoadAsync();
			ArcGISFeatureLayerInfo flInfo = fTab.LayerInfo;
			qLyr._fTab = fTab;
			qLyr._sName = flInfo.ServiceLayerName;
			qLyr._sID = Guid.NewGuid().ToString();
			qLyr.HasAttachments = flInfo.HasAttachments;
			qLyr._bPopup = false;
			qLyr._popDef = null;
			qLyr._sDisplayField = null;
			qLyr._Fields = null;
			if ((pPopupSrc != null) && (pPopupSrc.PopupDefinition != null))
			{
				qLyr._bPopup = true;
				qLyr._popDef = pPopupSrc.PopupDefinition;
			}
			else
			{
				string sDisplayField = flInfo.DisplayFieldName;
				if (!string.IsNullOrEmpty(sDisplayField))
					qLyr._sDisplayField = sDisplayField;
				else
					qLyr._sDisplayField = BasicQueryLayerHelper.GetDefaultDisplayField(qLyr._Fields);
			}
			qLyr._Fields = BasicQueryLayerHelper.GetQueryFieldInfos(flInfo, pPopupSrc);
			return qLyr;
		}

		// Overridden methods

		protected override async Task<List<Feature>> QueryGeneral(string sWhere = null, Geometry geom = null)
		{
			return await BasicQueryLayerHelper.QueryGeneral(_fTab, sWhere, geom);
		}

		protected override async Task<List<string>> QueryUniqueValues(string sField, string sWhereClause = null)
		{
			return await BasicQueryLayerHelper.QueryUniqueValues(_fTab, sField, sWhereClause);
		}

		protected override async Task<List<Attachment>> QueryAttachments(Feature Feat)
		{
			if (!(Feat is ArcGISFeature))
				return null;
			return await BasicQueryLayerHelper.QueryAttachments((ArcGISFeature)Feat);
		}

		protected override async Task<List<QueryRelatedResult>> QueryRelated(Feature Feat)
		{
			if (!(Feat is ArcGISFeature))
				return null;
			return await BasicQueryLayerHelper.QueryRelated((ArcGISFeature)Feat);
		}

		protected override bool CheckVisibility(double dScale)
		{
			if (_pLyr == null)
				return false;
			if (!_pLyr.IsVisible)
				return false;
			if (double.IsNaN(dScale))
				return false;
			if (dScale <= 0)
				return false;
			return _pLyr.IsVisibleAtScale(dScale);
		}

	}

}
