﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Windows;
using System.Windows.Input;

using Extensibility;

namespace QueryAddin
{

	// FindDialogModel is used for executing Find-style queries

	public class FindDialogModel : INotifyPropertyChanged, IWindowContent, ICommand
	{

		private static FindDialogModel _current;
		private FindEngine _Engine;
		private string _sFindText;
		private int _iLayerIndex;
		private bool _bPartialMatch;
		private bool _bAllFields;
		private bool _bFieldItemsEnabled;
		private int _iFieldIndex;
		private Visibility _visResults;
		private string _sResultText;
		private FindResult _SelectedItem;

		// Bindable properties

		public string FindText
		{
			get { return _sFindText; }
			set
			{
				_sFindText = value;
				this.RaisePropertyChanged();
			}
		}

		public ObservableCollection<string> LayerItemsSource { get; private set; }

		public int LayerItemsSelectedIndex
		{
			get { return _iLayerIndex; }
			set
			{
				_iLayerIndex = value;
				_Engine.SetLayerItem(_iLayerIndex);
				this.RaisePropertyChanged();
			}
		}

		public bool PartialMatch
		{
			get { return _bPartialMatch; }
			set
			{
				_bPartialMatch = value;
				this.RaisePropertyChanged();
			}
		}

		public bool AllFields
		{
			get { return _bAllFields; }
			set
			{
				_bAllFields = value;
				if (_bAllFields)
				{
					FieldItemsSelectedIndex = -1;
					CanSelectFieldItem = false;
				}
				else
				{
					FieldItemsSelectedIndex = 0;
					CanSelectFieldItem = true;
				}
				_Engine.SetFieldItem(FieldItemsSelectedIndex);
				this.RaisePropertyChanged();
			}
		}

		public bool CanSelectFieldItem
		{
			get { return _bFieldItemsEnabled; }
			set
			{
				_bFieldItemsEnabled = value;
				this.RaisePropertyChanged();
			}
		}

		public ObservableCollection<string> FieldItemsSource { get; private set; }

		public int FieldItemsSelectedIndex
		{
			get { return _iFieldIndex; }
			set
			{
				_iFieldIndex = value;
				_Engine.SetFieldItem(_iFieldIndex);
				this.RaisePropertyChanged();
			}
		}

		public ICommand FindCommand { get; private set; }

		public ICommand NewSearchCommand { get; private set; }

		public Visibility ResultsVisibility
		{
			get { return _visResults; }
			set
			{
				_visResults = value;
				this.RaisePropertyChanged();
			}
		}

		public ObservableCollection<FindResult> ResultItemsSource { get; private set; }

		public FindResult SelectedItem 
		{
			get { return _SelectedItem; }
			set
			{
				_SelectedItem = value;
				if (this.CanExecuteChanged != null)
					this.CanExecuteChanged(this, new EventArgs());
			}
		}

		public ICommand ContextMenuCommand { get { return this; } }

		public string ResultText
		{
			get { return _sResultText; }
			set
			{
				_sResultText = value;
				this.RaisePropertyChanged();
			}
		}

		// INotifyPropertyChanged

		public event PropertyChangedEventHandler PropertyChanged; 

		// IWindowContent

		public string Title { get { return "Find"; } }

		public event EventHandler<EventArgs> ShowWindow;

		public event EventHandler<EventArgs> HideWindow { add { } remove { } } // Not used

		public void ShutDown() { }

		// ICommand

		public bool CanExecute(object parameter)
		{
			return this.CanExecuteContextMenu(parameter as string);
		}

		public event EventHandler CanExecuteChanged;

		public void Execute(object parameter)
		{
			this.ExecuteContextMenu(parameter as string);
		}

		// Public methods

		public static void Open()
		{
			if (_current == null)
			{
				try
				{
					_current = new FindDialogModel();
					MapApplication.Current.CreateChildWindow("QueryAddin", "FindDialog", _current);
				}
				catch (Exception ex)
				{
					string sMsg = "Could not create Find dialog:\n" + ex.Message;
					MapApplication.Current.Error(sMsg);
					return;
				}
			}
			_current.RaiseShowWindow();
		}

		// Private constructors

		private FindDialogModel()
		{

			_Engine = QueryModelExt.Current.FindEngine;
			LayerItemsSource = _Engine.LayerItemList;
			FieldItemsSource = _Engine.FieldItemList;
			ResultItemsSource = new ObservableCollection<FindResult>();
			CommandHelper ch;
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.ExecuteFind);
			FindCommand = ch;
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.ExecuteNewSearch);
			NewSearchCommand = ch;
			AllFields = true;
			ResultsVisibility = Visibility.Collapsed;

		}

		// Private methods

		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private void RaiseShowWindow()
		{
			if (this.ShowWindow == null)
				return;
			this.ShowWindow(this, new EventArgs());
		}

		private async void ExecuteFind(object parameter)
		{
			if (string.IsNullOrEmpty(FindText))
			{
				MapApplication.Current.Error("No search text entered.");
				return;
			}
			ResultItemsSource.Clear();
			ResultsVisibility = Visibility.Visible;
			ResultText = "Querying...";
			void callback(string sMsg) { ResultText = sMsg; };
			List<FindResult> Results = await _Engine.FindAsync(FindText, PartialMatch, callback);
			int iCount = 0;
			if (Results != null)
			{
				foreach (FindResult result in Results)
				{
					ResultItemsSource.Add(result);
					iCount++;
				}
			}
			if (iCount == 0)
			{
				MapApplication.Current.Message("No features found.");
				ResultText = "OK";
				return;
			}
			else if (iCount == 1)
				ResultText = "One object found";
			else
				ResultText = iCount.ToString() + " objects found";
		}

		private void ExecuteNewSearch(object parameter)
		{
			ResultItemsSource.Clear();
			ResultText = "";
			ResultsVisibility = Visibility.Collapsed;
			FindText = "";
			LayerItemsSelectedIndex = 0;
			PartialMatch = true;
			AllFields = true;
		}

		private bool CanExecuteContextMenu(string sParameter)
		{
			if (SelectedItem == null)
				return false;
			if (sParameter == "Identify")
				return true;
			return SelectedItem.HasGeometry;
		}

		private void ExecuteContextMenu(string sParameter)
		{
			if (SelectedItem == null)
				return;
			bool bFlash = false;
			FlashParameters param = new FlashParameters();
			param.Geometry = SelectedItem.Feature.Geometry;
			param.Flash = true;
			param.ShowFlashLines = true;
			switch(sParameter)
			{
				case "Flash":
					bFlash = true;
					param.Navigate = FlashNavigation.None;
					break;
				case "ZoomTo":
					bFlash = true;
					param.Navigate = FlashNavigation.Zoom;
					break;
				case "PanTo":
					bFlash = true;
					param.Navigate = FlashNavigation.Pan;
					break;
				case "Identify":
					QueryTools.Current.IdentifyFeature(SelectedItem.Layer, SelectedItem.Feature, false);
					break;
			}
			if (!bFlash)
				return;
			QueryViewExt.Current.Flash(param);
		}

	}

}
