﻿using System;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Windows.Input;

using Esri.ArcGISRuntime.Geometry;

using Extensibility;

namespace QueryAddin
{

	// GoToXYDataItem encapsulates a location result

	public class GoToXYDataItem : INotifyPropertyChanged
	{
		private string _sHeader;
		public string Header 
		{
			get { return _sHeader; }
			set 
			{
				_sHeader = value;
				this.RaisePropertyChanged();
			}
		}
		public ICommand Command { get; set; }
		public MapPoint MapLocation { get; set; }
		public MapPoint GeoLocation { get; set; }
		public event PropertyChangedEventHandler PropertyChanged; // INotifyPropertyChanged
		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}
	}

	// GoToXYDialogModel is used for basic GoToXY functionality
	// Assumes that map spatial reference is projected

	public class GoToXYDialogModel : INotifyPropertyChanged, IWindowContent
	{

		private enum GoToXYUnits
		{
			MapUnits,
			DD,
			DMS
		}

		private static GoToXYDialogModel _current;
		private SpatialReference _srGeo = SpatialReferences.Wgs84;
		private GoToXYUnits _units = GoToXYUnits.DMS;
		private string _sUnits;
		private bool _bUnitsSelectedMap;
		private bool _bUnitsSelectedDD;
		private bool _bUnitsSelectedDMS = true;
		private string _sTitle = "Go To XY (Degrees Minutes Seconds)";
		private string _sXTitle = "Long:";
		private string _sX;
		private string _sYTitle = "Lat:";
		private string _sY;
		private GoToXYDataItem _CurrentResult;
		private ICommand _pRecentItemCommand;

		// Bindable dialog properties

		public ICommand FlashCommand { get; private set; }

		public ICommand AddCommand { get; private set; }

		public ICommand ClearCommand { get; private set; }

		public ObservableCollection<GoToXYDataItem> RecentItemsSource { get; private set; }

		public ICommand UnitsCommand { get; private set; }

		public string MapUnits
		{
			get { return _sUnits; }
			set
			{
				_sUnits = value;
			}
		}

		public bool UnitsSetMap
		{
			get { return _bUnitsSelectedMap; }
			set
			{
				_bUnitsSelectedMap = value;
				this.RaisePropertyChanged();
			}
		}

		public bool UnitsSetDD
		{
			get { return _bUnitsSelectedDD; }
			set
			{
				_bUnitsSelectedDD = value;
				this.RaisePropertyChanged();
			}
		}

		public bool UnitsSetDMS
		{
			get { return _bUnitsSelectedDMS; }
			set
			{
				_bUnitsSelectedDMS = value;
				this.RaisePropertyChanged();
			}
		}

		public string XTitle
		{
			get { return _sXTitle; }
			set
			{
				_sXTitle = value;
				this.RaisePropertyChanged();
			}
		}

		public string XText
		{
			get { return _sX; }
			set
			{
				_sX = value;
				this.RaisePropertyChanged();
			}
		}

		public string YTitle
		{
			get { return _sYTitle; }
			set
			{
				_sYTitle = value;
				this.RaisePropertyChanged();
			}
		}

		public string YText
		{
			get { return _sY; }
			set
			{
				_sY = value;
				this.RaisePropertyChanged();
			}
		}

		public ICommand TextEnteredCommand { get; private set; }

		// INotifyPropertyChanged

		public event PropertyChangedEventHandler PropertyChanged; 

		// IWindowContent

		public string Title
		{
			get { return _sTitle; }
			set
			{
				_sTitle = value;
				this.RaisePropertyChanged();
			}
		}

		public event EventHandler<EventArgs> ShowWindow;

		public event EventHandler<EventArgs> HideWindow { add { } remove { } } // Not used

		public void ShutDown() { }

		// Public methods

		public static void Open()
		{
			if (_current == null)
			{
				try
				{
					_current = new GoToXYDialogModel();
					MapApplication.Current.CreateChildWindow("QueryAddin", "GoToXYDialog", _current);
				}
				catch (Exception ex)
				{
					string sMsg = "Could not create Find dialog:\n" + ex.Message;
					MapApplication.Current.Error(sMsg);
					return;
				}
			}
			_current.RaiseShowWindow();
		}

		// Event handlers

		void AppViewExt_MapChanged(object sender, MapChangedArgs e)
		{
			this.RecentItemsSource.Clear();
			this.UpdateResults(true);
		}

		// Private constructors

		private GoToXYDialogModel()
		{
			_current = this;
			AppViewExt.Current.MapChanged += AppViewExt_MapChanged;
			RecentItemsSource = new ObservableCollection<GoToXYDataItem>();
			CommandHelper ch;
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.ExecuteFlash);
			FlashCommand = ch;
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.ExecuteAdd);
			AddCommand = ch;
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.ExecuteClear);
			ClearCommand = ch;
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.ExecuteRecentItem);
			_pRecentItemCommand = ch;
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.ExecuteUnits);
			UnitsCommand = ch;
			MapUnits = this.GetMapUnitText();
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.ExecuteTextEntered);
			TextEnteredCommand = ch;
		}

		// Private methods

		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private void RaiseShowWindow()
		{
			if (this.ShowWindow == null)
				return;
			this.ShowWindow(this, new EventArgs());
		}

		private void ExecuteFlash(object parameter)
		{
			string sOption = parameter as string;
			if (sOption == null)
				return;
			if (!this.GetEntry())
				return;
			FlashParameters fp = new FlashParameters();
			fp.Geometry = _CurrentResult.MapLocation;
			fp.Flash = true;
			fp.ShowFlashLines = true;
			fp.Navigate = FlashNavigation.None;
			switch (sOption)
			{
				case "PanTo":
					fp.Navigate = FlashNavigation.Pan;
					break;
				case "ZoomTo":
					fp.Navigate = FlashNavigation.Zoom;
					break;
			}
			QueryViewExt.Current.Flash(fp);
		}

		private void ExecuteAdd(object parameter)
		{
			string sOption = parameter as string;
			if (sOption == null)
				return;
			if (!this.GetEntry())
				return;
			GraphicAction action = GraphicAction.AddXYPoint;
			switch (sOption)
			{
				case "LabeledPoint":
					action = GraphicAction.AddXYLabeledPoint;
					break;
				case "Callout":
					action = GraphicAction.AddXYCallout;
					break;
			}
			GraphicParameters gp = new GraphicParameters();
			gp.Action = action;
			gp.Geometry = _CurrentResult.MapLocation;
			gp.Label = _CurrentResult.Header;
			QueryViewExt.Current.ProcessGraphic(gp);
		}

		private void ExecuteClear(object parameter)
		{
			string sOption = parameter as string;
			if (sOption == null)
				return;
			switch (sOption)
			{
				case "Graphics":
					GraphicParameters gp = new GraphicParameters();
					gp.Action = GraphicAction.ClearXY;
					QueryViewExt.Current.ProcessGraphic(gp);
					break;
				case "Results":
					RecentItemsSource.Clear();
					break;
			}
		}

		private void ExecuteRecentItem(object parameter)
		{
			GoToXYDataItem item = parameter as GoToXYDataItem;
			if (item == null)
				return;
			_CurrentResult = item;
			FlashParameters fp = new FlashParameters();
			fp.Geometry = _CurrentResult.MapLocation;
			fp.Flash = true;
			fp.ShowFlashLines = true;
			fp.Navigate = FlashNavigation.None;
			QueryViewExt.Current.Flash(fp);
			this.UpdateResults(false);
		}

		private void ExecuteUnits(object parameter)
		{
			string sUnits = parameter as string;
			if (sUnits == null)
				return;
			switch (sUnits)
			{
				case "MapUnits":
					_units = GoToXYUnits.MapUnits;
					UnitsSetMap = true;
					UnitsSetDD = false;
					UnitsSetDMS = false;
					break;
				case "DD":
					_units = GoToXYUnits.DD;
					UnitsSetMap = false;
					UnitsSetDD = true;
					UnitsSetDMS = false;
					break;
				case "DMS":
					_units = GoToXYUnits.DMS;
					UnitsSetMap = false;
					UnitsSetDD = false;
					UnitsSetDMS = true;
					break;
			}
			this.UpdateResults(true);
		}

		private void ExecuteTextEntered(object parameter)
		{
			if (!this.GetEntry())
				return;
			FlashParameters fp = new FlashParameters();
			fp.Geometry = _CurrentResult.MapLocation;
			fp.Flash = true;
			fp.ShowFlashLines = true;
			fp.Navigate = FlashNavigation.Pan;
			QueryViewExt.Current.Flash(fp);
		}

		private string GetMapUnitText()
		{
			string sUnits = "Map Units";
			SpatialReference sr = QueryModelExt.Current.SpatialReference;
			Unit unit = sr.Unit;
			if (unit == null)
				return sUnits;
			if (!string.IsNullOrEmpty(unit.Name))
				return unit.Name;
			LinearUnit lu;
			if (!UnitsHelper.GetStandardLinearUnit(sr, out lu))
				return sUnits;
			return lu.Name;
		}

		private bool GetEntry()
		{

			// Check for valid text values

			double dX = this.ToDouble(XText);
			if (double.IsNaN(dX))
			{
				MapApplication.Current.Error("Invalid X or Longitude value");
				return false;
			}
			double dY = this.ToDouble(YText);
			if (double.IsNaN(dY))
			{
				MapApplication.Current.Error("Invalid Y or Latitude value");
				return false;
			}

			// Check if result is redundant

			bool bGeo = (_units != GoToXYUnits.MapUnits);
			if (_CurrentResult != null)
			{
				MapPoint pt;
				if (bGeo)
					pt = _CurrentResult.GeoLocation;
				else
					pt = _CurrentResult.MapLocation;
				if ((pt.X == dX) && (pt.Y == dY))
					return true;
			}

			// Create result

			SpatialReference srMap = QueryModelExt.Current.SpatialReference;
			_CurrentResult = new GoToXYDataItem();
			if (bGeo)
			{
				_CurrentResult.GeoLocation = new MapPoint(dX, dY, _srGeo);
				_CurrentResult.MapLocation = GeometryEngine.Project(_CurrentResult.GeoLocation, srMap) as MapPoint;
			}
			else
			{
				_CurrentResult.MapLocation = new MapPoint(dX, dY, srMap);
				_CurrentResult.GeoLocation = GeometryEngine.Project(_CurrentResult.MapLocation, _srGeo) as MapPoint;
			}
			_CurrentResult.Command = _pRecentItemCommand;
			this.SetHeader(_CurrentResult);

			// Add to results

			RecentItemsSource.Insert(0, _CurrentResult);
			this.UpdateResults(false);
			return true;

		}

		private void UpdateResults(bool bUnitsChanged)
		{

			// Update text properties

			if (!bUnitsChanged)
			{
				string[] Tok = _CurrentResult.Header.Split(' ');
				XText = Tok[0];
				YText = Tok[1];
				return;
			}
			XText = "";
			YText = "";

			// Update title

			MapUnits = this.GetMapUnitText();
			string sTitle = "Go To XY (";
			bool bGeo = true;
			switch (_units)
			{
				case GoToXYUnits.DMS:
					sTitle += "Degrees Minutes Seconds)"; 
					break;
				case GoToXYUnits.DD:
					sTitle += "Decimal Degrees)";
					break;
				case GoToXYUnits.MapUnits:
					sTitle += MapUnits + ")";
					bGeo = false;
					break;
			}
			Title = sTitle;

			// Update labels

			if (bGeo)
			{
				XTitle = "Long:";
				YTitle = "Lat:";
			}
			else
			{
				XTitle = "X:";
				YTitle = "Y:";
			}

			// Update result headers

			foreach (GoToXYDataItem item in RecentItemsSource)
				this.SetHeader(item);

		}

		private double ToDouble(string sVal)
		{
			double d = double.NaN;
			if (string.IsNullOrEmpty(sVal))
				return d;
			if (_units == GoToXYUnits.DMS)
				d = GeometryHelper.FromDMS(sVal);
			else
				double.TryParse(sVal, out d);
			return d;
		}

		private void SetHeader(GoToXYDataItem item)
		{
			if (_units != GoToXYUnits.MapUnits)
			{
				MapPoint pt = item.GeoLocation;
				if (_units == GoToXYUnits.DMS)
					item.Header = GeometryHelper.ToDMS(pt.X, true) + " " + GeometryHelper.ToDMS(pt.Y, true);
				else
					item.Header = pt.X.ToString("N3") + " " + pt.Y.ToString("N3");
			}
			else
			{
				MapPoint pt = item.MapLocation;
				item.Header = pt.X.ToString("N3") + " " + pt.Y.ToString("N3");
			}
		}

	}

}
