﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;

using Esri.ArcGISRuntime.Data;
using Esri.ArcGISRuntime.Geometry;

using Extensibility;

namespace QueryAddin
{

	// IdentifyDialogModel is used to display Identify-style queries
	// Assumes that map spatial reference is projected

	class IdentifyDialogModel : INotifyPropertyChanged, IWindowContent
	{

		private bool _bInitialized = false;
		private ResultCollectionType _rcType = ResultCollectionType.Tree;
		private bool _bQueryRelated = true;
		private bool _bAttached = false;
		private IdentifyEngine _Engine;
		private int _iLayerIndex;
		private string _sLocation;
		private Visibility _visAttachments;
		private string _sAttachments;
		private string _sStatus;
		private bool _bFlashLines = false;
		private IdentifyUnits _Units;
		private MapPoint _ptLocation;
		private IdentifyResult _idResult;
		private IdentifyListHelper _idHelper;

		// Bindable properties

		public ObservableCollection<string> LayerItemsSource { get; private set; }

		public int LayerItemsSelectedIndex
		{
			get { return _iLayerIndex; }
			set
			{
				_iLayerIndex = value;
				_Engine.SetLayerItem(_iLayerIndex);
				this.RaisePropertyChanged();
			}
		}

		public ObservableCollection<IdentifyResult> IdentifyItemsSource { get; private set; }

		public ICommand SelectItemCommand { get { return _idHelper; } }

		public string FeatureLocation
		{
			get { return _sLocation; }
			set
			{
				_sLocation = value;
				this.RaisePropertyChanged();
			}
		}

		public ICommand LocationUnitsCommand { get; private set; }

		public Visibility AttachmentVisibility
		{
			get { return _visAttachments; }
			private set
			{
				_visAttachments = value;
				this.RaisePropertyChanged();
			}
		}

		public ICommand AttachmentManagerCommand { get; private set; }

		public ObservableCollection<AttachmentDataItem> AttachmentItemsSource { get; private set; }

		public string AttachmentsText
		{
			get { return _sAttachments; }
			private set
			{
				_sAttachments = value;
				this.RaisePropertyChanged();
			}
		}

		public ObservableCollection<Attribute> Attributes { get; private set; }

		public string Status
		{
			get { return _sStatus; }
			private set
			{
				_sStatus = value;
				this.RaisePropertyChanged();
			}
		}

		// INotifyPropertyChanged

		public event PropertyChangedEventHandler PropertyChanged; 

		// IWindowContent

		public string Title { get { return "Identify"; } }

		public event EventHandler<EventArgs> ShowWindow;

		public event EventHandler<EventArgs> HideWindow { add { } remove { } } // Not used

		public void ShutDown() { }

		// Constructors

		public IdentifyDialogModel()
		{
			this.Initialize();
		}

		// Public methods

		public void Open()
		{
			if (!_bInitialized)
				return;
			if (!_bAttached)
			{
				try
				{
					MapApplication.Current.CreateChildWindow("QueryAddin", "IdentifyDialog", this);
					_bAttached = true;
				}
				catch (Exception ex)
				{
					string sMsg = "Could not open Identify dialog:\n" + ex.Message;
					MapApplication.Current.Error(sMsg);
				}
			}
			this.RaiseShowWindow();
		}

		public void Identify(Geometry geom)
		{
			Task t = this.ExecuteIdentify(geom);
		}

		public void IdentifyFeature(IQueryableLayer pLyr, Feature Feat, bool QueryRelated)
		{
			Task t = this.ExecuteIdentifyFeature(pLyr, Feat, QueryRelated);
		}

		public async Task<Geometry> FindFeatureGeometry(Geometry geom)
		{
			return await _Engine.FindFeatureGeometry(geom);
		}

		// Event handlers

		private void RecordResult_IsSelectedChanged(object sender, EventArgs e)
		{
			IdentifyResult idResult = (IdentifyResult)sender;
			if (!idResult.IsSelected)
				return;
			this.Selected((IdentifyResult)sender);
		}

		// Private methods

		private void Initialize()
		{

			if (_bInitialized)
				return;
			try
			{

				// Initialize internal and bindable items

				_rcType = QueryModelExt.Current.ResultCollectionType;
				_bQueryRelated = QueryModelExt.Current.QueryRelated;
				_Engine = QueryModelExt.Current.IdentifyEngine;
				LayerItemsSource = _Engine.LayerItemList;
				LayerItemsSelectedIndex = 0;
				IdentifyItemsSource = new ObservableCollection<IdentifyResult>();
				CommandHelper ch;
				ch = new CommandHelper(true);
				ch.AddExecuteAction(this.ExecuteLocationUnits);
				LocationUnitsCommand = ch;
				AttachmentVisibility = Visibility.Collapsed;
				ch = new CommandHelper(true);
				ch.AddExecuteAction(this.ExecuteAttachmentManager);
				AttachmentManagerCommand = ch;
				AttachmentItemsSource = new ObservableCollection<AttachmentDataItem>();
				Attributes = new ObservableCollection<Attribute>();

				// Initialize helper for list-style display of identify results

				_idHelper = new IdentifyListHelper();
				bool bEnabled = (_rcType == ResultCollectionType.List);
				_idHelper.Initialize(IdentifyItemsSource, bEnabled);
				_bInitialized = true;

			}
			catch (Exception ex)
			{
				string sMsg = "Could not create Identify dialog:\n" + ex.Message;
				MapApplication.Current.Error(sMsg);
			}

		}

		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private void RaiseShowWindow()
		{
			if (this.ShowWindow == null)
				return;
			this.ShowWindow(this, new EventArgs());
		}

		private async Task ExecuteIdentify(Geometry geom)
		{
			if (_Engine.IsBusy)
				return;
			this.ClearResults();
			FeatureLocation = "";
			AttachmentItemsSource.Clear();
			Attributes.Clear();
			Status = "Querying...";
			_bFlashLines = false;
			_Engine.SetLayerItem(LayerItemsSelectedIndex);
			void callback(string sMsg) { Status = sMsg; };
			List<IdentifyResult> Results = await _Engine.IdentifyAsync(geom, _bQueryRelated, callback);
			if (Results.Count == 0)
			{
				Status = "Identified 0 features";
				return;
			}
			bool bFirst = true;
			int iFeatCount = 0;
			foreach (IdentifyResult idResult in Results)
			{

				IdentifyItemsSource.Add(idResult);
				if (_rcType == ResultCollectionType.List)
				{
					this.AddResult(idResult);
					iFeatCount++;
					if (bFirst)
					{
						idResult.IsSelected = true;
						bFirst = false;
					}
					continue;
				}
				foreach (IdentifyResult idRecordResult in idResult.RelatedRecords)
				{
					this.AddResult(idRecordResult);
					iFeatCount++;
				}
				if (bFirst)
				{
					idResult.IsExpanded = true;
					idResult.RelatedRecords[0].IsSelected = true;
					bFirst = false;
				}

			}
			if (iFeatCount == 1)
				Status = "Identified 1 feature";
			else
				Status = "Identified " + iFeatCount.ToString() + " feature(s)";
			this.Open();
			_bFlashLines = true;
		}

		private async Task ExecuteIdentifyFeature(IQueryableLayer pLyr, Feature Feat, bool QueryRelated)
		{
			if (_Engine.IsBusy)
				return;
			this.ClearResults();
			FeatureLocation = "";
			AttachmentItemsSource.Clear();
			Attributes.Clear();
			Status = "Querying...";
			_bFlashLines = false;
			IdentifyResult idLayerResult = await _Engine.IdentifyFeature(pLyr, Feat, QueryRelated);
			if (idLayerResult == null)
				return;
			idLayerResult.IsExpanded = true;
			idLayerResult.RelatedRecords[0].IsSelected = true;
			IdentifyItemsSource.Add(idLayerResult);
			this.Selected(idLayerResult.RelatedRecords[0]);
			Status = "Identified 1 feature";
			this.Open();
			_bFlashLines = true;
		}

		private void ExecuteLocationUnits(object parameter)
		{
			string sUnits = parameter as string;
			switch (sUnits)
			{
				case "Feet":
					_Units = IdentifyUnits.Feet;
					break;
				case "Meters":
					_Units = IdentifyUnits.Meters;
					break;
				case "DD":
					_Units = IdentifyUnits.DD;
					break;
				case "DMS":
					_Units = IdentifyUnits.DMS;
					break;
			}
			if (FeatureLocation == "")
				return;
			FeatureLocation = UnitsHelper.GetLocationString(_ptLocation, _Units);
		}

		private void ExecuteAttachmentManager(object parameter)
		{

			QueryModelExt.Current.OpenAttachmentDialog(_idResult);

			// Update attachment list

			AttachmentItemsSource.Clear();
			foreach (Attachment aiItem in _idResult.Attachments)
				AttachmentItemsSource.Add(new AttachmentDataItem(aiItem));
			AttachmentsText = "Attachments (" + _idResult.Attachments.Count.ToString() + ")";

		}

		private void Selected(IdentifyResult idRecordResult)
		{

			_idResult = idRecordResult;
			FeatureLocation = "";
			AttachmentVisibility = Visibility.Collapsed;
			AttachmentItemsSource.Clear();
			Attributes.Clear();
			if (idRecordResult.Feature == null)
				return;

			// Flash and set location

			if (idRecordResult.ContextMenuCommand.CanExecute(null))
			{
				Geometry geom = idRecordResult.Feature.Geometry;
				FlashParameters fp = new FlashParameters();
				fp.Geometry = geom;
				fp.Flash = true;
				fp.ShowFlashLines = _bFlashLines;
				fp.Navigate = FlashNavigation.None;
				QueryViewExt.Current.Flash(fp);
				_ptLocation = GeometryHelper.GetCenter(geom);
				FeatureLocation = UnitsHelper.GetLocationString(_ptLocation, _Units);
			}

			// Populate attachments

			if (idRecordResult.Attachments != null)
			{
				AttachmentVisibility = Visibility.Visible;
				foreach (Attachment aiItem in idRecordResult.Attachments)
					AttachmentItemsSource.Add(new AttachmentDataItem(aiItem));
				AttachmentsText = "Attachments (" + idRecordResult.Attachments.Count.ToString() + ")";
			}

			// Populate attribute list

			if (idRecordResult.Attributes != null)
			{
				foreach (Attribute a in idRecordResult.Attributes)
					Attributes.Add(a);
			}

			// Update status and navigation

			IdentifyResult idParent = idRecordResult.Parent;
			if (idParent != null)
				Status = idParent.Value;
			while (idParent != null)
			{
				idParent.IsExpanded = true;
				idParent = idParent.Parent;
			}
			_idHelper.ItemSelected(idRecordResult);

		}

		private void AddResult(IdentifyResult idResult)
		{
			if (idResult.ContextMenuCommand.CanExecute(null))
			{
				FlashParameters fp = new FlashParameters();
				fp.Geometry = idResult.Feature.Geometry;
				fp.Flash = true;
				fp.ShowFlashLines = false;
				fp.Navigate = FlashNavigation.None;
				QueryViewExt.Current.Flash(fp);
			}
			idResult.IsSelectedChanged += RecordResult_IsSelectedChanged;
			if (!_bQueryRelated)
				return;
			if (idResult.RelatedRecords != null)
			{
				foreach (IdentifyResult pRelLayer in idResult.RelatedRecords)
				{
					foreach (IdentifyResult pRelRecord in pRelLayer.RelatedRecords)
						pRelRecord.IsSelectedChanged += RecordResult_IsSelectedChanged;
				}
			}
		}

		private void ClearResults()
		{
			if (IdentifyItemsSource.Count == 0)
				return;
			foreach (IdentifyResult idResult in IdentifyItemsSource)
			{
				if (_rcType == ResultCollectionType.List)
				{
					this.ClearResult(idResult);
					continue;
				}
				foreach (IdentifyResult idRecordResult in idResult.RelatedRecords)
					this.ClearResult(idRecordResult);
			}
			IdentifyItemsSource.Clear();
		}

		private void ClearResult(IdentifyResult idResult)
		{
			idResult.IsSelectedChanged -= RecordResult_IsSelectedChanged;
			if (!_bQueryRelated)
				return;
			if (idResult.RelatedRecords != null)
			{
				foreach (IdentifyResult pRelLayer in idResult.RelatedRecords)
				{
					foreach (IdentifyResult pRelRecord in pRelLayer.RelatedRecords)
						pRelRecord.IsSelectedChanged += RecordResult_IsSelectedChanged;
				}
			}
		}

	}

}
