﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Text;
using System.Windows.Input;

using Esri.ArcGISRuntime.Geometry;

using Extensibility;

namespace QueryAddin
{

	// MeasureDialogModel is used to set measure options and display results
	// Assumes that map spatial reference is projected

	public class MeasureDialogModel : INotifyPropertyChanged, ICommand, IWindowContent
	{

		private bool _bAttached;
		private bool _bShowTotal;
		private string _sResultText;
		private MeasureLinearUnit _LinearUnit;
		private MeasureAreaUnit _AreaUnit;
		private MeasureType _Mode = MeasureType.Line;
		private List<Geometry> _GeomList = new List<Geometry>();

		// Bindable properties

		public ICommand Command { get { return this; } }

		public bool ShowTotal
		{
			get { return _bShowTotal; }
			set 
			{
				_bShowTotal = value;
				this.Reset();
			}
		}

		public string ResultText
		{
			get { return _sResultText; }
			set
			{
				_sResultText = value;
				this.RaisePropertyChanged();
			}
		}

		public CheckHelper DistIsChecked { get; private set; }

		public CheckHelper AreaIsChecked { get; private set; }

		// INotifyPropertyChanged

		public event PropertyChangedEventHandler PropertyChanged;

		// ICommand

		public bool CanExecute(object parameter) { return true; }

		public event EventHandler CanExecuteChanged { add { } remove { } } // Not used

		public void Execute(object parameter)
		{
			string sParam = parameter as string;
			if (string.IsNullOrEmpty(sParam))
				return;
			switch (sParam)
			{
				case "MeasureLine":
					_Mode = MeasureType.Line;
					QueryTools.Current.SetMeasureType(MeasureType.Line);
					break;
				case "MeasureArea":
					_Mode = MeasureType.Area;
					QueryTools.Current.SetMeasureType(MeasureType.Area);
					break;
				case "MeasureFeature":
					_Mode = MeasureType.Feature;
					QueryTools.Current.SetMeasureType(MeasureType.Feature);
					break;
				case "SetDistKilometers":
					_LinearUnit = new MeasureLinearUnit(MeasureLinearUnits.Kilometers);
					this.DistanceChecked();
					this.Reset();
					break;
				case "SetDistMeters":
					_LinearUnit = new MeasureLinearUnit(MeasureLinearUnits.Meters);
					this.DistanceChecked();
					this.Reset();
					break;
				case "SetDistDecimeters":
					_LinearUnit = new MeasureLinearUnit(MeasureLinearUnits.Decimeters);
					this.DistanceChecked();
					this.Reset();
					break;
				case "SetDistCentimeters":
					_LinearUnit = new MeasureLinearUnit(MeasureLinearUnits.Centimeters);
					this.DistanceChecked();
					this.Reset();
					break;
				case "SetDistMillimeters":
					_LinearUnit = new MeasureLinearUnit(MeasureLinearUnits.Millimeters);
					this.DistanceChecked();
					this.Reset();
					break;
				case "SetDistMiles":
					_LinearUnit = new MeasureLinearUnit(MeasureLinearUnits.Miles);
					this.DistanceChecked();
					this.Reset();
					break;
				case "SetDistNauticalMiles":
					_LinearUnit = new MeasureLinearUnit(MeasureLinearUnits.NauticalMiles);
					this.DistanceChecked();
					this.Reset();
					break;
				case "SetDistYards":
					_LinearUnit = new MeasureLinearUnit(MeasureLinearUnits.Yards);
					this.DistanceChecked();
					this.Reset();
					break;
				case "SetDistFeet":
					_LinearUnit = new MeasureLinearUnit(MeasureLinearUnits.Feet);
					this.DistanceChecked();
					this.Reset();
					break;
				case "SetDistInches":
					_LinearUnit = new MeasureLinearUnit(MeasureLinearUnits.Inches);
					this.DistanceChecked();
					this.Reset();
					break;
				case "SetAreaKilometers":
					_AreaUnit = new MeasureAreaUnit(MeasureAreaUnits.SquareKilometers);
					this.AreaChecked();
					this.Reset();
					break;
				case "SetAreaHectares":
					_AreaUnit = new MeasureAreaUnit(MeasureAreaUnits.Hectares);
					this.AreaChecked();
					this.Reset();
					break;
				case "SetAreaMeters":
					_AreaUnit = new MeasureAreaUnit(MeasureAreaUnits.SquareMeters);
					this.AreaChecked();
					this.Reset();
					break;
				case "SetAreaAres":
					_AreaUnit = new MeasureAreaUnit(MeasureAreaUnits.Ares);
					this.AreaChecked();
					this.Reset();
					break;
				case "SetAreaMiles":
					_AreaUnit = new MeasureAreaUnit(MeasureAreaUnits.SquareMiles);
					this.AreaChecked();
					this.Reset();
					break;
				case "SetAreaAcres":
					_AreaUnit = new MeasureAreaUnit(MeasureAreaUnits.Acres);
					this.AreaChecked();
					this.Reset();
					break;
				case "SetAreaFeet":
					_AreaUnit = new MeasureAreaUnit(MeasureAreaUnits.SquareFeet);
					this.AreaChecked();
					this.Reset();
					break;
				case "SetAreaYards":
					_AreaUnit = new MeasureAreaUnit(MeasureAreaUnits.SquareYards);
					this.AreaChecked();
					this.Reset();
					break;
				case "Clear":
					this.Reset(true);
					break;
			}
		}

		// IWindowContent

		public string Title { get { return "Measure"; } }

		public event EventHandler<EventArgs> ShowWindow;

		public event EventHandler<EventArgs> HideWindow { add { } remove { } } // Not used

		public void ShutDown() {}

		// Constructors

		public MeasureDialogModel()
		{
			LinearUnit lu = LinearUnits.Meters;
			SpatialReference sr = QueryModelExt.Current.SpatialReference;
			LinearUnit luStandard;
			bool bResult = UnitsHelper.GetStandardLinearUnit(sr, out luStandard);
			if (bResult == true)
				lu = luStandard;
			_LinearUnit = new MeasureLinearUnit(lu);
			DistIsChecked = new CheckHelper();
			DistIsChecked.PropertyChanged += DistIsChecked_PropertyChanged;
			this.DistanceChecked();
			_AreaUnit = new MeasureAreaUnit(AreaUnit.CreateFromLinearUnit(lu));
			AreaIsChecked = new CheckHelper();
			AreaIsChecked.PropertyChanged += AreaIsChecked_PropertyChanged;
			this.AreaChecked();
			QueryModelExt.Current.CanNavigateMap += QueryModelExt_CanNavigateMap;
			this.Reset();
		}

		// Public methods

		public void Open()
		{
			if (!_bAttached)
			{
				try
				{

					// Create parent window

					MapApplication.Current.CreateChildWindow("QueryAddin", "MeasureDialog", this);
					_bAttached = true;

				}
				catch (Exception ex)
				{
					string sMsg = "Could not create Measure dialog:\n" + ex.Message;
					MapApplication.Current.Error(sMsg);
					return;
				}
			}
			this.RaiseShowWindow();
		}

		public void MeasureProgress(Geometry g)
		{
			this.UpdateResults(g, false);
		}

		public async void MeasureComplete(Geometry g)
		{
			Geometry geom = null;
			if (_Mode == MeasureType.Feature)
			{
				IdentifyEngine Engine = QueryModelExt.Current.IdentifyEngine;
				geom = await Engine.FindFeatureGeometry(g);
			}
			else
				geom = g;
			if (geom != null)
			{
				bool bFlash = false;
				bool bGraphic = false;
				if (_Mode == MeasureType.Feature)
				{
					if (!_bShowTotal)
						bFlash = true;
					else if (geom is MapPoint)
						bFlash = true;
					else
						bGraphic = true;
				}
				else if (_bShowTotal)
					bGraphic = true;
				if (bFlash)
				{
					// Flash point geometry
					FlashParameters fp = new FlashParameters();
					fp.Geometry = geom;
					fp.Flash = true;
					fp.ShowFlashLines = false;
					fp.Navigate = FlashNavigation.None;
					QueryViewExt.Current.Flash(fp);
				}
				else if (bGraphic)
				{
					// Add measure graphic
					string sLabel = UnitsHelper.GetMeasureFeatureLabel(geom, _LinearUnit, _AreaUnit);
					GraphicParameters gp = new GraphicParameters();
					gp.Action = GraphicAction.AddMeasure;
					gp.Geometry = geom;
					gp.Label = sLabel;
					QueryViewExt.Current.ProcessGraphic(gp);
				}
			}
			this.UpdateResults(geom, true);
		}

		// Event handlers

		void DistIsChecked_PropertyChanged(object sender, PropertyChangedEventArgs e)
		{
			this.RaisePropertyChanged("DistIsChecked");
		}

		void AreaIsChecked_PropertyChanged(object sender, PropertyChangedEventArgs e)
		{
			this.RaisePropertyChanged("AreaIsChecked");
		}

		void QueryModelExt_CanNavigateMap(object sender, NotifyCanExecuteArgs e)
		{

			if (!e.CanExecute)
				return;

			// Set default measurement units according to map units

			SpatialReference sr = QueryModelExt.Current.SpatialReference;
			LinearUnit lu;
			bool? bResult = UnitsHelper.GetStandardLinearUnit(sr, out lu);
			if (bResult == null)
				return;
			_LinearUnit = new MeasureLinearUnit(lu);
			if (bResult == false)
				return;
			string sKey = lu.Name;
			if (sKey == "Statute Miles")
				sKey = "Miles";
			else if (sKey == "Nautical Miles")
				sKey = "NauticalMiles";
			DistIsChecked[sKey] = true;
			_AreaUnit = new MeasureAreaUnit(AreaUnit.CreateFromLinearUnit(lu));
			string[] Tok = _AreaUnit.Name.Split(' ');
			sKey = Tok[Tok.Length - 1];
			AreaIsChecked[sKey] = true;

		}

		// Private methods

		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private void RaiseShowWindow()
		{
			if (this.ShowWindow == null)
				return;
			this.ShowWindow(this, new EventArgs());
		}

		private void Reset(bool bClearGraphics = false)
		{
			_GeomList.Clear();
			StringBuilder sb = new StringBuilder();
			sb.Append("To measure a distance, draw a line.\n\n");
			sb.Append("To measure an area, click 'Measure An Area'\nthen draw a polygon.\n\n");
			sb.Append("To measure a feature, click 'Measure A Feature'\nthen click a feature.");
			ResultText = sb.ToString();
			if (!bClearGraphics)
				return;
			GraphicParameters gp = new GraphicParameters();
			gp.Action = GraphicAction.ClearMeasure;
			QueryViewExt.Current.ProcessGraphic(gp);
		}

		private void UpdateResults(Geometry geom, bool bComplete)
		{

			// Check for null or point geometry

			if (geom == null)
			{
				ResultText = "No feature found.";
				return;
			}
			StringBuilder sb = new StringBuilder();
			if (geom is MapPoint)
			{
				this.AddPointText(sb, (MapPoint)geom);
				ResultText = sb.ToString();
				return;
			}

			// Get current length and area

			Measurement m = UnitsHelper.GetMeasureLength(geom, _LinearUnit);
			double dLength = m.Value;
			string sLengthUnits = m.Units;
			m = UnitsHelper.GetMeasureArea(geom, _AreaUnit);
			double dArea = m.Value;
			string sAreaUnits = m.Units;
			bool bGeo = geom.SpatialReference.IsGeographic;
			bool bPoly = (geom.Dimension > GeometryDimension.Curve);

			// Get length and area sums
			// (Assumes that all geometries have the same spatial reference)

			double dSumLength = 0;
			double dSumArea = 0;
			if (_bShowTotal)
			{
				foreach (Geometry g in _GeomList)
				{
					if (bPoly && (g.Dimension < GeometryDimension.Area))
						continue;
					//else if (!bPoly && (g.Dimension > GeometryDimension.Curve))
					//	continue;
					m = UnitsHelper.GetMeasureLength(g, _LinearUnit);
					dSumLength += m.Value;
					m = UnitsHelper.GetMeasureArea(g, _AreaUnit);
					dSumArea += m.Value;
				}
				dSumLength += dLength;
				dSumArea += dArea;
			}

			// Add header

			this.AddHeaderText(sb, bGeo, bPoly);

			// Add segment length and bearing

			if (_Mode != MeasureType.Feature)
				this.AddSegmentText(sb, geom);

			// Add length/area text

			string sLen = "Length";
			if (bPoly)
				sLen = "Perimeter";
			sb.Append("\n" + sLen + ": " + dLength.ToString("N6") + " " + sLengthUnits);
			if (bPoly)
				sb.Append("\nArea: " + dArea.ToString("N6") + " " + sAreaUnits);
			if (_bShowTotal)
			{
				sb.Append("\n\nSum " + sLen + "s: " + dSumLength.ToString("N6") + " " + sLengthUnits);
				if (bPoly)
					sb.Append("\nSum Areas: " + dSumArea.ToString("N6") + " " + sAreaUnits);
			}

			// Finish up

			if (_bShowTotal && bComplete)
				_GeomList.Add(geom);
			ResultText = sb.ToString();

		}

		private void AddPointText(StringBuilder sb, MapPoint pt)
		{

			// Show coordinate

			double dX = pt.X;
			double dY = pt.Y;
			string sUnits = "";
			SpatialReference sr = pt.SpatialReference;
			if (sr.Unit is LinearUnit)
			{
				MeasureLinearUnit mlu = new MeasureLinearUnit((LinearUnit)sr.Unit);
				dX = mlu.ConvertTo(dX, _LinearUnit);
				dY = mlu.ConvertTo(dY, _LinearUnit);
				sUnits = _LinearUnit.Name;
			}
			sb.Append("Point feature");
			sb.Append("\nX: " + dX.ToString() + " " + sUnits);
			sb.Append("\nY: " + dY.ToString() + " " + sUnits);

		}

		private void AddHeaderText(StringBuilder sb, bool bGeo, bool bPoly)
		{
			string sMeasurement = "Planar";
			if (bGeo)
				sMeasurement = "Geodesic";
			switch (_Mode)
			{
				case MeasureType.Line:
					sb.Append("Line measurement (" + sMeasurement + ")");
					break;
				case MeasureType.Area:
					sb.Append("Area measurement (" + sMeasurement + ")");
					break;
				case MeasureType.Feature:
					if (!bPoly)
						sb.Append("Line feature (" + sMeasurement + ")");
					else
						sb.Append("Polygon feature (" + sMeasurement + ")");
					break;
			}
		}

		private void AddSegmentText(StringBuilder sb, Geometry g)
		{

			// Get segment geometry

			double dSegLen = 0;
			string sUnits = "";
			string sSegBearing = "";
			IReadOnlyCollection<ReadOnlyPart> PC;
			if (g.GeometryType == GeometryType.Polyline)
			{
				Polyline p = (Polyline)g;
				PC = p.Parts;
			}
			else if (g.GeometryType == GeometryType.Polygon)
			{
				Polygon p = (Polygon)g;
				PC = p.Parts;
			}
			else
				return;
			ReadOnlyPart sc = PC.ElementAt(0);
			int iCount = sc.Count;
			if (iCount == 0)
				return;
			Segment seg = sc[iCount - 1];
			Polyline pl = new Polyline(new Segment[] { seg });
			Measurement m = UnitsHelper.GetMeasureLength(pl, _LinearUnit);
			dSegLen = m.Value;
			if (dSegLen == 0)
				return;  // Ignore zero-length segments
			sUnits = m.Units;
			sSegBearing = GeometryHelper.Bearing(seg);

			// Add text

			sb.Append("\n Segment: " + dSegLen.ToString("N6") + " " + sUnits);
			sb.Append("\n Bearing: " + sSegBearing);

		}

		private void DistanceChecked()
		{
			string sKey = _LinearUnit.Name.Replace(" ","").Replace("Statute","");
			DistIsChecked[sKey] = true;
		}

		private void AreaChecked()
		{
			string sKey = _AreaUnit.Name.Replace(" ", "").Replace("Statute", "").Replace("Square", "");
			AreaIsChecked[sKey] = true;
		}

	}

	// CheckHelper manages the check state of the unit menu items

	public class CheckHelper : IDictionary<string, bool>, INotifyPropertyChanged
	{

		private Dictionary<string, bool> _dict = new Dictionary<string, bool>();

		public event PropertyChangedEventHandler PropertyChanged;

		public bool this[string key]
		{
			get
			{
				if (!_dict.ContainsKey(key))
					return false;
				return _dict[key];
			}
			set
			{
				_dict[key] = value;
				if (!value)
					return;
				List<string> KeyList = _dict.Keys.ToList();
				foreach (string sKey in KeyList)
				{
					if (sKey == key)
						continue;
					_dict[sKey] = false;
				}
				this.RaisePropertyChanged(key);
			}
		}

		public void Add(string key, bool value)
		{
			_dict.Add(key, value);
			this.RaisePropertyChanged(key);
		}

		public bool Remove(string key)
		{
			bool bResult = _dict.Remove(key);
			if (bResult)
				this.RaisePropertyChanged(key);
			return bResult;
		}

		public bool ContainsKey(string key)
		{
			return _dict.ContainsKey(key);
		}

		public bool TryGetValue(string key, out bool value)
		{
			return _dict.TryGetValue(key, out value);
		}

		public ICollection<string> Keys
		{
			get { return _dict.Keys; }
		}

		public ICollection<bool> Values
		{
			get { return _dict.Values; }
		}

		public int Count
		{
			get { return _dict.Count; }
		}

		public bool IsReadOnly
		{
			get { return false; }
		}

		public void Clear()
		{
			_dict.Clear();
		}

		public void Add(KeyValuePair<string, bool> item)
		{
			throw new NotImplementedException();
		}

		public bool Contains(KeyValuePair<string, bool> item)
		{
			throw new NotImplementedException();
		}

		public void CopyTo(KeyValuePair<string, bool>[] array, int arrayIndex)
		{
			throw new NotImplementedException();
		}

		public bool Remove(KeyValuePair<string, bool> item)
		{
			throw new NotImplementedException();
		}

		public IEnumerator<KeyValuePair<string, bool>> GetEnumerator()
		{
			throw new NotImplementedException();
		}

		System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
		{
			throw new NotImplementedException();
		}

		private void RaisePropertyChanged(string sKey)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sKey));
		}

	}

}
