﻿using System;
using System.Collections.Generic;
using System.Windows.Input;
using System.Xml.Linq;

using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Mapping;

using Extensibility;

namespace QueryAddin
{

	// QueryModelExt manages query engines,
	// and acts as a front end for QueryViewExt

	public class QueryModelExt : IConfigurable
	{

		public static QueryModelExt Current { get; private set; }
		public SpatialReference SpatialReference { get; private set; }
		public IQueryLayerHelper LayerHelper { get; private set; }
		public IdentifyEngine IdentifyEngine { get; private set; }
		public FindEngine FindEngine { get; private set; }
		public List<IQueryableLayer> QueryableLayers { get; private set; }
		public ResultCollectionType ResultCollectionType { get; private set; }
		public bool QueryRelated { get; private set; }

		private string _sConfig;
		private XElement _xAttachmentModel;
		private Map _map;

		// Bindable view properties

		public ICommand ViewSpatialRefChangedCommand { get; private set; }

		public ICommand ViewScaleChangedCommand { get; private set; }

		// IConfigurable

		public void LoadConfiguration(string sConfigData)
		{
			_sConfig = sConfigData;
		}

		// Public events

		public event EventHandler<NotifyCanExecuteArgs> CanNavigateMap;
		public event EventHandler<NotifyCanExecuteArgs> CanQueryMap;

		// Constructor

		public QueryModelExt()
		{
			Current = this;
			QueryableLayers = new List<IQueryableLayer>();
			ResultCollectionType = QueryAddin.ResultCollectionType.Tree;
			QueryRelated = true;
			MapApplication.Current.AppReady += MapApplication_AppReady;
			AppViewExt.Current.MapChanged += AppViewExt_MapChanged;
			CommandHelper ch;
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.ExecuteSpatialRefChanged);
			ViewSpatialRefChangedCommand = ch;
			ch = new CommandHelper(true);
			ch.AddExecuteAction(this.ExecuteScaleChanged);
			ViewScaleChangedCommand = ch;
		}

		// Public methods

		public void OpenAttachmentDialog(IdentifyResult idResult)
		{
			IAttachmentDialogModel pModel;
			if (_xAttachmentModel == null)
				pModel = new BasicAttachmentDialogModel();
			else
				pModel = ConfigHelper.CreateInstance(_xAttachmentModel.Attribute("Assembly").Value, _xAttachmentModel.Attribute("ClassName").Value) as IAttachmentDialogModel;
			pModel.Initialize(idResult);
			MapApplication.Current.CreateChildWindow("QueryAddin", "AttachmentDialog", pModel);
		}

		// Event handlers

		void MapApplication_AppReady(object sender, EventArgs e)
		{

			// Finalize configuration

			this.FinalizeConfiguration();

		}

		void AppViewExt_MapChanged(object sender, MapChangedArgs e)
		{
			if (_map != null)
				_map.OperationalLayers.CollectionChanged -= OperationalLayers_CollectionChanged;
			_map = e.Map;
			_map.OperationalLayers.CollectionChanged += OperationalLayers_CollectionChanged;
			this.GetQueryableLayers();
		}

		void OperationalLayers_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
		{
			this.GetQueryableLayers();
		}

		// Private methods

		private void ExecuteSpatialRefChanged(object parameter)
		{
			SpatialReference sr = parameter as SpatialReference;
			SpatialReference = sr;
			if (this.CanNavigateMap == null)
				return;
			bool bCanNavigateMap = (sr != null);
			NotifyCanExecuteArgs e = new NotifyCanExecuteArgs() { CanExecute = bCanNavigateMap };
			this.CanNavigateMap(this, e);
		}

		private void ExecuteScaleChanged(object parameter)
		{
			if (FindEngine != null)
				FindEngine.ScaleChanged();
		}

		private void FinalizeConfiguration()
		{

			QueryRelated = true;
			if (!string.IsNullOrEmpty(_sConfig))
			{
				XDocument xDoc = XDocument.Parse(_sConfig);
				XElement xRoot = xDoc.Root;
				XElement xElem;
				xElem = xRoot.Element("Identify");
				if (xElem != null)
				{
					XAttribute xAttr = xElem.Attribute("ResultCollection");
					if (xAttr != null)
					{
						switch (xAttr.Value)
						{
							case "Tree":
								xAttr = xElem.Attribute("QueryRelated");
								if (xAttr != null)
									QueryRelated = (xAttr.Value == "True");
								break;
							case "List":
								ResultCollectionType = ResultCollectionType.List;
								QueryRelated = false; // List type overrides QueryRelayed value
								break;
						}
					}
				}
				xElem = xRoot.Element("QueryLayerHelper");
				if (xElem == null)
					LayerHelper = new BasicQueryLayerHelper();
				else
					LayerHelper = ConfigHelper.CreateInstance(xElem.Attribute("Assembly").Value, xElem.Attribute("ClassName").Value) as IQueryLayerHelper;
				_xAttachmentModel = xRoot.Element("AttachmentDialogModel");
			}
			else
				LayerHelper = new BasicQueryLayerHelper();
			FindEngine = new FindEngine();
			IdentifyEngine = new IdentifyEngine();

		}

		private async void GetQueryableLayers()
		{

			// Check for queryable layers

			if (LayerHelper == null)
				return;
			QueryableLayers.Clear();
			List<IQueryableLayer> NewList = await LayerHelper.GetQueryableLayers(_map);
			if (NewList != null)
			{
				foreach (IQueryableLayer pLyr in NewList)
					QueryableLayers.Add(pLyr);
			}
			if (FindEngine != null)
				FindEngine.Initialize(QueryableLayers);
			if (IdentifyEngine != null)
				IdentifyEngine.Initialize(QueryableLayers, ResultCollectionType);
			if (this.CanQueryMap != null)
			{
				NotifyCanExecuteArgs args = new NotifyCanExecuteArgs() { CanExecute = (QueryableLayers.Count > 0) };
				this.CanQueryMap(this, args);
			}

		}

	}

}
