﻿using System.Collections.Generic;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Windows.Controls;
using System.Windows.Input;

//using Debug = System.Diagnostics.Debug;

using Esri.ArcGISRuntime.Data;
using Esri.ArcGISRuntime.Symbology;
using Esri.ArcGISRuntime.UI;

using Extensibility;

namespace QueryAddin
{

	// QueryTools manages Identify, Measure, Highlight, and Text tool functions

	public enum QueryToolType
	{
		None,
		Identify,
		Measure,
		Highlighter,
		Eraser,
		Text
	}

	public enum MeasureType
	{
		Line,
		Area,
		Feature
	}

	public enum HighlighterType
	{
		Yellow,
		Blue,
		Green
	}

	public enum EraserType
	{
		Entire,
		Small,
		Medium,
		Large
	}

	public class QueryTools : INotifyPropertyChanged
	{

		public const double PixelTolerance = 3;

		public static QueryTools Current { get; private set; }

		private bool _bIdentifyChecked = false;
		private bool _bMeasureChecked = false;
		private Image _imgHighlighterIcon;
		private bool _bHighlighterChecked = false;
		private string _sHighlighterText;
		private Image _imgEraserIcon;
		private bool _bEraserChecked = false;
		private string _sEraserText;
		private bool _bTextChecked = false;

		private QueryToolType _ToolType = QueryToolType.None;
		private MeasureType _MeasureType = MeasureType.Line;
		private HighlighterType _HighlighterType = HighlighterType.Yellow;
		private EraserType _EraserType = EraserType.Entire;
		private Dictionary<QueryToolType, ToolHelper> _QueryTools = new Dictionary<QueryToolType, ToolHelper>();
		private IdentifyDialogModel _IdentifyDialog;
		private MeasureDialogModel _MeasureDialog;

		// Bindable layout properties

		public ICommand IdentifyCommand { get; private set; }

		public bool IdentifyToolChecked
		{
			get { return _bIdentifyChecked; }
			set
			{
				_bIdentifyChecked = value;
				this.RaisePropertyChanged();
			}
		}

		public ICommand MeasureCommand { get; private set; }

		public bool MeasureToolChecked
		{
			get { return _bMeasureChecked; }
			set
			{
				_bMeasureChecked = value;
				this.RaisePropertyChanged();
			}
		}

		public ICommand HighlighterCommand { get; private set; }

		public Image HighlighterIcon
		{
			get { return _imgHighlighterIcon; }
			set
			{
				_imgHighlighterIcon = value;
				this.RaisePropertyChanged();
			}
		}

		public bool HighlighterToolChecked
		{
			get { return _bHighlighterChecked; }
			set
			{
				_bHighlighterChecked = value;
				this.RaisePropertyChanged();
			}
		}

		public string HighlighterText
		{
			get { return _sHighlighterText; }
			set
			{
				_sHighlighterText = value;
				this.RaisePropertyChanged();
			}
		}

		public ICommand HighlighterMenuCommand { get; private set; }

		public ICommand EraserCommand { get; private set; }

		public Image EraserIcon
		{
			get { return _imgEraserIcon; }
			set
			{
				_imgEraserIcon = value;
				this.RaisePropertyChanged();
			}
		}

		public bool EraserToolChecked
		{
			get { return _bEraserChecked; }
			set
			{
				_bEraserChecked = value;
				this.RaisePropertyChanged();
			}
		}

		public string EraserText
		{
			get { return _sEraserText; }
			set
			{
				_sEraserText = value;
				this.RaisePropertyChanged();
			}
		}

		public ICommand EraserMenuCommand { get; private set; }

		public ICommand TextCommand { get; private set; }

		public bool TextToolChecked
		{
			get { return _bTextChecked; }
			set
			{
				_bTextChecked = value;
				this.RaisePropertyChanged();
			}
		}

		// INotifyPropertyChanged

		public event PropertyChangedEventHandler PropertyChanged;

		// Constructor

		public QueryTools()
		{
			Current = this;
			this.Init();
		}

		// Public methods

		public void IdentifyFeature(IQueryableLayer pLyr, Feature Feat, bool QueryRelated)
		{
			if (_IdentifyDialog == null)
				_IdentifyDialog = new IdentifyDialogModel();
			_IdentifyDialog.Open();
			_IdentifyDialog.IdentifyFeature(pLyr, Feat, QueryRelated);
		}

		public void SetMeasureType(MeasureType mt)
		{
			_MeasureType = mt;
			ToolHelper th = _QueryTools[QueryToolType.Measure];
			ToolConfiguration tc = th.Configuration;
			switch (mt)
			{
				case MeasureType.Line:
					tc.SketchType = SketchCreationMode.Polyline;
					tc.SketchStyle.LineSymbol = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.MeasureLine);
					tc.SketchStyle.FillSymbol = null;
					break;
				case MeasureType.Area:
					tc.SketchType = SketchCreationMode.Polygon;
					tc.SketchStyle.LineSymbol = null;
					tc.SketchStyle.FillSymbol = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.MeasureArea);
					break;
				case MeasureType.Feature:
					tc.SketchType = SketchCreationMode.Point;
					tc.SketchStyle.LineSymbol = null;
					tc.SketchStyle.FillSymbol = null;
					break;
			}
			th.ExecuteCommand.Execute(null);
		}

		// Event handlers

		void ToolHelper_ToolProgress(object sender, ToolHelper.NotifyToolArgs e)
		{
			_MeasureDialog.MeasureProgress(e.Geometry);
		}

		void ToolHelper_ToolComplete(object sender, ToolHelper.NotifyToolArgs e)
		{
			GraphicParameters gp;
			switch (_ToolType)
			{
				case QueryToolType.Identify:
					_IdentifyDialog.Identify(e.Geometry);
					break;
				case QueryToolType.Measure:
					_MeasureDialog.MeasureComplete(e.Geometry);
					break;
				case QueryToolType.Highlighter:
					gp = new GraphicParameters();
					switch (_HighlighterType)
					{
						case HighlighterType.Yellow:
							gp.Action = GraphicAction.AddHighlightYellow;
							break;
						case HighlighterType.Blue:
							gp.Action = GraphicAction.AddHighlightBlue;
							break;
						case HighlighterType.Green:
							gp.Action = GraphicAction.AddHighlightGreen;
							break;
					}
					gp.Geometry = e.Geometry;
					QueryViewExt.Current.ProcessGraphic(gp);
					break;
				case QueryToolType.Eraser:
					gp = new GraphicParameters();
					switch (_EraserType)
					{
						case EraserType.Entire:
							gp.Action = GraphicAction.EraseHighlightEntire;
							break;
						case EraserType.Small:
							gp.Action = GraphicAction.EraseHighlightSmall;
							break;
						case EraserType.Medium:
							gp.Action = GraphicAction.EraseHighlightMedium;
							break;
						case EraserType.Large:
							gp.Action = GraphicAction.EraseHighlightLarge;
							break;
					}
					gp.Geometry = e.Geometry;
					QueryViewExt.Current.ProcessGraphic(gp);
					break;
				case QueryToolType.Text:
					TextDialogModel.OpenNew(e.Geometry);
					break;
			}
		}

		// Private methods

		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private void Init()
		{

			// Set up tools

			ToolHelper th;

			th = this.CreateTool(QueryToolType.Identify);
			IdentifyCommand = th.ExecuteCommand;

			th = this.CreateTool(QueryToolType.Measure);
			MeasureCommand = th.ExecuteCommand;

			th = this.CreateTool(QueryToolType.Highlighter);
			HighlighterCommand = th.ExecuteCommand;
			HighlighterIcon = QueryToolHelper.GetHighlighterIcon(_HighlighterType);
			HighlighterText = QueryToolHelper.GetHighlighterToolTip(_HighlighterType);

			th = this.CreateTool(QueryToolType.Eraser);
			EraserCommand = th.ExecuteCommand;
			EraserIcon = QueryToolHelper.GetEraserIcon(_EraserType);
			EraserText = QueryToolHelper.GetEraserToolTip(_EraserType);

			th = this.CreateTool(QueryToolType.Text);
			TextCommand = th.ExecuteCommand;

			// Highlighter/Eraser menu commands

			CommandHelper ch;
			ch = new CommandHelper(false);
			ch.AddCanExecuteEvent(QueryModelExt.Current, "CanNavigateMap");
			ch.AddExecuteAction(this.ExecuteHighlighterMenu);
			HighlighterMenuCommand = ch;
			ch = new CommandHelper(false);
			ch.AddCanExecuteEvent(QueryModelExt.Current, "CanNavigateMap");
			ch.AddExecuteAction(this.ExecuteEraserMenu);
			EraserMenuCommand = ch;

		}

		private void ExecuteActivate(object parameter)
		{
			_ToolType = (QueryToolType)parameter;
			switch (_ToolType)
			{
				case QueryToolType.Identify:
					if (_IdentifyDialog == null)
						_IdentifyDialog = new IdentifyDialogModel();
					_IdentifyDialog.Open();
					IdentifyToolChecked = true;
					break;
				case QueryToolType.Measure:
					if (_MeasureDialog == null)
						_MeasureDialog = new MeasureDialogModel();
					_MeasureDialog.Open();
					MeasureToolChecked = true;
					break;
				case QueryToolType.Highlighter:
					HighlighterToolChecked = true;
					break;
				case QueryToolType.Eraser:
					EraserToolChecked = true;
					break;
				case QueryToolType.Text:
					TextToolChecked = true;
					break;
			}
		}

		private void ExecuteDeactivate(object parameter)
		{
			QueryToolType tt = (QueryToolType)parameter;
			switch (tt)
			{
				case QueryToolType.Identify:
					IdentifyToolChecked = false;
					break;
				case QueryToolType.Measure:
					MeasureToolChecked = false;
					break;
				case QueryToolType.Highlighter:
					HighlighterToolChecked = false;
					break;
				case QueryToolType.Eraser:
					EraserToolChecked = false;
					break;
				case QueryToolType.Text:
					TextToolChecked = false;
					break;
			}
			_ToolType = QueryToolType.None;
		}

		private ToolHelper CreateTool(QueryToolType tt)
		{

			// Get tool cursor and draw symbol

			Cursor cur = null;
			Symbol sym = null;
			string sEnableEvent = "CanNavigateMap";
			switch (tt)
			{
				case QueryToolType.Identify:
					cur = QueryToolHelper.GetCursor(QueryToolHelperType.Identify);
					sym = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.Identify);
					sEnableEvent = "CanQueryMap";
					break;
				case QueryToolType.Measure:
					cur = QueryToolHelper.GetCursor(QueryToolHelperType.MeasureLine);
					sym = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.MeasureLine);
					break;
				case QueryToolType.Highlighter:
					cur = QueryToolHelper.GetCursor(QueryToolHelperType.HighlightYellow);
					sym = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.HighlightYellow);
					break;
				case QueryToolType.Eraser:
					cur = QueryToolHelper.GetCursor(QueryToolHelperType.EraseEntire);
					sym = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.EraseEntire);
					break;
				case QueryToolType.Text:
					cur = QueryToolHelper.GetCursor(QueryToolHelperType.Text);
					sym = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.Text);
					break;

			}

			// Create ToolHelper

			ToolHelper th = new ToolHelper(false);
			th.ExecuteCommand.AddCanExecuteEvent(QueryModelExt.Current, sEnableEvent);
			th.ActivateCommand.AddExecuteAction(this.ExecuteActivate, tt);
			th.DeactivateCommand.AddExecuteAction(this.ExecuteDeactivate, tt);
			th.ToolComplete += ToolHelper_ToolComplete;
			if (tt == QueryToolType.Measure)
				th.ToolProgress += ToolHelper_ToolProgress; // Used only for measure tool
			_QueryTools[tt] = th; // Used for measure, highlighter, and eraser

			// Set up initial tool configuration

			ToolConfiguration tc = th.Configuration;
			tc.Cursor = cur;
			tc.Editable = false;
			tc.EditConfiguration.AllowMove = false;
			tc.EditConfiguration.AllowRotate = false;
			tc.EditConfiguration.AllowVertexEditing = false;
			tc.EditConfiguration.ResizeMode = SketchResizeMode.None;
			tc.SketchStyle.FillSymbol = null;
			tc.SketchStyle.LineSymbol = null;
			tc.SketchStyle.MidVertexSymbol = null;
			tc.SketchStyle.VertexSymbol = null;
			tc.SketchStyle.SelectedVertexSymbol = null;
			tc.SelectionColor = System.Drawing.Color.Transparent;
			switch (tt)
			{
				case QueryToolType.Identify:
					tc.SketchType = SketchCreationMode.Rectangle;
					tc.SingleTapCreatesPoint = true;
					tc.PointPixelTolerance = PixelTolerance;
					tc.SketchStyle.FillSymbol = sym;
					break;
				case QueryToolType.Measure:
					tc.SketchType = SketchCreationMode.Polyline;
					tc.DoubleTapToFinish = true;
					tc.RubberBand = true;
					tc.SketchStyle.LineSymbol = sym;
					break;
				case QueryToolType.Highlighter:
					tc.SketchType = SketchCreationMode.FreehandLine;
					tc.SketchStyle.LineSymbol = sym;
					break;
				case QueryToolType.Eraser:
					tc.SketchType = SketchCreationMode.Rectangle;
					tc.SingleTapCreatesPoint = true;
					tc.PointPixelTolerance = PixelTolerance;
					tc.SketchStyle.FillSymbol = sym;
					break;
				case QueryToolType.Text:
					tc.SketchType = SketchCreationMode.Point;
					break;
			}
			return th;

		}

		private void ExecuteHighlighterMenu(object parameter)
		{
			string sParameter = parameter as string;
			ToolHelper th = _QueryTools[QueryToolType.Highlighter];
			ToolConfiguration tc = th.Configuration;
			switch (sParameter)
			{
				case "SelectYellow":
					_HighlighterType = HighlighterType.Yellow;
					tc.Cursor = QueryToolHelper.GetCursor(QueryToolHelperType.HighlightYellow);
					tc.SketchStyle.LineSymbol = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.HighlightYellow);
					break;
				case "SelectBlue":
					_HighlighterType = HighlighterType.Blue;
					tc.Cursor = QueryToolHelper.GetCursor(QueryToolHelperType.HighlightBlue);
					tc.SketchStyle.LineSymbol = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.HighlightBlue);
					break;
				case "SelectGreen":
					_HighlighterType = HighlighterType.Green;
					tc.Cursor = QueryToolHelper.GetCursor(QueryToolHelperType.HighlightGreen);
					tc.SketchStyle.LineSymbol = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.HighlightGreen);
					break;
			}
			HighlighterIcon = QueryToolHelper.GetHighlighterIcon(_HighlighterType);
			HighlighterText = QueryToolHelper.GetHighlighterToolTip(_HighlighterType);
			ICommand pExecute = th.ExecuteCommand;
			pExecute.Execute(null);
		}

		private void ExecuteEraserMenu(object parameter)
		{
			string sParameter = parameter as string;
			ToolHelper th = _QueryTools[QueryToolType.Eraser];
			ToolConfiguration tc = th.Configuration;
			switch (sParameter)
			{
				case "SelectEntire":
					_EraserType = EraserType.Entire;
					tc.Cursor = QueryToolHelper.GetCursor(QueryToolHelperType.EraseEntire);
					tc.SketchType = SketchCreationMode.Rectangle;
					tc.SketchStyle.LineSymbol = null;
					tc.SketchStyle.FillSymbol = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.EraseEntire);
					break;
				case "SelectSmall":
					_EraserType = EraserType.Small;
					tc.Cursor = QueryToolHelper.GetCursor(QueryToolHelperType.EraseSmall);
					tc.SketchType = SketchCreationMode.FreehandLine;
					tc.SketchStyle.LineSymbol = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.EraseSmall);
					tc.SketchStyle.FillSymbol = null;
					break;
				case "SelectMedium":
					_EraserType = EraserType.Medium;
					tc.Cursor = QueryToolHelper.GetCursor(QueryToolHelperType.EraseMedium);
					tc.SketchType = SketchCreationMode.FreehandLine;
					tc.SketchStyle.LineSymbol = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.EraseLarge);
					tc.SketchStyle.FillSymbol = null;
					break;
				case "SelectLarge":
					_EraserType = EraserType.Large;
					tc.Cursor = QueryToolHelper.GetCursor(QueryToolHelperType.EraseLarge);
					tc.SketchType = SketchCreationMode.FreehandLine;
					tc.SketchStyle.LineSymbol = QueryToolHelper.GetDrawSymbol(QueryToolHelperType.EraseLarge);
					tc.SketchStyle.FillSymbol = null;
					break;
			}
			EraserIcon = QueryToolHelper.GetEraserIcon(_EraserType);
			EraserText = QueryToolHelper.GetEraserToolTip(_EraserType);
			ICommand pExecute = th.ExecuteCommand;
			pExecute.Execute(null);
		}

	}

}
