﻿using System;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Windows;
using System.Windows.Input;

using Brush = System.Windows.Media.Brush;
using Brushes = System.Windows.Media.Brushes;
using FontFamily = System.Windows.Media.FontFamily;
using Fonts = System.Windows.Media.Fonts;
using Shape = System.Windows.Shapes.Shape;

using Esri.ArcGISRuntime.Geometry;

using Extensibility;

namespace QueryAddin
{

	// TextDialogModel creates and updates map text overlays

	public class TextDialogModel : INotifyPropertyChanged, ICommand, IDialogContent
	{

		private const string DefaultFontFamily = "Arial";
		private const double DefaultFontSize = 11;
		private static double[] FontSizes = new double[] { 8, 9, 10, 11, 12, 14, 16, 18, 20, 22, 24, 26, 28, 36, 48, 72 };

		private MapPoint _pt;
		private MapTextModel _mt;
		private bool _bExisting = false;
		private bool _bUpdated = false;

		private string _sText;
		private FontFamily _FontFamily;
		private double _dFontSize;
		private FontWeight _FontWeight;
		private FontStyle _FontStyle;
		private TextDecorationCollection _TextDecorations;
		private bool _bColorGrid;
		private Brush _brushForeground;
		private double _dAngle;
		private TextAlignment _TextAlignment;


		// Bindable properties

		public string Text
		{
			get { return _sText; }
			set
			{
				_sText = value;
				this.RaisePropertyChanged();
			}
		}

		public ObservableCollection<FontFamily> FontFamiliesSource { get; private set; }

		public FontFamily FontFamiliesSelectedItem
		{
			get { return _FontFamily; }
			set
			{
				_FontFamily = value;
				this.RaisePropertyChanged();
			}
		}

		public ObservableCollection<double> FontSizesSource { get; private set; }

		public double FontSizesSelectedItem
		{
			get { return _dFontSize; }
			set
			{
				_dFontSize = value;
				this.RaisePropertyChanged();
			}
		}

		public bool IsBold
		{
			get { return (_FontWeight == FontWeights.Bold); }
			set
			{
				if (value)
					_FontWeight = FontWeights.Bold;
				else
					_FontWeight = FontWeights.Normal;
				this.RaisePropertyChanged();
			}
		}

		public bool IsItalic
		{
			get { return (_FontStyle == FontStyles.Italic); }
			set
			{
				if (value)
					_FontStyle = FontStyles.Italic;
				else
					_FontStyle = FontStyles.Normal;
				this.RaisePropertyChanged();
			}
		}

		public bool IsUnderline
		{
			get { return (_TextDecorations == TextDecorations.Underline); }
			set
			{
				if (value)
					_TextDecorations = TextDecorations.Underline;
				else
					_TextDecorations = null;
				this.RaisePropertyChanged();
			}
		}

		public ICommand Command { get { return this; } }

		public bool ColorGridVisible
		{
			get { return _bColorGrid; }
			set
			{
				_bColorGrid = value;
				this.RaisePropertyChanged();
			}
		}

		public Brush FontColor
		{
			get { return _brushForeground; }
			set
			{
				_brushForeground = value;
				this.RaisePropertyChanged();
			}
		}

		public string TextAngle
		{
			get { return _dAngle.ToString(); }
			set
			{
				if (!double.TryParse(value, out _dAngle))
					return;
				this.RaisePropertyChanged();
			}
		}

		public bool IsAlignLeft
		{
			get { return (_TextAlignment == TextAlignment.Left);}
			set
			{
				if (value)
					this.SetTextAlignment(TextAlignment.Left);
			}
		}

		public bool IsAlignCenter
		{
			get { return (_TextAlignment == TextAlignment.Center); }
			set
			{
				if (value)
					this.SetTextAlignment(TextAlignment.Center);
			}
		}

		public bool IsAlignRight
		{
			get { return (_TextAlignment == TextAlignment.Right); }
			set
			{
				if (value)
					this.SetTextAlignment(TextAlignment.Right);
			}
		}

		public bool IsAlignFull
		{
			get { return (_TextAlignment == TextAlignment.Justify); }
			set
			{
				if (value)
					this.SetTextAlignment(TextAlignment.Justify);
			}
		}

		// INotifyPropertyChanged

		public event PropertyChangedEventHandler PropertyChanged;

		// ICommand

		public bool CanExecute(object parameter)
		{
			if (parameter is Shape)
				return true;
			string sParam = parameter as string;
			if (sParam == null)
				return false;
			switch (sParam)
			{
				case "SetColor":
				case "OK":
				case "Cancel":
					return true;
				case "Apply":
					return _bUpdated;
			}
			return false;
		}

		public event EventHandler CanExecuteChanged;

		public void Execute(object parameter)
		{
			if (parameter is Shape)
			{
				Shape shpParam = (Shape)parameter;
				FontColor = shpParam.Fill;
				ColorGridVisible = false;
			}
			string sParam = parameter as string;
			if (sParam == null)
				return;
			switch (sParam)
			{
				case "SetColor":
					ColorGridVisible = true;
					break;
				case "OK":
					this.OK();
					break;
				case "Cancel":
					this.Cancel();
					break;
				case "Apply":
					this.Apply();
					break;
			}
		}

		// IDialogContent

		public string Title
		{
			get { return "Add Text"; }
		}

		public bool Closable
		{
			get { return true; }
		}

		public event EventHandler<EventArgs> DialogOK;

		public event EventHandler<EventArgs> DialogCancel;

		// Public methods

		public static void OpenNew(Geometry geom)
		{
			MapPoint pt = geom as MapPoint;
			if (pt == null)
				return;
			TextDialogModel tm = new TextDialogModel(pt, null);
			MapApplication.Current.CreateChildWindow("QueryAddin", "TextDialog", tm);
		}

		public static void OpenExisting(MapTextModel mt)
		{
			if (mt == null)
				return;
			TextDialogModel tm = new TextDialogModel(null, mt);
			MapApplication.Current.CreateChildWindow("QueryAddin", "TextDialog", tm);
		}

		// Private constructors

		private TextDialogModel(MapPoint pt, MapTextModel mt)
		{

			_pt = pt;
			_mt = mt;
			_bExisting = (_mt != null);
			if (!_bExisting)
				_mt = new MapTextModel();
			_mt.MapTextIsSelected = true;

			// Set up font families and sizes

			FontFamily ffDefault = null;
			FontFamiliesSource = new ObservableCollection<FontFamily>();
			foreach (FontFamily ff in Fonts.SystemFontFamilies)
			{
				FontFamiliesSource.Add(ff);
				if (ff.ToString() != DefaultFontFamily)
					continue;
				ffDefault = ff;
			}
			FontSizesSource = new ObservableCollection<double>();
			foreach (double d in FontSizes)
				FontSizesSource.Add(d);

			// Initialize values

			if (_bExisting)
			{
				_sText = _mt.MapTextValue;
				_FontFamily = _mt.MapTextFontFamily;
				_dFontSize = _mt.MapTextFontSize;
				_FontWeight = _mt.MapTextFontWeight;
				_FontStyle = _mt.MapTextFontStyle;
				_TextDecorations = _mt.MapTextDecorations;
				_brushForeground = _mt.MapTextForeground;
				_dAngle = -(_mt.MapTextAngle);
				_TextAlignment = _mt.MapTextAlignment;
			}
			else
			{
				_sText = "Text";
				_FontFamily = ffDefault;
				_dFontSize = DefaultFontSize;
				_FontWeight = FontWeights.Normal;
				_FontStyle = FontStyles.Normal;
				_TextDecorations = null;
				_brushForeground = Brushes.Black;
				_dAngle = 0;
				_TextAlignment = TextAlignment.Left;
				this.ApplyValues();
			}
			_bColorGrid = false;
			
		}

		// Private methods

		private void RaisePropertyChanged([CallerMemberName]string sPropertyName = null)
		{
			_bUpdated = true;
			this.RaiseCanExecuteChanged();
			if (this.PropertyChanged == null)
				return;
			this.PropertyChanged(this, new PropertyChangedEventArgs(sPropertyName));
		}

		private void RaiseCanExecuteChanged()
		{
			if (this.CanExecuteChanged == null)
				return;
			this.CanExecuteChanged(this, new EventArgs());
		}

		private void ApplyValues()
		{
			_mt.MapTextValue = _sText;
			_mt.MapTextFontFamily = _FontFamily;
			_mt.MapTextFontSize = _dFontSize;
			_mt.MapTextFontWeight = _FontWeight;
			_mt.MapTextFontStyle = _FontStyle;
			_mt.MapTextDecorations = _TextDecorations;
			_mt.MapTextForeground = _brushForeground;
			_mt.MapTextAngle = -_dAngle;
			_mt.MapTextAlignment = _TextAlignment;
		}

		private void SetTextAlignment(TextAlignment ta)
		{
			_TextAlignment = ta;
			this.RaisePropertyChanged("IsAlignLeft");
			this.RaisePropertyChanged("IsAlignCenter");
			this.RaisePropertyChanged("IsAlignRight");
			this.RaisePropertyChanged("IsAlignFull");
		}

		private void OK()
		{
			this.ApplyValues();
			_mt.MapTextIsSelected = false;
			if (!_bExisting)
				this.Create();
			else
				_mt.UpdateLayout();
			if (this.DialogOK == null)
				return;
			this.DialogOK(this, new EventArgs());
		}

		private void Cancel()
		{
			_mt.MapTextIsSelected = false;
			if (this.DialogCancel == null)
				return;
			this.DialogCancel(this, new EventArgs());
		}

		private void Apply()
		{
			this.ApplyValues();
			_mt.MapTextIsSelected = true;
			if (!_bExisting)
				this.Create();
			else
				_mt.UpdateLayout();
			_bUpdated = false;
			this.RaiseCanExecuteChanged();
		}

		private void Create()
		{
			GraphicParameters gp = new GraphicParameters();
			gp.Action = GraphicAction.AddMapText;
			gp.Geometry = _pt;
			gp.MapText = _mt;
			QueryViewExt.Current.ProcessGraphic(gp);
			_bExisting = true;
		}

	}

}
