﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;

using Esri.ArcGISRuntime.ArcGISServices;
using Esri.ArcGISRuntime.Data;
using Esri.ArcGISRuntime.Geometry;
using Esri.ArcGISRuntime.Mapping;

namespace QueryAddin
{

	// WebQueryableLayer provides IQueryableLayer support for ArcGISMapImageLayers

	public interface IWebQueryableLayer
	{
		string MapImageLayerId { get; }
		long SublayerId { get; }
		WebSublayerInfo Info { get; }
	}

	public class WebQueryableLayer : BasicQueryableLayer, IWebQueryableLayer
	{

		private WebQueryLayerHelper _wh;

		// IWebQueryableLayer

		public string MapImageLayerId { get; private set; }

		public long SublayerId { get ; private set; }

		public WebSublayerInfo Info { get; private set; }

		// Public methods

		public static async Task<WebQueryableLayer> CreateFromSublayer(ArcGISMapImageSublayer subLyr, WebQueryLayerHelper wh)
		{
			long iID = subLyr.Id;
			WebQueryableLayer wqLyr = await WebQueryableLayer.CreateFromSublayerId(iID, wh);
			if (wqLyr == null)
				return null;
			wqLyr._pLyr = subLyr;
			return wqLyr;
		}

		public static async Task<WebQueryableLayer> CreateFromSublayerId(long iID, WebQueryLayerHelper wh)
		{
			WebSublayerInfo wsInfo = await wh.GetSublayerInfo(iID);
			ArcGISMapServiceSublayerInfo slInfo = wsInfo.ServiceInfo;
			if (slInfo == null)
				return null;
			WebQueryableLayer wqLyr = new WebQueryableLayer();
			wqLyr.MapImageLayerId = wh.MapImageLayerId;
			wqLyr.SublayerId = iID;
			wqLyr.Info = wsInfo;
			wqLyr._wh = wh;
			wqLyr._sName = slInfo.ServiceLayerName;
			wqLyr._sID = Guid.NewGuid().ToString();
			wqLyr.HasAttachments = slInfo.HasAttachments;
			wqLyr._Fields = WebQueryableLayer.GetQueryFieldInfos(slInfo);
			string sDisplayField = slInfo.DisplayFieldName;
			if (!string.IsNullOrEmpty(sDisplayField))
				wqLyr._sDisplayField = sDisplayField;
			else
				wqLyr._sDisplayField = BasicQueryLayerHelper.GetDefaultDisplayField(wqLyr._Fields);
			return wqLyr;
		}

		public static List<QueryFieldInfo> GetQueryFieldInfos(ArcGISMapServiceSublayerInfo slInfo)
		{
			return BasicQueryLayerHelper.GetQueryFieldInfos(slInfo.Fields, slInfo.SubtypeField, slInfo.FeatureSubtypes);
		}

		// Overridden methods

		protected override async Task<List<Feature>> QueryGeneral(string sWhere = null, Geometry geom = null)
		{
			return await _wh.QueryGeneral(this, sWhere, geom);
		}

		protected override async Task<List<string>> QueryUniqueValues(string sField, string sWhereClause = null)
		{
			return await _wh.QueryUniqueValues(this, sField, sWhereClause);
		}

		protected override async Task<List<Attachment>> QueryAttachments(Feature feat)
		{
			if (!Info.ServiceInfo.HasAttachments)
				return new List<Attachment>();
			if (feat is ArcGISFeature)
				return await BasicQueryLayerHelper.QueryAttachments((ArcGISFeature)feat);
			ArcGISFeature agFeat = await this.GetArcGISFeature(feat);
			if (agFeat == null)
				return null;
			return await BasicQueryLayerHelper.QueryAttachments(agFeat);
		}

		protected override async Task<List<QueryRelatedResult>> QueryRelated(Feature feat)
		{
			if (Info.ServiceInfo.RelationshipInfos.Count == 0)
				return new List<QueryRelatedResult>();
			if (feat is ArcGISFeature)
				return await _wh.QueryRelated((ArcGISFeature)feat);
			ArcGISFeature agFeat = await this.GetArcGISFeature(feat);
			if (agFeat == null)
				return null;
			return await _wh.QueryRelated(agFeat);
		}

		protected override bool CheckVisibility(double dScale)
		{
			if (_pLyr == null)
				return false;
			if (!_pLyr.IsVisible)
				return false;
			if (double.IsNaN(dScale))
				return false;
			if (dScale <= 0)
				return false;
			if (!_pLyr.IsVisibleAtScale(dScale))
				return false;
			return _wh.ParentVisible(SublayerId, dScale);
		}

		// Private methods

		private async Task<ArcGISFeature> GetArcGISFeature(Feature feat)
		{
			ServiceFeatureTable sfTab = await _wh.GetSublayerTable(SublayerId);
			string sOIDField = sfTab.ObjectIdField;
			if (!feat.Attributes.ContainsKey(sOIDField))
				return null;
			long iOID = -1;
			if (!long.TryParse(feat.Attributes[sOIDField].ToString(), out iOID))
				return null;
			QueryParameters qp = new QueryParameters();
			qp.ObjectIds.Add(iOID);
			qp.ReturnGeometry = true;
			FeatureQueryResult fqResult = await sfTab.QueryFeaturesAsync(qp, QueryFeatureFields.LoadAll);
			foreach (ArcGISFeature agFeat in fqResult)
				return agFeat;
			return null;
		}


	}

}
